/*
 * Copyright (C) <2002-2008> <Funkwerk Enterprise Communications GmbH> 
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * bsp.h - board support package header
 * ------------------------------------
 *
 */

#ifndef _INC_modulardsl_bsp_h_
#define _INC_modulardsl_bsp_h_

/*
** ----------------------------------------------------------------------------
** --- INCLUDES ---------------------------------------------------------------
** ----------------------------------------------------------------------------
*/

#include "p2001/p2001ok.h"



/*
** ----------------------------------------------------------------------------
** --- PLL / CLOCK ------------------------------------------------------------
** ----------------------------------------------------------------------------
*/

#define OSCILLATOR_CLK		12288000	/* 12.288 MHz */
#define SYSCLK			(6 * OSCILLATOR_CLK)



/*
** ----------------------------------------------------------------------------
** --- WATCHDOG ---------------------------------------------------------------
** ----------------------------------------------------------------------------
*/
#if !defined(_ASMLANGUAGE) && !defined(__ASSEMBLER__)
#define WD_TIMER_ADDR	ABS_Adr_WatchDog_Timer
#define WD_TRIGGER	do { \
				*(volatile unsigned long *)WD_TIMER_ADDR = 0; \
			} while(0)
#endif

/*
** ----------------------------------------------------------------------------
** --- HARDWARE IDENTIFICATION ------------------------------------------------
** ----------------------------------------------------------------------------
*/

#define HWID_ADDR		0x00007ffc
#define HWID_HWREV_MSK		0x00000007
#define HWID_MSK		0x000000f8

#define HWID_EUMEX820		0x30
#define HWID_T484		0x50
#define HWID_HWREV_0		0x0
#define HWID_HWREV_1		0x1

#if !defined(_ASMLANGUAGE) && !defined(__ASSEMBLER__)
#define GET_HWID()		\
	((*(volatile unsigned long *)(HWID_ADDR)) & (HWID_MSK))
#define GET_HWREV()		\
	((*(volatile unsigned long *)(HWID_ADDR)) & (HWID_HWREV_MSK))
#endif	/* _ASMLANGUAGE */



/*
** ----------------------------------------------------------------------------
** --- MEMORY MAP -------------------------------------------------------------
** ----------------------------------------------------------------------------
**
** ------------------------------ 0x00000000
** |                            |
** | internal SRAM 32KB         |
** |                            |
** |----------------------------|
**              ...
** |----------------------------| 0x00040000
** |                            |
** | internal ROM               |
** |                            |
** |----------------------------|
**              ...
** |----------------------------| 0x00100000
** |                            |
** | internal ASIC registers    |
** |                            |
** |----------------------------|
**              ...
** |----------------------------| 0x02000000
** |                            |
** | Flash - 2/4Mbyte/16Bit     |
** |                            |
** |----------------------------| 0x02(2|4)00000
** |                            |
** | Flash - 1/2/4Mbyte/16Bit   |
** |                            |
** |----------------------------|
**              ...
** |----------------------------| 0x04000000
** |                            |
** | Expansion Slot Modul 1     |
** |                            |
** |----------------------------| 0x06000000
** |                            |
** | Expansion Slot Modul 2     |
** |                            |
** |----------------------------| 0x08000000
** |                            |
** | CF-Card Interface          |
** |                            |
** |----------------------------|
**              ...
** |----------------------------| 0x40000000
** |                            |
** | SDRAM - 8/16Mbyte/32bit    |
** |                            |
** |----------------------------|
**
*/

#define SDRAM_BASE_ADDR		CS_SDRAM_BASE_ADDR
#if defined(PROJECT_VARIANT_16MB)
#define _SDRAM_SIZE		0x01000000
#else
#define _SDRAM_SIZE		0x00800000
#endif

#if defined(PROJECT_VARIANT_4MB_FLASH)
#define FLASH_SIZE		0x00400000
#else
#define FLASH_SIZE		0x00200000
#endif

#if defined(PROJECT_VARIANT_FLASHSIM)
#define SDRAM_SIZE		(_SDRAM_SIZE - FLASH_SIZE)
#define FLASH_BASE_ADDR		(CS_SDRAM_BASE_ADDR + SDRAM_SIZE)
#else
#define SDRAM_SIZE		_SDRAM_SIZE
#define FLASH_BASE_ADDR		CS_1_BASE_ADDR
#endif	/* PROJECT_VARIANT_FLASHSIM */

#define FLASH2_BASE_ADDR	CS_2_BASE_ADDR
#define FLASH2_SIZE		0x00100000



/*
** ----------------------------------------------------------------------------
** --- FLASH LAYOUT -----------------------------------------------------------
** ----------------------------------------------------------------------------
** NOTE!!! if you are going to change something do so in bsp/$(ARCH)/Makefile
** ----------------------------------------------------------------------------
**
** ------------------------------ 0x0200.0000 (16k)
** |                            |
** | 1st stage bootloader       |
** |                            |
** ------------------------------ 0x0200.4000 (8k)
** |                            |
** | serno, mac addresses ...   | 0x0200.5000
** |                            |
** ------------------------------ 0x0200.6000 (8k)
** |                            |
** | not used                   |
** |                            |
** ------------------------------ 0x0200.8000 (32k)
** |                            |
** | not used                   |
** |                            |
** ------------------------------ 0x0201.0000 (31x 64k)
** | flash status               |             (16bytes)
** | -------------------------- | 0x0201.0010
** | bootrecord                 |             (1k - 16bytes)
** | -------------------------- | 0x0201.0400
** | fs flashimage              |
** |                            |
** |             .              |
** |             .              |
** |             .              |
** |                            |
** ------------------------------
*/

#define FLASH_SECTOR_SIZE	0x10000			/* XXX todo CFI */
#define FLASH_FSIMAGE		(FLASH_BASE_ADDR + FLASH_SECTOR_SIZE)
#define FLASH_FSIMAGE_END	(FLASH_FSIMAGE + ((31*FLASH_SECTOR_SIZE) - 1))

#if !defined(_ASMLANGUAGE) && !defined(__ASSEMBLER__)
typedef struct {
	/* 0x0200.0000 */
	unsigned char		_res_0[16384];

	/* 0x0200.4000 */
	unsigned char		_res_1[4096];
	unsigned char		hw_pbx_data[4096];

	/* 0x0200.6000 */
	unsigned char		_res_2[8192];

	/* 0x0200.8000 */
	unsigned char		_res_3[32768];
	
	/* 0x0201.0000 */
	unsigned short		status_update_force;
	unsigned short		status_update_started;
	unsigned short		status_update_done;
	unsigned short		status_mps;
	unsigned char           _res_4[8];

	/* 0x0201.0010 */
	unsigned short		bootrecord;
	unsigned short		_pad_bootrecord;
	unsigned char		_res_5[1024 - 16 - 4];

	/* 0x0201.0400 */
	unsigned short		fsimage;
	unsigned short		_pad_fsimage;
} flash_layout_t;
#endif	/* _ASMLANGUAGE */


/*
** ----------------------------------------------------------------------------
** --- INTERRUPTS -------------------------------------------------------------
** ----------------------------------------------------------------------------
*/



/*
** ----------------------------------------------------------------------------
** --- LINUX STUFF ------------------------------------------------------------
** ----------------------------------------------------------------------------
*/

#define LINUX_PAGESIZE		4096
#define LINUX_PARAMS_ADDR	(SDRAM_BASE_ADDR + 0x7000)



/*
** ----------------------------------------------------------------------------
** --- ETHERNET STUFF ---------------------------------------------------------
** ----------------------------------------------------------------------------
*/

#if defined(PROJECT_T484)
#define MAC0_INITSTRING		\
	"0:0x180000:0x180000:0x1:AUTO:13:5:2:256:00094f000001:0"
#define MAC1_INITSTRING		\
	"1:0x181000:0x180000:0x0:AUTO:14:5:2:256:00094f000002:0"
#define MAC2_INITSTRING		"NOT USED"
#define MAC3_INITSTRING		"NOT USED"

#else	/* Eumex820 */
#define ETH_MAC0_INITIALIZATION	\
{ "eth0",0,0,"0x180000:0x180000:13:5:32:256:00094f000001:0:6:AUTOSW:16:4" }, \
{ "eth1",1,0,"0x180000:0x180000:13:5:32:256:00094f000002:0:6:AUTOSW:16:4" },
#endif

#endif	/* _INC_modulardsl_bsp_h_ */
