/*
 * Copyright (C) <2002-2008> <Funkwerk Enterprise Communications GmbH> 
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * display.h - eumex 820 lan led interface
 * ------------------------------------
 *
 */

#ifndef _INC_modulardsl_display_h_
#define _INC_modulardsl_display_h_

/* System */
#define DISP_SYS_NOTREADY		0
#define DISP_SYS_READY			1
#define DISP_SYS_READY_CHARGES_FULL	2
#define DISP_SYS_READY_CHARGES_OK	3
#define DISP_SYS_READY_MESSAGE_WAITING	4
#define DISP_SYS_READY_MESSAGE_OFF	5
#define DISP_SYS_ALARM			6
#define DISP_SYS_ALARM_OFF		7
#define DISP_SYS_LED_ON			8
#define DISP_SYS_LED_OFF		9

/* Isdn */
#define DISP_ISDN_BUSNO_EXT		0
#define DISP_ISDN_BUSNO_INT		1
#define DISP_ISDN_D			0
#define DISP_ISDN_B1			1
#define DISP_ISDN_B2			2
#define DISP_ISDN_B3			4

#define DISP_ISDN_DEACTIVATED		0
#define DISP_ISDN_ACTIVATED		1
#define DISP_ISDN_ISP_DOWN		0
#define DISP_ISDN_ISP_UP		1

/* Ethernet */
#define DISP_ENET_BUSNO_LAN1		0		/* eth0 */
#define DISP_ENET_BUSNO_LAN2		0		/* eth0 */
#define DISP_ENET_BUSNO_DSL		1		/* eth1 */
#define DISP_ENET_BUSNO_HOMEPNA		-1
#define DISP_ENET_NOLINK		0
#define DISP_ENET_10BTHD		1
#define DISP_ENET_10BTFD		2
#define DISP_ENET_100BTHD		3
#define DISP_ENET_100BTFD		4

/* Dsl */
#define DISP_DSL_BUSNO			DISP_ENET_BUSNO_DSL	/* 1 */
#define DISP_DSL_NOLINK			DISP_ENET_NOLINK	/* 0 */
#define DISP_DSL_LINK			DISP_ENET_10BTHD	/* 1 */
#define DISP_DSL_HL_DOWN		5			/* > 100BTFD */
#define DISP_DSL_HL_UP			6

/* Usb */
#define DISP_USB_BUSNO_0		0
#define DISP_USB_NOLINK			0
#define DISP_USB_ESTABLISHED		1

/* Bluetooth */
#define DISP_BT_BUSNO			DISP_ENET_BUSNO_HOMEPNA	/* 3 */
#define DISP_BT_NOLINK			DISP_ENET_NOLINK	/* 0 */
#define DISP_BT_LINK			DISP_ENET_100BTFD	/* 4 green */
#define DISP_BT_HL_DOWN			DISP_ENET_100BTFD	/* 4 green */
#define DISP_BT_HL_UP			DISP_ENET_10BTHD	/* 1 yellow */

/* Compact Flash Card */
#define DISP_CF_NOTREADY		0
#define DISP_CF_READY			1

/*
** ----------------------------------------------------------------------------
** --- KERNEL -----------------------------------------------------------------
** ----------------------------------------------------------------------------
*/
#ifdef __KERNEL__

#define DISPLAY_SYSTEM_STATE(state)				\
	(*_func_disp_cb_2)( DISP_ID_system,state )

#define DISPLAY_ISDN_STATE_L1(bus,state)			\
	(*_func_disp_cb_3)( DISP_ID_isdn_state_l1,bus,state )
#define DISPLAY_ISDN_STATE(bus,channel,state)			\
	(*_func_disp_cb_4)( DISP_ID_isdn_state,bus,channel,state )
#define DISPLAY_ISDN_XMIT(bus,channel,size)
#define DISPLAY_ISDN_RECV(bus,channel,size)
#define DISPLAY_ISDN_ISP(bus,channel,state)

#define DISPLAY_DSL_STATE(bus,state)				\
	(*_func_disp_cb_3)( DISP_ID_dsl_state,bus,state )

#define DISPLAY_ENET_STATE(bus,state)				\
	(*_func_disp_cb_3)( DISP_ID_ethernet_state,bus,state )
#define DISPLAY_ENET_RECV(bus,size)
#define DISPLAY_ENET_XMIT(bus,size)
#define DISPLAY_ENET_COLL(bus)

#define DISPLAY_USB_STATE(bus,state)				\
	(*_func_disp_cb_3)( DISP_ID_usb_state,bus,state )
#define DISPLAY_USB_RECV(bus,size)				\
	(*_func_disp_cb_3)( DISP_ID_usb_recv,bus,size )
#define DISPLAY_USB_XMIT(bus,size)				\
	(*_func_disp_cb_3)( DISP_ID_usb_xmit,bus,size )

#define DISPLAY_BT_STATE(state)
#define DISPLAY_BT_RECV(size)
#define DISPLAY_BT_XMIT(size)

#define DISPLAY_CF_STATE(unit,state)				\
	(*_func_disp_cb_3)( DISP_ID_cf_state,unit,state )
#define DISPLAY_CF_READ(unit,size)				\
	(*_func_disp_cb_3)( DISP_ID_cf_read,unit,size )
#define DISPLAY_CF_WRITE(unit,size)				\
	(*_func_disp_cb_3)( DISP_ID_cf_write,unit,size )

#define DISPLAY_CAPI_STATE(bus,state)

/*
 * DO NOT USE THE FOLLOWING KERNEL STUFF DIRECTLY !!!
 */

extern void	(*_func_disp_cb_1)( int );
extern void	(*_func_disp_cb_2)( int,int );
extern void	(*_func_disp_cb_3)( int,int,int );
extern void	(*_func_disp_cb_4)( int,int,int,int );

/*
** ----------------------------------------------------------------------------
** --- USERMODE ---------------------------------------------------------------
** ----------------------------------------------------------------------------
*/
#else	/* !__KERNEL__ => eg. must be in usermode */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>

#define DISP_DEV	"/dev/display"


#if defined(INCLUDE_SW_ANALOG_TRUNK_LED)
#define DISPLAY_ANALOG_TRUNK_EXIST() do {				\
	int	fd;							\
	if ((fd = open( DISP_DEV,O_RDWR )) != -1) {			\
		ioctl( fd,DISP_ID_analog_trunk );			\
		close( fd );						\
	}								\
} while (0);
#else
#define DISPLAY_ANALOG_TRUNK_EXIST()
#endif


#define DISPLAY_SYSTEM_STATE(state) do {				\
	int	fd;							\
	if ((fd = open( DISP_DEV,O_RDWR )) != -1) {			\
		ioctl( fd,DISP_ID_system,state );			\
		close( fd );						\
	}								\
} while (0);

#define DISPLAY_ISDN_STATE_L1(bus,state) do {				\
	int	fd;							\
	if ((fd = open( DISP_DEV,O_RDWR )) != -1) {			\
		ioctl( fd,DISP_ID_isdn_state_l1,(state << 8) | bus );	\
		close( fd );						\
	}								\
} while (0);

#define DISPLAY_ISDN_STATE(bus,channel,state) do {			\
	int	fd;							\
	if ((fd = open( DISP_DEV,O_RDWR )) != -1) {			\
	ioctl( fd,DISP_ID_isdn_state,(state << 16) | (channel << 8) | bus );\
		close( fd );						\
	}								\
} while (0);

#define DISPLAY_ISDN_XMIT(bus,channel,size)
#define DISPLAY_ISDN_RECV(bus,channel,size)
#define DISPLAY_ISDN_ISP(bus,channel,state)
#define DISPLAY_BT_STATE(state)

#define DISPLAY_DSL_STATE(bus,state) do {				\
	int	fd;							\
	if ((fd = open( DISP_DEV,O_RDWR )) != -1) {			\
		ioctl( fd,DISP_ID_dsl_state,(state << 8) | bus );	\
		close( fd );						\
	}								\
} while (0);

#define DISPLAY_DIALER_STATE(state)

#endif	/* !__KERNEL */

#endif	/* _INC_modulardsl_display_h_ */
