/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH> 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

/*~-*/
/*~XSF_LANGUAGE: C/C++*/
/*~A*/
/*~+:rcs-header*/
/*~T*/

/*~E*/
/*~I*/
#ifndef P2001_H

/*~T*/
#define P2001_H

/*~T*/
/*
FILE CONTENTS
01 defines
02 direct memory/register access
03 GPIO
04 PCM Highway
05 Watchdog
06 Waitstates
07 PLL 12.288 MHz
08 Chip Revision
09 Interrupt
*/

/*~A*/
/*~+:01 defines*/
/*~T*/
/*********************************************************************
01 defines
*********************************************************************/

#define MASK_UPPER_16   0xffff0000
#define MASK_LOWER_16   0x0000ffff

/*~E*/
/*~A*/
/*~+:02 direct memory/register access*/
/*~T*/
/*********************************************************************
02 direct memory/register access
*********************************************************************/

/*~I*/
// defines Kiel
/* 32 bit access to memory/register */
#ifndef READ_REG

/*~T*/
#define READ_REG(addr)          (* (volatile unsigned int *) (addr))

/*~-*/
#endif
/*~E*/
/*~I*/
#define READ_U32                READ_REG

#ifndef WRITE_REG

/*~T*/
#define WRITE_REG(addr, val)    (* (volatile unsigned int *) (addr) = (volatile unsigned int)(val))

/*~-*/
#endif
/*~E*/
/*~I*/
#define WRITE_U32               WRITE_REG

#ifndef REG_CLR32_BITS

/*~T*/
#define REG_CLR32_BITS(addr,bits) (READ_REG(addr) &= (volatile unsigned int)~(bits))

/*~-*/
#endif
/*~E*/
/*~I*/
#ifndef REG_SET32_BITS

/*~T*/
#define REG_SET32_BITS(addr,bits) (READ_REG(addr) |= (volatile unsigned int)(bits))

/*~-*/
#endif
/*~E*/
/*~T*/
/* 16 bit access to memory/register */
#define READ_U16(addr)          (* (volatile unsigned short int *) (addr))
#define WRITE_U16(addr, val)    (* (volatile unsigned short int *) (addr) = (volatile unsigned short)(val&0xffff))


/* 8 bit access to memory/register */
#define READ_BYTE(addr)         (* (volatile unsigned char *) (addr))
#define WRITE_BYTE(addr, val)   (* (volatile unsigned char *) (addr) = (volatile unsigned char)(val&0xff))

#define READ_U8                 READ_BYTE
#define WRITE_U8                WRITE_BYTE

/*~T*/
// defines (Peine)
#define M_REG_RD_BYTE(_REG_)         (*(volatile unsigned char  *) (_REG_) )
#define M_REG_RD_HWORD(_REG_)        (*(volatile unsigned short *) (_REG_) )
#define M_REG_RD_WORD(_REG_)         (*(volatile unsigned int   *) (_REG_) )

#define M_NOP()  
/*~T*/
#define M_REG_WR_BYTE(_REG_,_VAL_)   ( *(volatile unsigned char  *) (_REG_)  =  (volatile unsigned char)_VAL_ )
#define M_REG_WR_HWORD(_REG_,_VAL_)  ( *(volatile unsigned short *) (_REG_)  =  (volatile unsigned short)_VAL_ )
#define M_REG_WR_WORD(_REG_,_VAL_)   ( *(volatile unsigned int   *) (_REG_)  =  (volatile unsigned int )_VAL_ )

#define M_REG_SET_8BITS(_REG_,_VAL_)	 ( (M_REG_RD_BYTE(_REG_)) |= (volatile unsigned char)_VAL_ )
#define M_REG_CLR_8BITS(_REG_,_VAL_)	 ( (M_REG_RD_BYTE(_REG_)) &= (volatile unsigned char)~(_VAL_))

#define M_REG_SET_16BITS(_REG_,_VAL_)	 ( (M_REG_RD_HWORD(_REG_)) |= (volatile unsigned short) _VAL_ )
#define M_REG_CLR_16BITS(_REG_,_VAL_)	 ( (M_REG_RD_HWORD(_REG_)) &= (volatile unsigned short)~(_VAL_))

#define M_REG_SET_32BITS(_REG_,_VAL_)	 ( (M_REG_RD_WORD(_REG_)) |= (volatile unsigned int)  _VAL_ )
#define M_REG_CLR_32BITS(_REG_,_VAL_)	 ( (M_REG_RD_WORD(_REG_)) &= (volatile unsigned int)~(_VAL_))
/*~E*/
/*~A*/
/*~+:03 PLL 12.288 MHz*/
/*~T*/
/*********************************************************************
07 PLL 12.288 MHz
*********************************************************************/

/*~T*/
// defines for System clock


#define CFG_SYSCLOCK_00		(0*1228800)
#define CFG_SYSCLOCK_01		(1*1228800)
#define CFG_SYSCLOCK_02		(2*CFG_SYSCLOCK_01)
#define CFG_SYSCLOCK_03		(3*CFG_SYSCLOCK_01)
#define CFG_SYSCLOCK_04		(4*CFG_SYSCLOCK_01)
#define CFG_SYSCLOCK_05		(5*CFG_SYSCLOCK_01)
#define CFG_SYSCLOCK_06		(6*CFG_SYSCLOCK_01)
#define CFG_SYSCLOCK_07		(7*CFG_SYSCLOCK_01)
#define CFG_SYSCLOCK_08		(8*CFG_SYSCLOCK_01)
#define CFG_SYSCLOCK_09		(9*CFG_SYSCLOCK_01)


/*~T*/
// Attention HW_SYSCLOCK must fit different clocks (PLL, SIO, So ...)
#define HW_OSCCLOCK                     12288000
#define HW_SYSCLOCK                     ((HW_OSCCLOCK*CFG_SYSCLOCKFACTOR)+CFG_SYSCLOCK_00)


#define DEF_PLL_12288_CONFIG_1          0x04000000  /* = reset value */
#define DEF_DIV_12288_CONFIG_1          0x00000101  /* = reset value, PLL inactive */

#define DEF_PLL_12288_CONFIG_CALC(p)    ( 0x00004000 | (p-2)<<2 )
#define DEF_DIV_12288_CONFIG_CALC(p)    ( 0x00010001 | (p<<8) )

#define DEF_PLL_12288_CONFIG_INIT       0x00004110

/*~I*/
#if   CFG_SYSCLOCKFACTOR == 1   /* 12.288 MHz */                              

/*~T*/
#define DEF_PLL_12288_CONFIG  DEF_PLL_12288_CONFIG_1               
#define DEF_DIV_12288_CONFIG  DEF_DIV_12288_CONFIG_1 

/*~O*/
/*~-*/
#else
/*~I*/
#if   CFG_SYSCLOCKFACTOR <= 8   /* 24.576 MHz - 98.304 Mhz  */

/*~T*/
#define DEF_PLL_12288_CONFIG  DEF_PLL_12288_CONFIG_CALC(CFG_SYSCLOCKFACTOR)              
#define DEF_DIV_12288_CONFIG  DEF_DIV_12288_CONFIG_CALC(CFG_SYSCLOCKFACTOR) 


/*~O*/
/*~-*/
#else
/*~T*/

#error Clockfactor not supported ( PLL ), check CFG_SYSCLOCKFACTOR               

/*~-*/
#endif
/*~E*/
/*~-*/
#endif
/*~E*/
/*~E*/
/*~A*/
/*~+:04 Watchdog*/
/*~T*/
/*********************************************************************
05 Watchdog
*********************************************************************/

/*~T*/
#define B_WATCHDOG_RESET        0x80000000      /* Bit 31: Watchdog Reset */

#define HW_WATCHDOG_RESET()  REG_SET32_BITS( ABS_Adr_Timer12_PreDiv_24Bit, B_WATCHDOG_RESET) /* trigger watchdog */

/*~T*/
// watchdog
#define CFG_WDG_PRESCAL   0xBB8
#define CFG_WDG_TIME      2

/*~E*/
/*~A*/
/*~+:05 GPIO*/
/*~T*/
/*********************************************************************
03 GPIO
*********************************************************************/

/*~A*/
/*~+:GPIO*/
/*~K*/
/*~+:I/O*/
/*~+:*/
/*~T*/
/* config input/output */
#define HW_GPIO_CONFIG_INOUT(p) \
        WRITE_REG( ABS_Adr_GPIO_En, (p) & MASK_LOWER_16)  

/* configure specified bit as input */
#define HW_GPIO_CONFIG_IN(p) \
        REG_CLR32_BITS( ABS_Adr_GPIO_En, 1U << (p) )

/* configure specified bit as output */
#define HW_GPIO_CONFIG_OUT(p) \
        REG_SET32_BITS( ABS_Adr_GPIO_En, 1U << (p) )
/*~T*/
/* configure specified bit as multi function */
#define HW_GPIO_CONFIG_MF_ON(p) \
        REG_SET32_BITS( ABS_Adr_GPIO_En, 1U << ((p)+16) )

/* configure specified bit as GPIO */
#define HW_GPIO_CONFIG_MF_OFF(p) \
        REG_CLR32_BITS( ABS_Adr_GPIO_En, 1U << ((p)+16) )

/*~T*/
/* set output */
#define HW_GPIO_OUT(p) \
        WRITE_REG( ABS_Adr_GPIO_Out, MASK_UPPER_16 | ((p)&MASK_LOWER_16) )   

/* set specified bit */
#define HW_GPIO_SET(p) \
        WRITE_REG( ABS_Adr_GPIO_Out, 1U << ((p)+16) | 0xffff)   

/* reset specified bit */
#define HW_GPIO_RESET(p) \
        WRITE_REG( ABS_Adr_GPIO_Out, 1U << ((p)+16) | 0x0000)  
/*~T*/
/* read input */
#define HW_GPIO_IN() \
        ( READ_REG( ABS_Adr_GPIO_IN) & MASK_LOWER_16 )

/*~K*/
/*~+:Interrupt*/
/*~+:*/
/*~T*/
/* set interrupt mask */
#define HW_GPIO_IRQ_SET(p) \
          WRITE_REG( ABS_Adr_GPIO_Config, \
          ( READ_REG(ABS_Adr_GPIO_Config) & MASK_UPPER_16 ) |  ((p) & MASK_LOWER_16) ) 

/* enable interrupt for specified pin */
#define HW_GPIO_IRQ_ENABLE(p) \
          WRITE_REG( ABS_Adr_GPIO_Config, \
          ( READ_REG( ABS_Adr_GPIO_Config) |  (1U << (p)) ) & MASK_LOWER_16)  

/* disable interrupt for specified pin */
#define HW_GPIO_IRQ_DISABLE(p) \
          WRITE_REG( ABS_Adr_GPIO_Config, \
          ( READ_REG( ABS_Adr_GPIO_Config) & ~(1U << p) ) & MASK_LOWER_16)  

/*~T*/
/* configure trigger */
#define HW_GPIO_IRQ_CONFIG(p) \
          WRITE_REG( ABS_Adr_GPIO_INT, \
          ( READ_REG(ABS_Adr_GPIO_INT) & MASK_LOWER_16 ) | (((p)&MASK_LOWER_16)<<16) ) 

/* configure trigger (rising edge / high level ) for specified pin */
#define HW_GPIO_IRQ_CONFIG_RISING(p) \
          WRITE_REG( ABS_Adr_GPIO_INT, \
          READ_REG( ABS_Adr_GPIO_INT) |  ( 1U << ((p)+16)) ) 

/* configure trigger (falling edge / low level ) for specified pin */
#define HW_GPIO_IRQ_CONFIG_FALLING(p) \
          WRITE_REG( ABS_Adr_GPIO_INT, \
          READ_REG( ABS_Adr_GPIO_INT) & ~( 1U << ((p)+16)) ) 

/*~T*/
/* select interrupt type: FIQ */
#define HW_GPIO_IRQ_TYPE_FIQ(p) \
          REG_SET32_BITS( ABS_Adr_GPIO_INT_SEL, 1U << ((p)+16) ) 

/* select interrupt type: IRQ */
#define HW_GPIO_IRQ_TYPE_IRQ(p) \
          REG_CLR32_BITS( ABS_Adr_GPIO_INT_SEL, 1U << ((p)+16) ) 

/*~T*/
/* select interrupt sense: level triggered */
#define HW_GPIO_IRQ_SENSE_LEVEL(p) \
          REG_SET32_BITS( ABS_Adr_GPIO_INT_SEL, 1U << (p) ) 

/* select interrupt sense: edge triggered */
#define HW_GPIO_IRQ_SENSE_EDGE(p) \
          REG_CLR32_BITS( ABS_Adr_GPIO_INT_SEL, 1U << (p) ) 

/*~T*/
/* confirm Interrupt for specified pin */
#define HW_GPIO_IRQ_CONFIRM(p) \
          WRITE_REG( ABS_Adr_GPIO_INT, \
          READ_REG( ABS_Adr_GPIO_INT) & ~( 1U << (p) ) ) 
/*~E*/
/*~A*/
/*~+:GPIO2*/
/*~K*/
/*~+:I/O*/
/*~+:*/
/*~T*/
/* config input/output */
#define HW_GPIO2_CONFIG_INOUT(p) \
        WRITE_REG( ABS_Adr_GPIO2_En, (p) & MASK_LOWER_16)  

/* configure specified bit as input */
#define HW_GPIO2_CONFIG_IN(p) \
        REG_CLR32_BITS( ABS_Adr_GPIO2_En, 1U << (p) )

/* configure specified bit as output */
#define HW_GPIO2_CONFIG_OUT(p) \
        REG_SET32_BITS( ABS_Adr_GPIO2_En, 1U << (p) )

/*~T*/
/* configure specified bit as multi function */
#define HW_GPIO2_CONFIG_MF_ON(p)  /* empty */

/* configure specified bit as GPIO */
#define HW_GPIO2_CONFIG_MF_OFF(p) /* empty */
/*~T*/
/* set output */
#define HW_GPIO2_OUT(p) \
        WRITE_REG( ABS_Adr_GPIO2_Out, MASK_UPPER_16 | ((p)&MASK_LOWER_16) )   

/* set specified bit */
#define HW_GPIO2_SET(p) \
        WRITE_REG( ABS_Adr_GPIO2_Out, 1U << (p+16) | 0xffff)   

/* reset specified bit */
#define HW_GPIO2_RESET(p) \
        WRITE_REG( ABS_Adr_GPIO2_Out, 1U << (p+16) | 0x0000)  
/*~T*/
/* read input */
#define HW_GPIO2_IN() \
        ( READ_REG( ABS_Adr_GPIO2_IN) & MASK_LOWER_16 )

/*~T*/

/*~E*/
/*~T*/

/*~E*/
/*~A*/
/*~+:06 PCM Highway*/
/*~T*/
/*********************************************************************
04 PCM Highway
*********************************************************************/

/*~E*/
/*~A*/
/*~+:07 Waitstates*/
/*~T*/
/*********************************************************************
06 Waitstates
*********************************************************************/

/*~T*/
/* Register WaitState_Ext */

/*~I*/
#if   CFG_SYSCLOCKFACTOR == 1   /* 12.288 MHz */                              

/*~T*/
#define B_WS_FLASH1  3  //2  /* 0 - 15 cycles */
#define B_WS_FLASH2           3       /* 0 - 15 cycles */
#define B_WS_FLASH3           3       /* 0 - 15 cycles */
#define B_WS_FLASH4           3       /* 0 - 15 cycles */

/*~O*/
/*~-*/
#else
/*~I*/
#if   CFG_SYSCLOCKFACTOR <= 4   /* up to 49MHz */

/*~T*/
#define B_WS_FLASH1  6       /* 90ns Flash -> 4 */
#define B_WS_FLASH2  6
#define B_WS_FLASH3  6
#define B_WS_FLASH4  6

/*~O*/
/*~-*/
#else
/*~T*/
#define B_WS_FLASH1 10
      /* 120ns */
#define B_WS_FLASH2 15
      /* 120ns */
#define B_WS_FLASH3 15
      /* 120ns */
#define B_WS_FLASH4 15
      /* 120ns */

/*~-*/
#endif
/*~E*/
/*~-*/
#endif
/*~E*/
/*~T*/
#define B_WS_FLASH_PAUSE        3UL
#define B_WS_PRE_PAUSE          1

#define DEF_WAITSTATE_EXT       (( B_WS_FLASH_PAUSE    & 0x03) << 30) | \
                                (( B_WS_PRE_PAUSE      & 0x01) << 29) | \
                                (( B_WS_FLASH4         & 0x0f) << 12) | \
                                (( B_WS_FLASH3         & 0x0f) <<  8) | \
                                (( B_WS_FLASH2         & 0x0f) <<  4) | \
                                (( B_WS_FLASH1         & 0x0f) <<  0)

/*~T*/
/* Register WaitState_Asic */

/*~I*/
#if   CFG_SYSCLOCKFACTOR == 1   /* 12.288 MHz */                              

/*~T*/
#define B_WS_S0_HDLC          3       /* 0 - 15 cycles */
#define B_WS_UART             0       /* 0 - 15 cycles */
#define B_WS_USB              2       /* 0 - 15 cycles */


/*~O*/
/*~-*/
#else
/*~I*/
#if   CFG_SYSCLOCKFACTOR <= 4   /* up to 49MHz */

/*~T*/
#define B_WS_S0_HDLC          15      /* max */
#define B_WS_UART             15      /* max */
#define B_WS_USB              7      /* max */

/*~O*/
/*~-*/
#else
/*~T*/
#define B_WS_S0_HDLC          15      /* max */
#define B_WS_UART             15      /* max */
#define B_WS_USB              7      /* max */

/*~-*/
#endif
/*~E*/
/*~-*/
#endif
/*~E*/
/*~T*/

#define DEF_WAITSTATE_ASIC      (( B_WS_USB     & 0x0f) <<  8) | \
                                (( B_WS_UART    & 0x0f) <<  4) | \
                                (( B_WS_S0_HDLC & 0x0f) <<  0)



/*~T*/
/* Register ExtMem_Ctrl */

/*~I*/


#ifdef CFG_FLASHDEVICE_8BIT

/*~T*/
#define B_WIDTH_FLASH1        0       /* 0: width 8bit, 1: 16bit, 2: 32bit */

/*~O*/
/*~-*/
#else
/*~I*/
#if   defined CFG_FLASHDEVICE_16BIT

/*~T*/
#define B_WIDTH_FLASH1        1

/*~O*/
/*~-*/
#else
/*~I*/
#if  defined CFG_FLASHDEVICE_32BIT

/*~T*/
#define B_WIDTH_FLASH1        2

/*~O*/
/*~-*/
#else
/*~T*/
#error "no buswitdh for flash devices defined"
/*~-*/
#endif
/*~E*/
/*~-*/
#endif
/*~E*/
/*~-*/
#endif
/*~E*/
/*~T*/
#define B_WIDTH_FLASH2          0
#define B_WIDTH_FLASH3          0  //0   /* ext. SRAM on P2001-EvaBoard */

/*~I*/
#ifdef G_FAC_BIT_BY_MPS
/*~T*/
#define B_WIDTH_FLASH4          2  // HW Kennung


/*~O*/
/*~-*/
#else
#ifdef PROJECT_FAMILY_MODULARDSL
/* Compact Flash 16 bittig */
#define B_WIDTH_FLASH4          1
#else
#define B_WIDTH_FLASH4          0 
#endif
#endif
/*~E*/
/*~T*/
#define DEF_EXTMEM_CTRL         (( B_WIDTH_FLASH4  & 0x03) << 6) | \
                                (( B_WIDTH_FLASH3  & 0x03) << 4) | \
                                (( B_WIDTH_FLASH2  & 0x03) << 2) | \
                                (( B_WIDTH_FLASH1  & 0x03) << 0)



/*~E*/
/*~A*/
/*~+:08 Chip Revision*/
/*~T*/
/*********************************************************************
08 Chip Revision
*********************************************************************/

/*~E*/
/*~A*/
/*~+:09 Interrupt*/
/*~T*/
/*********************************************************************
09 Interrupt
*********************************************************************/

/*~K*/
/*~+:Interrupt Sources*/
/*~+:*/
/*~I*/
#ifndef LINUX_MPS
/*~T*/
#define IRQ_USB         0
#define IRQ_UART        1
#define IRQ_PWM         2
#define IRQ_FSC         3
#define IRQ_GPIO        4
#define IRQ_PFS0        5
#define IRQ_PFS1        6
#define IRQ_PFS2        7
#define IRQ_PFS3        8
#define IRQ_S01         9
#define IRQ_S02         10
#define IRQ_DTMF        11
#define IRQ_WATCHDOG    12
#define IRQ_PCMHW       13
#define IRQ_HDLC_MTS    14
#define IRQ_HDLC_ERR    15
#define IRQ_HDLC_DATA   16
#define IRQ_EU0_DATA    17
#define IRQ_EU0_ERR     18
#define IRQ_EU1_DATA    19
#define IRQ_EU1_ERR     20
#define IRQ_EU2_DATA    21
#define IRQ_EU2_ERR     22
#define IRQ_EU3_DATA    23
#define IRQ_EU3_ERR     24
#define IRQ_TIMER1      25
#define IRQ_TIMER2      26

/*~-*/
#endif
/*~E*/
/*~T*/

#define IRQ_TIMER       IRQ_TIMER1    /* defined for compatibility reasons */

/*~T*/
#define MAX_IRQ_SOURCE  (IRQ_TIMER2+1)
/*~T*/

/*~E*/
/*~T*/
/* ...P2001_H */

/*~-*/
#endif
/*~E*/
