/*
 * Copyright (C) <2002-2008> <Funkwerk Enterprise Communications GmbH> 
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * p2001.h - p2001 processor header
 * --------------------------------
 *
 * Initial Version	Mon Feb  4 17:41:16 CET 2002 - jse
 * Written
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */


#ifndef _INC_p2001_h_
#define _INC_p2001_h_

/*
** ----------------------------------------------------------------------------
** --- MAZ asic headers -------------------------------------------------------
** ----------------------------------------------------------------------------
*/

#include "p2001/asic1.h"
#include "p2001/asic1_ab.h"



/*
** ----------------------------------------------------------------------------
** --- P2001 ADDRESS MAP ------------------------------------------------------
** ----------------------------------------------------------------------------
**
** ------------------------------ 0x00000000
** |                            |
** | internal SRAM 32KB         |
** |                            |
** |----------------------------|
**              ...
** |----------------------------| 0x00040000
** |                            |
** | internal ROM               |
** |                            |
** |----------------------------|
**              ...
** |----------------------------| 0x00100000
** |                            |
** | internal ASIC registers    |
** |                            |
** |----------------------------|
**              ...
** |----------------------------| 0x02000000
** |                            |
** | external CS_1              |
** |                            |
** |----------------------------| 0x04000000
** |                            |
** | external CS_2              |
** |                            |
** |----------------------------| 0x06000000
** |                            |
** | external CS_3              |
** |                            |
** |----------------------------| 0x08000000
** |                            |
** | external CS_4              |
** |                            |
** |----------------------------|
**              ...
** |----------------------------| 0x40000000
** |                            |
** | external CS_SDRAM          |
** |                            |
** |----------------------------|
**
*/

/* internal SRAM */
#define SRAM_BASE_ADDR		ABS_Adr_RAM_BASE
#define SRAM_SIZE		0x00008000		/* 32 KB */

/* internal ROM */
#define ROM_BASE_ADDR		ABS_Adr_ROM_BASE
#define ROM_SIZE		0x00002000		/* 8KB XXX */

/* internal asic features */
#define SYS_BASE_ADDR		ABS_Adr_SYS_BASE
#define TIMER_BASE_ADDR		ABS_Adr_TIMER_BASE
#define GPIO_BASE_ADDR		ABS_Adr_GPIO_BASE
#define INTCNTL_BASE_ADDR	ABS_Adr_INT_CTRL_BASE
#define UART_BASE_ADDR		ABS_Adr_UART_BASE
#define PCMHW_BASE_ADDR		ABS_Adr_PCM_HW_BASE
#define DTMF_BASE_ADDR		ABS_Adr_DTMF_BASE
#define S0_BASE_ADDR		ABS_Adr_S0_BASE
#define INBUS_BASE_ADDR		ABS_Adr_INBUS_BASE
#define EU0_BASE_ADDR		ABS_Adr_EU0_BASE
#define EU1_BASE_ADDR		ABS_Adr_EU1_BASE
#define EU2_BASE_ADDR		ABS_Adr_EU2_BASE
#define EU3_BASE_ADDR		ABS_Adr_EU3_BASE
#define HDLC_BASE_ADDR		ABS_Adr_HDLC_BASE
#define USB_BASE_ADDR		ABS_Adr_USB_REG_BASE

/* external chipselects */
#define CS_1_BASE_ADDR		ABS_Adr_Flash1_BASE
#define CS_2_BASE_ADDR		ABS_Adr_Flash2_BASE
#define CS_3_BASE_ADDR		ABS_Adr_Flash3_BASE
#define CS_4_BASE_ADDR		ABS_Adr_Flash4_BASE
#define CS_SDRAM_BASE_ADDR	ABS_Adr_SDRAM_BASE



/*
** ----------------------------------------------------------------------------
** --- SYSTEM -----------------------------------------------------------------
** ----------------------------------------------------------------------------
*/



/*
** ----------------------------------------------------------------------------
** --- TIMER ------------------------------------------------------------------
** ----------------------------------------------------------------------------
*/

#if !defined(_ASMLANGUAGE) && !defined(__ASSEMBLER__)
typedef struct {
	volatile unsigned long	timer1;
	volatile unsigned long	timer2;
	volatile unsigned long	timer_preload;
	volatile unsigned long	timer12_prediv;
	volatile unsigned long	timer_int;
	volatile unsigned long	freerun_timer;
	volatile unsigned long	watchdog_timer;

	unsigned char		reserved_0[4];

	volatile unsigned long	pwm_cnt;
	volatile unsigned long	pwm_cnt2;

	unsigned char		reserved_1[8];

	volatile unsigned long	pll_12000_config;
	volatile unsigned long	pll_12288_config;
	volatile unsigned long	div_12288_config;
	volatile unsigned long	mod_cnt_768;
	volatile unsigned long	fsc_irq_status;
	volatile unsigned long	fsc_config;
	volatile unsigned long	fsc_construct;
	volatile unsigned long	fsc_base_clk_reg;
	volatile unsigned long	sysclk_shape;
	volatile unsigned long	sdramclk_shape;
	volatile unsigned long	ring_oszi;
} p2001_tm_t;
#endif	/* _ASMLANGUAGE */

/* TIMER1 */
#define TIMER_TIMER1_MASK		0x0000ffff
#define TIMER_TIMER1(x)			((x) & TIMER_TIMER1_MASK)

/* TIMER2 */
#define TIMER_TIMER2_MASK		0x0000ffff
#define TIMER_TIMER2(x)			((x) & TIMER_TIMER2_MASK)

/* TIMER_PRELOAD */
#define TIMER_PRELOAD_1_SHIFT		0
#define TIMER_PRELOAD_1_MASK		(0xFFFF << TIMER_PRELOAD_1_SHIFT)
#define TIMER_PRELOAD_2_SHIFT		16
#define TIMER_PRELOAD_2_MASK		(0xFFFF << TIMER_PRELOAD_2_SHIFT)

/* TIMER12_PREDIV */
#define TIMER_PREDIV_PREDIV_1_SHIFT	0
#define TIMER_PREDIV_PREDIV_1_MASK	(0xFF << TIMER_PREDIV_PREDIV_1_SHIFT)
#define TIMER_PREDIV_PREDIV_2_SHIFT	8
#define TIMER_PREDIV_PREDIV_2_MASK	(0xFF << TIMER_PREDIV_PREDIV_2_SHIFT)
#define TIMER_PREDIV_PREDIV_WD_SHIFT	16
#define TIMER_PREDIV_PREDIV_WD_MASK	(0xFFF << TIMER_PREDIV_PREDIV_WD_SHIFT)
#define TIMER_PREDIV_DISABLE_1		(1 << 28)
#define TIMER_PREDIV_DISABLE_2		(1 << 29)
#define TIMER_PREDIV_DISABLE_WD		(1 << 30)
#define TIMER_PREDIV_RESET_WD		(1 << 31)

/* TIMER_INT */
#define TIMER_INT_TIMER1_INT		(1 << 0)
#define TIMER_INT_TIMER2_INT		(1 << 1)
#define TIMER_INT_WD_INT		(1 << 2)
#define TIMER_INT_PWM_INT		(1 << 3)
#define TIMER_INT_TIMER1_EN		(1 << 4)
#define TIMER_INT_TIMER2_EN		(1 << 5)
#define TIMER_INT_WD_EN			(1 << 6)
#define TIMER_INT_PWM_EN		(1 << 7)
#define TIMER_INT_WD_LEVEL_SHIFT	8
#define TIMER_INT_WD_LEVEL_MASK		(0xFFFF << TIMER_INT_WD_LEVEL_SHIFT)



/*
** ----------------------------------------------------------------------------
** --- GPIO -------------------------------------------------------------------
** ----------------------------------------------------------------------------
*/



/*
** ----------------------------------------------------------------------------
** --- INTERRUPT CONTROL ------------------------------------------------------
** ----------------------------------------------------------------------------
*/

#if !defined(_ASMLANGUAGE) && !defined(__ASSEMBLER__)
typedef struct {
	volatile unsigned long	fiq_cntl;
	volatile unsigned long	irq_cntl;
	volatile unsigned long	fiq_stat;
	volatile unsigned long	irq_stat;
} p2001_int_t;
#endif	/* _ASMLANGUAGE */

/* MAIN_NFIQ_INT_CRTL */
#define INT_FIQ_CNTL_USB		(1 <<  0)
#define INT_FIQ_CNTL_UART		(1 <<  1)
#define INT_FIQ_CNTL_PWM		(1 <<  2)
#define INT_FIQ_CNTL_FSC		(1 <<  3)
#define INT_FIQ_CNTL_GPIO		(1 <<  4)
#define INT_FIQ_CNTL_PFS0		(1 <<  5)
#define INT_FIQ_CNTL_PFS1		(1 <<  6)
#define INT_FIQ_CNTL_PFS2		(1 <<  7)
#define INT_FIQ_CNTL_PFS3		(1 <<  8)
#define INT_FIQ_CNTL_S0_0		(1 <<  9)
#define INT_FIQ_CNTL_S0_1		(1 << 10)
#define INT_FIQ_CNTL_DTMF		(1 << 11)
#define INT_FIQ_CNTL_WATCHDOG		(1 << 12)
#define INT_FIQ_CNTL_PCMHW		(1 << 13)
#define INT_FIQ_CNTL_HDLC_MTS		(1 << 14)
#define INT_FIQ_CNTL_HDLC_ERR		(1 << 15)
#define INT_FIQ_CNTL_HDLC_DATA		(1 << 16)
#define INT_FIQ_CNTL_EU0_DATA		(1 << 17)
#define INT_FIQ_CNTL_EU0_ERR		(1 << 18)
#define INT_FIQ_CNTL_EU1_DATA		(1 << 19)
#define INT_FIQ_CNTL_EU1_ERR		(1 << 20)
#define INT_FIQ_CNTL_EU2_DATA		(1 << 21)
#define INT_FIQ_CNTL_EU2_ERR		(1 << 22)
#define INT_FIQ_CNTL_EU3_DATA		(1 << 23)
#define INT_FIQ_CNTL_EU3_ERR		(1 << 24)
#define INT_FIQ_CNTL_TIMER1		(1 << 25)
#define INT_FIQ_CNTL_TIMER2		(1 << 26)

/* MAIN_NIRQ_INT_CRTL */
#define INT_IRQ_CNTL_USB		(1 <<  0)
#define INT_IRQ_CNTL_UART		(1 <<  1)
#define INT_IRQ_CNTL_PWM		(1 <<  2)
#define INT_IRQ_CNTL_FSC		(1 <<  3)
#define INT_IRQ_CNTL_GPIO		(1 <<  4)
#define INT_IRQ_CNTL_PFS0		(1 <<  5)
#define INT_IRQ_CNTL_PFS1		(1 <<  6)
#define INT_IRQ_CNTL_PFS2		(1 <<  7)
#define INT_IRQ_CNTL_PFS3		(1 <<  8)
#define INT_IRQ_CNTL_S0_0		(1 <<  9)
#define INT_IRQ_CNTL_S0_1		(1 << 10)
#define INT_IRQ_CNTL_DTMF		(1 << 11)
#define INT_IRQ_CNTL_WATCHDOG		(1 << 12)
#define INT_IRQ_CNTL_PCMHW		(1 << 13)
#define INT_IRQ_CNTL_HDLC_MTS		(1 << 14)
#define INT_IRQ_CNTL_HDLC_ERR		(1 << 15)
#define INT_IRQ_CNTL_HDLC_DATA		(1 << 16)
#define INT_IRQ_CNTL_EU0_DATA		(1 << 17)
#define INT_IRQ_CNTL_EU0_ERR		(1 << 18)
#define INT_IRQ_CNTL_EU1_DATA		(1 << 19)
#define INT_IRQ_CNTL_EU1_ERR		(1 << 20)
#define INT_IRQ_CNTL_EU2_DATA		(1 << 21)
#define INT_IRQ_CNTL_EU2_ERR		(1 << 22)
#define INT_IRQ_CNTL_EU3_DATA		(1 << 23)
#define INT_IRQ_CNTL_EU3_ERR		(1 << 24)
#define INT_IRQ_CNTL_TIMER1		(1 << 25)
#define INT_IRQ_CNTL_TIMER2		(1 << 26)

/* STATUS_NFIQ */
#define INT_FIQ_STAT_USB		(1 <<  0)
#define INT_FIQ_STAT_UART		(1 <<  1)
#define INT_FIQ_STAT_PWM		(1 <<  2)
#define INT_FIQ_STAT_FSC		(1 <<  3)
#define INT_FIQ_STAT_GPIO		(1 <<  4)
#define INT_FIQ_STAT_PFS0		(1 <<  5)
#define INT_FIQ_STAT_PFS1		(1 <<  6)
#define INT_FIQ_STAT_PFS2		(1 <<  7)
#define INT_FIQ_STAT_PFS3		(1 <<  8)
#define INT_FIQ_STAT_S0_0		(1 <<  9)
#define INT_FIQ_STAT_S0_1		(1 << 10)
#define INT_FIQ_STAT_DTMF		(1 << 11)
#define INT_FIQ_STAT_WATCHDOG		(1 << 12)
#define INT_FIQ_STAT_PCMHW		(1 << 13)
#define INT_FIQ_STAT_HDLC_MTS		(1 << 14)
#define INT_FIQ_STAT_HDLC_ERR		(1 << 15)
#define INT_FIQ_STAT_HDLC_DATA		(1 << 16)
#define INT_FIQ_STAT_EU0_DATA		(1 << 17)
#define INT_FIQ_STAT_EU0_ERR		(1 << 18)
#define INT_FIQ_STAT_EU1_DATA		(1 << 19)
#define INT_FIQ_STAT_EU1_ERR		(1 << 20)
#define INT_FIQ_STAT_EU2_DATA		(1 << 21)
#define INT_FIQ_STAT_EU2_ERR		(1 << 22)
#define INT_FIQ_STAT_EU3_DATA		(1 << 23)
#define INT_FIQ_STAT_EU3_ERR		(1 << 24)
#define INT_FIQ_STAT_TIMER1		(1 << 25)
#define INT_FIQ_STAT_TIMER2		(1 << 26)

/* STATUS_NIRQ */
#define INT_IRQ_STAT_USB		(1 <<  0)
#define INT_IRQ_STAT_UART		(1 <<  1)
#define INT_IRQ_STAT_PWM		(1 <<  2)
#define INT_IRQ_STAT_FSC		(1 <<  3)
#define INT_IRQ_STAT_GPIO		(1 <<  4)
#define INT_IRQ_STAT_PFS0		(1 <<  5)
#define INT_IRQ_STAT_PFS1		(1 <<  6)
#define INT_IRQ_STAT_PFS2		(1 <<  7)
#define INT_IRQ_STAT_PFS3		(1 <<  8)
#define INT_IRQ_STAT_S0_0		(1 <<  9)
#define INT_IRQ_STAT_S0_1		(1 << 10)
#define INT_IRQ_STAT_DTMF		(1 << 11)
#define INT_IRQ_STAT_WATCHDOG		(1 << 12)
#define INT_IRQ_STAT_PCMHW		(1 << 13)
#define INT_IRQ_STAT_HDLC_MTS		(1 << 14)
#define INT_IRQ_STAT_HDLC_ERR		(1 << 15)
#define INT_IRQ_STAT_HDLC_DATA		(1 << 16)
#define INT_IRQ_STAT_EU0_DATA		(1 << 17)
#define INT_IRQ_STAT_EU0_ERR		(1 << 18)
#define INT_IRQ_STAT_EU1_DATA		(1 << 19)
#define INT_IRQ_STAT_EU1_ERR		(1 << 20)
#define INT_IRQ_STAT_EU2_DATA		(1 << 21)
#define INT_IRQ_STAT_EU2_ERR		(1 << 22)
#define INT_IRQ_STAT_EU3_DATA		(1 << 23)
#define INT_IRQ_STAT_EU3_ERR		(1 << 24)
#define INT_IRQ_STAT_TIMER1		(1 << 25)
#define INT_IRQ_STAT_TIMER2		(1 << 26)



/*
** ----------------------------------------------------------------------------
** --- UART -------------------------------------------------------------------
** ----------------------------------------------------------------------------
*/

#if !defined(_ASMLANGUAGE) && !defined(__ASSEMBLER__)
struct uart_rd_regs {
	volatile unsigned long	rx1;
	volatile unsigned long	rx2;
	volatile unsigned long	rx3;
	volatile unsigned long	rx4;
	volatile unsigned long	unused[4];
	volatile unsigned long	pre_status;
	volatile unsigned long	status;
	volatile unsigned long	unused2;
	volatile unsigned long	irq_status;
};

struct uart_wr_regs {
	volatile unsigned long	tx1;
	volatile unsigned long	tx2;
	volatile unsigned long	tx3;
	volatile unsigned long	tx4;
	volatile unsigned long	baudrate;
	volatile unsigned long	unused[3];
	volatile unsigned long	config;
	volatile unsigned long	clear;
	volatile unsigned long	echo_en;
	volatile unsigned long	irq_status;
};

union uart_regs {
	volatile struct uart_rd_regs r;
	volatile struct uart_wr_regs w;
};
#endif	/* _ASMLANGUAGE */

/* Baudrate = (system clock / 8) * (M / N)
 * M / N    <=   1/4
 */

#define BAUD_MUL_DIV(M,N)	((N) << 16 | (M))

/* UART config register */
#define UART_CONFIG_STOPBIT_SHIFT	0
#define UART_CONFIG_STOPBIT_MASK	(3 << UART_CONFIG_STOPBIT_SHIFT)
#define UART_CONFIG_PARITY_SHIFT	2
#define UART_CONFIG_PARITY_MASK		(7 << UART_CONFIG_PARITY_SHIFT)
#define UART_CONFIG_WDLEN_SHIFT		5
#define UART_CONFIG_WDLEN_MASK		(0x1F << UART_CONFIG_WDLEN_SHIFT)
#define UART_CONFIG_USE_CTS		(1 << 10)
#define UART_CONFIG_RX_HIWAT_SHIFT	11
#define UART_CONFIG_RX_HIWAT_MASK	(0x3F << UART_CONFIG_RX_HIWAT_SHIFT)
#define UART_CONFIG_TX_LOWAT_SHIFT	17
#define UART_CONFIG_TX_LOWAT_MASK	(0x3F << UART_CONFIG_TX_LOWAT_SHIFT)
#define UART_CONFIG_TX_INV		(1 << 23)
#define UART_CONFIG_RX_INV		(1 << 24)

/* UART RX and pre status register */
#define UART_RX_SHIFT			0
#define UART_RX_MASK			(0x1FF << UART_RX_SHIFT)
#define UART_TX_DELTA_SHIFT		16
#define UART_TX_DELTA_MASK		(0x3F << UART_TX_DELTA_SHIFT)
#define UART_RX_DELTA_SHIFT		22
#define UART_RX_DELTA_MASK		(0x3F << UART_RX_DELTA_SHIFT)
#define UART_RX_PARITY_ERR		(1 << 28)

/* UART Status register */
#define UART_STAT_TX_DELTA_SHIFT	0
#define UART_STAT_TX_DELTA_MASK		(0x3F << UART_STAT_TX_DELTA_SHIFT)
#define UART_STAT_RX_DELTA_SHIFT	6
#define UART_STAT_RX_DELTA_MASK		(0x3F << UART_STAT_RX_DELTA_SHIFT)

/* UART IRQ status register */
#define UART_IRQ_TX_SEND		(1 << 0)
#define UART_IRQ_TX_LOWAT		(1 << 1)
#define UART_IRQ_TX_LAST_DATA		(1 << 2)
#define UART_IRQ_RX_DATA		(1 << 3)
#define UART_IRQ_RX_HIWAT		(1 << 4)
#define UART_IRQ_RX_FIFO_FULL		(1 << 5)
#define UART_IRQ_RX_FIFO_RAR_ERR	(1 << 6)
#define UART_IRQ_RX_BRK			(1 << 7)
#define UART_IRQ_ECHO_ERR		(1 << 8)
#define UART_IRQ_RX_LOST		(1 << 9)
#define UART_IRQ_EN_FIQ_SHIFT		10
#define UART_IRQ_EN_FIQ_MASK		(0x3FF << UART_IRQ_UART_EN_FIQ_SHIFT)
#define UART_IRQ_EN_IRQ_SHIFT		20
#define UART_IRQ_EN_IRQ_MASK		(0x3FF << UART_IRQ_EN_IRQ_SHIFT)

#define UART_IRQ_IRQ_MASK		0x3FF

#define UART_TX_FIFO_SIZE		32
#define UART_RX_FIFO_SIZE		32



/*
** ----------------------------------------------------------------------------
** --- PCM-HW -----------------------------------------------------------------
** ----------------------------------------------------------------------------
*/


/*
** ----------------------------------------------------------------------------
** --- DTMF -------------------------------------------------------------------
** ----------------------------------------------------------------------------
*/

/*
** ----------------------------------------------------------------------------
** --- S0 ---------------------------------------------------------------------
** ----------------------------------------------------------------------------
*/

/*
** ----------------------------------------------------------------------------
** --- INBUS (EU/HDLC) --------------------------------------------------------
** ----------------------------------------------------------------------------
*/

#if !defined(_ASMLANGUAGE) && !defined(__ASSEMBLER__)
typedef struct {
	volatile unsigned long	rmac_dma_desc;
	volatile unsigned long	rmac_dma_cntl;
	volatile unsigned long	rmac_dma_stat;
	volatile unsigned long	rmac_dma_en;
	volatile unsigned long	rmac_cntl;
	volatile unsigned long	rmac_tlen;
	volatile unsigned long	rmac_phyu;
	volatile unsigned long	rmac_phyl;
	volatile unsigned long	rmac_pfm[8];
	volatile unsigned long	rmac_mib0;
	volatile unsigned long	rmac_mib1;
	volatile unsigned long	rmac_mib2;
	volatile unsigned long	rmac_mib3;
	volatile unsigned long	rmac_mib4;
	volatile unsigned long	rmac_mib5;

	unsigned char		reserved_0[1960];

	/* 0x800 */
	volatile unsigned long	tmac_dma_desc;
	volatile unsigned long	tmac_dma_cntl;
	volatile unsigned long	tmac_dma_stat;
	volatile unsigned long	tmac_dma_en;
	volatile unsigned long	tmac_cntl;
	volatile unsigned long	tmac_mib6;
	volatile unsigned long	tmac_mib7;

	unsigned char		reserved_1[16];

	/* 0x82c */
	volatile unsigned long	rmii_cntl;
} p2001_eu_t;


typedef struct {
	unsigned char		reserved_0[0x820];

	/* 0x820 */
	volatile unsigned long	mu_cntl;
	volatile unsigned long	mu_data;
	volatile unsigned long	mu_div;
} p2001_mii_t;
#endif	/* _ASMLANGUAGE */

/* RMAC_DMA_DESC0 */
#define EU_RX_DMA_DESC0_RXLEN_MSK	0x0000ffff
#define EU_RX_DMA_DESC0_RXLEN_GET(x)	((x) & EU_RX_DMA_DESC0_RXLEN_MSK)
#define EU_RX_DMA_DESC0_PHY		(1 << 18)
#define EU_RX_DMA_DESC0_FPHY		(1 << 19)
#define EU_RX_DMA_DESC0_MULTI		(1 << 20)
#define EU_RX_DMA_DESC0_BROAD		(1 << 21)
#define EU_RX_DMA_DESC0_ILEN		(1 << 22)
#define EU_RX_DMA_DESC0_ISE		(1 << 23)
#define EU_RX_DMA_DESC0_COL		(1 << 24)
#define EU_RX_DMA_DESC0_CRCERR		(1 << 25)
#define EU_RX_DMA_DESC0_NOBYTE		(1 << 26)
#define EU_RX_DMA_ERR_MSK		(EU_RX_DMA_DESC0_ILEN	| \
					 EU_RX_DMA_DESC0_ISE	| \
					 EU_RX_DMA_DESC0_COL	| \
					 EU_RX_DMA_DESC0_CRCERR	| \
					 EU_RX_DMA_DESC0_NOBYTE)

/* TMAC_DMA_DESC0 */
#define EU_TX_DMA_DESC0_EC		(1 <<  0)
#define EU_TX_DMA_DESC0_OWC		(1 <<  1)
#define EU_TX_DMA_DESC0_ED		(1 <<  2)
#define EU_TX_DMA_DESC0_CRS		(1 <<  3)
#define EU_TX_DMA_DESC0_NOCRC		(1 << 26)
#define EU_TX_DMA_ERR_MSK		(EU_TX_DMA_DESC0_EC	| \
					 EU_TX_DMA_DESC0_OWC	| \
					 EU_TX_DMA_DESC0_ED	| \
					 EU_TX_DMA_DESC0_CRS	| \
					 EU_TX_DMA_DESC0_NOCRC)

/* RMAC_DMA_CNTL */
#define EU_RX_DMA_CNTL_MSK		0x00000003
#define EU_RX_DMA_CNTL(x)		((x) & EU_RX_DMA_CNTL_MSK)

/* RMAC_DMA_STAT & TMAC_DMA_STAT*/
#define EU_DMA_STAT_TYPE		(1 << 0)
#define EU_DMA_STAT_ALIGN		(1 << 1)
#define EU_DMA_STAT_DSC_R		(1 << 2)
#define EU_DMA_STAT_DSC_W		(1 << 3)
#define EU_DMA_STAT_BUF			(1 << 4)
#define EU_DMA_STAT_FIFO		(1 << 5)
#define EU_DMA_STAT_END			(1 << 6)
#define EU_DMA_STAT_ERR			(1 << 7)
#define EU_DMA_STAT_DATA		(1 << 8)
#define EU_DMA_STAT_MEND		(1 << 9)
#define EU_DMA_STAT_MSTART		(1 <<10)
#define EU_DMA_STAT_MSK			0x000007ff

/* RMAC_CNTL */
#define EU_RX_CNTL_MIB			(1 << 0)
#define EU_RX_CNTL_PHY			(1 << 1)
#define EU_RX_CNTL_MULTI		(1 << 2)
#define EU_RX_CNTL_BROAD		(1 << 3)
#define EU_RX_CNTL_ARP			(1 << 4)
#define EU_RX_CNTL_FPHY			(1 << 5)
#define EU_RX_CNTL_MSK			0x0000003f

/* RMAC_TLEN */
#define EU_RX_TLEN_MAXRLEN_MSK		0x000007ff
#define EU_RX_TLEN_MAXRLEN(x)		((x) & EU_RX_TLEN_MAXRLEN_MSK)
#define EU_RX_TLEN_MSK			0x000007ff

/* TMAC_CNTL */
#define EU_TX_CNTL_IFG_MSK		0x00000003
#define EU_TX_CNTL_IFG(x)		((x) & EU_TX_CNTL_IFG_MSK)
#define EU_TX_CNTL_ATP			(1 << 2)
#define EU_TX_CNTL_CSI			(1 << 3)
#define EU_TX_CNTL_COI			(1 << 4)
#define EU_TX_CNTL_MSK			0x0000001f

/* MU_CNTL */
#define EU_MU_CNTL_REGADR_MSK		0x0000001f
#define EU_MU_CNTL_REGADR(x)		((x) & EU_MU_CNTL_REGADR_MSK)
#define EU_MU_CNTL_PHYADR_MSK		0x000003e0
#define EU_MU_CNTL_PHYADR(x)		(((x) << 5) & EU_MU_CNTL_PHYADR_MSK)
#define EU_MU_CNTL_MODE_MSK		0x00000c00
#define EU_MU_CNTL_MODE(x)		(((x) << 10) & EU_MU_CNTL_MODE_MSK)
#define EU_MU_CNTL_MODE_WRITE		EU_MU_CNTL_MODE(1)
#define EU_MU_CNTL_MODE_READ		EU_MU_CNTL_MODE(2)
#define EU_MU_CNTL_MODE_RESET		EU_MU_CNTL_MODE(3)
#define EU_MU_CNTL_MRE			(1 << 14)
#define EU_MU_CNTL_MIU			(1 << 15)

/* MU_DIV */
#define EU_MU_DIV_MSK			0x000000ff
#define EU_MU_DIV(x)			((x) & EU_MU_DIV_MSK)

/* RMII_CNTL */
#define EU_RMII_CNTL_IFACE_MII		0x00000000
#define EU_RMII_CNTL_IFACE_RMII		(1 << 0)
#define EU_RMII_CNTL_RATE_10MBIT	0x00000000
#define EU_RMII_CNTL_RATE_100MBIT	(1 << 1)
#define EU_RMII_CNTL_RESET		(1 << 2)



/*
** ----------------------------------------------------------------------------
** --- DMA --------------------------------------------------------------------
** ----------------------------------------------------------------------------
*/

#if !defined(_ASMLANGUAGE) && !defined(__ASSEMBLER__)
typedef struct p2001_dma_desc_s {
	volatile unsigned long		stat;
	volatile unsigned long		cntl;
	volatile unsigned char		*buff;
	volatile struct p2001_dma_desc_s *next;
} p2001_dma_desc_t;
#endif	/* _ASMLANGUAGE */

/* DCS0 - status */
#define DMA_DESC0_OFFSET_MSK		0x18000000
#define DMA_DESC0_OFFSET_SET(x)		(((x) << 27) & DMA_DESC0_OFFSET_MSK)
#define DMA_DESC0_OFFSET_GET(x)		(((x) & DMA_DESC0_OFFSET_MSK) >> 27)
#define DMA_DESC0_END			(1u << 29)
#define DMA_DESC0_START			(1u << 30)
#define DMA_DESC0_OWN			(1u << 31)
#define DMA_DESC0_OWN_DMA		DMA_DESC0_OWN
#define DMA_DESC0_OWN_CPU		0

/* DSC1 - control */
#define DMA_DESC1_LEN_MSK		0x0000ffff
#define DMA_DESC1_LEN(x)		((x) & DMA_DESC1_LEN_MSK)
#define DMA_DESC1_CHANNEL_MSK		0x007f0000
#define DMA_DESC1_CHANNEL(x)		(((x) << 16) & DMA_DESC1_CHANNEL_MSK)
#define DMA_DESC1_TYPE			(1 << 23)
#define DMA_DESC1_TYPE_RX		DMA_DESC1_TYPE
#define DMA_DESC1_TYPE_TX		0
#define DMA_DESC1_INT			(1 << 30)
#define DMA_DESC1_LOOP			(1 << 31)

/* DCS2 - buffer ptr */
#define DMA_DESC2_BUF_MSK		0x0fffffff
#define DMA_DESC2_BUF(x)		((x) & DMA_DESC2_BUF_MSK)

/* DSC3 - next descptr */
#define DMA_DESC3_NEXT_MSK		0x0fffffff
#define DMA_DESC3_NEXT(x)		((x) & DMA_DESC3_NEXT_MSK)



/*
** ----------------------------------------------------------------------------
** --- USB --------------------------------------------------------------------
** ----------------------------------------------------------------------------
*/

#endif	/* _INC_p2001_h_ */
