/*
 * Copyright (C) 2004 elmeg GmbH
 * Written by Juergen Selent (j.selent@elmeg.de)
 *
 * arp.c
 * -----
 *
 * $RCSfile: arp.c,v $
 * $Source: /opt/funkwerk/cvsroot/source/cmd/busybox/busybox-0.60.3/arp.c,v $
 * $Author: selent $
 * $Revision: 1.1 $
 * $Date: 2004/01/12 20:41:49 $
 * $State: Exp $
 * $Name:  $
 *
 * $Log: arp.c,v $
 * Revision 1.1  2004/01/12 20:41:49  selent
 * mini applet "arp -s ip hwaddr dev" added
 *
 * Initial Version	Mon Jan 12 18:47:11 CET 2004 - jse
 * Written
 */

/* ************************************************************************* */
/* *** INCLUDES ************************************************************ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stddef.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <linux/if_ether.h>
#include "busybox.h"

#define SET_SA_FAMILY(addr, family)			\
    memset ((char *) &(addr), '\0', sizeof(addr));	\
    addr.sa_family = (family);



/* ************************************************************************* */
/* *** STATIC FUNCTIONS **************************************************** */

static int
get_ether( unsigned char *dp,unsigned char *sp )
{
	int		i;
	int		j;
	unsigned char	val;
	unsigned char	c;

	for (i=0; i<6; i++) {
		val = 0;
		if (i && (*sp == ':')) {
			sp++;
		}
		for (j=0; j<2; j++) {
			c = *sp;
			if (c >= '0' && c <= '9') {
				c -= '0';
			} else if (c >= 'a' && c <= 'f') {
				c -= ('a' - 10);
			} else if (c >= 'A' && c <= 'F') {
				c -= ('A' - 10);
			} else if (j && (c == ':' || c == 0)) {
				break;
			} else {
				return( -1 );
			}
			sp++;
			val <<= 4;
			val += c;
		}
		*dp++ = val;
	}
	return( 0 );
}



/* ************************************************************************* */
/* *** GLOBAL FUNCTIONS **************************************************** */

int
arp_main( int argc,char *argv[] )
{
	int		fd;
	unsigned char	hwaddr[6];
	struct arpreq	arpreq;
	struct in_addr	in;

	if (argc != 5) {
		return( -1 );
	}

	if (argv[1][0] != '-' || argv[1][1] != 's') {
		return( -2 );
	}

	memset( &arpreq,0,sizeof( arpreq ));

	if (inet_aton( argv[2],&in ) == 0) {
		return( -3 );
	}

	if (get_ether( &hwaddr[0],argv[3] )) {
		return( -4 );
	}

	if ((fd = socket( AF_INET,SOCK_DGRAM,0 )) < 0) {
		return( -5 );
	}

	SET_SA_FAMILY( arpreq.arp_ha,1 );
	memcpy( arpreq.arp_ha.sa_data,hwaddr,6 );
	SET_SA_FAMILY( arpreq.arp_pa,AF_INET );
	((struct sockaddr_in *) &arpreq.arp_pa)->sin_addr.s_addr = in.s_addr;
	strcpy( arpreq.arp_dev,argv[4] );
	arpreq.arp_flags = ATF_PERM | ATF_COM;

	if (ioctl( fd,SIOCSARP,(caddr_t)&arpreq ) < 0) {
		close( fd );
		return( -5 );
	}

	close( fd );
	return( 0 );
}
