/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH> 
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * arpping.c
 * ---------
 *
 * Initial Version	Thu Mar 28 14:53:48 CET 2002 - jse
 * Rewritten based on uDhcpd
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

/* ************************************************************************* */
/* *** INCLUDES ************************************************************ */

#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/if_ether.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include "dhcpd.h"
#include "debug.h"
#include "arpping.h"



/* ************************************************************************* */
/* *** STATIC FUNCTIONS **************************************************** */

static void
mkArpMsg( int opcode,
	  u_long tInaddr,
	  u_char *tHaddr,
	  u_long sInaddr,
	  u_char *sHaddr,
	  struct arpMsg *msg )
{
	memset( msg,0,sizeof( *msg ));
	memcpy( msg->ethhdr.h_dest,MAC_BCAST_ADDR,6 );
	memcpy( msg->ethhdr.h_source,sHaddr,6 );
	msg->ethhdr.h_proto = htons( ETH_P_ARP );
	msg->htype = htons( ARPHRD_ETHER );	/* hardware type */
	msg->ptype = htons( ETH_P_IP );		/* protocol type (ARP message)*/
	msg->hlen = 6;				/* hardware address length */
	msg->plen = 4;				/* protocol address length */
	msg->operation = htons( opcode );	/* ARP op code */
	memcpy( msg->sInaddr,(u_char *)&sInaddr,4 );/* source IP address */
	memcpy( msg->sHaddr,sHaddr,6 );		/* source hardware address */
	memcpy( msg->tInaddr,(u_char *)&tInaddr,4 );/* dest IP address */
	if (opcode == ARPOP_REPLY) {
		memcpy( msg->tHaddr,tHaddr,6);	/* target hardware address */
	}
}

static int
arpCheck( u_long inaddr,struct ifinfo *ifbuf,long timeout )
{
	static char	*_f = "arpCheck:";
	int		s;
	int		rv;
	struct sockaddr	addr;
	struct arpMsg	arp;
	fd_set		fdset;
	struct timeval	tm;
	time_t		prevTime;

	rv = 1;
	openRawSocket( &s,ETH_P_ARP );
	if (s == -1) {
		/* assume valid arp reply */
		WRN( "%s cannot open rawsocket - assume valid arp reply\n",_f );
		return( 0 );
	}

	/* send arp request */
	mkArpMsg( ARPOP_REQUEST,inaddr,NULL,ifbuf->addr,ifbuf->haddr,&arp );
	memset( &addr,0,sizeof( addr ));
	strcpy( addr.sa_data,ifbuf->ifname );
	if (sendto( s,&arp,sizeof( arp ),0,&addr,sizeof( addr )) < 0) {
		WRN( "%s sendto failed - assume valid arp reply\n",_f );
		close( s );
		return( 0 );
	}

	tm.tv_usec = 0;
	time( &prevTime );
	while (timeout > 0) {
		FD_ZERO( &fdset );
		FD_SET( s,&fdset );
		tm.tv_sec  = timeout;
		if (select( s+1,&fdset,NULL,NULL,&tm ) < 0 ) {
			if (errno != EINTR) {
				WRN( "%s select error - assume valid arp reply\n",_f );
				rv = 0;
				break;
			}
		} else if (FD_ISSET( s,&fdset )) {
			if (recv( s,&arp,sizeof( arp ),0) < 0) {
				WRN( "%s recv failed - assume valid arp reply\n",_f );
				rv = 0;
				break;
			}
			if (arp.operation == htons( ARPOP_REPLY ) && 
			    memcmp( arp.tHaddr,ifbuf->haddr,6 ) == 0 && 
			    *((u_int *)arp.sInaddr) == inaddr )
			{
				DEB( "%s valid arp reply received\n",_f );
				rv = 0;
				break;
			}
		}
		timeout -= time( NULL ) - prevTime;
		time( &prevTime );
	}
	close( s );
	return( rv );
}



/* ************************************************************************* */
/* *** GLOBAL FUNCTIONS **************************************************** */

int
arpping( u_int32_t yiaddr,u_int32_t ip,char *arp,char *interface )
{
	struct ifinfo	ifbuf;

	strcpy( ifbuf.ifname,interface );
	ifbuf.addr = ip;
	ifbuf.mask = 0x0;
	ifbuf.bcast = 0x0;
	memcpy( ifbuf.haddr,arp,6 );
	ifbuf.flags = 0;

	/*
	 * returns: 1 if yiaddr is free
	 *          0 if yiaddr is used
	 */
	return arpCheck( yiaddr,&ifbuf,2 );
}

int openRawSocket( int *s,u_short type )
{
	int	optval = 1;

	if ((*s = socket( PF_INET,SOCK_PACKET,htons( type ))) == -1) {
		INF( "openRawSocket: socket failed\n" );
		return( -1 );
	}

	if (setsockopt( *s,SOL_SOCKET,SO_BROADCAST,&optval,sizeof( optval )) == -1) {
		INF( "openRawSocket: setsockopt failed\n" );
		close( *s );
		return( -1 );
	}
	return( 0 );
}
