/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH> 
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * dhcpd.c
 * -------
 *
 * Initial Version	Thu Mar 28 14:53:48 CET 2002 - jse
 * Rewritten based on uDhcpd
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

/* ************************************************************************* */
/* *** INCLUDES ************************************************************ */

#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <time.h>
#include <sys/time.h>

#include "debug.h"
#include "dhcpd.h"
#include "arpping.h"
#include "socket.h"
#include "options.h"
#include "files.h"
#include "leases.h"
#include "packet.h"
#include "serverpacket.h"
#include "pidfile.h"
#include "version.h"



/* ************************************************************************* */
/* *** GLOBAL VARIABLES **************************************************** */

/* globals */
struct dhcpOfferedAddr *leases;
struct server_config_t server_config;



/* ************************************************************************* */
/* *** STATIC FUNCTIONS **************************************************** */

static void
exit_server( int retval )
{
#ifdef WITH_PIDFILE
	pidfile_delete( server_config.pidfile );
#endif
	write_leases( 0 );
	LOG( "terminating with status=%s\n",strerror( retval ));
	CLOSELOG();
	exit( retval );
}

static
void dhcpd_killed( int sig )
{
	sig = 0;
	LOG( "SIGTERM received\n" );
	exit_server( 0 );
}



/* ************************************************************************* */
/* *** GLOBAL FUNCTIONS **************************************************** */

int
main( int argc,char *argv[] )
{
	static char		*_f = "main:";
	fd_set			rfds;
	int			server_socket;
	int			bytes;
	int			retval;
	struct dhcpMessage	packet;
	unsigned char		*state;
	char			*server_id;
	char			*requested;
	u_int32_t		server_id_align;
	u_int32_t		requested_align;
	struct option_set	*option;
	struct dhcpOfferedAddr	*lease;
	struct sockaddr_in	*sin;
#ifdef WITH_PIDFILE
	int			pid_fd;
#endif
	int			fd = -1;
	struct ifreq		ifr;

#if defined(FIX_RELOCATION_BUG)
	init_config_keywords();
#endif

	OPENLOG( "dhcpd",LOG_PID|LOG_CONS,LOG_DAEMON );
	LOG( "version %s started\n",dhcpdVersion );
	
#ifdef WITH_PIDFILE
	pid_fd = pidfile_acquire( server_config.pidfile );
	pidfile_write_release( pid_fd );
#endif

	memset( &server_config,0,sizeof( struct server_config_t ));

	read_config( DHCPD_CONF_FILE );
	if ((option = find_option( server_config.options,DHCP_LEASE_TIME ))) {
		memcpy( &server_config.lease,option->data+2,4 );
		server_config.lease = ntohl( server_config.lease );
	} else {
		server_config.lease = LEASE_TIME;
	}

	leases = malloc( server_config.max_leases *
			 sizeof( struct dhcpOfferedAddr ));
	if (!leases) {
		retval = errno;
		ERR( "%s malloc lease table (size=%ld)\n",
		      _f,server_config.max_leases );
		exit_server( errno );
	}
	memset( leases,0,server_config.max_leases *
		sizeof( struct dhcpOfferedAddr ));
	read_leases( server_config.lease_file );

	if ((fd = socket( AF_INET,SOCK_RAW,IPPROTO_RAW )) >= 0) {
		ifr.ifr_addr.sa_family = AF_INET;
		strcpy( ifr.ifr_name,server_config.interface );
		if (ioctl( fd,SIOCGIFADDR,&ifr ) == 0) {
			sin = (struct sockaddr_in *) &ifr.ifr_addr;
			server_config.server = sin->sin_addr.s_addr;
			DEB( "%s using %s = %s\n",
			     _f,ifr.ifr_name,inet_ntoa( sin->sin_addr ));
		} else {
			retval = errno;
			ERR( "%s ioctl SIOCGIFADDR\n",_f );
			exit_server( retval );
		}
		if (ioctl( fd,SIOCGIFHWADDR,&ifr) == 0) {
			memcpy( server_config.arp,ifr.ifr_hwaddr.sa_data,6 );
			DEB( "%s using hwaddr %02x:%02x:%02x:%02x:%02x:%02x\n",
			      _f,
			      server_config.arp[0],
			      server_config.arp[1],
			      server_config.arp[2],
			      server_config.arp[3],
			      server_config.arp[4],
			      server_config.arp[5] );
		} else {
			retval = errno;
			ERR( "%s ioctl SIOCGIFHWADDR\n",_f );
			exit_server( retval );
		}
	} else {
		retval = errno;
		ERR( "%s socket failed",_f );
		exit_server( retval );
	}
	close( fd );

	signal( SIGUSR1,write_leases );
	signal( SIGUSR2,dump_leases );
	signal( SIGTERM,dhcpd_killed );
	signal( SIGQUIT,dhcpd_killed );
	signal( SIGINT,dhcpd_killed );

	while(1) {
		server_socket = listen_socket( INADDR_ANY,SERVER_PORT,server_config.interface );
		if(server_socket == -1) {
			retval = errno;
			ERR( "%s cannot create server socket\n",_f );
			exit_server( retval );
		}

		FD_ZERO( &rfds );
		FD_SET( server_socket,&rfds );
		retval = select( server_socket + 1,
				 &rfds,NULL,NULL,NULL );
		if (retval <= 0) {
			INF( "%s select returns %d\n",_f,retval );
			close(server_socket);
			continue;
		}

		bytes = get_packet( &packet,server_socket );
		close( server_socket );
		if (bytes < 0) {
			continue;
		}

		if ((state = get_option( &packet,DHCP_MESSAGE_TYPE )) == NULL) {
			INF( "%s get_option failed\n",_f );
			continue;
		}
		
		lease = find_lease_by_chaddr( packet.chaddr );
		switch (state[0]) {
		case DHCPDISCOVER:
			DEB( "%s DISCOVER\n",_f );
			if (sendOffer( &packet ) < 0) {
				WRN( "%s sendOffer failed\n",_f );
			}
			break;
 		case DHCPREQUEST:
			DEB( "%s REQUEST\n",_f );

			requested = get_option( &packet,DHCP_REQUESTED_IP );
			server_id = get_option( &packet,DHCP_SERVER_ID );

			if (requested) {
				memcpy( &requested_align,requested,4 );
			}
			if (server_id) {
				memcpy( &server_id_align,server_id,4 );
			}

			if (!lease) {
				/* remain silent */				
				continue;
			}
			if (server_id) {
				/* SELECTING State */
				DEB( "%s server_id=0x%08x",ntohl( server_id_align ));
				if (server_id_align == server_config.server &&
				    requested && 
				    requested_align == lease->yiaddr)
				{
					sendACK( &packet,lease->yiaddr );
				}
			} else {
				if (requested) {
					/* INIT-REBOOT State */
					if (lease->yiaddr == requested_align) {
						sendACK( &packet,lease->yiaddr);
					} else {
						sendNAK( &packet );
					}
				} else {
					/* RENEWING or REBINDING State */
					if (lease->yiaddr == packet.ciaddr) {
						sendACK( &packet,lease->yiaddr);
					} else {
						/* don't know what to do!!!! */
						sendNAK( &packet );
					}
				}
			}
			write_leases( 0 );
			break;
		case DHCPDECLINE:
			DEB( "%s DECLINE\n",_f );
			if (lease) {
				memset( lease->chaddr,0,16 );
				lease->expires = time( 0 ) + server_config.decline_time;
			}
			write_leases( 0 );
			break;
		case DHCPRELEASE:
			DEB( "%s RELEASE\n",_f );
			if (lease) {
				lease->expires = time( 0 );
			}
			write_leases( 0 );
			break;
		case DHCPINFORM:
			DEB( "%s INFORM\n",_f );
			send_inform( &packet );
			break;
		default:
			WRN( "%s unsupported DHCP message 0x%02x\n",
			      _f,state[0] );
			break;
		}
	}
	exit_server( 0 );
}
