/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH> 
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * files.c
 * -------
 *
 * Initial Version	Thu May 16 17:36:09 CEST 2002 - jse
 * Rewritten based on uDhcpd
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

/* ************************************************************************* */
/* *** INCLUDES ************************************************************ */

#include <stdio.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>

#include "debug.h"
#include "dhcpd.h"
#include "files.h"
#include "options.h"
#include "leases.h"



/* ************************************************************************* */
/* *** STATIC FUNCTIONS **************************************************** */

static int
read_ip( char *line,void *arg )
{
	inet_aton( line,(struct in_addr *)arg );
	return( 1 );
}


static int
read_str( char *line,void *arg )
{
	char	**dst = arg;
	int i;
	
	if (*dst != NULL) {
		free( *dst );
	}
	if ((*dst = strdup( line )) == NULL) {
		return( 0 );
	}
	
	/* elimate trailing whitespace */
	for (i = strlen( *dst ) - 1; i > 0 && isspace(( *dst )[i]); i--);
	if (i > 0) {
		(*dst)[i+1] = '\0';
	} else {
		(*dst)[0] = '\0';
	}
	return( 1 );
}


static int read_u32(char *line, void *arg)
{
	u_int32_t	*dst = arg;

	*dst = strtoul( line,NULL,0 );
	return( 1 );
}


static int
read_yn( char *line,void *arg )
{
	char	*dst = arg;

	if (!strcasecmp( "yes",line ) ||
	    !strcmp( "1",line ) ||
	    !strcasecmp( "true",line ))
	{
		*dst = 1;
	} else if (!strcasecmp( "no",line ) ||
	           !strcmp( "0",line ) ||
	           !strcasecmp( "false",line ))
	{
		*dst = 0;
	} else {
		return( 0 );
	}
	return( 1 );
}

static int
read_opt( char *line,void *arg )
{
	struct option_set	**opt_list = arg;
	char			*opt;
	char			*val;
	struct dhcp_option	*option = NULL;
	int			length = 0;
	char			buffer[255];
	char			fail;
	u_int16_t		result_u16;
	int16_t			result_s16;
	u_int32_t		result_u32;
	int32_t			result_s32;
	int			i;
	
	if (!(opt = strtok( line," \t=" ))) {
		return( 0 );
	}
	
	for (i=0; options[i].code; i++) {
		if (!strcmp( options[i].name,opt)) {
			option = &(options[i]);
			break;
		}
	}

	if (!option) {
		return( 0 );
	}
	
	do {
		val = strtok( NULL,", \t" );
		if (val) {
			fail = 0;
			length = 0;
			switch (option->flags & TYPE_MASK) {
			case OPTION_IP:
				read_ip( val,buffer );
				break;
			case OPTION_IP_PAIR:
				read_ip( val,buffer );
				if ((val = strtok( NULL,", \t/-" ))) {
					read_ip( val,buffer+4 );
				} else {
					fail = 1;
				}
				break;
			case OPTION_STRING:
				length = strlen( val );
				if (length > 254) {
					length = 254;
				}
				memcpy( buffer,val,length );
				break;
			case OPTION_BOOLEAN:
				if (!read_yn( val,buffer )) {
					fail = 1;
				}
				break;
			case OPTION_U8:
				buffer[0] = strtoul( val,NULL,0 );
				break;
			case OPTION_U16:
				result_u16 = htons( strtoul( val,NULL,0 ));
				memcpy( buffer,&result_u16,2 );
				break;
			case OPTION_S16:
				result_s16 = htons( strtol( val,NULL,0 ));
				memcpy( buffer,&result_s16,2 );
				break;
			case OPTION_U32:
				result_u32 = htonl( strtoul( val,NULL,0 ));
				memcpy( buffer,&result_u32,4 );
				break;
			case OPTION_S32:
				result_s32 = htonl( strtol( val,NULL,0 ));	
				memcpy( buffer,&result_s32,4 );
				break;
			default:
				break;
			}
			length += option_lengths[option->flags & TYPE_MASK];
			if (!fail) {
				attach_option( opt_list,option,buffer,length );
			}
		} else {
			fail = 1;
		}
	} while (!fail && option->flags & OPTION_LIST);
	return( 1 );
}

static struct config_keyword keywords[] = {
	/* keyword	handler   variable address		default */
	{"start",	read_ip,  &(server_config.start),	""},
	{"end",		read_ip,  &(server_config.end),		""},
	{"interface",	read_str, &(server_config.interface),	""},
	{"option",	read_opt, &(server_config.options),	""},
	{"opt",		read_opt, &(server_config.options),	""},
	{"max_leases",	read_u32, &(server_config.max_leases),	"254"},
	{"remaining",	read_yn,  &(server_config.remaining),	"yes"},
	{"decline_time",read_u32, &(server_config.decline_time),"3600"},
	{"conflict_time",read_u32,&(server_config.conflict_time),"3600"},
	{"offer_time",	read_u32, &(server_config.offer_time),	"60"},
	{"min_lease",	read_u32, &(server_config.min_lease),	"60"},
	{"lease_file",	read_str, &(server_config.lease_file),	"/etc/dhcpd.leases"},
	{"pidfile",	read_str, &(server_config.pidfile),	"/var/run/dhcpd.pid"},
	{"siaddr",	read_ip,  &(server_config.siaddr),	"0.0.0.0"},
	{"sname",	read_str, &(server_config.sname),	""},
	{"boot_file",	read_str, &(server_config.boot_file),	""},
	{"",		NULL, 	  NULL,				""}
};



/* ************************************************************************* */
/* *** GLOBAL FUNCTIONS **************************************************** */

#if defined(FIX_RELOCATION_BUG)
void
init_config_keywords( void )
{
	keywords[0].var = &(server_config.start);
	keywords[1].var = &(server_config.end);
	keywords[2].var = &(server_config.interface);
	keywords[3].var = &(server_config.options);
	keywords[4].var = &(server_config.options);
	keywords[5].var = &(server_config.max_leases);
	keywords[6].var = &(server_config.remaining);
	keywords[7].var = &(server_config.auto_time);
	keywords[8].var = &(server_config.decline_time);
	keywords[9].var = &(server_config.conflict_time);
	keywords[10].var = &(server_config.offer_time);
	keywords[11].var = &(server_config.min_lease);
	keywords[12].var = &(server_config.lease_file);
	keywords[13].var = &(server_config.pidfile);
	keywords[14].var = &(server_config.siaddr);
	keywords[15].var = &(server_config.sname);
	keywords[16].var = &(server_config.boot_file);
}
#endif

int
read_config( char *file )
{
	FILE	*in;
	char	*token;
	char	*line;
	char	*p;
	int	i;
	char	buffer[80];

	for (i=0; strlen( keywords[i].keyword ); i++) {
		if (strlen(keywords[i].def)) {
			keywords[i].handler( keywords[i].def,keywords[i].var );
		}
	}

	if (!(in = fopen( file,"r" ))) {
		ERR( "open config file '%s' errno=%s",file,strerror( errno ));
		return( 0 );
	}
	
	while (fgets( buffer,80,in )) {
		if ((p = strchr( buffer,'\n' )) != NULL) {
			*p = '\0';
		}
		if ((p = strchr( buffer,'#' )) != NULL) {
			*p = '\0';
		}
		token = buffer + strspn( buffer," \t" );
		if (*token == '\0') {
			continue;
		}
		line = token + strcspn( token," \t=" );
		if (*line == '\0') {
			continue;
		}
		*line = '\0';
		line++;
		line = line + strspn( line," \t=" );
		if (*line == '\0') {
			continue;
		}

		for (i=0; strlen( keywords[i].keyword ); i++) {
			if (!strcasecmp( token,keywords[i].keyword )) {
				keywords[i].handler( line,keywords[i].var );
			}
		}
	}
	fclose( in );
	return( 1 );
}

void
write_leases( int dummy )
{
	FILE		*fp;
	unsigned int	i;
	time_t		curr = time(0);
	unsigned long	lease_time;
	
	dummy = 0;
	if (!(fp = fopen( server_config.lease_file,"w" ))) {
		ERR( "open '%s' for writing errno=%s",
		     server_config.lease_file,strerror( errno ));
		return;
	}

	for (i=0; i<server_config.max_leases; i++) {
		if (leases[i].yiaddr != 0) {
			if (server_config.remaining) {
				if (lease_expired(&(leases[i]))) {
					lease_time = 0;
				} else {
					lease_time = leases[i].expires - curr;
				}
			} else {
				lease_time = leases[i].expires;
			}
			lease_time = htonl( lease_time );
			fwrite( leases[i].chaddr,16,1,fp );
			fwrite( &(leases[i].yiaddr),4,1,fp );
			fwrite( &lease_time,4,1,fp );
		}
	}
	fclose( fp );
}

void
read_leases( char *file )
{
	FILE			*fp;
	unsigned int		i = 0;
	time_t			curr = time(0);
	struct dhcpOfferedAddr	lease;

	if (!(fp = fopen(file, "r"))) {
		INF( "open '%s' for reading errno=%s",file,strerror( errno ));
		return;
	}

	while ((i < server_config.max_leases) &&
	       (fread( &lease,sizeof( lease ),1,fp ) == 1))
	{
		if (ntohl( lease.yiaddr ) >= ntohl( server_config.start ) &&
		    ntohl( lease.yiaddr ) <= ntohl( server_config.end ))
		{
			leases[i].yiaddr = lease.yiaddr;
			leases[i].expires = ntohl( lease.expires );	
			if (server_config.remaining) {
				leases[i].expires += curr;
			}
			memcpy( leases[i].chaddr,lease.chaddr,sizeof( lease.chaddr ));
			i++;
		}
	}

	INF( "read %d leases",i );

	if (i == server_config.max_leases) {
		if (fgetc( fp ) == EOF) {
			/* might be helpfull to drop expired leases */
			LOG( "too many leases while loading %s\n",file );
		}
	}
	fclose( fp );
}

void
dump_leases( int dummy )
{
	time_t		curr;
	char		*curr_date;
	unsigned long	lease_time;
	unsigned int	i;
	unsigned long	ipaddr;
	char		ipbuf[32];

	dummy = 0;
	curr = time( 0 );
	curr_date = ctime( &curr );
	*(curr_date + strlen( curr_date ) - 1) = 0;
	LOG( "%s lease table:\n",curr_date );
	LOG( "======================================\n" );
	for (i=0; i<server_config.max_leases; i++) {
		if (leases[i].yiaddr != 0) {
			if (server_config.remaining) {
				if (lease_expired(&(leases[i]))) {
					lease_time = 0;
				} else {
					lease_time = leases[i].expires - curr;
				}
			} else {
				lease_time = leases[i].expires;
			}
			ipaddr = ntohl( leases[i].yiaddr );
			sprintf( ipbuf,"%ld.%ld.%ld.%ld",
				 (ipaddr >> 24) & 0xff,
				 (ipaddr >> 16) & 0xff,
				 (ipaddr >> 8) & 0xff,
				 ipaddr & 0xff );

			LOG( "hw=%02x:%02x:%02x:%02x:%02x:%02x ip=%s lease_time=%ld",
				leases[i].chaddr[0],
				leases[i].chaddr[1],
				leases[i].chaddr[2],
				leases[i].chaddr[3],
				leases[i].chaddr[4],
				leases[i].chaddr[5],
				ipbuf,
				lease_time );
		}
	}
}
