/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH> 
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * leases.c
 * --------
 *
 * Initial Version	Thu May 16 17:36:09 CEST 2002 - jse
 * Rewritten based on uDhcpd
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

/* ************************************************************************* */
/* *** INCLUDES ************************************************************ */

#include <time.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "debug.h"
#include "dhcpd.h"
#include "files.h"
#include "options.h"
#include "leases.h"
#include "arpping.h"



/* ************************************************************************* */
/* *** GLOBAL FUNCTIONS **************************************************** */

/* clear every lease out that chaddr OR yiaddr matches and is nonzero */
void
clear_lease( u_int8_t *chaddr,u_int32_t yiaddr )
{
	unsigned int	blank_chaddr = 0;
	unsigned int	blank_yiaddr = 0;
	unsigned int	i;

	for ( i=0; i<16 && !chaddr[i]; i++);
	if (i == 16) {
		blank_chaddr = 1;
	}
	blank_yiaddr = (yiaddr == 0);

	for (i=0; i<server_config.max_leases; i++) {
		if ((!blank_chaddr && !memcmp(leases[i].chaddr, chaddr, 16)) ||
		    (!blank_yiaddr && leases[i].yiaddr == yiaddr))
		{
			memset( &(leases[i]),0,sizeof( struct dhcpOfferedAddr ));
		}
	}
}

/* add a lease into the table, clearing out any old ones */
struct dhcpOfferedAddr *
add_lease( u_int8_t *chaddr,u_int32_t yiaddr,unsigned long lease )
{
	struct dhcpOfferedAddr	*oldest;

	/* clean out any old ones */
	clear_lease( chaddr,yiaddr );
	oldest = oldest_expired_lease();

	if (oldest) {
		memcpy( oldest->chaddr,chaddr,16 );
		oldest->yiaddr = yiaddr;
		oldest->expires = time( 0 ) + lease;
	}
	return( oldest );
}

/* true if a lease has expired */
int
lease_expired( struct dhcpOfferedAddr *lease )
{
	return( lease->expires < (unsigned long)time( 0 ));
}	

/* Find the oldest expired lease, NULL if there are no expired leases */
struct dhcpOfferedAddr *
oldest_expired_lease( void )
{
	struct dhcpOfferedAddr	*oldest = NULL;
	unsigned long		oldest_lease = time(0);
	unsigned int		i;

	for (i=0; i<server_config.max_leases; i++) {
		if (oldest_lease > leases[i].expires) {
			oldest_lease = leases[i].expires;
			oldest = &(leases[i]);
		}
	}
	return( oldest );
}


/* Find the first lease that matches chaddr, NULL if no match */
struct dhcpOfferedAddr *
find_lease_by_chaddr( u_int8_t *chaddr )
{
	unsigned int	i;

	for (i=0; i<server_config.max_leases; i++) {
		if (!memcmp( leases[i].chaddr,chaddr,16 )) {
			return( &(leases[i]) );
		}
	}
	return( NULL );
}

/* Find the first lease that matches yiaddr, NULL is no match */
struct dhcpOfferedAddr *
find_lease_by_yiaddr( u_int32_t yiaddr )
{
	unsigned int	i;

	for (i=0; i<server_config.max_leases; i++) {
		if (leases[i].yiaddr == yiaddr) {
			return( &(leases[i]) );
		}
	}
	return( NULL );
}

/*
 * find an assignable address, it check_expired is true,
 * we check all the expired leases as well.
 * Maybe this should try expired leases by age...
 */
u_int32_t
find_address( int check_expired ) 
{
	u_int32_t		addr;
	u_int32_t		ret = 0;
	struct dhcpOfferedAddr	*lease = NULL;		

	addr = server_config.start;
	for (;ntohl(addr) < ntohl(server_config.end); addr = htonl(ntohl(addr) + 1)) {
		/* ie, 192.168.55.0 */
		if (!(ntohl( addr ) & 0xFF)) {
			continue;
		}

		/* ie, 192.168.55.255 */
		if ((ntohl( addr ) & 0xFF) == 0xFF) {
			continue;
		}

		/* lease is not taken */
		if ((!(lease = find_lease_by_yiaddr( addr )) ||

		     /* or it expired and we are checking for expired leases */
		     (check_expired  && lease_expired( lease ))) &&

		     /* and it isn't on the network */
	    	     !check_ip( addr )) {
			ret = addr;
			break;
		}
	}
	return( ret );
}

/* check is an IP is taken, if it is, add it to the lease table */
int
check_ip( u_int32_t addr )
{
	struct in_addr	temp;
	char		blank_chaddr[] = {[0 ... 15] = 0};
	
	if (!arpping( addr,
	              server_config.server,
	              server_config.arp,
	              server_config.interface ))
	{
		temp.s_addr = addr;
	 	LOG( "%s belongs to someone, reserving it for %ld seconds\n", 
 		     inet_ntoa( temp ),server_config.conflict_time );
		add_lease( blank_chaddr,addr,server_config.conflict_time );
		return( 1 );
	}
	return( 0 );
}
