/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH> 
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * options.c
 * ---------
 *
 * Initial Version	Thu May 16 17:36:09 CEST 2002 - jse
 * Rewritten based on uDhcpd
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

/* ************************************************************************* */
/* *** INCLUDES ************************************************************ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "debug.h"
#include "dhcpd.h"
#include "files.h"
#include "options.h"
#include "leases.h"



/* ************************************************************************* */
/* *** GLOBAL VARIABLES **************************************************** */

/* supported options are easily added here */
struct dhcp_option options[] = {
	/* name[10]	flags				code */
	{"subnet",	OPTION_IP,			0x01},
	{"timezone",	OPTION_S32,			0x02},
	{"router",	OPTION_IP | OPTION_LIST,	0x03},
	{"timesvr",	OPTION_IP | OPTION_LIST,	0x04},
	{"namesvr",	OPTION_IP | OPTION_LIST,	0x05},
	{"dns",		OPTION_IP | OPTION_LIST,	0x06},
	{"logsvr",	OPTION_IP | OPTION_LIST,	0x07},
	{"cookiesvr",	OPTION_IP | OPTION_LIST,	0x08},
	{"lprsvr",	OPTION_IP | OPTION_LIST,	0x09},
	{"hostname",	OPTION_STRING,			0x0c},
	{"bootsize",	OPTION_U16,			0x0d},
	{"domain",	OPTION_STRING,			0x0f},
	{"swapsvr",	OPTION_IP,			0x10},
	{"rootpath",	OPTION_STRING,			0x11},
	{"ipttl",	OPTION_U8,			0x17},
	{"mtu",		OPTION_U16,			0x1a},
	{"broadcast",	OPTION_IP,			0x1c},
	{"ntpsrv",	OPTION_IP | OPTION_LIST,	0x2a},
	{"wins",	OPTION_IP | OPTION_LIST,	0x2c},
	{"requestip",	OPTION_IP,			0x32},
	{"lease",	OPTION_U32,			0x33},
	{"dhcptype",	OPTION_U8,			0x35},
	{"serverid",	OPTION_IP,			0x36},
	{"tftp",	OPTION_STRING,			0x42},
	{"bootfile",	OPTION_STRING,			0x43},
	{"",		0x00,				0x00}
};

/* Lengths of the different option types */
int option_lengths[] = {
	[OPTION_IP] =		4,
	[OPTION_IP_PAIR] =	8,
	[OPTION_BOOLEAN] =	1,
	[OPTION_STRING] =	0,
	[OPTION_U8] =		1,
	[OPTION_U16] =		2,
	[OPTION_S16] =		2,
	[OPTION_U32] =		4,
	[OPTION_S32] =		4
};



/* ************************************************************************* */
/* *** GLOBAL FUNCTIONS **************************************************** */

/* get an option with bounds checking (warning, not aligned). */
unsigned char *
get_option( struct dhcpMessage *packet,int code )
{
	unsigned char	*optionptr;
	int		length;
	int		curr;
	int		over;
	int		done;
	int		i;

	optionptr = packet->options;
	curr = OPTION_FIELD;
	over = 0;
	done = 0;
	i = 0;
	length = 308;
	while (!done) {
		if (i >= length) {
			WRN( "bogus dhcp packet (option fields too long)\n" );
			return( NULL );
		}
		if (optionptr[i + OPT_CODE] == code) {
			if (i + 1 + optionptr[i + OPT_LEN] >= length) {
				WRN( "bogus dhcp packet (option fields too long)\n" );
				return( NULL );
			}
			return( optionptr + i + 2 );
		}			
		switch (optionptr[i + OPT_CODE]) {
		case DHCP_PADDING:
			i++;
			break;
		case DHCP_OPTION_OVER:
			if (i + 1 + optionptr[i + OPT_LEN] >= length) {
				WRN( "bogus dhcp packet (option fields too long)\n" );
				return( NULL );
			}
			over = optionptr[i + 3];
			i += optionptr[OPT_LEN] + 2;
			break;
		case DHCP_END:
			if (curr == OPTION_FIELD && over & FILE_FIELD) {
				optionptr = packet->file;
				i = 0;
				length = 128;
				curr = FILE_FIELD;
			} else if (curr == FILE_FIELD && over & SNAME_FIELD) {
				optionptr = packet->sname;
				i = 0;
				length = 64;
				curr = SNAME_FIELD;
			} else {
				done = 1;
			}
			break;
		default:
			i += optionptr[OPT_LEN + i] + 2;
		}
	}
	return( NULL );
}

/* return the position of the 'end' option (no bounds checking) */
int
end_option( unsigned char *optionptr ) 
{
	int	i = 0;

	while (optionptr[i] != DHCP_END) {
		if (optionptr[i] == DHCP_PADDING) {
			i++;
		} else {
			i += optionptr[i + OPT_LEN] + 2;
		}
	}
	return( i );
}

/*
 * add an option string to the options
 * (an option string contains an option code, length, then data)
 */
int
add_option_string( unsigned char *optionptr,unsigned char *string )
{
	int	i;
	int	end = end_option( optionptr );
	
	/* end position + string length + option code/length + end option */
	if (end + string[OPT_LEN] + 2 + 1 >= 308) {
		for (i=0; options[i].code && options[i].code != string[OPT_CODE]; i++);
		ERR( "option %s (0x%02x) did not fit into the packet\n", 
		      options[i].code ? options[i].name : "unknown",
		      string[OPT_CODE] );
		return( 0 );
	}
	DEB( "adding option 0x%02x",string[OPT_CODE] );
	memcpy( optionptr+end,string,string[OPT_LEN]+2 );
	optionptr[ end+string[OPT_LEN]+2 ] = DHCP_END;
	return( string[OPT_LEN] + 2 );
}

/* add a one to four byte option to a packet */
int
add_simple_option( unsigned char *optionptr,unsigned char code,u_int32_t data )
{
	char	buffer[4]; /* Cant copy straight to optionptr, it might not be aligned */
	char	length = 0;
	int	end;
	int	i;

	for (i=0; options[i].code; i++) {
		if (options[i].code == code) {
			length = option_lengths[options[i].flags & TYPE_MASK];
			break;
		}
	}

	if (!length) {
		WRN( "could not add option 0x%02x\n",code );
		return( 0 );
	}

	DEB( "adding option 0x%02x\n",code );
	end = end_option( optionptr );
	optionptr[end + OPT_CODE] = code;
	optionptr[end + OPT_LEN] = length;

	switch (length) {
	case 1: buffer[0] = (char) data;
		break;
	case 2: *((u_int16_t *) buffer) = htons( data );
		break;
	case 4: *((u_int32_t *) buffer) = htonl( data );
		break;
	}
	memcpy( &optionptr[end + 2],buffer,length );
	optionptr[end + length + 2] = DHCP_END;

	return( length );
}

/* find option 'code' in opt_list */
struct option_set *
find_option( struct option_set *opt_list,char code )
{
	while (opt_list && opt_list->data[OPT_CODE] < code) {
		opt_list = opt_list->next;
	}

	if (opt_list && opt_list->data[OPT_CODE] == code) {
		return( opt_list );
	} else {
		return( NULL );
	}
}

/* add an option to the opt_list */
void
attach_option( struct option_set **opt_list,
               struct dhcp_option *option,
               char *buffer,
               int length )
{
	struct option_set	*existing;
	struct option_set	*new;
	struct option_set	**curr;

	/* add it to an existing option */
	if ((existing = find_option( *opt_list,option->code ))) {
		DEB( "attaching option %s to existing member of list\n",
		      option->name );
		if (option->flags & OPTION_LIST) {
			if (existing->data[OPT_LEN] + length <= 255) {
				existing->data = realloc( existing->data,existing->data[OPT_LEN] + length + 2 );
				memcpy( existing->data + existing->data[OPT_LEN] + 2,buffer,length );
				existing->data[OPT_LEN] += length;
			} /* else, ignore the data, we could put this in a second option in the future */
		} /* else, ignore the new data */
	} else {
		DEB( "attaching option %s to list",option->name );
		if (!(new = malloc( sizeof( struct option_set )))) {
			ERR( "attach_option(%s): malloc failed\n",option->name );
			return;
		}
		if (!(new->data = malloc( length+2 ))) {
			free( new );
			ERR( "attach_option(%s): malloc failed\n",option->name );
			return;
		}
		new->data[OPT_CODE] = option->code;
		new->data[OPT_LEN] = length;
		memcpy( new->data+2,buffer,length );

		curr = opt_list;
		while (*curr && (*curr)->data[OPT_CODE] < option->code) {
			curr = &(*curr)->next;
		}
		new->next = *curr;
		*curr = new;		
	}
}
