/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH> 
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * pidfile.c
 * ---------
 *
 * Initial Version	Thu May 16 17:36:09 CEST 2002 - jse
 * Rewritten based on uDhcpd
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

/* ************************************************************************* */
/* *** INCLUDES ************************************************************ */

#ifdef WITH_PIDFILE

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>

#include "debug.h"



/* ************************************************************************* */
/* *** GLOBAL FUNCTIONS **************************************************** */

int
pidfile_acquire( char *pidfile )
{
	int	pid_fd;

	if (pidfile == NULL) {
		return( -1 );
	}

	pid_fd = open( pidfile,O_CREAT|O_WRONLY,0644 );
	if (pid_fd < 0) {
		ERR( "unable to open pidfile '%s' errno=%s\n",
		      pidfile,strerror( errno ));
	} else {
		lockf( pid_fd,F_LOCK,0 );
	}

	return( pid_fd );
}


void
pidfile_write_release( int pid_fd )
{
	FILE	*out;

	if (pid_fd < 0) {
		return;
	}

	if ((out = fdopen( pid_fd,"w" )) != NULL) {
		fprintf( out,"%d\n",getpid() );
		fclose( out );
	}
	lockf( pid_fd,F_UNLCK,0 );
	close( pid_fd );
}

void
pidfile_delete( char *pidfile )
{
	if (pidfile) {
		unlink( pidfile );
	}
}

#endif	/* WITH_PIDFILE */
