/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH> 
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * socket.c
 * --------
 *
 * Initial Version	Thu May 16 17:36:09 CEST 2002 - jse
 * Rewritten based on uDhcpd
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

/* ************************************************************************* */
/* *** INCLUDES ************************************************************ */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <string.h>
#include <net/if.h>
#include <errno.h>
#include <features.h>
#if __GLIBC__ >=2 && __GLIBC_MINOR >= 1
#include <netpacket/packet.h>
#include <net/ethernet.h>
#else
#include <asm/types.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>
#endif


#include "debug.h"



/* ************************************************************************* */
/* *** GLOBAL FUNCTIONS **************************************************** */

int
listen_socket( unsigned int ip,int port,char *inf )
{
	static char		*_f = "listen_socket:";
	struct ifreq		interface;
	int			fd;
	struct sockaddr_in	addr;
	int			n = 1;

	if ((fd = socket( PF_INET,SOCK_DGRAM,IPPROTO_UDP )) < 0) {
		ERR( "%s socket failed errno=%s",_f,strerror( errno ));
		return( -1 );
	}
	
	memset( &addr,0,sizeof( addr ));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = ip;

	if (setsockopt( fd,SOL_SOCKET,SO_REUSEADDR,(char *)&n,sizeof(n)) == -1) {
		ERR( "%s setsockopt(SO_REUSEADDR) failed errno=%s\n",
		      _f,strerror( errno ));
		close( fd );
		return( -1 );
	}
	if (setsockopt( fd,SOL_SOCKET,SO_BROADCAST,(char *)&n,sizeof(n)) == -1) {
		ERR( "%s setsockopt(SO_BROADCAST) failed errno=%s\n",
		      _f,strerror( errno ));
		close( fd );
		return( -1 );
	}

	strncpy( interface.ifr_ifrn.ifrn_name,inf,IFNAMSIZ );
	if (setsockopt( fd,SOL_SOCKET,SO_BINDTODEVICE,(char *)&interface,sizeof( interface )) < 0) {
		ERR( "%s setsockopt(SO_BINDTODEVICE) failed errno=%s\n",
		      _f,strerror( errno ));
		close( fd );
		return( -1 );
	}

	if (bind( fd,(struct sockaddr *)&addr,sizeof(struct sockaddr)) == -1) {
		ERR( "%s bind failed errno=%s\n",_f,strerror( errno ));
		close( fd );
		return( -1 );
	}
	return( fd );
}
