/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH> 
 * Rewritten by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * ipportfw.c - Port Forwarding Table Editing
 * ------------------------------------------
 *
 * $Log: ipportfw.c,v $
 * Revision 1.3  2003-01-22 14:08:19  selent
 * rewritten to support port ranges
 *
 * Initial Version	Wed Jan 22 15:07:39 CET 2003 - jse
 * Rewritten
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

/*
 *
 *	portfw - Port Forwarding Table Editing v1.11
 *
 *	See the accompanying manual page portfw(8) for information
 *	about proper usage of this program. [ Not yet available ]
 *
 *
 *	Copyright (c) 1997 Steven Clarke
 *	All rights reserved.
 *
 *	Author: Steven Clarke <steven@monmouth.demon.co.uk>
 *
 *		Keble College
 *		Oxford
 *		OX1 3PG
 *
 *		WWW:    http://www.monmouth.demon.co.uk/
 *
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *	Change history:
 * 		1.11	Changed things around so that it would work
 *                      under GLIBC
 *              1.1     Renamed to ipportfw, added address specific forwarding
 *                      with the "-t" and "-u" options and added a new option "-L"
 *		1.0	Initial release
 *		
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef WITH_SYSLOG
#include <syslog.h>
#endif

#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#if 0 /* XXX -jse too much compiler warnings */
#include <netinet/ip_fw.h>
#else
#define IP_FW_BASE_CTL  	64	/* base for firewall socket options */
#define IP_FW_SHIFT		8	/* shift count for type (chain) */
#define IP_FW_INSERT		(IP_FW_BASE_CTL)
#define IP_FW_APPEND		(IP_FW_BASE_CTL+1)
#define IP_FW_DELETE		(IP_FW_BASE_CTL+2)
#define IP_FW_FLUSH		(IP_FW_BASE_CTL+3)
#define IP_FW_ZERO		(IP_FW_BASE_CTL+4)
#define IP_FW_POLICY		(IP_FW_BASE_CTL+5)
#define IP_FW_CHECK		(IP_FW_BASE_CTL+6)
#define IP_FW_MASQ_TIMEOUTS	(IP_FW_BASE_CTL+7)
#endif

typedef unsigned int __u32;
typedef unsigned short __u16;

/* These really shouldn't be here as they're kernel dependant
   but glibc leaves us no choice :-( */
#define IP_FW_PORTFW            6

#define IP_PORTFW_ADD           (IP_FW_APPEND | (IP_FW_PORTFW << IP_FW_SHIFT))
#define IP_PORTFW_DEL           (IP_FW_DELETE | (IP_FW_PORTFW << IP_FW_SHIFT))
#define IP_PORTFW_FLUSH         (IP_FW_FLUSH  | (IP_FW_PORTFW << IP_FW_SHIFT))

#define IP_PORTFW_PORT_MIN 1
#define IP_PORTFW_PORT_MAX 60999

struct ip_portfw_edits {
       __u16           protocol;       /* Which protocol are we talking? */
       __u32           laddr, raddr;   /* Local and remote address */
       __u16           lport, rport;   /* Local and remote port */
       __u16           dummy;          /* Make up to multiple of 4 */
};

#if 0
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/icmp.h>
#include <linux/if.h>
#include <linux/timer.h>
#include <linux/ip_fw.h>
#endif

#include <sys/param.h>

#define IP_PORTFW_NONE	0
#ifndef IPPROTO_NONE
#define IPPROTO_NONE	65535
#endif

long string_to_number( char *,int,int );
int parse_addressport( char *,__u32 *,__u16 *,__u16 * );
int do_setsockopt( int,struct ip_portfw_edits *,int );
void exit_error( int,char * );
void exit_display_help( void );

char *program;

int main(int argc, char *argv[])
{
	int			c;
	int			command = IP_PORTFW_NONE;
	unsigned short		lport;
	unsigned short		lportmax;
	unsigned short		rport;
	unsigned short		rportmax;
	struct ip_portfw_edits	pfw;

	pfw.protocol = IPPROTO_NONE;
	pfw.raddr = 0;
	rport = 0;
	rportmax = 0;
	lport = 0;
	lportmax = 0;

	program = argv[0];

#ifdef WITH_SYSLOG
	openlog( "ipportfw",LOG_PID,LOG_DAEMON );
#endif

	while ((c = getopt(argc, argv, "ADFhtul:r:")) != -1) {
		switch (c) {
		case 'A':
			if (command != IP_PORTFW_NONE) {
				exit_error( 2,"multiple commands specified" );
			}
			command = IP_PORTFW_ADD;
			break;

		case 'D':
			if (command != IP_PORTFW_NONE) {
				exit_error( 2,"multiple commands specified" );
			}
			command = IP_PORTFW_DEL;
			break;

		case 'F':
			if (command != IP_PORTFW_NONE) {
				exit_error( 2,"multiple commands specified" );
			}
			command = IP_PORTFW_FLUSH;
			break;

		case 't':
		case 'u':
			if (pfw.protocol != IPPROTO_NONE) {
				exit_error( 2,"multiple protocols specified" );
			}
			pfw.protocol = (c == 't' ? IPPROTO_TCP : IPPROTO_UDP);
			break;

		case 'l':
			if (pfw.laddr != 0 || lport != 0) {
				exit_error( 2,"multiple sources specified" );
			}
			if (parse_addressport( optarg,&pfw.laddr,&lport,&lportmax ) == -1) {
				exit_error( 2,"illegal local address/port specified" );
			}
			break;

		case 'r':
			if (pfw.raddr != 0 || rport != 0) {
				exit_error( 2,"multiple destinations specified" );
			}
			if (parse_addressport( optarg,&pfw.raddr,&rport,&rportmax ) == -1) {
				exit_error( 2,"illegal destination specified" );
			}
			break;

		default:
			exit_display_help();
		}
	}

        if (optind < argc) {
                exit_error( 2,"unknown arguments found on commandline" );
	}

	if (command == IP_PORTFW_NONE) {
		exit_display_help();
	} else if (command == IP_PORTFW_ADD && (pfw.protocol == IPPROTO_NONE || lport == 0 || rport == 0 || pfw.laddr == 0 || pfw.raddr == 0)) {
		exit_error( 2,"insufficient options specified" );
	} else if (command == IP_PORTFW_DEL && (pfw.protocol == IPPROTO_NONE || pfw.laddr == 0 || lport == 0)) {
		exit_error( 2,"insufficient options specified" );
	} else if (command == IP_PORTFW_DEL && (rport != 0 || pfw.raddr != 0)) {
		exit_error( 2,"incompatible options specified" );
	} else if (command == IP_PORTFW_FLUSH && (pfw.protocol != IPPROTO_NONE || lport != 0 || rport != 0 || pfw.laddr != 0 || pfw.raddr != 0)) {
		exit_error( 2,"incompatible options specified" );
	}

	if (lport != lportmax || rport != rportmax) {
		if ((lport > lportmax || rport > rportmax) ||
		    ((lportmax - lport) != (rportmax - rport)))
		{
			exit_error( 2,"incompatible portranges specified" );
		}
		for (;lport <= lportmax; lport++, rport++) {
			pfw.lport = lport;
			pfw.rport = rport;
			c = do_setsockopt( command,&pfw,sizeof( pfw ));
		}
	} else {
		pfw.lport = lport;
		pfw.rport = rport;
		c = do_setsockopt( command,&pfw,sizeof( pfw ));
	}
	exit( c );
}


long
string_to_number( char *str,int min,int max )
{
	char	*end;
	long	number;

	number = strtol( str,&end,10 );
	if (*end == '\0' && end != str) {
		if (min <= number && number <= max) {
                        return( number );
		} else {
                        return( -1 );
		}
        } else {
                return( -1 );
	}
}


int
parse_addressport( char *name,__u32 *addr,__u16 *port,__u16 *portmax )
{
	char		*b;
	char		*p;
	long		l;
	struct in_addr	in;

	if ((b = strchr( name,'/' )) == NULL) {
		return( -1 );
	}

	*b = '\0';
	if (inet_aton( name,&in ) == 0) {
		*b = '/';
		return( -1 );
	}
	*addr = htonl( in.s_addr );

	*b++ = '/';
	if (*b == '\0') {
		return( -1 );
	}

	if ((p = strchr( b,':' )) != NULL) {
		if (*(p+1) == '\0') {
			return( -1 );
		}
		*p = '\0';
		if ((l = string_to_number( b,IP_PORTFW_PORT_MIN,IP_PORTFW_PORT_MAX )) == -1) {
			*p = ':';
			return( -1 );
		}
		*port = l;
		*p++ = ':';
		if ((l = string_to_number( p,IP_PORTFW_PORT_MIN,IP_PORTFW_PORT_MAX )) == -1) {
			return( -1 );
		}
		*portmax = l;
	} else {
		if ((l = string_to_number( b,IP_PORTFW_PORT_MIN,IP_PORTFW_PORT_MAX )) == -1) {
			return( -1 );
		}
		*port = l;
		*portmax = l;
	}
	return( 0 );
}


int
do_setsockopt( int cmd,struct ip_portfw_edits *data,int length )
{
	static int	sockfd = -1;
	int		ret;

	if (sockfd == -1) {
		if ((sockfd = socket( AF_INET,SOCK_RAW,IPPROTO_RAW )) == -1) {
			exit_error( errno,"socket creation failed" );
		}
	}
	
	ret = setsockopt( sockfd,IPPROTO_IP,cmd,(char *)data,length );
	if (ret) {
#ifdef WITH_SYSLOG
		syslog( LOG_WARNING,"[WRN] setsockopt failed" );
#else
		perror( "ipportfw: setsockopt failed" );
#endif
	}
	return( ret );
}

void
exit_error( int status,char *msg )
{
#ifdef WITH_SYSLOG
	syslog( LOG_ERR,"[ERR] %s",msg );
	closelog();
#else
	fprintf( stderr,"%s: %s\n",program,msg );
#endif
	exit( status );
}

void
exit_display_help( void )
{
#ifdef WITH_SYSLOG
	syslog( LOG_NOTICE,"usage: ipportfw -[F|A|D] -[t|u] -l localip/p[:p] -r remoteip/p[:p]\n" );
	closelog();
#else
	fprintf( stderr,"usage: ipportfw -[F|A|D] -[t|u] -l localip/p[:p] -r remoteip/p[:p]\n" );
#endif
	exit( 0 );
}
