/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * blkmtd.c - block access to memory spaces via mtd layer
 * ------------------------------------------------------
 *
 * $RCSfile: blkmtd.c,v $
 * $Source: /opt/funkwerk/cvsroot/source/linux/uclinux2.0.38/arch/armnommu/drivers/block/blkmtd.c,v $
 * $Author: jselent $
 * $Revision: 1.4 $
 * $Date: 2008-05-19 06:44:53 $
 * $State: Exp $
 * $Name:  $
 *
 * $Log: blkmtd.c,v $
 * Revision 1.4  2008-05-19 06:44:53  jselent
 * CHG: copyright info
 *
 * Revision 1.3  2002-05-30 15:12:47  selent
 * changes for new flash layout
 *
 * Revision 1.2  2002/05/24 13:34:41  selent
 * added mtd support
 *
 * Revision 1.1  2002/05/21 18:35:31  selent
 * imported
 *
 * Initial Version	Tue May 21 16:16:53 CEST 2002 - jse
 * Written mainly based on blkmem.c & ftl.c
 */

/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */



/* ************************************************************************* */
/* *** INCLUDES ************************************************************ */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/major.h>
#include <linux/malloc.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/romfs_fs.h>
#include <linux/mtd.h>

#include <asm/segment.h>
#include <asm/byteorder.h>

#define MAJOR_NR	BLKMTD_MAJOR
#include <linux/blk.h>

#include "bsp.h"



/* ************************************************************************** */
/* *** DEBUG **************************************************************** */

#if 0
#define DEB(fmt, devname, args...) \
	printk( KERN_DEBUG/**/"%s: [DEB] "/**/fmt, devname, ##args );
#define INF(fmt, devname, args...) \
	printk( KERN_NOTICE/**/"%s: [INF] "/**/fmt, devname, ##args );
#else
#define DEB(fmt, devname, args...)
#define INF(fmt, devname, args...)
#endif

#define WRN(fmt, devname, args...) \
	printk( KERN_WARNING/**/"%s: [WRN] "/**/fmt, devname, ##args );
#define ERR(fmt, devname, args...) \
	printk( KERN_ERR/**/"%s: [ERR] "/**/fmt, devname, ##args );



/* ************************************************************************* */
/* *** DEFINES ************************************************************* */

#define BLOCKSIZE		1024

/* partition state flags */
#define ST_UNFORMATTED		0
#define ST_FORMATTED		1



/* ************************************************************************* */
/* *** TYPEDEFS ************************************************************ */

typedef struct {
	struct mtd_info	*mtd;
	unsigned long	offset;
	int		size;
	int 		state;
	int		refcnt;
	char		name[16];
} partition_t;



/* ************************************************************************* */
/* *** PROTOTYPES ********************************************************** */

static int	blkmtd_ioctl( struct inode *,struct file *,
			      unsigned int,unsigned long );
static int	blkmtd_open( struct inode *,struct file * );
static void	blkmtd_release( struct inode *,struct file * );
static void	blkmtd_notify( struct mtd_info *,int );



/* ************************************************************************* */
/* *** LOCAL DATA ********************************************************** */

static struct file_operations blkmtd_fops = {
	NULL,			/* lseek			*/
	block_read,		/* read - generic read		*/
	block_write,		/* write - generic write	*/
	NULL,			/* readdir			*/
	NULL,			/* select			*/
	blkmtd_ioctl,		/* ioctl			*/
	NULL,			/* mmap				*/
	blkmtd_open,		/* open				*/
	blkmtd_release		/* release			*/
};

static struct mtd_notifier	blkmtd_notifier = { blkmtd_notify,NULL };
static partition_t		partition_table[MAX_MTD_DEVICES];




/* ************************************************************************* */
/* *** LOCAL FUNCTIONS ***************************************************** */

static void
do_blkmtd_request( void )
{
	static char	*_f = "blkmtd_request:";
	partition_t	*pp;
	unsigned long	offset;
	unsigned long	len;
	unsigned long	rlen;
	int		major;
	int		minor;
	int		rc;

	while (1) {
		INIT_REQUEST;
		major = MAJOR( CURRENT->rq_dev );
		minor = MINOR( CURRENT->rq_dev );

		if (major != MAJOR_NR) {
			ERR( "%s #%d/%d wrong major\n",DEVICE_NAME,
			      _f,major,minor );
			end_request( 0 );
			continue;
		}

		if (minor >= MAX_MTD_DEVICES) {
			ERR( "%s #%d/%d wrong minor\n",DEVICE_NAME,
			      _f,major,minor );
			end_request( 0 );
			continue;
		}

		pp = &partition_table[minor];
		if (pp->state != ST_FORMATTED) {
			ERR( "%s no partition\n",pp->name,_f );
			end_request( 0 );
			continue;
		}

		if (CURRENT->cmd != READ) {
			ERR( "%s wrong command\n",pp->name,_f );
			end_request( 0 );
			continue;
		}

		offset = CURRENT->sector << 9;
		len  = CURRENT->current_nr_sectors << 9;

#if 0		/* XXX */
		if ((offset + len) > pp->size) {
			ERR( "%s wrong size\n",pp->name,_f );
			end_request( 0 );
			continue;
		}
#endif
		rc = MTD_READ( pp->mtd,pp->offset + offset,
		               len,&rlen,CURRENT->buffer,0 );
		if (rc || rlen != len) {
			ERR( "%s read failed rc=%d rlen=%ld len=%ld\n",
			      pp->name,_f,rc,rlen,len );
			end_request( 0 );
			continue;
		}
		end_request(1);
	}
}

static int
blkmtd_ioctl( struct inode *inode,
	      struct file *file,
	      unsigned int cmd,
	      unsigned long arg )
{
	static char	*_f = "blkmtd_ioctl:";
	partition_t	*pp;
	int		major;
	int		minor;
	int		rc;

	if (!inode) {
		ERR( "%s no inode given\n",DEVICE_NAME,_f );
		return( -EINVAL );
	}
	major = MAJOR( inode->i_rdev );
	minor = MINOR( inode->i_rdev );

	DEB( "%s #%d/%d called\n",DEVICE_NAME,_f,major,minor );

	if (major != MAJOR_NR) {
		ERR( "%s #%d/%d wrong major\n",DEVICE_NAME,_f,major,minor );
		return( -ENODEV );
	}

	if (minor >= MAX_MTD_DEVICES) {
		ERR( "%s #%d/%d wrong minor\n",DEVICE_NAME,_f,major,minor );
		return( -ENODEV );
	}
	pp = &partition_table[minor];
	if (pp->state != ST_FORMATTED) {
		return( -EINVAL );
	}

	switch (cmd) {
	case BLKGETSIZE:	/* Return device size */
		if (!arg) {
			ERR( "%s #%d/%d !arg\n",pp->name,_f,major,minor );
			return( -EINVAL );
		}
		rc = verify_area( VERIFY_WRITE,(long *)arg,sizeof( long ));
		if (rc != 0) {
			ERR( "%s #%d/%d verify_area failed\n",pp->name,
			      _f,major,minor );
			return( rc );
		}
		put_fs_long( pp->size >> 9,(long *)arg );
		break;
	default:
		WRN( "%s #%d/%d wrong command\n",pp->name,_f,major,minor );
		return( -EINVAL );
	}
	return( 0 );
}

static int
blkmtd_open( struct inode *inode,struct file *file )
{
	static char	*_f = "blkmtd_open:";
	partition_t	*pp;
	int		major;
	int		minor;

	if (!inode) {
		ERR( "%s no inode given\n",DEVICE_NAME,_f );
		return( -EINVAL );
	}
	major = MAJOR( inode->i_rdev );
	minor = MINOR( inode->i_rdev );

	DEB( "%s #%d/%d called\n",DEVICE_NAME,_f,major,minor );

	if (major != MAJOR_NR) {
		ERR( "%s #%d/%d wrong major\n",DEVICE_NAME,_f,major,minor );
		return( -ENODEV );
	}

	if (minor >= MAX_MTD_DEVICES) {
		ERR( "%s #%d/%d wrong minor\n",DEVICE_NAME,_f,major,minor );
		return( -ENODEV );
	}

	pp = &partition_table[minor];
	if (pp->state != ST_FORMATTED) {
		return( -ENODEV );
	}
	pp->refcnt ++;
	MOD_INC_USE_COUNT;
	return( 0 );
}

static void
blkmtd_release( struct inode *inode,struct file *file )
{
	static char	*_f = "blkmtd_release:";
	partition_t	*pp;
	int		major;
	int		minor;

	if (!inode) {
		ERR( "%s no inode given\n",DEVICE_NAME,_f );
		return;
	}
	major = MAJOR( inode->i_rdev );
	minor = MINOR( inode->i_rdev );

	DEB( "%s #%d/%d called\n",DEVICE_NAME,_f,major,minor );

	if (major != MAJOR_NR) {
		ERR( "%s #%d/%d wrong major\n",DEVICE_NAME,_f,major,minor );
		return;
	}

	if (minor >= MAX_MTD_DEVICES) {
		ERR( "%s #%d/%d wrong minor\n",DEVICE_NAME,_f,major,minor );
		return;
	}
	pp = &partition_table[minor];

	fsync_dev( inode->i_rdev );
	if (pp->state == ST_FORMATTED) {
		if (pp->refcnt > 0) {
			pp->refcnt --;
			MOD_DEC_USE_COUNT;
		}
	}
}

static void
blkmtd_adddev( partition_t *pp,struct mtd_info *mtd )
{
	static char			*_f = "blkmtd_adddev:";
	struct romfs_super_block	*rsb;
	unsigned long			offset;
	unsigned long			len;
	unsigned char			buf[16];
	int				rc;

	DEB( "%s add mtd device\n",pp->name,_f );

	if (pp->state == ST_FORMATTED) {
		return;
	}

	for (offset=0, rsb=NULL; offset<mtd->size; offset+=1024) {
		rc = MTD_READ( mtd,offset,sizeof( buf ),&len,buf,0 );
		if (rc || len != sizeof( buf )) {
			ERR( "%s reading partition\n",pp->name,_f );
			return;
		}
		rsb = (struct romfs_super_block *) buf;
		if (rsb->word0 == ROMSB_WORD0 && rsb->word1 == ROMSB_WORD1) {
			break;
		}
	}

	if (offset >= mtd->size) {
		ERR( "%s no RomFs partition found\n",pp->name,_f );
		return;
	}

	pp->mtd = mtd;
	pp->offset = offset;
	pp->size = ntohl( rsb->size );
	pp->state = ST_FORMATTED;

	INF( "%s found partition at offset 0x%lx size=%d\n",
	      pp->name,_f,offset,pp->size );
}

static void
blkmtd_remdev( partition_t *pp )
{
	DEB( "blkmtd_remdev: remove mtd device\n",pp->name );
	pp->mtd = NULL;
	pp->offset = 0;
	pp->size = 0;
	pp->state = ST_UNFORMATTED;
}

static void
blkmtd_notify( struct mtd_info *mtd,int dev )
{
	partition_t	*pp;

	if (dev >= MAX_MTD_DEVICES) {
		ERR( "blkmtd_notify: device number %d out of range\n",
		      DEVICE_NAME,dev );
		return;
	}
	pp = &partition_table[dev];

	if (mtd != NULL) {
		blkmtd_adddev( pp,mtd );
	} else {
		blkmtd_remdev( pp );
	}
}



/* ************************************************************************* */
/* *** GLOBAL FUNCTIONS **************************************************** */

int
blkmtd_init( void )
{
	static char	*_f = "blkmtd_init:";
	struct mtd_info	*mtd;
	int		i;

	DEB( "%s called\n",DEVICE_NAME,_f );

	if (register_blkdev( MAJOR_NR,DEVICE_NAME,&blkmtd_fops )) {
		ERR( "%s register_blkdev for major %d failed\n",DEVICE_NAME,
		      _f,MAJOR_NR );
		return( -EIO );
	}
	memset( partition_table,0,sizeof( partition_table ));

	printk( "%s: [INF] registered device at major %d\n",
	         DEVICE_NAME,MAJOR_NR );

	blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;

	register_mtd_notifier( &blkmtd_notifier );
	for (i=0; i< MAX_MTD_DEVICES; i++) {
		sprintf( partition_table[i].name,"%s%d",DEVICE_NAME,i );
		if ((mtd = get_mtd_device( i )) != NULL) {
			blkmtd_notify( mtd,i );
		}
	}

	return( 0 );
}

#ifdef MODULE
int
init_module( void )
{
	DEB( "init module\n",DEVICE_NAME );
	return( blkmtd_init() );
}

void
cleanup_module( void )
{
	int	rc;

	DEB( "cleanup module\n",DEVICE_NAME );
	if ((rc = unregister_blkdev( MAJOR_NR,DEVICE_NAME )) != 0) {
		ERR( "unregister_blkdev failed (%d)\n",DEVIVE_NAME,rc );
	}
}
#endif
