/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/config.h>
#include <linux/major.h>
#include <linux/blkdev.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/mm.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/malloc.h>

#include <asm/segment.h>
#include <asm/arch/irqs.h>
#include <asm/arch/irq.h>
#include <asm/proc/system.h>
#include <asm/arch/hardware.h>
#include <display.h>

#include "cf.h"
#include <linux/cfioctl.h>
#include "ide.h"

#define DEBUG_CF 0

#if DEBUG_CF
#define DEBOUT(X) OUTPUT("cf", X)
#define ERROUT(X) OUTPUT("cf_err", X)
#define REGOUT(X)
#else
#define DEBOUT(X)
#define ERROUT(X) OUTPUT("cf_err", X)
#define REGOUT(X)
#endif

#define OUTPUT(X,Y) do { \
			printk("%s (%d): ", X, (int) __LINE__); \
			printk Y; \
		    } while (0)

#if DEBUG_CF
#warning CF being compiled
#endif

unsigned short wr_control_shadow;

static inline void write_wr_control(unsigned short val)
{
	wr_control_shadow = val;
	CF_WR_CONTROL = val;
}

#if 0

/* TODO Assembly draus machen */

unsigned short read_one(unsigned char addr)
{
	unsigned long flags;
	unsigned short shadow;
	unsigned short data;
	volatile unsigned short dummy;

	if (wr_control_shadow != 0x1F8) {
		ERROUT(("read_one called while shadow = 0x%04X\n",
			wr_control_shadow));
		return (unsigned short) -1;
	}
	if (addr & 0xE0) {
		ERROUT(("read_one called with illegal addr: 0x%02X\n",
			addr));
		return (unsigned short) -1;
	}
	save_flags_clif(flags);
	shadow = 0x1E0 | addr;
	CF_WR_CONTROL = shadow;
	shadow &= ~0x80;
	CF_WR_CONTROL = shadow;
	read_nops;
#if 0
	dummy = CF_RD_DATA;
	data = CF_LATCH_RD_DATA;
#else
	dummy = CF_LATCH_RD_DATA;
	data = CF_RD_DATA;
#endif
	shadow |= 0x80;
	CF_WR_CONTROL = shadow;
	CF_WR_CONTROL = 0x1F8;
	restore_flags(flags);
	REGOUT(("read_one from 0x%02X returns 0x%04X\n", addr, data));
	return data;
}

/* TODO Assembly draus machen */

void write_one(unsigned char addr, unsigned short data)
{
	unsigned long flags;
	unsigned short shadow;

	if (wr_control_shadow != 0x1F8) {
		ERROUT(("write_one called while shadow = 0x%04X\n",
			wr_control_shadow));
		return;
	}
	if (addr & 0xE0) {
		ERROUT(("write_one called with illegal addr: 0x%02X\n",
			addr));
		return;
	}
	REGOUT(("write_one: 0x%04X to 0x%02X\n", data, addr));
	save_flags_clif(flags);
	shadow = 0x1E0 | addr;
	CF_WR_CONTROL = shadow;
	CF_LATCH_WR_DATA = data;
	shadow &= ~0x40;
	CF_WR_CONTROL = shadow;
	shadow &= ~0x20;
	CF_WR_CONTROL = shadow;
	write_nops;
	shadow |= 0x40;
	CF_WR_CONTROL = shadow;
	CF_WR_CONTROL = 0x1F8;
	restore_flags(flags);
}

#endif

static struct slot_data sd_global;

static void reset_timer_cb(unsigned long arg);
static void activate_timer_cb(unsigned long arg);

static void init_slot_data(struct slot_data *sd)
{
	sd->empty = 1;
	sd->index = -1;
	sd->active = 0;
	init_timer(&sd->reset_timer);
	sd->reset_timer.data = (unsigned long) sd;
	sd->reset_timer.function = reset_timer_cb;
	init_timer(&sd->activate_timer);
	sd->activate_timer.data = (unsigned long) sd;
	sd->activate_timer.function = activate_timer_cb;
	sd->cf_wait = NULL;
}

static void activate_timer_cb(unsigned long arg)
{
	struct slot_data *sd = (struct slot_data *) arg;

	DEBOUT(("Setting card active\n"));
	sd->active = 1;
	wake_up_interruptible(&sd->cf_wait);
}

static void reset_timer_cb(unsigned long arg)
{
	struct slot_data *sd = (struct slot_data *) arg;

	DEBOUT(("Disabling reset\n"));
	write_wr_control(0x1F8); /* Disable Reset */
	sd->activate_timer.expires = jiffies + HZ / 2;
	add_timer(&sd->activate_timer);
#if 0
	DEBOUT(("Calling ide_register\n"));
	if ((sd->index = ide_register_cf(0x10, 0xE, IRQ_GPIO)) < 0) {
		ERROUT(("ide_register failed\n"));
	} else {
		DEBOUT(("ide_register succeeds\n"));
	}
#endif
}

static void activate_card_slot(struct slot_data *sd)
{
	write_wr_control(0x0F8); /* Strom an */
/*	sd->reset_timer.expires = jiffies + HZ / 4; */
	sd->reset_timer.expires = jiffies + HZ / 2;
	add_timer(&sd->reset_timer);
}

static void deactivate_card_slot(struct slot_data *sd)
{
	del_timer(&sd->reset_timer);
	del_timer(&sd->activate_timer);
	sd->active = 0;
	wake_up_interruptible(&sd->cf_wait);
	write_wr_control(0x2F8); /* Strom aus */
}

static struct timer_list card_timer;

static void card_timer_cb(unsigned long arg)
{
	unsigned stat = CF_RD_STATUS;
	struct slot_data *sd = &sd_global;

	if (sd->empty && ((stat & 3) == 0)) {
		sd->empty = 0;
		DEBOUT(("CF card detected.\n"));
		DISPLAY_CF_STATE(0, DISP_CF_READY);
		activate_card_slot(sd);
	} else if (!sd->empty && (stat & 3)) {
		sd->empty = 1;
		DEBOUT(("CF card removed.\n"));
		DISPLAY_CF_STATE(0, DISP_CF_NOTREADY);
		deactivate_card_slot(sd);
	}
	card_timer.expires = jiffies + HZ / 4;
	add_timer(&card_timer);
}

static void start_card_timer(void)
{
	init_timer(&card_timer);
	card_timer.data = 0;
	card_timer.function = card_timer_cb;
	card_timer_cb(0);
}
#ifdef MODULE
static void stop_timers(void)
{
	del_timer(&card_timer);
	del_timer(&sd_global.reset_timer);
	del_timer(&sd_global.activate_timer);
}
#endif


static int cf_open(struct inode *ino, struct file *filep)
{
	struct slot_data *sd = &sd_global;

	MOD_INC_USE_COUNT;
	filep->private_data = sd;
	    /* HACK LED Update (Modul ist evtl. erst jetzt geladen */
	if (!sd->empty) {
		DISPLAY_CF_STATE(0, DISP_CF_READY);
	} else {
		DISPLAY_CF_STATE(0, DISP_CF_NOTREADY);
	}
	return 0;
}

static void cf_close(struct inode *ino, struct file *filep)
{
	MOD_DEC_USE_COUNT;
}

static int cf_select(struct inode *inode, struct file *filep,
		     int sel_type, select_table *wait)
{
	struct slot_data *sd = filep->private_data;
	unsigned long flags;

	save_flags_cli(flags);
	switch (sel_type) {
	case SEL_IN: /* Wait for card */
		if (sd->active) {
			restore_flags(flags);
			return 1;
		}
		select_wait(&sd->cf_wait, wait);
		restore_flags(flags);
		break;
	case SEL_OUT: /* Wait for empty */
		if (!sd->active) {
			restore_flags(flags);
			return 1;
		}
		select_wait(&sd->cf_wait, wait);
		restore_flags(flags);
		break;
	default:
		break;
	}
	return 0;
}

static int wait_for_card(struct slot_data *sd, struct file *filep,
			 int active)
{
	struct wait_queue wait = { current, NULL };
	int retval = 0;

	add_wait_queue(&sd->cf_wait, &wait);
	for (;;) {
		current->state = TASK_INTERRUPTIBLE;
		if (!active == !sd->active)
			break;
		if (filep->f_flags & O_NONBLOCK) {
			retval = -EAGAIN;
			break;
		}
		if (current->signal & ~current->blocked) {
			retval = -ERESTARTSYS;
			break;
		}
		schedule();
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&sd->cf_wait, &wait);
	return retval;
}

static int cf_ioctl(struct inode *ino, struct file *filep,
		     unsigned int cmd, unsigned long arg)
{
	struct slot_data *sd = filep->private_data;

	switch (cmd) {
	case CFIOCTL_READ_STAT:
	{
		struct cf_regs *parg = (struct cf_regs *) arg;
		struct cf_regs loc;
		unsigned short val;

		if (sd_global.empty)
			return -ENODEV;
		if (verify_area(VERIFY_WRITE, parg, sizeof *parg))
			return -EINVAL;
		val = CF_RD_STATUS;

		loc.cf_status = val & 0xFF;
#if 0
		loc.error = read_one(CFLOG_ERROR);
		loc.feature = read_one(CFLOG_FEATURES);
		loc.status = read_one(CFLOG_STATUS);
#endif
		memcpy_tofs(parg, &loc, sizeof *parg);
		break;
	}
	case CFIOCTL_REGISTER:
		if (!sd->active)
			return -ENODEV;
		DEBOUT(("Calling ide_register\n"));
		if ((sd->index = ide_register_cf(0x10, 0xE, IRQ_GPIO)) < 0) {
			ERROUT(("ide_register failed\n"));
			return -EIO;
		} else {
			DEBOUT(("ide_register succeeds\n"));
		}
		break;
	case CFIOCTL_UNREGISTER:
		if (sd->index >= 0) {
			DEBOUT(("Calling ide_unregister\n"));
			ide_unregister_cf(sd->index);
			sd->index = -1;
		} else {
			return -ENODEV;
		}
		break;
	case CFIOCTL_WAIT_CARD:
		return wait_for_card(sd, filep, 1);
	case CFIOCTL_WAIT_EMPTY:
		return wait_for_card(sd, filep, 0);
	case CFIOCTL_PROBE_CARD:
	{
		int *parg = (int *) arg;
		int card;

		if (verify_area(VERIFY_WRITE, parg, sizeof *parg))
			return -EINVAL;
		if (!sd->active) {
			card = 0;
		} else {
			unsigned stat = CF_RD_STATUS;

			if (stat & 3) {
				card = 0;
			} else {
				card = 1;
			}
		}
		memcpy_tofs(parg, &card, sizeof *parg);
		break;
	}
	default:
		return -EINVAL;
	}
	return 0;
}

static struct file_operations cf_fops = {
        NULL,		/* lseek */
	NULL,		/* read */
	NULL,		/* write */
	NULL,		/* readdir */
	cf_select,      /* select */
	cf_ioctl,	/* ioctl */
	NULL,		/* mmap */
        cf_open,
        cf_close	/* release */
};

static void init_pio_irq(void)
{
	CF_CLEAR_IRQ();
}

#ifdef MODULE
int init_module(void)
#else
int cf_init(void)
#endif
{
	if (register_chrdev(CFDEV_MAJOR, "cf", &cf_fops)) {
		printk(KERN_ERR"cf: [ERR] unable to get major %d for cf device\n",
		       CFDEV_MAJOR);
		return -EIO;
	}
#if 1
	write_wr_control(0x2F8);
	init_slot_data(&sd_global);
	init_pio_irq();
	start_card_timer();
#endif
	return 0;
}

#ifdef MODULE
void cleanup_module( void )
{
	stop_timers();
}
#endif
