/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

#ifndef CF_H_INCLUDED__
#define CF_H_INCLUDED__

#define CF_BASE 0x8000000

#define CF_LATCH_RD_DATA (*(volatile unsigned short *) (CF_BASE))
#define CF_RD_DATA (*(volatile unsigned short *) (CF_BASE + 0x80))
#define CF_RD_STATUS (*(volatile unsigned short *) (CF_BASE + 0x100))
#define CF_LATCH_WR_DATA (*(volatile unsigned short *)(CF_BASE))
#define CF_WR_CONTROL (*(volatile unsigned short *) (CF_BASE + 0x80))

struct slot_data {
	int empty;
	int index;
	int active;
	struct timer_list reset_timer;
	struct timer_list activate_timer;
	struct wait_queue *cf_wait;
};

#define save_flags_clif(x)		\
	do {				\
	  unsigned long temp;		\
	  asm volatile (		\
	"mrs	%1, cpsr\n"		\
"	and	%0, %1, #192\n"		\
"	orr	%1, %1, #192\n"		\
"	msr	cpsr, %1"		\
	  : "=&r" (x), "=r" (temp)	\
	  :				\
	  : "memory");			\
	} while (0)


#define read_nops \
	asm volatile ( \
	"nop\n" \
"	nop\n" \
"	nop\n" \
"	nop\n" \
"	nop\n" \
"	nop\n" \
"	nop\n" \
"	nop")

#define write_nops read_nops

#define CFLOG_RD_DATA		0x10
#define CFLOG_WR_DATA		0x10
#define CFLOG_ERROR		0x11
#define CFLOG_FEATURES		0x11
#define CFLOG_SECTOR_CNT	0x12
#define CFLOG_SECTOR_NR		0x13
#define CFLOG_CYL_LOW		0x14
#define CFLOG_CYL_HIGH		0x15
#define CFLOG_SELECT		0x16
#define CFLOG_STATUS		0x17
#define CFLOG_COMMAND		0x17
#define CFLOG_ALT_STATUS	0x0E
#define CFLOG_DEV_CTRL		0x0E
#define CFLOG_DRIVE_ADDR	0x0F

#if 0
unsigned short read_one(unsigned char addr);
void write_one(unsigned char addr, unsigned short data);
#endif

#define CF_PIO_PIN 10
#define CF_PIO_PIN_PRESENT 11
#define CF_PIO_PIN_IORDY 12

#define CF_SET_BIT(addr, bit) do { \
				volatile unsigned long old; \
				old = inl_t(addr); \
				old |= (1 << (bit)); \
				outl_t(old, addr); \
				} while (0)

#define CF_CLR_BIT(addr, bit) do { \
				volatile unsigned long old; \
				old = inl_t(addr); \
				old &= ~(1 << (bit)); \
				outl_t(old, addr); \
				} while (0)

#define CF_CLEAR_IRQ() do { \
			CF_CLR_BIT(ABS_Adr_GPIO_Config, CF_PIO_PIN); \
			CF_CLR_BIT(ABS_Adr_GPIO_INT, CF_PIO_PIN); \
			CF_CLR_BIT(ABS_Adr_GPIO_INT, CF_PIO_PIN_PRESENT); \
			CF_CLR_BIT(ABS_Adr_GPIO_INT, CF_PIO_PIN_IORDY); \
			CF_CLR_BIT(ABS_Adr_GPIO_En, CF_PIO_PIN); \
			CF_CLR_BIT(ABS_Adr_GPIO_En, CF_PIO_PIN_PRESENT); \
			CF_CLR_BIT(ABS_Adr_GPIO_En, CF_PIO_PIN_IORDY); \
 			} while (0)

#define CF_ENA_IRQ() do { \
			CF_CLR_BIT(ABS_Adr_GPIO_En, CF_PIO_PIN); \
			CF_SET_BIT(ABS_Adr_GPIO_INT, CF_PIO_PIN + 16); \
			CF_CLR_BIT(ABS_Adr_GPIO_INT, CF_PIO_PIN); \
			CF_SET_BIT(ABS_Adr_GPIO_Config, CF_PIO_PIN); \
			} while (0)

extern unsigned short wr_control_shadow;

static inline unsigned short read_one(unsigned char addr)
{
	unsigned long flags;
	unsigned short shadow;
	unsigned short data;
	volatile unsigned short dummy;
#if 0
	if (wr_control_shadow != 0x1F8) {
		ERROUT(("read_one called while shadow = 0x%04X\n",
			wr_control_shadow));
		return (unsigned short) -1;
	}
	if (addr & 0xE0) {
		ERROUT(("read_one called with illegal addr: 0x%02X\n",
			addr));
		return (unsigned short) -1;
	}
#endif
	save_flags_clif(flags);
	shadow = 0x1E0 | addr;
	CF_WR_CONTROL = shadow;
	shadow &= ~0x80;
	CF_WR_CONTROL = shadow;
	read_nops;
#if 0
	dummy = CF_RD_DATA;
	data = CF_LATCH_RD_DATA;
#else
	dummy = CF_LATCH_RD_DATA;
	data = CF_RD_DATA;
#endif
	shadow |= 0x80;
	CF_WR_CONTROL = shadow;
	CF_WR_CONTROL = 0x1F8;
	restore_flags(flags);
	return data;
}

/* TODO Assembly draus machen */

static inline void write_one(unsigned char addr, unsigned short data)
{
	unsigned long flags;
	unsigned short shadow;
#if 0
	if (wr_control_shadow != 0x1F8) {
		ERROUT(("write_one called while shadow = 0x%04X\n",
			wr_control_shadow));
		return;
	}
	if (addr & 0xE0) {
		ERROUT(("write_one called with illegal addr: 0x%02X\n",
			addr));
		return;
	}
#endif
	save_flags_clif(flags);
	shadow = 0x1E0 | addr;
	CF_WR_CONTROL = shadow;
	CF_LATCH_WR_DATA = data;
	shadow &= ~0x40;
	CF_WR_CONTROL = shadow;
	shadow &= ~0x20;
	CF_WR_CONTROL = shadow;
	write_nops;
	shadow |= 0x40;
	CF_WR_CONTROL = shadow;
	CF_WR_CONTROL = 0x1F8;
	restore_flags(flags);
}

#endif /* CF_H_INCLUDED__ */
