/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

#include <linux/module.h>

#include <linux/sysdep-2.1.h>

#ifdef LINUX21
#include <asm/uaccess.h>
#endif
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/major.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <stdarg.h>
#include <linux/delay.h>

#include <linux/mtd.h>

#undef MONSTER_DEBUG

/*
 * Debug macros
 */
#ifdef MONSTER_DEBUG
#define DEB(fmt, args...) \
	printk( KERN_DEBUG/**/"am29: [DEB] "/**/fmt, ##args );
#else
#define DEB(fmt, args...)
#endif
#define INF(fmt, args...) \
	printk( KERN_NOTICE/**/"am29: [INF] "/**/fmt, ##args );
#define WRN(fmt, args...) \
	printk( KERN_WARNING/**/"am29: [WRN] "/**/fmt, ##args );
#define ERR(fmt, args...) \
	printk( KERN_ERR/**/"am29: [ERR] "/**/fmt, ##args );


#define SECERASE_CMD		0x30
#define CHIPERASE_CMD		0x10
#define PROG_CMD		0xA0
#define SETUP_CMD		0x80
#if 1
#define PROG_ADDR		(0x5555*2)
#define SETUP_ADDR		(0x5555*2)
#define ID_ADDR			(0x5555*2)
#define UNLOCK_ADDR1		(0x5555*2)
#define UNLOCK_ADDR2		(0x2AAA*2)
#else
#define PROG_ADDR		(0x555*2)		/* XXX jse */
#define SETUP_ADDR		(0x555*2)
#define ID_ADDR			(0x555*2)
#define UNLOCK_ADDR1		(0x555*2)
#define UNLOCK_ADDR2		(0x2AA*2)
#endif
#define UNLOCK_CMD1		0xAA
#define UNLOCK_CMD2		0x55
#define ERASE_SUSPEND_CMD	(0xB0)
#define ERASE_RESUME_CMD	(0x30)
#define RESET_CMD		(0xF0)
#define ID_CMD			(0x90)

#define DQ2			0x04
#define DQ3			(DQ2*2)
#define DQ5			(DQ3*4)
#define DQ6			(DQ5*2)


#define WRITE_UNLOCK(a) { *(volatile __u16*)(a+UNLOCK_ADDR1) = UNLOCK_CMD1; *(volatile __u16*)(a+UNLOCK_ADDR2) = UNLOCK_CMD2; }

#define MAX_FLASH_DEVICES	1

/*
 *	Assume that 'normal' flash sectors have 0x10000 bytes and
 *	has only one continous region of special sectors.
 *	If this is not true, the soft must be modified
 */	

struct special_sector{
	unsigned long addr;
	int size;
};

static struct special_sector am2916bot[5]={
	{0x0, 0x4000},
	{0x4000, 0x2000},
	{0x6000, 0x2000},
	{0x8000, 0x8000},
	{0,0}
};

static struct special_sector am2916top[5]={
	{0x1f0000, 0x8000},
	{0x1f8000, 0x2000},
	{0x1fa000, 0x2000},
	{0x1fc000, 0x4000},
	{0,0}
};

#if 0
static struct special_sector at4916bot[11]={
	{0x0, 0x2000},
	{0x2000, 0x2000},
	{0x4000, 0x2000},
	{0x6000, 0x2000},
	{0x8000, 0x2000},
	{0xa000, 0x2000},
	{0xc000, 0x2000},
	{0xe000, 0x2000},
	{0x10000, 0x8000},
	{0x18000, 0x8000},
	{0,0}
};
	
static struct special_sector at4916top[11]={
	{0x1e0000, 0x8000},
	{0x1e8000, 0x8000},
	{0x1f0000, 0x2000},
	{0x1f2000, 0x2000},
	{0x1f4000, 0x2000},
	{0x1f6000, 0x2000},
	{0x1f8000, 0x2000},
	{0x1fa000, 0x2000},
	{0x1fc000, 0x2000},
	{0x1fe000, 0x2000},
	{0,0}
};
#endif

#define AMD			1
#define AMD29LV160Bottom	0x49
#define AMD29LV160Top		0xc4

#define AMD29F800Bottom		0xd6
#define AMD29F800Top		0x58
	
#define ATMEL			0x1f
#define AT49BV16X4Bottom	0xc0
#define AT49BV16X4Top		0xc2

#define TOSHIBA			0x98	
#define	TC58FVBottom160FT	0x43	
#define	TC58FVTop160FT		0xc2	

#define FUJITSU			0x04
#define MBM29LV160Bottom	0x49
#define MBM29LV160Top		0xc4

#define ST			0x20
#define M29W3200Bottom		0xcb
#define M29W160Bottom		0x49
#define M29W160Top		0xc4
#define M29W800Bottom		0x5b
#define M29W800Top		0xd7

#define HYNIX			0xad
#define HY29LV320Bottom		0x7d
#define HY29LV320Top		0x7e

#define EXCEL			0x4a
#define ES29LV160Bottom		0x49
#define ES29LV160Top		0xc4
#define ES29LV800Bottom		0x5b
#define ES29LV800Top		0xda

typedef struct {
	unsigned long		base;
	unsigned long		special_base;
	unsigned long		special_base2;
	unsigned long		normal_base;
	unsigned long		dev_size;
	unsigned long		cell_size;
	unsigned short		flash_mId;
	unsigned short		flash_devId;
	struct special_sector	*smap;
	int			bottom;
	int			erase_timeout;
	int			retry_limit;
	int			sect_nb;
	int			start_index;
	int			step;
	char			chip_name[64];
	__u16			toggling_bits;
} flash_desc_t;


/* A flash region is composed of one or more "cells", where we allow
   simultaneous erases if they are in different cells */

struct mypriv {
	unsigned char		*baseAddr;
	flash_desc_t		*flashDesc;
	int			numDevices;
	int			simulation;
	struct erase_info	*cur_erases;
	struct wait_queue	*wq;
#if 0
	unsigned short		curpage;
	unsigned char		locked;
	unsigned char		interleave;
#endif
	unsigned char		devstat[MAX_FLASH_DEVICES];
	struct timer_list	flashmem_timer;
};

static struct mtd_info *mymtd[MAX_MTD_DEVICES];
static flash_desc_t	flashDesc[MAX_MTD_DEVICES];






static void flashmem_periodic(unsigned long data);
static int flashmem_erase (struct mtd_info *mtd, struct erase_info *instr);
static int flashmem_point (struct mtd_info *mtd, unsigned long from, unsigned long len, unsigned long *retlen, unsigned char **mtdbuf);
static void flashmem_unpoint (struct mtd_info *mtd, unsigned long addr);
static int flashmem_read (struct mtd_info *mtd, unsigned long from, unsigned long len, unsigned long *retlen, unsigned char *buf, int userspace);
static int flashmem_write (struct mtd_info *mtd, unsigned long to, unsigned long len, unsigned long *retlen, const unsigned char *buf, int userspace);
static void flashmem_openclose (struct mtd_info *mtd, int which);
static void flashmem_sync (struct mtd_info *mtd);

static inline void resume_erase(volatile unsigned char *addr,int simulation);
static inline int suspend_erase(volatile unsigned char *addr,int simulation);
static inline int byte_write (struct mypriv *priv,volatile unsigned char *addr, unsigned char byte);
static inline int word_write (struct mypriv *priv,volatile unsigned char *addr, __u16 word);
static inline int check_erase(struct mypriv *priv,volatile unsigned char *addr, unsigned long size);

#define pagein(mtd, addr) (((struct mypriv*)mtd->priv)->baseAddr + (unsigned long)(addr))


static int
flashmem_getId( flash_desc_t *fp,int simulation )
{
	if (simulation) {
#ifdef PROJECT_VARIANT_4MB_FLASH
		fp->flash_mId = HYNIX;
		fp->flash_devId = HY29LV320Bottom;
#else
		fp->flash_mId = AMD;
		fp->flash_devId = AMD29LV160Bottom;
#endif
		fp->toggling_bits = DQ2 | DQ6;
	} else {
		WRITE_UNLOCK( fp->base );
		*(__u16*)(fp->base+ID_ADDR) = ID_CMD;
		fp->flash_mId = *(__u16*)(fp->base);
		*(__u16*)(fp->base+ID_ADDR) = RESET_CMD;
		
		WRITE_UNLOCK( fp->base );
		*(__u16*)(fp->base+ID_ADDR) = ID_CMD;
		fp->flash_devId = *(__u16*)(fp->base+2) & 0xff;
		*(__u16*)(fp->base+ID_ADDR) = RESET_CMD;
		INF( "flashid: 0x%x, 0x%x\n",fp->flash_mId,fp->flash_devId );
		if (fp->flash_mId == TOSHIBA) {
			fp->toggling_bits = (DQ6);
		} else {
			fp->toggling_bits = (DQ2|DQ6);
		}
		INF( "toggling_bits=0x%x\n",fp->toggling_bits );
	}
	memset( fp->chip_name,0,sizeof( fp->chip_name ));

	switch (fp->flash_mId) {
	case AMD:
	case TOSHIBA:
	case FUJITSU:
	case ST:
	case EXCEL:
		switch (fp->flash_devId) {
		case M29W3200Bottom:
			fp->bottom = 1;
			fp->smap = am2916bot;
			fp->special_base = fp->base;
			fp->special_base2 = fp->base;
			fp->normal_base = fp->base + 0x10000;
			fp->dev_size = 4096 * 1024;	/* 4M */
			fp->cell_size = 64 * 1024;
			strcpy( fp->chip_name,"M29W3200Bottom" );
			break;
		case AMD29LV160Bottom:
		case TC58FVBottom160FT:	 
		/*case MBM29LV160Bottom:	duplicate case value */
		/*case M29W160Bottom:		duplicate case value */
		/*case ES29LV160Bottom:		duplicate case value */
			fp->bottom = 1;
			fp->smap = am2916bot;
			fp->special_base = fp->base;
			fp->special_base2 = fp->base;
			fp->normal_base = fp->base + 0x10000;
			fp->dev_size = 2048 * 1024;	/* 2M */
			fp->cell_size = 64 * 1024;
			switch (fp->flash_mId) {
			case AMD:
				strcpy( fp->chip_name,"AMD29LV160Bottom" );
				break;
			case TOSHIBA:
				strcpy( fp->chip_name,"TC58FVBottom160FT" );
				break;
			case FUJITSU:
				strcpy( fp->chip_name,"MBM29LV160Bottom" );
				break;
			case ST:
				strcpy( fp->chip_name,"M29W160Bottom" );
				break;
			case EXCEL:
				strcpy( fp->chip_name,"ES29LV160Bottom" );
				break;
			}
			break;

		case AMD29LV160Top:
		case TC58FVTop160FT: 
		/*case MBM29LV160Top:		duplicate case value */
		/*case M29W160Top:		duplicate case value */
		/*case ES29LV160Top:		duplicate case value */
			fp->bottom = 0;
			fp->smap = am2916top;
			fp->special_base = fp->base + 0x1F0000;
			fp->special_base2 = fp->base + 0x1F0000;
			fp->normal_base = fp->base;
			fp->dev_size = 2048 * 1024;	/* 2M */
			fp->cell_size = 64 * 1024;
			switch (fp->flash_mId) {
			case AMD:
				strcpy( fp->chip_name,"AMD29LV160Top" );
				break;
			case TOSHIBA:
				strcpy( fp->chip_name,"TC58FVTop160FT" );
				break;
			case FUJITSU:
				strcpy( fp->chip_name,"MBM29LV160Top" );
				break;
			case ST:
				strcpy( fp->chip_name,"M29W160Top" );
				break;
			case EXCEL:
				strcpy( fp->chip_name,"ES29LV160Top" );
				break;
			}
			break;

		case M29W800Bottom:
		/*case ES29LV800Bottom:		duplicate case value */
			fp->bottom = 1;
			fp->smap = am2916bot;
			fp->special_base = fp->base;
			fp->special_base2 = fp->base;
			fp->normal_base = fp->base + 0x10000;
			fp->dev_size = 1024 * 1024;	/* 1M */
			fp->cell_size = 64 * 1024;
			switch (fp->flash_mId) {
			case ST:
				strcpy( fp->chip_name,"M29W800Bottom" );
				break;
			case EXCEL:
				strcpy( fp->chip_name,"ES29LV800Bottom" );
				break;
			}
			break;

		case M29W800Top:
		case ES29LV800Top:
			fp->bottom = 0;
			fp->smap = am2916top;
			fp->special_base = fp->base + 0xf0000;
			fp->special_base2 = fp->base + 0xf0000;
			fp->normal_base = fp->base;
			fp->dev_size = 1024 * 1024;	/* 1M */
			fp->cell_size = 64 * 1024;
			switch (fp->flash_mId) {
			case ST:
				strcpy( fp->chip_name,"M29W800Top" );
				break;
			case EXCEL:
				strcpy( fp->chip_name,"ES29LV800Top" );
				break;
			}
			break;

		default:
			ERR( "unknown flash organization\n" );
			return( ENODEV );
		}
		break;
#if 0
	case ATMEL:
		fp->dev_size = XXX;
		fp->cell_size = 64 * 1024;
		switch (fp->flash_devId) {
		case AT49BV16X4Bottom:	
			fp->bottom = 1;
			fp->smap = at4916bot;
			fp->special_base = fp->base;
			fp->special_base2 = fp->base + 0x10000;
			fp->normal_base = fp->base + 0x20000;
			strcpy( fp->chip_name,"Atmel29BV16X4Bottom");
			break;

		case AT49BV16X4Top:	
			fp->bottom = 0;
			fp->smap = at4916top;
			fp->special_base2 = fp->base + 0x1E0000;
			fp->special_base = fp->base + 0x1F0000;
			fp->normal_base = fp->base;
			strcpy( fp->chip_name,"Atmel29BV16X4Top" );
			break;
		}
		break;
#endif
	case HYNIX:
		switch (fp->flash_devId) {
		case HY29LV320Bottom:
			fp->bottom = 1;
			fp->smap = am2916bot;
			fp->special_base = fp->base;
			fp->special_base2 = fp->base;
			fp->normal_base = fp->base + 0x10000;
			fp->dev_size = 4096 * 1024;	/* 4M */
			fp->cell_size = 64 * 1024;
			strcpy( fp->chip_name,"HY29LV320Bottom" );
			break;
		case HY29LV320Top:
			fp->bottom = 0;
			fp->smap = am2916top;
			fp->special_base = fp->base + 0x3f0000;
			fp->special_base2 = fp->base + 0x3f0000;
			fp->normal_base = fp->base;
			fp->dev_size = 4096 * 1024;	/* 4M */
			fp->cell_size = 64 * 1024;
			strcpy( fp->chip_name,"HY29LV320Top" );
			break;
		default:
			ERR( "unknown flash organization\n" );
			return( ENODEV );
		}
		break;
	default:
		ERR( "unknown flash\n" );
		return( ENODEV );
	}

	/*
	 * find special sectors total nb and the index of
	 * first segment of second region
	 */
	INF( "special_base=0x%lx special_base2=0x%lx\n",
	      fp->special_base,fp->special_base2 );
	INF( "special sectors map:\n" );
	fp->sect_nb = 0; 
	while(fp->smap[fp->sect_nb].size){
		INF( "start=0x%lx size=0x%x\n",
		      fp->smap[fp->sect_nb].addr,
		      fp->smap[fp->sect_nb].size );
		fp->sect_nb++;
	}

	INF( "total=%d sectors in %d region(s)\n",
	      fp->sect_nb,(fp->special_base2 == fp->special_base)?1:2 );

	/* find start index to compare and the direction for that index */
	if (fp->bottom){
		fp->start_index = fp->sect_nb - 1;
		fp->step = -1;
	} else{
		fp->start_index = 0;
		fp->step = 1;
	}
	INF( "org='%s' start_index=%d step=%d\n",
	      (fp->bottom) ? "BOTTOM" : "TOP",fp->start_index,fp->step );
	return( 0 );
}

#if 0
static int special( flash_desc_t *fp,unsigned long addr,int *index )
{
	int	i = 0;
	if ((fp->bottom && (addr < fp->normal_base)) ||
	   (!fp->bottom && (addr > fp->special_base2)))
	{
		while (fp->smap[i].size) {
			if ((fp->base + fp->smap[i].addr) == addr) {
				*index = i;
				break;
			}
			i++;
		}
		if (!fp->smap[i].size) {
			return( 0 );
		}
#ifdef MONSTER_DEBUG
		printk( "am29: [DEB] special: addr=0x%lx special_base=0x%lx special_base2=0x%lx index=%d\n",addr,fp->special_base,fp->special_base2,*index ); 
#endif
		return( 1 );
	}
	return( 0 );
}
#endif

void flashmem_sync (struct mtd_info *mtd)
{
	struct mypriv	*priv = mtd->priv;

	flashmem_periodic( (unsigned long)mtd );

	DEB( "starting sync ...\n" );
	if (priv->cur_erases) {
		DEB( "sync goes to sleep...\n" );
		interruptible_sleep_on( &priv->wq );
	}
	DEB( "sync done\n" );
}

int
flashmem_erase( struct mtd_info *mtd,struct erase_info *instr )
{
	unsigned char		*pageaddr;
	struct mypriv		*priv = mtd->priv;
	struct erase_info	**tmp = &priv->cur_erases;
	unsigned long		flags;

	DEB( "am29: [DEB] erasing flash ...\n" );

#if 0
	if (instr->len != mtd->erasesize) {
		return( -EINVAL );
	}
#else
	if (instr->addr + instr->len > mtd->size) {
		return( -EINVAL );
	}
#endif

	pageaddr = pagein( mtd,instr->addr );
	instr->mtd = mtd;
	instr->dev = instr->addr / mtd->devsize;
	instr->cell = (instr->addr & (mtd->devsize - 1)) / mtd->erasesize;
	instr->next = NULL;
	instr->state = MTD_ERASE_PENDING;
	DEB( "erase step1 ...\n" );
	
	save_flags_cli( flags );
	while (*tmp) {
		tmp = &((*tmp)->next);
	}
	*tmp = instr;
	restore_flags( flags );

	flashmem_periodic( (unsigned long)mtd );
	return( 0 );
}

int flashmem_point( struct mtd_info *mtd,
		    unsigned long from,
		    unsigned long len,
		    unsigned long *retlen,
		    unsigned char **mtdbuf )
{
	unsigned char	*pageaddr = pagein( mtd,from );
	unsigned char	device = from / mtd->devsize;
	struct mypriv	*priv = mtd->priv;
       
	*mtdbuf = pageaddr;
	*retlen = len;
	if (priv->devstat[device]) {
		/* There is an erase in progress or pending.
		 * Make them use flashmem_read()
		 */
		return( -EBUSY );
	}
	return( 0 );
}

void flashmem_unpoint( struct mtd_info *mtd,unsigned long addr )
{
}

int flashmem_read( struct mtd_info *mtd,
		   unsigned long from,
		   unsigned long len,
		   unsigned long *retlen,
		   unsigned char *buf,
		   int userspace )
{
	unsigned long		flags;
 	unsigned char		*pageaddr = pagein(mtd,from);
	unsigned char		device = from / mtd->devsize;
	unsigned char		cell = (from & (mtd->devsize - 1)) / mtd->erasesize;
	struct mypriv		*priv=mtd->priv;
	struct erase_info	*erase = priv->cur_erases;
	int			ret = 0;
	int			timeron = 0;
	
	*retlen = len;

	if (priv->devstat[device]) {
		DEB( "device currently erasing\n" );
		if (priv->cur_erases && priv->cur_erases->cell == cell) {
			DEB( "cell %d currently erasing - setting to all 0xff\n",cell );
			if (userspace) {
				unsigned long l = len;
				while(l--) {
					put_user( 0xff,buf );
					buf++;
				}
				return( 0 );
			} else {
				memset( buf,0xff,len );
				return( 0 );
			}
		}
		save_flags_cli( flags );
		if (priv->devstat[device] == MTD_ERASING) {
			/* There is an erase in progress or pending
			 * for this device. Suspend it.
			 */
			timeron = del_timer( &priv->flashmem_timer );
			restore_flags( flags );
			ret = suspend_erase( pagein( mtd,erase->addr ),priv->simulation );
			if (ret) {
				priv->devstat[device] = MTD_ERASE_SUSPEND;
			}
		} else if (priv->devstat[device] != 0) {
			restore_flags( flags );
			ERR( ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n" );
			ERR( "unknown devstat=0x%x\n",priv->devstat[device] );
			ERR( ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n" );
			return -EAGAIN;
		} else {
			restore_flags( flags );
		}
	}

	ret = 0;
	if (userspace) {
		if (copy_to_user( buf,pageaddr,len ) < 0) {
			ret = -EFAULT;
		}
	} else {
		memcpy( buf,pageaddr,len );
	}


	if (priv->devstat[device] & MTD_ERASE_SUSPEND) {
		resume_erase( pagein( mtd,erase->addr ),priv->simulation );
		priv->devstat[device] = MTD_ERASING;
		if (!timeron) {
			WRN( "XXX jse - MTD_ERASE_SUSPEND && !timeron\n" );
		}
	}

	if (timeron) {
		add_timer( &priv->flashmem_timer );
	}

	return( ret );
}

int flashmem_write( struct mtd_info *mtd,
		    unsigned long to,
		    unsigned long len,
		    unsigned long *retlen,
		    const unsigned char *buf,
		    int userspace )
{
	struct mypriv	*priv = (struct mypriv *)mtd->priv;
 	unsigned char	*endaddr, *startaddr;
	register unsigned char *pageaddr;
	unsigned char	device = to / mtd->devsize;
	int		ret = 0;
	int 		i=0;

#if 0
	DEB( "flashmem_write called\n" );
#endif

	while (priv->devstat[device]) {
		DEB( "flashmem_write calling sync ...\n" );
		flashmem_sync( mtd );
	}

	*retlen = len;
	pageaddr = pagein( mtd,to );
	startaddr = (unsigned char *)((unsigned long)pageaddr & ~1);
	endaddr = pageaddr + (*retlen);

#if 0
	DEB( "to=0x%lx len=%ld pageaddr=0x%lx startaddr=0x%lx\n",
	      to,len,(unsigned long)pageaddr,(unsigned long)startaddr );
#endif

	/* Make sure it's aligned by reading the first byte if necessary */
	if (to & 1) {		/* Unaligned access */
		unsigned char	cbuf;
#if 0
		if (userspace) {
			cbuf = get_user( (char*)buf );
		} else {
			cbuf = *buf;
		}

		if (!((unsigned long)pageaddr & 0xf)) {
			schedule();
		}

		ret = byte_write( pageaddr,cbuf );
		if (ret) {
			return( ret );
		}
		pageaddr++; buf++;
#endif
		/* copying by bytes */
		if (userspace) {
			for ( ; pageaddr<=endaddr; buf+=1, pageaddr+=1) {
				cbuf = get_user( (unsigned char*)buf );
				ret = byte_write( priv,pageaddr,cbuf );
				if (ret) {
					return( ret );
				}
				if (!((unsigned long)pageaddr & 0xf)) {
					DEB( "schedule in flashmem_write odd\n" );
					schedule();
				}
			}
		} else {
			for ( ; pageaddr<=endaddr; buf+=1, pageaddr+=1) {
#if 0
				if ((unsigned long)pageaddr & 0xf) {
					schedule();
				}
#endif
				ret = byte_write( priv,pageaddr,*(unsigned char*)buf );
				if (ret) {
					return( ret );
				}
			}
		}
	} else {
		if (userspace) {
			__u16	sbuf;
/* fsch */			i=0;
				for ( ; pageaddr+1 < endaddr; buf+=2, pageaddr+= 2) {
				sbuf = get_user((__u16*)buf);
				ret = word_write( priv,pageaddr,sbuf );
/* fsch */			i++;
				if (ret) {
					return( ret );
				}
				/* fsch schedule only all 256 byte*/
				if (!(i & 0xff)) {
					DEB( "schedule in flashmem_write even\n" );
					schedule();
				}
			}
		} else {
			for ( ; pageaddr+1 < endaddr; buf+=2, pageaddr+=2) {
#if 0
				if ((unsigned long)pageaddr & 0xf) {
					schedule();
				} 
#endif				ret = word_write( priv,pageaddr,*(__u16 *)buf );
				if (ret) {
					return( ret );
				}
			}
		}
	}
	if (pageaddr != endaddr) {	/* One more byte to write at the end */
		unsigned char	cbuf;

		if (userspace) {
			cbuf = get_user( (char*)buf );
		} else {
			cbuf = *buf;
		}
		ret = byte_write( priv,pageaddr,cbuf );
		if (ret) {
			return( ret );
		}
	}
	return( 0 );
}

void flashmem_openclose( struct mtd_info *mtd,int which )
{
	if (which) {
		MOD_INC_USE_COUNT;
	} else {
		MOD_DEC_USE_COUNT;
	}
}


/*====================================================================*/

static int byte_write( struct mypriv *priv,
		       volatile unsigned char *addr,
		       unsigned char byte )
{
	register __u16	*waddr = (__u16*)((unsigned long)addr & ~1);
	union {
		__u16  w;
		unsigned char b[2];
	} u;

	u.w = *waddr;
	if((unsigned long)addr & 1){
		/* lower byte */
		u.b[1] = byte;
	}else{
		/* higher byte */
		u.b[0] = byte;
	}

	return( word_write( priv,(unsigned char*)waddr,u.w ));
}

static inline int word_write( struct mypriv *priv,
			      volatile unsigned char *addr,
			      __u16 word )
{
	flash_desc_t	*fp = priv->flashDesc;
	unsigned long	flags;
	register __u16	i, i2;
	unsigned char	*base;

	base = (unsigned char *)((unsigned long)addr & ~(fp->dev_size - 1));
	if (priv->simulation) {
		i = *((volatile __u16*)addr);
		if (~i & word) {
			ERR( "word_write(sim) failed at %p old=0x%04x new=0x%04x\n",addr,i,word );
			return( -EIO );
		}
		*((volatile __u16*)addr) = word;
		return( 0 );
	}

	if (*((volatile __u16*)addr) == word) {
		return( 0 );
	}
	
	save_flags_cli( flags );

	WRITE_UNLOCK(base);
	*((volatile __u16*)(base + PROG_ADDR)) = PROG_CMD;
	*((volatile __u16*)addr) = word;
	restore_flags( flags );

	while(1) {
		save_flags_cli( flags );
		i = *((volatile __u16*)addr);
		i2 = *((volatile __u16*)addr);
		restore_flags( flags );

		if (((i ^ i2) & DQ6) == 0) {		/* programming done */
			i = *((volatile __u16*)addr);
			if (word != i) {
				ERR( "word_write(1) failed at %p old=0x%04x new=0x%04x\n",addr,i,word );
				return( -EIO );
			}
			return( 0 );
		} else if (((i | i2) & DQ5) == DQ5) {	/* program failure */
			*((volatile __u16*)addr) = RESET_CMD;
			i = *((volatile __u16*)addr);
			if (word != i) {
				ERR( "word_write(2) failed at %p old=0x%04x new=0x%04x\n",addr,i,word );

				return( -EIO );
			}
			return( 0 );
		}
	}
	return( 0 );
}

static unsigned long
sector_size( flash_desc_t *fp,
             unsigned char *base,
             unsigned char *addr,
             unsigned char **sector_start )
{
	unsigned long	offset;
	int		i;

	offset = addr - base;
	if (fp->bottom) {
		if ((unsigned long)addr >= fp->normal_base){
			*sector_start = (unsigned char*) ((unsigned long)addr & 0xffff0000);
			return( 0x10000 );
		}
		i = fp->start_index;
		while (i>=0 && fp->smap[i].size) {
			if(offset >= fp->smap[i].addr){
				*sector_start = (unsigned char *)(base + fp->smap[i].addr);
				return( fp->smap[i].size );
			}
			i += fp->step;	/* jse changed */
		}
	}
	/* XXX jse */
	WRN( "assuming default sector size in special sector area\n" );
	*sector_start = (unsigned char*) ((unsigned long)addr & 0xffff0000);
	return( 0x10000 );
}

static  void block_erase( struct mypriv *priv,
			  volatile unsigned char *addr,
			  unsigned long size )
{
	flash_desc_t		*fp = priv->flashDesc;
	volatile unsigned char	*base;
	unsigned long		flags;
#if 0
	int			index;
	unsigned long		top_addr;
#endif

	base = (unsigned char*)((unsigned long)addr & ~(fp->dev_size - 1));
	if (priv->simulation) {
		unsigned long	maxsize;
		unsigned char	*sector_start = NULL;

		maxsize = sector_size( fp,(unsigned char *)base,
		                       (unsigned char *)addr,&sector_start );

		/* XXX jse - this is a hack but what's the reason ? */
		if (addr == base && size == 0x10000) {
			maxsize = 0x10000;
		}

		DEB( ">>> erasing 0x%lx bytes at 0x%lx\n",
		      maxsize,(unsigned long)sector_start );
		memset( sector_start,0xff,maxsize );
		return;
	}
	
	DEB( "erasing 0x%lx bytes at 0x%lx\n",size,(unsigned long)addr );
	save_flags_cli( flags );
	
	WRITE_UNLOCK(base);
	*(volatile __u16*) (base + SETUP_ADDR) = SETUP_CMD;
	WRITE_UNLOCK(base);
/* fsch */
#if 0
	if ((addr = 0) && (size == 0x200000)) {
		printk("am29: [DEB] Chip erase ...\n");
		*(volatile __u16*) (base + UNLOCK_ADDR1) = CHIPERASE_CMD;
	}
	else
#endif
		*(__u16*) addr = SECERASE_CMD;
        

/* fsch */
#if 0
	while(check_erase( priv,addr,size )) {;}

#ifdef MONSTER_DEBUG
	printk( "am29: [DEB] ...erased\n" );
#endif
	if (special( fp,(unsigned long)addr,&index )) {
		top_addr = (unsigned long)addr + size - (unsigned long)base;
#ifdef MONSTER_DEBUG
		printk("am29: [DEB] top_addr=0x%lx\n",top_addr );
		printk("am29: [DEB] erasing special - index=%d\n",index );
		printk("am29: [DEB] 0x%lx of 0x%x\n",
			fp->smap[index].addr,fp->smap[index].size);
#endif
		while(fp->smap[++index].size){
			if(top_addr > fp->smap[index].addr){
#ifdef MONSTER_DEBUG
				printk("am29: [DEB] erasing special index=%d\n",
					index);
				printk("am29: [DEB] 0x%lx of 0x%x\n",
					fp->smap[index].addr,
					fp->smap[index].size);
#endif
				WRITE_UNLOCK(base);
				*(volatile __u16*) (base + SETUP_ADDR) = SETUP_CMD;
				WRITE_UNLOCK(base);
				*(__u16*)(base + fp->smap[index].addr) = SECERASE_CMD;
				while(check_erase( priv,
				                   base + fp->smap[index].addr,
				                   fp->smap[index].size )) {;}
#ifdef MONSTER_DEBUG
				printk( "am29: [DEB] ... erased\n" );
#endif
			} else {
				break;
			}
		}
	}
#endif
/* fsch */
	restore_flags( flags );
}

static inline int erase_check0( flash_desc_t *fp,volatile __u16* addr )
{
	__u16	v1,v2;

	v1 = *addr;
	v2 = *addr;

	if (((v1 ^ v2) & fp->toggling_bits) == fp->toggling_bits) {
		if (v2 & DQ5) {
			v1 = *addr;
			v2 = *addr;
			if (((v1 ^ v2) & fp->toggling_bits) == fp->toggling_bits) {
				*addr = RESET_CMD;
				return( -EIO );
			}
		}
		return( -EBUSY );
	}
	return( 0 );
}

static inline int check_erase( struct mypriv *priv,
			       volatile unsigned char *addr,
			       unsigned long size )
{
	flash_desc_t		*fp = priv->flashDesc;
	volatile unsigned char	*base;
	unsigned long		flags;
#if 0
	unsigned long		top_addr;
	int			index;
#endif
	int			ret;

	base = (unsigned char*)((unsigned long)addr & ~(fp->dev_size - 1));

	if (priv->simulation) {
		return( 0 );
	}
	
	save_flags_cli( flags );

	ret = erase_check0( priv->flashDesc,(__u16*)addr );
	if (ret) goto done;

	DEB( "check_erase: addr=0x%lx erased\n",(unsigned long)addr );

	/* XXX jse fehlt hier noch was ?!? */
#if 0
	if (special((unsigned long)addr, &index)) {
		top_addr = (unsigned long)addr + size - (unsigned long)base;
		while(smap[++index].size){
			if(top_addr > smap[index].addr){
				ret = erase_check0((__u16*)(base + smap[index].addr));
				if (ret) goto done;
			}else{
				break;
			}
		}
		printk("\ncheck_erase: %d special erased\n",index-1 );
	}
#endif

done:
	restore_flags( flags );
	return( ret );
}


static inline int suspend_erase( volatile unsigned char *addr,int simulation )
{
	unsigned long	flags;
	__u16		v1,v2;
	int		erasing;
	int		loopcnt;

	if (simulation) {
		return( 0 );
	}

	DEB( "suspend_erase: addr=0x%lx\n",(unsigned long)addr );

	save_flags_cli( flags );
	erasing = 0;
	v1 = *(volatile __u16*)addr;
	v2 = *(volatile __u16*)addr;

	if (((v1 ^ v2) & (DQ2|DQ6)) == (DQ6|DQ2)) {

		*(__u16*)addr = ERASE_SUSPEND_CMD;
		erasing = 1;
	}
	restore_flags( flags );

	DEB( "suspend_erase: erasing=%d\n",erasing );
        DEB( "v1= 0x%8.8X | v2= 0x%8.8X  at Addr= 0x%8.8X\n",v1,v2,addr );

	loopcnt = 0;
	while(erasing) {
		save_flags_cli( flags );

		v1 = *(volatile __u16*)addr;
		v2 = *(volatile __u16*)addr;
		restore_flags( flags );
		
		if (((v1 ^ v2) & (DQ6|DQ2)) == DQ2) {
			break;
		}

		loopcnt++;

		if (loopcnt > 1000000) {
			ERR( "suspend_erase: max loop count reached\n" );
		}
	}
	DEB( "suspend_erase: DONE\n" );
	return( erasing );
}

static inline void resume_erase( volatile unsigned char *addr,int simulation )
{
	unsigned long	flags;
	__u16		v1,v2;

	if (simulation) {
		return;
	}

	DEB( "resume_erase: addr=0x%lx\n",(unsigned long)addr );
	save_flags_cli( flags );
	v1 = *(volatile __u16*)addr;
	v2 = *(volatile __u16*)addr;
	if ((v1 ^ v2) & 0x04) {
		*(volatile __u16*)addr = ERASE_RESUME_CMD;
	}
	restore_flags( flags );
	return;
}

/*====================================================================*/

static void flashmem_periodic( unsigned long data )
{
	struct mtd_info		*mtd = (struct mtd_info *)data;
	struct mypriv		*priv = mtd->priv;
	struct erase_info	*erase = priv->cur_erases;
	flash_desc_t		*fp = priv->flashDesc;
	unsigned char		*pageaddr;

	del_timer( &priv->flashmem_timer );

	if (!erase) {
		return;
	}
	pageaddr = pagein( mtd,erase->addr );
	
	if (erase->state == MTD_ERASE_PENDING) {
		block_erase( priv,pageaddr,erase->len );
		priv->devstat[erase->dev] = erase->state = MTD_ERASING;
		erase->time = jiffies;
		erase->retries = 0;
	} else if (erase->state == MTD_ERASING) {
		/* It's trying to erase. Check whether it's finished */
		int	ret;

		ret = check_erase( priv,pageaddr,erase->len );
		if (!ret) {
			/* It's finished OK */
			priv->devstat[erase->dev] = 0;
			priv->cur_erases = erase->next;
			erase->state = MTD_ERASE_DONE;
			if (erase->callback) {
				(*(erase->callback))( erase );
			} else {
				kfree( erase );
			}
		} else if (ret == -EIO) {
			if (++erase->retries > fp->retry_limit) {
				ERR( "erase retry timeout - giving up\n" );
				priv->cur_erases = erase->next;
				priv->devstat[erase->dev] = 0;
				erase->state = MTD_ERASE_FAILED;
				if (erase->callback) {
					(*(erase->callback))( erase );
				} else {
					kfree( erase );
				}
			} else {
				priv->devstat[erase->dev] = MTD_ERASE_PENDING;
				erase->state = MTD_ERASE_PENDING;
			}
		} else if (erase->time + fp->erase_timeout < jiffies) {
			ERR( "flash erase timeout\n" );

			/* Just ignore and hope it goes away.
			 * For a while, read ops will give the CSR
			 * and writes won't work
			 */

			priv->cur_erases = erase->next;
			priv->devstat[erase->dev] = 0;
			erase->state = MTD_ERASE_FAILED;
			if (erase->callback) {
				(*(erase->callback))(erase);
			} else {
				kfree(erase);
			}
		}
	}

	if (priv->cur_erases) {
		priv->flashmem_timer.expires = jiffies + HZ/40+1;
		add_timer( &priv->flashmem_timer );
	} else {
		DEB( "sync wakes up again ...\n" );
		wake_up_interruptible( &priv->wq );
	}
}


int flashmem_init_real( int devNo,flash_desc_t *fp,int simulation )
{
	struct mtd_info *mtdp;
	struct mypriv	*priv;

	INF( "flashmem_init (simulation=%s)\n",(simulation)?"on":"off" );

	if (!(mtdp = kmalloc( sizeof( struct mtd_info ),GFP_KERNEL ))) {
		ERR( "!kmalloc mtd_info\n" );
		return( -ENOMEM );
	}

	memset( mtdp,0,sizeof( struct mtd_info ));
	mtdp->priv = kmalloc( sizeof( struct mypriv ),GFP_KERNEL );
	if (!mtdp->priv) {
		kfree( mtdp );
		ERR( "!kmalloc priv\n" );
		return( -ENOMEM );
	}
	priv = mtdp->priv;
	memset( priv,0,sizeof( struct mypriv ));
	priv->baseAddr = phys_to_virt( fp->base );
	priv->flashDesc = fp;
	priv->simulation = simulation;
	priv->numDevices = MAX_FLASH_DEVICES;

	if (flashmem_getId( fp,simulation ) != 0) {
		ERR( "unable to get flash info\n" );
		kfree( mtdp->priv );
		kfree( mtdp );
		return( -ENODEV );
	}

	strcpy( &mtdp->name[0],&fp->chip_name[0] );
	mtdp->id		= fp->flash_devId;
	mtdp->flags		= MTD_CLEAR_BITS | MTD_ERASEABLE;
	mtdp->erase		= flashmem_erase;
	mtdp->point		= flashmem_point;
	mtdp->unpoint		= flashmem_unpoint;
	mtdp->read		= flashmem_read;
	mtdp->write		= flashmem_write;
	mtdp->openclose		= flashmem_openclose;
	mtdp->sync		= flashmem_sync;
	mtdp->erasesize		= fp->cell_size;
	mtdp->interleave	= 2;
	mtdp->size		= fp->dev_size;		/* AVAIL_SIZE */
	mtdp->devsize		= fp->dev_size;
	mtdp->devtype		= MTD_FLASH;

	if (add_mtd_device( mtdp )) {
		ERR( "register mtd device failed\n" );
		kfree( mtdp->priv );
		kfree( mtdp );
		return( -EAGAIN );
	}
	mymtd[devNo] = mtdp;

	init_timer( &priv->flashmem_timer );
	priv->flashmem_timer.function	= flashmem_periodic;
	priv->flashmem_timer.data	= (unsigned long)mtdp;
	return( 0 );
}

int flashmem_init( void )
{
	flash_desc_t	*fp;
	int		i;

	for (i=0; i<MAX_MTD_DEVICES; i++) {
		mymtd[i] = NULL;
	}

#ifdef CONFIG_ARCH_P2001
#if defined(PROJECT_FAMILY_COMPACTDSL) || defined(PROJECT_FAMILY_MODULARDSL)
#ifdef PROJECT_VARIANT_FLASHSIM
#warning ##############################
#warning >>> using flash simulation !!!
#warning ##############################
	i = 1;		/* device 0 - simulated flash in sdram */
#else
	i = 0;		/* device 0 - real flash */
#endif

/*
 *	device 0 - 2MB/4MB
 */
	fp = &flashDesc[0];
	fp->base		= FLASH_BASE_ADDR;
	fp->special_base	= 0;
	fp->special_base2	= 0;
	fp->normal_base		= 0;
	fp->flash_mId		= 0;
	fp->flash_devId		= 0;
	fp->smap		= NULL;
	fp->bottom		= 1;
	fp->erase_timeout	= 10 * HZ;	/* in ticks */
	fp->retry_limit		= 4;		/* write retries */
	fp->sect_nb		= 0;
	fp->start_index		= 0;
	fp->step		= 0;
	fp->chip_name[0]	= '\0';
	fp->toggling_bits	= 0;
	if ((i = flashmem_init_real( 0,fp,i )) != 0) {
		return( i );
	}

/*
 *	Device 1 - 1/2MB (if present)
 */
	fp = &flashDesc[1];
	fp->base		= FLASH2_BASE_ADDR;
	fp->special_base	= 0;
	fp->special_base2	= 0;
	fp->normal_base		= 0;
	fp->flash_mId		= 0;
	fp->flash_devId		= 0;
	fp->smap		= NULL;
	fp->bottom		= 1;
	fp->erase_timeout	= 10 * HZ;	/* in ticks */
	fp->retry_limit		= 4;		/* write retries */
	fp->sect_nb		= 0;
	fp->start_index		= 0;
	fp->step		= 0;
	fp->chip_name[0]	= '\0';
	fp->toggling_bits	= 0;
	flashmem_init_real( 1,fp,0 );		/* ignore exit status */
	return( 0 );
#elif defined(PROJECT_FAMILY_ROUTERMODUL)
#ifdef PROJECT_VARIANT_FLASHSIM
#warning ##############################
#warning >>> using flash simulation !!!
#warning ##############################
	i = 1;		/* device 0 - simulated flash in sdram */
#else
	i = 0;		/* device 0 - real flash */
#endif
/*
 *	Device 0 - real flash
 */
	fp = &flashDesc[0];
	fp->base		= FLASH_BASE_ADDR;
	fp->special_base	= 0;
	fp->special_base2	= 0;
	fp->normal_base		= 0;
	fp->flash_mId		= 0;
	fp->flash_devId		= 0;
	fp->smap		= NULL;
	fp->bottom		= 1;
	fp->erase_timeout	= 10 * HZ;	/* in ticks */
	fp->retry_limit		= 4;		/* write retries */
	fp->sect_nb		= 0;
	fp->start_index		= 0;
	fp->step		= 0;
	fp->chip_name[0]	= '\0';
	fp->toggling_bits	= 0;

	return( flashmem_init_real( 0,fp,i ));
#else
#error NO FLASH DEVICES SPECIFIED !!!
#endif
#endif
	return( -ENODEV );
}

#ifdef MODULE
int init_module( void )
{
	return( flashmem_init() );
}

void cleanup_module( void )
{
	struct mypriv	*priv;
	int		i;

	for (i=0; i<MAX_MTD_DEVICES; i++) {
		if (mymtd[i] != NULL) {
			priv = mymtd[i]->priv;
			del_timer( &priv->flashmem_timer );
			kfree( priv );
			del_mtd_device( mymtd[i] );
			kfree( mymtd[i] );
			mymtd[i] = NULL;
		}
	}
}
#endif
