#include <asm/hardware.h>
#include <linux/proc_fs.h>
#include <asm/segment.h>

#define AMD					1
#define AMD29LV160Bottom	0x49
#define AMD29LV160Top		0xc4

#define AMD29F800Bottom		0xD6
#define AMD29F800Top		0x58

#define ATMEL				0x1F
#define AT49BV16X4Bottom	0xC0
#define AT49BV16X4Top		0xC2

#define TOSHIBA				0x98	
#define	TC58FVBottom160FT	0x43	
#define	TC58FVTop160FT		0xC2	


#define ID_ADDR (0x5555*2)
#define UNLOCK_ADDR1 (0x5555*2)
#define UNLOCK_ADDR2 (0x2AAA*2)
#define UNLOCK_CMD1 0xAA
#define UNLOCK_CMD2 0x55
#define RESET_CMD (0xF0)
#define ID_CMD		(0x90)


#define WRITE_UNLOCK(a) { *(volatile __u16*)(a+UNLOCK_ADDR1) = UNLOCK_CMD1; *(volatile __u16*)(a+UNLOCK_ADDR2) = UNLOCK_CMD2; }


static void getFlashType(u_32 base, u_32 *offset, u_32 *length){
	__u16 flash_mId, flash_devId;
	
	WRITE_UNLOCK(base);
	*(volatile __u16*)(base+ID_ADDR) = ID_CMD;
	flash_mId = *(__u16*)(base);
	flash_devId = *(__u16*)(base+2) & 0xff;
	*(volatile __u16*)(base+ID_ADDR) = RESET_CMD;

	switch(flash_mId){
		default:
		case AMD:
		case TOSHIBA:	
			switch(flash_devId){
				default:
				case AMD29LV160Bottom:
				case TC58FVBottom160FT:	 
					*offset = 0xE000;
					*length = 0x2000;		  
					break;
				case AMD29LV160Top:
				case TC58FVTop160FT: 
					*offset = 0x1FE000;
					*length = 0x2000;		  
					break;
			}
			break;
		case ATMEL:	 
			switch(flash_devId){
				case AT49BV16X4Bottom:
					*offset = 0xE000;
					*length = 0x2000;		  
					break;
				case AT49BV16X4Top:
					*offset = 0x1FE000;
					*length = 0x2000;		  
					break;
			}
			break;
	}
}

int aplio_get_cfg(char *buf, char **start, off_t offset, int len, int unused){
	char cfgbuf[24];
	int i;
	u_32 offset, length;
	getFlashType((u_32)P_BOOT_SECTOR, &offset, &length);
	sprintf(cfgbuf,"%x %x",offset, length);
	for(i=0; i<strlen(cfgbuf); i++){
		if (cfgbuf[i]==0)
			break;
		put_user(cfgbuf[i],&buf[i]);
	}
	return (i);
}
struct proc_dir_entry apliocfg_proc_entry = {
	0,					/* dynamic inode */
	8, "apliocfg",		/* name length and name */
	S_IFREG | S_IRUGO,	/* mode */
	1,0,0,				/* nlinks, owner, group */
	0,					/* size unused */
	NULL,				/* default operations */
	&aplio_get_cfg,		/* function used to read */
};
