/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * davos44iad_relais.c - relais driver
 * -----------------------------------
 *
 * $RCSfile: davos44iad_relais.c,v $
 * $Source: /opt/funkwerk/cvsroot/source/linux/uclinux2.0.38/arch/armnommu/drivers/char/davos44iad_relais.c,v $
 * $Author: jselent $
 * $Revision: 1.5 $
 * $Date: 2008-05-19 06:44:53 $
 * $State: Exp $
 * $Name:  $
 *
 * $Log: davos44iad_relais.c,v $
 * Revision 1.5  2008-05-19 06:44:53  jselent
 * CHG: copyright info
 *
 * Revision 1.4  2003-05-05 14:23:16  selent
 * initialize relais to state OFF
 *
 * Revision 1.3  2003/05/05 14:16:22  selent
 * moved p2001_GPO4 to arch/armnommu/setup.c
 *
 * Revision 1.2  2003/02/18 14:47:21  selent
 * using GPO4 config as set by entry2.S
 *
 * Revision 1.1  2003/02/17 19:28:00  selent
 * davos44iad relais driver & GPO4 multiplexer
 *
 * Initial Version	Mon Feb 17 15:01:02 CET 2003 - jse
 * Written
 */

/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */



/* ************************************************************************* */
/* *** INCLUDES ************************************************************ */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>
#include <linux/major.h>
#include <linux/timer.h>
#include <linux/errno.h>

#include "bsp.h"
#include "conf.h"



#if defined(PROJECT_DAVOS44IAD)

/* ************************************************************************* */
/* *** PROTOTYPES ********************************************************** */

static int	relais_open( struct inode *,struct file * );
static void	relais_close( struct inode *,struct file * );
static int	relais_write( struct inode *,struct file *,const char *,int );
extern void	p2001_GPO4( unsigned long,unsigned long );



/* ************************************************************************* */
/* *** LOCAL VARIABLES ***************************************************** */

static struct file_operations	relais_fops = {
	NULL,		/* lseek   */
	NULL,		/* read    */
	relais_write,	/* write   */
	NULL,		/* readdir */
	NULL,		/* select  */
	NULL,		/* ioctl   */
	NULL,		/* mmap    */
	relais_open,	/* open    */
	relais_close	/* release */
};



/* ************************************************************************* */
/* *** STATIC FUNCTIONS **************************************************** */

static int
relais_open( struct inode *ino,struct file *filep )
{
	return( 0 );
}

static void
relais_close( struct inode *ino,struct file *filep )
{
}

static int
relais_write( struct inode *ino,struct file *filep,const char *buf,int count )
{
	unsigned long	msk;
	int		minor;
	int		rc;

	minor = MINOR( ino->i_rdev );
	if (minor == 0) {
		msk = 1;
	} else if (minor == 1) {
		msk = 2;
	} else {
		return( -EINVAL );
	}
	if (count < 0) {
		return( -EINVAL );
	}
	if (count == 0) {
		return( 0 );
	}
	if ((rc = verify_area( VERIFY_READ,buf,count )) != 0) {
		return( -EINVAL );
	}
	switch (*buf) {
	case '0':
		p2001_GPO4( msk,0 );
		break;
	case '1':
		p2001_GPO4( msk,msk );
		break;
	default:
		return( -EINVAL );
		break;
	}
	return( count );
}



/* ************************************************************************* */
/* *** GLOBAL FUNCTIONS **************************************************** */

int
davos44iad_relais_init( void )
{
	p2001_GPO4( 0x00000003,0x00000000 );
	if (register_chrdev( DAVOS_RELAIS_MAJOR,"relais",&relais_fops )) {
		printk( KERN_ERR"relais: [ERR] failed to register cdev major=%d\n",DAVOS_RELAIS_MAJOR );
		return( -EAGAIN );
	}
	return( 0 );
}
#endif	/* PROJECT_DAVOS44IAD */
