/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */
 
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/config.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/mm.h>
#include <linux/kernel.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/arch-p2001/irq.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <asm/bitops.h>
#include <asm/delay.h>
#if 0
#include <asm/kdebug.h>
#endif

//#include "drv/int/fiq_handler.h"

#include "p2001serial.h"

typedef unsigned long u_32;


/*
  CFG_P2001_SERIAL_FIQ		
  	HW-Fifo is filled/emptied from FIQ, set outside 
*/

#if defined(PROJECT_VARIANT_SERIAL_CONSOLE)	/* workaround: driver actually does not support console */
  #undef CFG_P2001_SERIAL_FIQ
#endif

#ifdef CFG_P2001_SERIAL_FIQ 

//  #define CFG_P2001_SERIAL_DIAG		/* diagnose counters, default=OFF */
  #define CFG_P2001_SERIAL_FIQ_DYN	/* dynamic registration of FIQ-function */
  #define CFG_P2001_SERIAL_FIQ__SOFTWARE_RTS /* RTS assertion evaluated by FIQ-Handler */
  #define CFG_P2001_SERIAL_FIQ_RX_NOTIFY_INDIRECT /* */
  #define CFG_P2001_SERIAL_FIQ__RUN_IN_SRAM /* run FIQ-part from internal SRAM */

  #define READ_REG(addr)          (* (volatile unsigned int *) (addr))
  #define WRITE_REG(addr, val)    (* (volatile unsigned int *) (addr) = (volatile unsigned int)(val))
  #define REG_CLR32_BITS(addr,bits) (READ_REG(addr) &= (volatile unsigned int)~(bits))
  #define REG_SET32_BITS(addr,bits) (READ_REG(addr) |= (volatile unsigned int)(bits))

#endif

#ifdef CFG_P2001_SERIAL_FIQ__SOFTWARE_RTS

  #define M_RTS_SET()		REG_CLR32_BITS(ABS_Adr_PIN_MUX, 0x20);
  #define M_RTS_RESET()		REG_SET32_BITS(ABS_Adr_PIN_MUX, 0x20);

#else

  #define M_RTS_SET()	
  #define M_RTS_RESET() 

#endif

#ifdef CFG_P2001_SERIAL_FIQ__RUN_IN_SRAM
  #define M_ATTRIBUTE_FOR_FIQ		__attribute__ ((section ("sram")))
#else
  #define M_ATTRIBUTE_FOR_FIQ
#endif  


void p2001_serial_diag_reset(void);


#define USE_INTS	1
#define UART_CLOCK	((SYSCLK) / 8)


#define XMIT_SERIAL_SIZE	PAGE_SIZE
//#define RX_SERIAL_SIZE		PAGE_SIZE
#define RX_SERIAL_SIZE		256


static int p2001_serial_fiqno;
#ifdef CFG_P2001_SERIAL_FIQ_RX_NOTIFY_INDIRECT
static void p2001_serial_fiq_tc1_init(void);
static void p2001_serial_fiq_tc1_deinit(void);
static void p2001_serial_fiq_tc1_start(void)	M_ATTRIBUTE_FOR_FIQ;
static void p2001_serial_fiq_tc1(void)		M_ATTRIBUTE_FOR_FIQ;
#endif
static void p2001_serial_fiq(void)		M_ATTRIBUTE_FOR_FIQ;


#if 0
static volatile union uart_regs *g_uart =
	(volatile union uart_regs *) UART_BASE_ADDR;
#endif
static struct p2001_serial p2001_info;
#if 0
struct tty_struct p2001_ttys
#endif
/* Console hooks... */
/*static int m68k_cons_chanout = 0;
static int m68k_cons_chanin = 0;*/
#if 0
struct p2001_serial *p2001_consinfo = 0;
#endif
#if 0
static unsigned char kgdb_regs[16] = {
	0, 0, 0,                     /* write 0, 1, 2 */
	(Rx8 | RxENABLE),            /* write 3 */
	(X16CLK | SB1 | PAR_EVEN),   /* write 4 */
	(Tx8 | TxENAB),              /* write 5 */
	0, 0, 0,                     /* write 6, 7, 8 */
	(NV),                        /* write 9 */
	(NRZ),                       /* write 10 */
	(TCBR | RCBR),               /* write 11 */
	0, 0,                        /* BRG time constant, write 12 + 13 */
	(BRSRC | BRENABL),           /* write 14 */
	(DCDIE)                      /* write 15 */
};
#endif

DECLARE_TASK_QUEUE(tq_serial);

struct tq_struct serialpoll;

struct tty_driver serial_driver, callout_driver;
static int serial_refcount;

/* serial subtype definitions */
#define SERIAL_TYPE_NORMAL	1
#define SERIAL_TYPE_CALLOUT	2
  
/* number of characters left in xmit buffer before we ask for more */
#define WAKEUP_CHARS 256

/* Debugging... DEBUG_INTR is bad to use when one of the zs
 * lines is your console ;(
 */
#undef SERIAL_DEBUG_INTR
#undef SERIAL_DEBUG_OPEN
#undef SERIAL_DEBUG_FLOW

#define RS_ISR_PASS_LIMIT 256

#define _INLINE_ inline

static void serpoll(void *data);

static void change_speed(struct p2001_serial *info);

static struct tty_struct *serial_table;
static struct termios *serial_termios;
static struct termios *serial_termios_locked;

#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

static char prompt0;
static void xmit_char(struct p2001_serial* info, char ch);
static void xmit_string(struct p2001_serial *info, char *p, int len);
static void start_rx(struct p2001_serial *info);
static void wait_EOT(volatile union uart_regs*);
static void uart_init(struct p2001_serial *info);
static void uart_speed(struct p2001_serial *info, unsigned cflag);

static void tx_enable(volatile union uart_regs *uart);
static void rx_enable(volatile union uart_regs *uart);
static void tx_disable(volatile union uart_regs *uart);
static void rx_disable(volatile union uart_regs *uart);
static void tx_stop(volatile union uart_regs *uart);
static void tx_start(volatile union uart_regs *uart, int ints);
static void rx_stop(volatile union uart_regs *uart);
static void rx_start(volatile union uart_regs *uart, int ints);
static void set_ints_mode(int yes, struct p2001_serial *info);
static void rs_interrupt(struct p2001_serial *info);
extern void show_net_buffers(void);
extern void hard_reset_now(void);

static int global;

void coucou1(void)
{
	global = 0;
}

void coucou2(void)
{
	global = 1;
}
#ifdef CFG_P2001_SERIAL_FIQ
  #define UART_TX_INTS (0)
#else
  #define UART_TX_INTS (UART_IRQ_TX_LOWAT | UART_IRQ_TX_LAST_DATA)
#endif

#define UART_TX_INTS_SHIFT (UART_TX_INTS << UART_IRQ_EN_IRQ_SHIFT)

#ifdef CFG_P2001_SERIAL_FIQ
  #define UART_RX_INTS (0)
#else
  #define UART_RX_INTS (UART_IRQ_RX_DATA  \
		      | UART_IRQ_RX_FIFO_RAR_ERR | UART_IRQ_RX_BRK \
		      | UART_IRQ_RX_LOST | UART_IRQ_ECHO_ERR)
#endif

#define UART_RX_INTS_SHIFT (UART_RX_INTS << UART_IRQ_EN_IRQ_SHIFT)

#define UART_ERROR_INTS ( UART_IRQ_RX_FIFO_RAR_ERR | UART_IRQ_RX_BRK \
		        | UART_IRQ_ECHO_ERR | UART_IRQ_RX_LOST )




static void _INLINE_ tx_enable(volatile union uart_regs *uart)
{
#ifndef CFG_P2001_SERIAL_FIQ 
	uart->w.irq_status = (uart->r.irq_status & ~UART_IRQ_IRQ_MASK)
		| UART_TX_INTS_SHIFT;
#endif
}

static void _INLINE_ rx_enable(volatile union uart_regs *uart)
{
#ifndef CFG_P2001_SERIAL_FIQ 
	uart->w.irq_status = (uart->r.irq_status & ~UART_IRQ_IRQ_MASK)
		| UART_RX_INTS_SHIFT;
#endif
}

static void _INLINE_ tx_disable(volatile union uart_regs *uart)
{
#ifndef CFG_P2001_SERIAL_FIQ 
	uart->w.irq_status = (uart->r.irq_status & ~UART_IRQ_IRQ_MASK)
		& ~UART_TX_INTS_SHIFT;
#endif
}

static void _INLINE_ rx_disable(volatile union uart_regs *uart)
{
#ifndef CFG_P2001_SERIAL_FIQ 
	uart->w.irq_status = (uart->r.irq_status & ~UART_IRQ_IRQ_MASK)
		& ~UART_RX_INTS_SHIFT;
#endif
}

static void _INLINE_ tx_stop(volatile union uart_regs *uart)
{
	tx_disable(uart);
}

static void _INLINE_ tx_start(volatile union uart_regs *uart, int ints)
{
	if (ints)
		tx_enable(uart);
}

static void _INLINE_ rx_stop(volatile union uart_regs *uart)
{
	rx_disable(uart);
}

static void _INLINE_ rx_start(volatile union uart_regs *uart, int ints)
{
	if (ints)
		rx_enable(uart);
}

static void _INLINE_ reset_status(volatile union uart_regs *uart, u_32 status)
{
#ifndef CFG_P2001_SERIAL_FIQ 
	uart->w.irq_status = uart->r.irq_status & ~status;
#endif
}

static void set_ints_mode(int yes, struct p2001_serial *info)
{
#ifndef CFG_P2001_SERIAL_FIQ 
	info->use_ints = yes;
	yes ? unmask_irq(info->irq) : mask_irq(info->irq);
#endif
}

/*
 * tmp_buf is used as a temporary buffer by serial_write.  We need to
 * lock it in case the memcpy_fromfs blocks while swapping in a page,
 * and some other program tries to do a serial write at the same time.
 * Since the lock will only come under contention when the system is
 * swapping and available memory is low, it makes sense to share one
 * buffer across all the serial ports, since it significantly saves
 * memory if large numbers of serial ports are open.
 */
static unsigned char tmp_buf[XMIT_SERIAL_SIZE]; /* This is cheating */
static struct semaphore tmp_buf_sem = MUTEX;

#if 0
static unsigned char rx_buf1[RX_SERIAL_SIZE];
static unsigned char rx_buf2[RX_SERIAL_SIZE];
#endif

#if 0
static void p2001_cts_off(struct p2001_serial *info)
{
#if 0
	volatile union uart_regs *uart;

	uart = info->uart;
	info->save_config &= ~UART_CONFIG_USE_CTS;
	uart->w.config = info->save_config;
	info->cts_state = 0;
#endif
}
static void p2001_cts_on(struct p2001_serial *info)
{
#if 0
	volatile union uart_regs *uart;

	uart = info->uart;
	info->save_config |= UART_CONFIG_USE_CTS;
	uart->w.config = info->save_config;
	info->cts_state = 1;
#endif
}
#endif

static inline int serial_paranoia_check(struct p2001_serial *info,
					dev_t device, const char *routine)
{
#ifdef SERIAL_PARANOIA_CHECK
	static const char *badmagic =
		"Warning: bad magic number for serial struct (%d, %d) in %s\n";
	static const char *badinfo =
		"Warning: null p2001_serial for (%d, %d) in %s\n";

	if (!info) {
		printk(badinfo, MAJOR(device), MINOR(device), routine);
		return 1;
	}
	if (info->magic != SERIAL_MAGIC) {
		printk(badmagic, MAJOR(device), MINOR(device), routine);
		return 1;
	}
#endif
	return 0;
}
/* Sets or clears DTR/RTS on the requested line */
static inline void p2001_rtsdtr(struct p2001_serial *ss, int set)
{
#if 0
	volatile union uart_regs *uart;
	uart = ss->uart;
	if(set) {
		uart->mc |= US_DTR | US_RTS;
	} else {
		uart->mc &= ~(u_32)(US_DTR | US_RTS);
	}
#endif
	return;
}

static inline void kgdb_chaninit(struct p2001_serial *ss, int intson, int bps)
{
#if 0
	int brg;

	if(intson) {
		kgdb_regs[R1] = INT_ALL_Rx;
		kgdb_regs[R9] |= MIE;
	} else {
		kgdb_regs[R1] = 0;
		kgdb_regs[R9] &= ~MIE;
	}
	brg = BPS_TO_BRG(bps, ZS_CLOCK/16);
	kgdb_regs[R12] = (brg & 255);
	kgdb_regs[R13] = ((brg >> 8) & 255);
	load_zsregs(ss->p2001_channel, kgdb_regs);
#endif
}


/*
 * ------------------------------------------------------------
 * rs_stop() and rs_start()
 *
 * This routines are called before setting or resetting tty->stopped.
 * They enable or disable transmitter interrupts, as necessary.
 * ------------------------------------------------------------
 */
static void rs_stop(struct tty_struct *tty)
{
	struct p2001_serial *info = (struct p2001_serial *)tty->driver_data;
	unsigned long flags;

	if (serial_paranoia_check(info, tty->device, "rs_stop"))
		return;
	
	save_flags(flags); cli();
	tx_stop(info->uart);
	rx_stop(info->uart);
	restore_flags(flags);
}


static void rs_put_char(struct p2001_serial *info, char ch)
{
        int flags = 0;
        save_flags(flags); cli();
		wait_EOT(info->uart);
		xmit_char(info,ch);
		wait_EOT(info->uart);
        restore_flags(flags);
}


static void rs_start(struct tty_struct *tty)
{
	struct p2001_serial *info = (struct p2001_serial *)tty->driver_data;
	unsigned long flags;
	
	if (serial_paranoia_check(info, tty->device, "rs_start"))
		return;
	
	save_flags(flags); cli();
	tx_start(info->uart, info->use_ints);
	start_rx(info);
	restore_flags(flags);
}

/* Drop into either the boot monitor or kadb upon receiving a break
 * from keyboard/console input.
 */
static void batten_down_hatches(void)
{
	/* If we are doing kadb, we call the debugger
	 * else we just drop into the boot monitor.
	 * Note that we must flush the user windows
	 * first before giving up control.
	 */
#if 0
	if((((unsigned long)linux_dbvec)>=DEBUG_FIRSTVADDR) &&
	   (((unsigned long)linux_dbvec)<=DEBUG_LASTVADDR))
		sp_enter_debugger();
	else
		panic("p2001_serial: batten_down_hatches");
	return;
#endif
}

/*
 * ----------------------------------------------------------------------
 *
 * Here starts the interrupt handling routines.  All of the following
 * subroutines are declared as inline and are folded into
 * rs_interrupt().  They were separated out for readability's sake.
 *
 * Note: rs_interrupt() is a "fast" interrupt, which means that it
 * runs with interrupts turned off.  People who may want to modify
 * rs_interrupt() should try to keep the interrupt handler as fast as
 * possible.  After you are done making modifications, it is not a bad
 * idea to do:
 * 
 * gcc -S -DKERNEL -Wall -Wstrict-prototypes -O6 -fomit-frame-pointer serial.c
 *
 * and look at the resulting assemble code in serial.s.
 *
 * 				- Ted Ts'o (tytso@mit.edu), 7-Mar-93
 * -----------------------------------------------------------------------
 */

/*
 * This routine is used by the interrupt handler to schedule
 * processing in the software interrupt portion of the driver.
 */
static _INLINE_ void rs_sched_event(struct p2001_serial *info,
				    int event)
{
	info->event |= 1 << event;
	queue_task_irq_off(&info->tqueue, &tq_serial);
	mark_bh(SERIAL_BH);
}

extern void breakpoint(void);  /* For the KGDB frame character */

#ifndef CFG_P2001_SERIAL_FIQ

static _INLINE_ void receive_chars(struct p2001_serial *info, u_32 status)
{
/*	unsigned char ch;*/
	int count, i;
	volatile union uart_regs *uart = info->uart;
#if 0
	// hack to receive chars by polling from anywhere
	struct p2001_serial * info1 = &p2001_info;
	struct tty_struct *tty = info1->tty;
	if (!(info->flags & S_INITIALIZED))
		return;
#else
	struct tty_struct *tty = info->tty;
	if (!(info->flags & S_INITIALIZED))
		return;
#endif
	count = (uart->r.status & UART_STAT_RX_DELTA_MASK)
		>> UART_STAT_RX_DELTA_SHIFT;
	// hack to receive chars by polling only BD fields
	if (!count){
		return;
	}
#if 0
	ch = info->rx_buf[0];
	if(info->is_cons) {
		if(status & US_RXBRK) { /* whee, break received */
			batten_down_hatches();
			/*rs_recv_clear(info->p2001_channel);*/
			return;
		} else if (ch == 0x10) { /* ^P */
			show_state();
			show_free_areas();
			show_buffers();
			show_net_buffers();
			return;
		} else if (ch == 0x12) { /* ^R */
			hard_reset_now();
			return;
		}
		/* It is a 'keyboard interrupt' ;-) */
		wake_up(&keypress_wait);
	}
#endif		
	/* Look for kgdb 'stop' character, consult the gdb documentation
	 * for remote target debugging and arch/sparc/kernel/sparc-stub.c
	 * to see how all this works.
	 */
	/*if((info->kgdb_channel) && (ch =='\003')) {
		breakpoint();
		goto clear_and_exit;
	}*/

	if(!tty)
		goto clear_and_exit;

	if (tty->flip.count >= TTY_FLIPBUF_SIZE)
		queue_task_irq_off(&tty->flip.tqueue, &tq_timer);
	
	if((count + tty->flip.count) >= TTY_FLIPBUF_SIZE){
/*		p2001_cts_off(info);*/
		serialpoll.data = (void *)info;		
		queue_task_irq_off(&serialpoll, &tq_timer);
	}
	memset(tty->flip.flag_buf_ptr, 0, count);
	for (i = count; i; i--) {
		*tty->flip.char_buf_ptr++ = (char) uart->r.rx1;
	}
	
	if(status & UART_IRQ_RX_FIFO_RAR_ERR)
		*(tty->flip.flag_buf_ptr-1) = TTY_PARITY;
	else if(status & UART_IRQ_RX_LOST)
		*(tty->flip.flag_buf_ptr-1) = TTY_OVERRUN;
#if 0
	else if(status & US_FRAME)
		*(tty->flip.flag_buf_ptr-1) = TTY_FRAME;
#endif
	tty->flip.count += count;

	queue_task_irq_off(&tty->flip.tqueue, &tq_timer);

clear_and_exit:
/*	start_rx(info);*/
	return;
}
#endif

#ifndef CFG_P2001_SERIAL_FIQ

static _INLINE_ void transmit_chars(struct p2001_serial *info)
{
	volatile union uart_regs *uart = info->uart;
	int count;

	if (info->x_char) {
		/* Send next char */
		xmit_char(info, info->x_char);
		info->x_char = 0;
		goto clear_and_return;
	}

	if((info->xmit_cnt <= 0) || info->tty->stopped) {
		/* That's peculiar... */
		tx_stop(info->uart);
		goto clear_and_return;
	}

	count = UART_TX_FIFO_SIZE - ((uart->r.status & UART_STAT_TX_DELTA_MASK)
				     >> UART_STAT_TX_DELTA_SHIFT);

	if (count > info->xmit_cnt)
		count = info->xmit_cnt;

	if (info->xmit_tail + count < SERIAL_XMIT_SIZE) {
		xmit_string(info, info->xmit_buf + info->xmit_tail, count);
		info->xmit_tail += count;
		info->xmit_cnt -= count;
	} else {
		xmit_string(info, info->xmit_buf + info->xmit_tail,
			    SERIAL_XMIT_SIZE - info->xmit_tail);
		info->xmit_cnt -= SERIAL_XMIT_SIZE - info->xmit_tail;
		info->xmit_tail = 0;
	}
#if 0
	if(info->xmit_tail + info->xmit_cnt < SERIAL_XMIT_SIZE){
		xmit_string(info, info->xmit_buf+info->xmit_tail, info->xmit_cnt);
		info->xmit_tail = (info->xmit_tail + info->xmit_cnt) & (SERIAL_XMIT_SIZE-1);
		info->xmit_cnt=0;
	}else{
		coucou1();
		xmit_string(info, info->xmit_buf+info->xmit_tail, SERIAL_XMIT_SIZE - info->xmit_tail);
//					xmit_string(info, info->xmit_buf, info->xmit_tail + info->xmit_cnt - SERIAL_XMIT_SIZE);
		info->xmit_cnt = info->xmit_cnt - (SERIAL_XMIT_SIZE - info->xmit_tail);
		info->xmit_tail = 0;
	}
#endif
#if 0		
	/* Send char */
	xmit_char(info, info->xmit_buf[info->xmit_tail++]);
	info->xmit_tail = info->xmit_tail & (SERIAL_XMIT_SIZE-1);
	info->xmit_cnt--;
#endif
	
	if (info->xmit_cnt < WAKEUP_CHARS)
		rs_sched_event(info, RS_EVENT_WRITE_WAKEUP);

	if(info->xmit_cnt <= 0) {
//		tx_stop(info->uart);
		goto clear_and_return;
	}

clear_and_return:
	/* Clear interrupt (should be auto)*/
	return;
}

#endif


static _INLINE_ void status_handle(struct p2001_serial *info, u_32 status)
{
#if 0
	if(status & DCD) {
		if((info->tty->termios->c_cflag & CRTSCTS) &&
		   ((info->curregs[3] & AUTO_ENAB)==0)) {
			info->curregs[3] |= AUTO_ENAB;
			info->pendregs[3] |= AUTO_ENAB;
			write_zsreg(info->p2001_channel, 3, info->curregs[3]);
		}
	} else {
		if((info->curregs[3] & AUTO_ENAB)) {
			info->curregs[3] &= ~AUTO_ENAB;
			info->pendregs[3] &= ~AUTO_ENAB;
			write_zsreg(info->p2001_channel, 3, info->curregs[3]);
		}
	}
#endif
	/* Whee, if this is console input and this is a
	 * 'break asserted' status change interrupt, call
	 * the boot prom.
	 */
#if 0
	/* jse: was ist AIC_UA beim P2001 ? */
	if((status & US_RXBRK) && info->break_abort)
		batten_down_hatches();
#else
	if((status & (1<<2)) && info->break_abort)
		batten_down_hatches();
#endif
	/* XXX Whee, put in a buffer somewhere, the status information
	 * XXX whee whee whee... Where does the information go...
	 */
	reset_status(info->uart, status);
	return;
}


/*
 * This is the serial driver's generic interrupt routine
 */
static void rs_interrupta(int irq, void *dev_id, struct pt_regs * regs){
	rs_interrupt(dev_id);
}

#if 0
void rs_interruptb(int irq, void *dev_id, struct pt_regs * regs){
	rs_interrupt(&p2001_info[1]);
}
#endif

static void rs_interrupt(struct p2001_serial *info)
{
#ifndef CFG_P2001_SERIAL_FIQ
	u_32 status;

	status = info->uart->r.irq_status & UART_IRQ_IRQ_MASK;
	if (status & UART_RX_INTS) {
		receive_chars(info, status);
	}
	if (status & UART_TX_INTS) {
		transmit_chars(info);
	}
	status_handle(info, status);
#if 0
	if(!info->cts_state){
		if(info->tty->flip.count < TTY_FLIPBUF_SIZE-RX_SERIAL_SIZE){
			p2001_cts_on(info);
		}
	}
#endif
	if(!info->use_ints){
//	if(1){
		serialpoll.data = (void *)info;		
		queue_task_irq_off(&serialpoll, &tq_timer);
	}
	return;
#endif
}
static void serpoll(void *data)
{
	struct p2001_serial * info = data;
	rs_interrupt(info);
}

/*
 * -------------------------------------------------------------------
 * Here ends the serial interrupt routines.
 * -------------------------------------------------------------------
 */

/*
 * This routine is used to handle the "bottom half" processing for the
 * serial driver, known also the "software interrupt" processing.
 * This processing is done at the kernel interrupt level, after the
 * rs_interrupt() has returned, BUT WITH INTERRUPTS TURNED ON.  This
 * is where time-consuming activities which can not be done in the
 * interrupt driver proper are done; the interrupt driver schedules
 * them using rs_sched_event(), and they get done here.
 */
static void do_serial_bh(void)
{
	run_task_queue(&tq_serial);
}

static void do_softint(void *private_)
{
	struct p2001_serial	*info = (struct p2001_serial *) private_;
	struct tty_struct	*tty;
	
	tty = info->tty;
	if (!tty)
		return;

	if (clear_bit(RS_EVENT_WRITE_WAKEUP, &info->event)) {
		if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
		    tty->ldisc.write_wakeup)
			(tty->ldisc.write_wakeup)(tty);
		wake_up_interruptible(&tty->write_wait);
	}
}

/*
 * This routine is called from the scheduler tqueue when the interrupt
 * routine has signalled that a hangup has occurred.  The path of
 * hangup processing is:
 *
 * 	serial interrupt routine -> (scheduler tqueue) ->
 * 	do_serial_hangup() -> tty->hangup() -> rs_hangup()
 * 
 */
static void do_serial_hangup(void *private_)
{
	struct p2001_serial	*info = (struct p2001_serial *) private_;
	struct tty_struct	*tty;
	
	tty = info->tty;
	if (!tty)
		return;

	tty_hangup(tty);
}


/*
 * This subroutine is called when the RS_TIMER goes off.  It is used
 * by the serial driver to handle ports that do not have an interrupt
 * (irq=0).  This doesn't work at all for 16450's, as a sun has a Z8530.
 */
 
static void rs_timer(void)
{
	panic("rs_timer called\n");
	return;
}
static u_32 calcCD(u_32 br)
{
	u_32 n, m;

	if (br) {
#if 0
		/*
		 * UPDATE - Thu Feb  5 17:00:31 CET 2004
		 * it seems that the P2001 have a problem with high m, n
		 * values and high baudrates. Actually 921600 Baud does not
		 * work, for a fucking reason nobody knows.
		 * -jse
		 */

		/*
		 * NOTE: with n == 42240 and SYSCLK == 5.5 or 6 * 12288MHz
		 * common baudrates in the range 2400 - 921600 are working
		 * -jse
		 */
		n = 42240;
#else
#if defined(PROJECT_FAMILY_ROUTERMODUL)
		n = 42240;
#else
		n = 960;
#endif
#endif
		m = ( br * n ) / (UART_CLOCK);
	} else {
		n = m = 0;
	}
	return (n << 16) + m;
}

static void uart_init(struct p2001_serial *info)
{
	volatile union uart_regs *uart;
	u_32 tmp, flags;

	if (info) {
		uart = info->uart;
	} else {
		uart = (volatile union uart_regs *) UART_BASE_ADDR;
	}
	save_flags(flags); cli();

	uart->w.clear = 0;
	uart->w.irq_status = 0;

	tmp = calcCD(57600);
	info->save_baudrate = tmp;
	uart->w.baudrate = tmp;

	tmp = 	  0 << UART_CONFIG_STOPBIT_SHIFT
		| 0 << UART_CONFIG_PARITY_SHIFT
		| 8 << UART_CONFIG_WDLEN_SHIFT
#if defined(PROJECT_FAMILY_COMPACTDSL) || defined(PROJECT_FAMILY_MODULARDSL)
#if !defined(PROJECT_VARIANT_SERIAL_CONSOLE)
		| UART_CONFIG_USE_CTS
#endif
#endif
		| 24 << UART_CONFIG_RX_HIWAT_SHIFT
		| 6 << UART_CONFIG_TX_LOWAT_SHIFT
/*		| UART_CONFIG_TX_INV*/
/*		| UART_CONFIG_RX_INV*/
		;
	uart->w.config = tmp;
	info->save_config = tmp;

	uart->w.echo_en = 0;

	restore_flags(flags);
#if 0
	uart->cr	= US_RSTRX|US_RSTTX|US_RSTSTA;
	uart->mr	= US_USCLKS(0)|US_CLK0|US_CHMODE(0)|US_NBSTOP(0)|US_PAR(4)|US_CHRL(3);
	uart->ier	= 0;
	uart->idr	= US_ALL_INTS;
	uart->brgr	= calcCD(57600);
#if FORCE_57600	
	uart->brgr	= calcCD(57600);
#endif	
	uart->rtor	= 20;	// timeout = value * 4 * bit period
	uart->ttgr	= 0;	// no guard time
	uart->rpr	= 0;
	uart->rcr	= 0;
	uart->tpr	= 0;
	uart->tcr	= 0;
	uart->mc	= 0;
#endif
}

static void uart_speed(struct p2001_serial *info, unsigned cflag)
{
	unsigned baud = info->baud;
	volatile union uart_regs *uart = info->uart;
	unsigned long config, tmp;

	uart->w.irq_status = 0;
	tmp = calcCD(baud);
	info->save_baudrate = tmp;
	uart->w.baudrate = tmp;
	if (cflag != 0xffff) {
		config = info->save_config;
		config &= ~(UART_CONFIG_STOPBIT_MASK | UART_CONFIG_PARITY_MASK
			    | UART_CONFIG_WDLEN_MASK);

		if ((cflag & CSIZE) == CS8)
			config |= 8 << UART_CONFIG_WDLEN_SHIFT;
		else config |= 7 << UART_CONFIG_WDLEN_SHIFT;

		if (cflag & CSTOPB)
			config |= 3 << UART_CONFIG_STOPBIT_SHIFT;

		if (cflag & PARENB) {
			if (cflag & PARODD)
				config |= 2 << UART_CONFIG_PARITY_SHIFT;
			else config |= 1 << UART_CONFIG_PARITY_SHIFT;
		}
		info->save_config = config;
		uart->w.config = config;
	}
#if 0
	uart->cr	= US_TXDIS|US_RXDIS;
	uart->ier	= 0;
	uart->idr	= US_ALL_INTS;
	uart->brgr	= calcCD(baud);
#if FORCE_57600	
	uart->brgr	= calcCD(57600);
#endif	
	uart->rtor	= 20;	// timeout = value * 4 *bit period	
	uart->ttgr	= 0;	// no guard time	
	uart->rpr	= 0;
	uart->rcr	= 0;
	uart->tpr	= 0;
	uart->tcr	=0;
	uart->mc	= 0;
	if (cflag != 0xffff){
		uart->mr	= US_USCLKS(0)|US_CLK0|US_CHMODE(0)|US_PAR(0);
		
		if ((cflag & CSIZE) == CS8)
			uart->mr |= US_CHRL(3);		// 8 bit char
		else
			uart->mr |= US_CHRL(2);		// 7 bit char
		
		if (cflag & CSTOPB)
			uart->mr |= US_NBSTOP(2);	// 2 stop bits
		
		if (!(cflag & PARENB))
			uart->mr |= US_PAR(4);		// parity disabled
		else
			if (cflag & PARODD)
				uart->mr |= US_PAR(1);	// odd parity
	}
#endif
	tx_start(uart, info->use_ints);
	start_rx(info);
}
static void wait_EOT(volatile union uart_regs *uart)
{
	u_32 count;

#if 0
	for (;;) {
		count = UART_TX_FIFO_SIZE -
			((uart->r.status & UART_STAT_TX_DELTA_MASK)
			 >> UART_STAT_TX_DELTA_SHIFT);
		if (count > 0)
			break;
	}
#else
	for (;;) {
		count = (uart->r.status & UART_STAT_TX_DELTA_MASK)
			>> UART_STAT_TX_DELTA_SHIFT;
		if (count == 0)
			break;
	}
#endif
}

static int startup(struct p2001_serial * info)
{
	unsigned long flags;

	if (info->flags & S_INITIALIZED)
		return 0;

	if (!info->xmit_buf) {
		info->xmit_buf = (unsigned char *) get_free_page(GFP_KERNEL);
		if (!info->xmit_buf)
			return -ENOMEM;
	}
#if 0
	if (!info->rx_buf) {
//		info->rx_buf = (unsigned char *) get_free_page(GFP_KERNEL);
		info->rx_buf = rx_buf1;
		if (!info->rx_buf)
			return -ENOMEM;
	}
#endif
	save_flags(flags); cli();
#ifdef SERIAL_DEBUG_OPEN
	printk("starting up ttyS%d (irq %d)...\n", info->line, info->irq);
#endif
	/*
	 * Clear the FIFO buffers and disable them
	 * (they will be reenabled in change_speed())
	 */
	
	if (info->tty)
		clear_bit(TTY_IO_ERROR, &info->tty->flags);
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;

	/*
	 * and set the speed of the serial port
	 */

	uart_init(info);
	set_ints_mode(1, info);
	change_speed(info);
	info->flags |= S_INITIALIZED;
	restore_flags(flags);
	return 0;
}

/*
 * This routine will shutdown a serial port; interrupts are disabled, and
 * DTR is dropped if the hangup on close termio flag is on.
 */
static void shutdown(struct p2001_serial * info)
{
	unsigned long	flags;

	tx_disable(info->uart);
	rx_disable(info->uart);
	rx_stop(info->uart); /* All off! */
	if (!(info->flags & S_INITIALIZED))
		return;

#ifdef SERIAL_DEBUG_OPEN
	printk("Shutting down serial port %d (irq %d)....\n", info->line,
	       info->irq);
#endif
	
	save_flags(flags); cli(); /* Disable interrupts */
	
	if (info->xmit_buf) {
		free_page((unsigned long) info->xmit_buf);
		info->xmit_buf = 0;
	}

	if (info->tty)
		set_bit(TTY_IO_ERROR, &info->tty->flags);
	
	info->flags &= ~S_INITIALIZED;
	restore_flags(flags);
}

/* rate = 1036800 / ((65 - prescale) * (1<<divider)) */

static int baud_table[] = {
	0, 50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800,
	9600, 19200, 38400,
/*16.*/ 57600, 115200, 230400, 460800, 500000, 576000, 921600, 0
};


/*
 * This routine is called to set the UART divisor registers to match
 * the specified baud rate for a serial port.
 */
static void change_speed(struct p2001_serial *info)
{
	unsigned cflag;
	int	i;

	if (!info->tty || !info->tty->termios)
		return;
	cflag = info->tty->termios->c_cflag;

	/* First disable the interrupts */
	tx_stop(info->uart);
	rx_stop(info->uart);
	/* set the baudrate */
	i = cflag & CBAUD;
	if (i & CBAUDEX) {
		i &= ~CBAUDEX;
		if (i < 1 || i > 7) 
			info->tty->termios->c_cflag &= ~CBAUDEX;
		else
			i += 15;
	}
	info->baud = baud_table[i];
	uart_speed(info, cflag);
	start_rx(info);
	tx_start(info->uart, info->use_ints);
	return;
}
static void start_rx(struct p2001_serial *info)
{
	volatile union uart_regs *uart = info->uart;
#if 0
	rx_stop(uart);
	if(info->rx_buf == rx_buf1){
		info->rx_buf = rx_buf2;
	}else{
		info->rx_buf = rx_buf1;
	}
	uart->rpr = (u_32)info->rx_buf;
	uart->rcr = (u_32)RX_SERIAL_SIZE;
#endif
	rx_start(uart, info->use_ints);
}
static void xmit_char(struct p2001_serial *info, char ch)
{
	prompt0 = ch;
	xmit_string(info, &prompt0, 1);
}
static void xmit_string(struct p2001_serial *info, char *p, int len)
{
	volatile union uart_regs *uart = info->uart;

	while (len-- > 0) {
		uart->w.tx1 = (unsigned char) *p++;
	}
#if 0
	info->uart->tcr = 0;
	info->uart->tpr = (u_32)p;	
	tx_start(info->uart, info->use_ints);
	info->uart->tcr = (u_32)len;
#endif
}

#if 0
/* These are for receiving and sending characters under the kgdb
 * source level kernel debugger.
 */
void putDebugChar(char kgdb_char)
{
	struct sun_zschannel *chan = p2001_kgdbchan;

	while((chan->control & Tx_BUF_EMP)==0)
		udelay(5);

	chan->data = kgdb_char;
}

char getDebugChar(void)
{
	struct sun_zschannel *chan = p2001_kgdbchan;

	while((chan->control & Rx_CH_AV)==0)
		barrier();
	return chan->data;
}
#endif

/*
 * Fair output driver allows a process to speak.
 */
#if 0
static void rs_fair_output(	struct p2001_serial *info)
{
	int left;		/* Output no more than that */
	unsigned long flags;
	char c;

	if (info == 0) return;
	if (info->xmit_buf == 0) return;

	save_flags(flags);  cli();
	left = info->xmit_cnt;
	while (left != 0) {
		c = info->xmit_buf[info->xmit_tail];
		info->xmit_tail = (info->xmit_tail+1) & (SERIAL_XMIT_SIZE-1);
		info->xmit_cnt--;
		restore_flags(flags);

		rs_put_char(info, c);

		save_flags(flags);  cli();
		left = MIN(info->xmit_cnt, left-1);
	}

	/* Last character is being transmitted now (hopefully). */
//	udelay(20);
	wait_EOT(info->uart);
	restore_flags(flags);
	return;
}
#endif
/*
 * p2001_console_print is registered for printk.
 */
static int console_initialized = 0;
static void init_console(void)
{
	struct p2001_serial *info;
	info = &p2001_info;
	memset(info, 0, sizeof(struct p2001_serial));
#if 0	
	info->uart = uarts[0];
#else	
	info->uart = (volatile union uart_regs*)UART_BASE_ADDR;
#endif
	info->tty = 0;
	info->irqmask = (1<<3);		/* jse: was ist AIC_UA beim P2001 ? */
	info->irq = IRQ_UART;
	info->port = 1;
	info->use_ints = 0;
	info->cts_state = 1;
	info->is_cons = 1;
	console_initialized = 1;
}
void console_print_p2001(const char *p)
{
	char c;
	struct p2001_serial *info;
	info = &p2001_info;

//	if (!(info->flags & S_INITIALIZED)){
	if(!console_initialized){
		init_console();
		uart_init(info);
		info->baud = 57600;
		uart_speed(info,0xffff);
	}
	while((c=*(p++)) != 0) {
		if(c == '\n')
			rs_put_char(info, '\r');
		rs_put_char(info, c);
	}

	/* Comment this if you want to have a strict interrupt-driven output */
//	if (!info->use_ints)
//		rs_fair_output(info);

	return;
}

static void rs_set_ldisc(struct tty_struct *tty)
{
	struct p2001_serial *info = (struct p2001_serial *)tty->driver_data;

	if (serial_paranoia_check(info, tty->device, "rs_set_ldisc"))
		return;

	info->is_cons = (tty->termios->c_line == N_TTY);
	
	printk("ttyS%d console mode %s\n", info->line, info->is_cons ? "on" : "off");
}

static void rs_flush_chars(struct tty_struct *tty)
{
	struct p2001_serial *info = (struct p2001_serial *)tty->driver_data;
	unsigned long flags;

	if (serial_paranoia_check(info, tty->device, "rs_flush_chars"))
		return;
	if(!info->use_ints){	
		for(;;) {
			if (info->xmit_cnt <= 0 || tty->stopped || tty->hw_stopped ||
				!info->xmit_buf)
				return;

			/* Enable transmitter */
			save_flags(flags); cli();
			tx_start(info->uart,info->use_ints);
		}
	}else{
		if (info->xmit_cnt <= 0 || tty->stopped || tty->hw_stopped ||
			!info->xmit_buf)
			return;

			/* Enable transmitter */
		save_flags(flags); cli();
		tx_start(info->uart, info->use_ints);
	}

	if(!info->use_ints)	
		wait_EOT(info->uart);
		/* Send char */
	xmit_char(info, info->xmit_buf[info->xmit_tail++]);
	info->xmit_tail = info->xmit_tail & (SERIAL_XMIT_SIZE-1);
	info->xmit_cnt--;
	
	restore_flags(flags);
}

extern void console_printn(const char * b, int count);

static int rs_write(struct tty_struct * tty, int from_user,
		    const unsigned char *buf, int count)
{
	int	c, total = 0;
	struct p2001_serial *info = (struct p2001_serial *)tty->driver_data;
	unsigned long flags;

	if (serial_paranoia_check(info, tty->device, "rs_write"))
		return 0;

	if (!tty || !info->xmit_buf)
		return 0;
	
	/*buf = "123456";
	count = 6;
	
	printk("Writing '%s' to serial port\n", buf);*/

	/*printk("rs_write of %d bytes\n", count);*/


	save_flags(flags);
	while (1) {
#ifdef CFG_P2001_SERIAL_FIQ 
		clf();
#endif
		cli();
		c = MIN(count, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1,
				   SERIAL_XMIT_SIZE - info->xmit_head));
		if (c <= 0)
			break;

		if (from_user) {
			down(&tmp_buf_sem);
			memcpy_fromfs(tmp_buf, buf, c);
			memcpy(info->xmit_buf + info->xmit_head, tmp_buf, c);
			up(&tmp_buf_sem);
		} else{
			memcpy(info->xmit_buf + info->xmit_head, buf, c);
		}
		info->xmit_head = (info->xmit_head + c) & (SERIAL_XMIT_SIZE-1);
		info->xmit_cnt += c;
		restore_flags(flags);
		buf += c;
		count -= c;
		total += c;
	}

#ifndef CFG_P2001_SERIAL_FIQ 

	if (info->xmit_cnt && !tty->stopped && !tty->hw_stopped ){
		/* Enable transmitter */
		
		cli();		
		/*printk("Enabling transmitter\n");*/

		if(!info->use_ints){	
			while(info->xmit_cnt) {
				wait_EOT(info->uart);
				/* Send char */
				xmit_char(info, info->xmit_buf[info->xmit_tail++]);
				wait_EOT(info->uart);
				info->xmit_tail = info->xmit_tail & (SERIAL_XMIT_SIZE-1);
				info->xmit_cnt--;
			}
		}else{
			if (info->xmit_cnt){
				/* Send char */
				wait_EOT(info->uart);
				if(info->xmit_tail + info->xmit_cnt < SERIAL_XMIT_SIZE){
					xmit_string(info, info->xmit_buf+info->xmit_tail, info->xmit_cnt);
					info->xmit_tail = (info->xmit_tail + info->xmit_cnt) & (SERIAL_XMIT_SIZE-1);
					info->xmit_cnt=0;
				}else{
					coucou2();
					xmit_string(info, info->xmit_buf+info->xmit_tail, SERIAL_XMIT_SIZE - info->xmit_tail);
//					xmit_string(info, info->xmit_buf, info->xmit_tail + info->xmit_cnt - SERIAL_XMIT_SIZE);
					info->xmit_cnt = info->xmit_cnt - (SERIAL_XMIT_SIZE - info->xmit_tail);
					info->xmit_tail = 0;
				}
			}
		}
	} else {
		/*printk("Skipping transmit\n");*/
	}
#endif


#if 0
		printk("Enabling stuff anyhow\n");
		tx_start(0);

		if (SCC_EOT(0,0)) {
			printk("TX FIFO empty.\n");
			/* Send char */
			p2001_xmit_char(info->uart, info->xmit_buf[info->xmit_tail++]);
			info->xmit_tail = info->xmit_tail & (SERIAL_XMIT_SIZE-1);
			info->xmit_cnt--;
		}
#endif

	restore_flags(flags);
	return total;
}

static int rs_write_room(struct tty_struct *tty)
{
	struct p2001_serial *info = (struct p2001_serial *)tty->driver_data;
	int	ret;
				
	if (serial_paranoia_check(info, tty->device, "rs_write_room"))
		return 0;
	ret = SERIAL_XMIT_SIZE - info->xmit_cnt - 1;
	if (ret < 0)
		ret = 0;
	return ret;
}

static int rs_chars_in_buffer(struct tty_struct *tty)
{
	struct p2001_serial *info = (struct p2001_serial *)tty->driver_data;

	if (serial_paranoia_check(info, tty->device, "rs_chars_in_buffer"))
		return 0;
	return info->xmit_cnt;
}

static void rs_flush_buffer(struct tty_struct *tty)
{
	unsigned long flags;
	struct p2001_serial *info = (struct p2001_serial *)tty->driver_data;
				
	if (serial_paranoia_check(info, tty->device, "rs_flush_buffer"))
		return;
	save_flags(flags);cli();
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
	restore_flags(flags);
	wake_up_interruptible(&tty->write_wait);
	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
	    tty->ldisc.write_wakeup)
		(tty->ldisc.write_wakeup)(tty);
}

/*
 * ------------------------------------------------------------
 * rs_throttle()
 * 
 * This routine is called by the upper-layer tty layer to signal that
 * incoming characters should be throttled.
 * ------------------------------------------------------------
 */
static void rs_throttle(struct tty_struct * tty)
{
	struct p2001_serial *info = (struct p2001_serial *)tty->driver_data;
#ifdef SERIAL_DEBUG_THROTTLE
	char	buf[64];
	
	printk("throttle %s: %d....\n", _tty_name(tty, buf),
	       tty->ldisc.chars_in_buffer(tty));
#endif

	if (serial_paranoia_check(info, tty->device, "rs_throttle"))
		return;
	
	if (I_IXOFF(tty))
		info->x_char = STOP_CHAR(tty);

	/* Turn off RTS line (do this atomic) */
}

static void rs_unthrottle(struct tty_struct * tty)
{
	struct p2001_serial *info = (struct p2001_serial *)tty->driver_data;
#ifdef SERIAL_DEBUG_THROTTLE
	char	buf[64];
	
	printk("unthrottle %s: %d....\n", _tty_name(tty, buf),
	       tty->ldisc.chars_in_buffer(tty));
#endif

	if (serial_paranoia_check(info, tty->device, "rs_unthrottle"))
		return;
	
	if (I_IXOFF(tty)) {
		if (info->x_char)
			info->x_char = 0;
		else
			info->x_char = START_CHAR(tty);
	}

	/* Assert RTS line (do this atomic) */
}

/*
 * ------------------------------------------------------------
 * rs_ioctl() and friends
 * ------------------------------------------------------------
 */

static int get_serial_info(struct p2001_serial * info,
			   struct serial_struct * retinfo)
{
	struct serial_struct tmp;
  
	if (!retinfo)
		return -EFAULT;
	memset(&tmp, 0, sizeof(tmp));
	tmp.type = info->type;
	tmp.line = info->line;
	tmp.irq = info->irq;
	tmp.port = info->port;
	tmp.flags = info->flags;
	tmp.baud_base = info->baud_base;
	tmp.close_delay = info->close_delay;
	tmp.closing_wait = info->closing_wait;
	tmp.custom_divisor = info->custom_divisor;
	memcpy_tofs(retinfo,&tmp,sizeof(*retinfo));
	return 0;
}

static int set_serial_info(struct p2001_serial * info,
			   struct serial_struct * new_info)
{
	struct serial_struct new_serial;
	struct p2001_serial old_info;
	int 			retval = 0;

	if (!new_info)
		return -EFAULT;
	memcpy_fromfs(&new_serial,new_info,sizeof(new_serial));
	old_info = *info;

	if (!suser()) {
		if ((new_serial.baud_base != info->baud_base) ||
		    (new_serial.type != info->type) ||
		    (new_serial.close_delay != info->close_delay) ||
		    ((new_serial.flags & ~S_USR_MASK) !=
		     (info->flags & ~S_USR_MASK)))
			return -EPERM;
		info->flags = ((info->flags & ~S_USR_MASK) |
			       (new_serial.flags & S_USR_MASK));
		info->custom_divisor = new_serial.custom_divisor;
		goto check_and_exit;
	}

	if (info->count > 1)
		return -EBUSY;

	/*
	 * OK, past this point, all the error checking has been done.
	 * At this point, we start making changes.....
	 */

	info->baud_base = new_serial.baud_base;
	info->flags = ((info->flags & ~S_FLAGS) |
			(new_serial.flags & S_FLAGS));
	info->type = new_serial.type;
	info->close_delay = new_serial.close_delay;
	info->closing_wait = new_serial.closing_wait;

check_and_exit:
	retval = startup(info);
	return retval;
}

/*
 * get_lsr_info - get line status register info
 *
 * Purpose: Let user call ioctl() to get info when the UART physically
 * 	    is emptied.  On bus types like RS485, the transmitter must
 * 	    release the bus after transmitting. This must be done when
 * 	    the transmit shift register is empty, not be done when the
 * 	    transmit holding register is empty.  This functionality
 * 	    allows an RS485 driver to be written in user space. 
 */
static int get_lsr_info(struct p2001_serial * info, unsigned int *value)
{
	unsigned count;

	cli();
	count = (info->uart->r.status & UART_STAT_TX_DELTA_MASK)
		>> UART_STAT_TX_DELTA_SHIFT;
	sti();
	put_user(!count, value);
	return 0;
#if 0
	unsigned char status;

	cli();
	status = info->uart->csr;
	status &= US_TXEMPTY;
	sti();
	put_user(status,value);
	return 0;
#endif
}

/*
 * This routine sends a break character out the serial port.
 */
static void send_break(	struct p2001_serial * info, int duration)
{
#if 0
	current->state = TASK_INTERRUPTIBLE;
	current->timeout = jiffies + duration;
	cli();
	info->uart->cr = US_STTBRK;
	if(!info->use_ints){
		while(US_TXRDY != (info->uart->csr & US_TXRDY)){
			;			// this takes max 2ms at 9600
		}
		info->uart->cr = US_STPBRK;
	}
	sti();
#endif
}

static int rs_ioctl(struct tty_struct *tty, struct file * file,
		    unsigned int cmd, unsigned long arg)
{
	int error;
	struct p2001_serial * info = (struct p2001_serial *)tty->driver_data;
	int retval;

	if (serial_paranoia_check(info, tty->device, "rs_ioctl"))
		return -ENODEV;

	if ((cmd != TIOCGSERIAL) && (cmd != TIOCSSERIAL) &&
	    (cmd != TIOCSERCONFIG) && (cmd != TIOCSERGWILD)  &&
	    (cmd != TIOCSERSWILD) && (cmd != TIOCSERGSTRUCT)) {
		if (tty->flags & (1 << TTY_IO_ERROR))
		    return -EIO;
	}
	
	switch (cmd) {

		case TCSBRK:	/* SVID version: non-zero arg --> no break */
			retval = tty_check_change(tty);
			if (retval)
				return retval;
			tty_wait_until_sent(tty, 0);
			if (!arg)
				send_break(info, HZ/4);	/* 1/4 second */
			return 0;
		case TCSBRKP:	/* support for POSIX tcsendbreak() */
			retval = tty_check_change(tty);
			if (retval)
				return retval;
			tty_wait_until_sent(tty, 0);
			send_break(info, arg ? arg*(HZ/10) : HZ/4);
			return 0;
		case TIOCGSOFTCAR:
			error = verify_area(VERIFY_WRITE, (void *) arg,sizeof(long));
			if (error)
				return error;
			put_fs_long(C_CLOCAL(tty) ? 1 : 0,
				    (unsigned long *) arg);
			return 0;
		case TIOCSSOFTCAR:
			arg = get_fs_long((unsigned long *) arg);
			tty->termios->c_cflag =
				((tty->termios->c_cflag & ~CLOCAL) |
				 (arg ? CLOCAL : 0));
			return 0;


		case TIOCGSERIAL:
			error = verify_area(VERIFY_WRITE, (void *) arg,
						sizeof(struct serial_struct));
			if (error)
				return error;
			return get_serial_info(info,
					       (struct serial_struct *) arg);

		case TIOCSSERIAL:
			return set_serial_info(info,
					       (struct serial_struct *) arg);
		case TIOCSERGETLSR: /* Get line status register */
			error = verify_area(VERIFY_WRITE, (void *) arg,
				sizeof(unsigned int));
			if (error)
				return error;
			else
			    return get_lsr_info(info, (unsigned int *) arg);

		case TIOCSERGSTRUCT:
			error = verify_area(VERIFY_WRITE, (void *) arg,
						sizeof(struct p2001_serial));
			if (error)
				return error;
			memcpy_tofs((struct p2001_serial *) arg,
				    info, sizeof(struct p2001_serial));
			return 0;
			
		default:
			return -ENOIOCTLCMD;
		}
	return 0;
}

static void rs_set_termios(struct tty_struct *tty, struct termios *old_termios)
{
	struct p2001_serial *info = (struct p2001_serial *)tty->driver_data;

	if (tty->termios->c_cflag == old_termios->c_cflag)
		return;

	change_speed(info);

	if ((old_termios->c_cflag & CRTSCTS) &&
	    !(tty->termios->c_cflag & CRTSCTS)) {
		tty->hw_stopped = 0;
		rs_start(tty);
	}
	
}

/*
 * ------------------------------------------------------------
 * rs_close()
 * 
 * This routine is called when the serial port gets closed.  First, we
 * wait for the last remaining data to be sent.  Then, we unlink its
 * S structure from the interrupt chain if necessary, and we free
 * that IRQ if nothing is left in the chain.
 * ------------------------------------------------------------
 */
static void rs_close(struct tty_struct *tty, struct file * filp)
{
	struct p2001_serial * info = (struct p2001_serial *)tty->driver_data;
	unsigned long flags;

	if (!info || serial_paranoia_check(info, tty->device, "rs_close"))
		return;
	
	save_flags(flags); cli();
	
	if (tty_hung_up_p(filp)) {
		restore_flags(flags);
		return;
	}
	
#ifdef SERIAL_DEBUG_OPEN
	printk("rs_close ttyS%d, count = %d\n", info->line, info->count);
#endif
	if ((tty->count == 1) && (info->count != 1)) {
		/*
		 * Uh, oh.  tty->count is 1, which means that the tty
		 * structure will be freed.  Info->count should always
		 * be one in these conditions.  If it's greater than
		 * one, we've got real problems, since it means the
		 * serial port won't be shutdown.
		 */
		printk("rs_close: bad serial port count; tty->count is 1, "
		       "info->count is %d\n", info->count);
		info->count = 1;
	}
	if (--info->count < 0) {
		printk("rs_close: bad serial port count for ttyS%d: %d\n",
		       info->line, info->count);
		info->count = 0;
	}
	if (info->count) {
		restore_flags(flags);
		return;
	}
	// closing port so disable interrupts
	set_ints_mode(0, info);
	
	info->flags |= S_CLOSING;
	/*
	 * Save the termios structure, since this port may have
	 * separate termios for callout and dialin.
	 */
	if (info->flags & S_NORMAL_ACTIVE)
		info->normal_termios = *tty->termios;
	if (info->flags & S_CALLOUT_ACTIVE)
		info->callout_termios = *tty->termios;
	/*
	 * Now we wait for the transmit buffer to clear; and we notify 
	 * the line discipline to only process XON/XOFF characters.
	 */
	tty->closing = 1;
	if (info->closing_wait != S_CLOSING_WAIT_NONE)
		tty_wait_until_sent(tty, info->closing_wait);
	/*
	 * At this point we stop accepting input.  To do this, we
	 * disable the receive line status interrupts, and tell the
	 * interrupt driver to stop checking the data ready bit in the
	 * line status register.
	 */

	shutdown(info);
	if (tty->driver.flush_buffer)
		tty->driver.flush_buffer(tty);
	if (tty->ldisc.flush_buffer)
		tty->ldisc.flush_buffer(tty);
	tty->closing = 0;
	info->event = 0;
	info->tty = 0;
	if (tty->ldisc.num != ldiscs[N_TTY].num) {
		if (tty->ldisc.close)
			(tty->ldisc.close)(tty);
		tty->ldisc = ldiscs[N_TTY];
		tty->termios->c_line = N_TTY;
		if (tty->ldisc.open)
			(tty->ldisc.open)(tty);
	}
	if (info->blocked_open) {
		if (info->close_delay) {
			current->state = TASK_INTERRUPTIBLE;
			current->timeout = jiffies + info->close_delay;
			schedule();
		}
		wake_up_interruptible(&info->open_wait);
	}
	info->flags &= ~(S_NORMAL_ACTIVE|S_CALLOUT_ACTIVE|
			 S_CLOSING);
	wake_up_interruptible(&info->close_wait);

#ifdef CFG_P2001_SERIAL_FIQ_DYN
{
  /* free FIQ-function */

  unsigned long flags=0;

  save_flags(flags);
  clf();
  free_fscfiq( p2001_serial_fiqno );
  restore_flags(flags);
}
#endif

#ifdef CFG_P2001_SERIAL_FIQ_RX_NOTIFY_INDIRECT
  p2001_serial_fiq_tc1_deinit();
#endif

  M_RTS_RESET();

	restore_flags(flags);
}

/*
 * rs_hangup() --- called by tty_hangup() when a hangup is signaled.
 */
void rs_hangup(struct tty_struct *tty)
{
	struct p2001_serial * info = (struct p2001_serial *)tty->driver_data;
	
	if (serial_paranoia_check(info, tty->device, "rs_hangup"))
		return;
	
	rs_flush_buffer(tty);
	shutdown(info);
	info->event = 0;
	info->count = 0;
	info->flags &= ~(S_NORMAL_ACTIVE|S_CALLOUT_ACTIVE);
	info->tty = 0;
	wake_up_interruptible(&info->open_wait);
}

/*
 * ------------------------------------------------------------
 * rs_open() and friends
 * ------------------------------------------------------------
 */
static int block_til_ready(struct tty_struct *tty, struct file * filp,
			   struct p2001_serial *info)
{
	unsigned long flags;
	struct wait_queue wait = { current, NULL };
	int		retval;
	int		do_clocal = 0;

	/*
	 * If the device is in the middle of being closed, then block
	 * until it's done, and then try again.
	 */
	if (info->flags & S_CLOSING) {
		interruptible_sleep_on(&info->close_wait);
#ifdef SERIAL_DO_RESTART
		if (info->flags & S_HUP_NOTIFY)
			return -EAGAIN;
		else
			return -ERESTARTSYS;
#else
		return -EAGAIN;
#endif
	}

	/*
	 * If this is a callout device, then just make sure the normal
	 * device isn't being used.
	 */
	if (tty->driver.subtype == SERIAL_TYPE_CALLOUT) {
		if (info->flags & S_NORMAL_ACTIVE)
			return -EBUSY;
		if ((info->flags & S_CALLOUT_ACTIVE) &&
		    (info->flags & S_SESSION_LOCKOUT) &&
		    (info->session != current->session))
		    return -EBUSY;
		if ((info->flags & S_CALLOUT_ACTIVE) &&
		    (info->flags & S_PGRP_LOCKOUT) &&
		    (info->pgrp != current->pgrp))
		    return -EBUSY;
		info->flags |= S_CALLOUT_ACTIVE;
		return 0;
	}
	
	/*
	 * If non-blocking mode is set, or the port is not enabled,
	 * then make the check up front and then exit.
	 */
	if ((filp->f_flags & O_NONBLOCK) ||
	    (tty->flags & (1 << TTY_IO_ERROR))) {
		if (info->flags & S_CALLOUT_ACTIVE)
			return -EBUSY;
		info->flags |= S_NORMAL_ACTIVE;
		return 0;
	}

	if (info->flags & S_CALLOUT_ACTIVE) {
		if (info->normal_termios.c_cflag & CLOCAL)
			do_clocal = 1;
	} else {
		if (tty->termios->c_cflag & CLOCAL)
			do_clocal = 1;
	}
	
	/*
	 * Block waiting for the carrier detect and the line to become
	 * free (i.e., not in use by the callout).  While we are in
	 * this loop, info->count is dropped by one, so that
	 * rs_close() knows when to free things.  We restore it upon
	 * exit, either normal or abnormal.
	 */
	retval = 0;
	add_wait_queue(&info->open_wait, &wait);
#ifdef SERIAL_DEBUG_OPEN
	printk("block_til_ready before block: ttyS%d, count = %d\n",
	       info->line, info->count);
#endif
	info->count--;
	info->blocked_open++;
	while (1) {
		save_flags(flags);cli();
		if (!(info->flags & S_CALLOUT_ACTIVE))
			p2001_rtsdtr(info, 1);
		restore_flags(flags);
		current->state = TASK_INTERRUPTIBLE;
		if (tty_hung_up_p(filp) ||
		    !(info->flags & S_INITIALIZED)) {
#ifdef SERIAL_DO_RESTART
			if (info->flags & S_HUP_NOTIFY)
				retval = -EAGAIN;
			else
				retval = -ERESTARTSYS;	
#else
			retval = -EAGAIN;
#endif
			break;
		}
		if (!(info->flags & S_CALLOUT_ACTIVE) &&
		    !(info->flags & S_CLOSING) && do_clocal)
			break;
		if (current->signal & ~current->blocked) {
			retval = -ERESTARTSYS;
			break;
		}
#ifdef SERIAL_DEBUG_OPEN
		printk("block_til_ready blocking: ttyS%d, count = %d\n",
		       info->line, info->count);
#endif
		schedule();
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&info->open_wait, &wait);
	if (!tty_hung_up_p(filp))
		info->count++;
	info->blocked_open--;
#ifdef SERIAL_DEBUG_OPEN
	printk("block_til_ready after blocking: ttyS%d, count = %d\n",
	       info->line, info->count);
#endif
	if (retval)
		return retval;
	info->flags |= S_NORMAL_ACTIVE;
	if(!info->use_ints){	
		serialpoll.data = (void *)info;		
		queue_task(&serialpoll, &tq_timer);
	}
	return 0;
}	

/*
 * This routine is called whenever a serial port is opened.  It
 * enables interrupts for a serial port, linking in its S structure into
 * the IRQ chain.   It also performs the serial-specific
 * initialization for the tty structure.
 */
int rs_open(struct tty_struct *tty, struct file * filp)
{
	struct p2001_serial	*info;
	int 			retval, line;

	line = MINOR(tty->device) - tty->driver.minor_start;
	
	if (line != 0) /* we have exactly one */
		return -ENODEV;

	info = &p2001_info;
#if 0
	/* Is the kgdb running over this line? */
	if (info->kgdb_channel)
		return -ENODEV;
#endif
	if (serial_paranoia_check(info, tty->device, "rs_open"))
		return -ENODEV;
#ifdef SERIAL_DEBUG_OPEN
	printk("rs_open %s%d, count = %d\n", tty->driver.name, info->line,
	       info->count);
#endif
	info->count++;
	tty->driver_data = info;
	info->tty = tty;

	/*
	 * Start up serial port
	 */
	retval = startup(info);
	if (retval)
		return retval;
	
	retval = block_til_ready(tty, filp, info);
	if (retval) {
#ifdef SERIAL_DEBUG_OPEN
		printk("rs_open returning after block_til_ready with %d\n",
		       retval);
#endif
		return retval;
	}

	if ((info->count == 1) && (info->flags & S_SPLIT_TERMIOS)) {
		if (tty->driver.subtype == SERIAL_TYPE_NORMAL)
			*tty->termios = info->normal_termios;
		else 
			*tty->termios = info->callout_termios;
		change_speed(info);
	}

	info->session = current->session;
	info->pgrp = current->pgrp;



#ifdef CFG_P2001_SERIAL_DIAG

  p2001_serial_diag_reset();

#endif


#ifdef CFG_P2001_SERIAL_FIQ_DYN
{
  /* register FIQ-function */

  unsigned long flags=0;

  save_flags(flags);
  clf();
  p2001_serial_fiqno = register_fscfiq( p2001_serial_fiq );
  restore_flags(flags);
}
#endif

#ifdef CFG_P2001_SERIAL_FIQ_RX_NOTIFY_INDIRECT
  p2001_serial_fiq_tc1_init();
#endif

#ifdef CFG_P2001_SERIAL_FIQ__SOFTWARE_RTS
  /* Use Highway 5 as PIO for assertion of RTS */
  WRITE_REG( 0x152040, 0 );
  WRITE_REG( 0x152044, 0 );
#endif

  M_RTS_RESET();

#ifdef SERIAL_DEBUG_OPEN
	printk("rs_open ttyS%d successful...\n", info->line);
#endif
	return 0;
}

extern void register_console(void (*proc)(const char *));
#if 0

static inline void
rs_cons_check(struct p2001_serial *ss, int channel)
{
	int i, o, io;
	static consout_registered = 0;
	static msg_printed = 0;

	i = o = io = 0;

	/* Is this one of the serial console lines? */
	if((p2001_cons_chanout != channel) &&
	   (p2001_cons_chanin != channel))
		return;
	p2001_conschan = ss->p2001_channel;
	p2001_consinfo = ss;

	/* Register the console output putchar, if necessary */
	if((p2001_cons_chanout == channel)) {
		o = 1;
		/* double whee.. */
		if(!consout_registered) {
			register_console(p2001_console_print);
			consout_registered = 1;
		}
	}

	/* If this is console input, we handle the break received
	 * status interrupt on this line to mean prom_halt().
	 */
	if(p2001_cons_chanin == channel) {
		ss->break_abort = 1;
		i = 1;
	}
	if(o && i)
		io = 1;
	if(ss->baud != 9600)
		panic("Console baud rate weirdness");

	/* Set flag variable for this port so that it cannot be
	 * opened for other uses by accident.
	 */
	ss->is_cons = 1;

	if(io) {
		if(!msg_printed) {
			printk("zs%d: console I/O\n", ((channel>>1)&1));
			msg_printed = 1;
		}
	} else {
		printk("zs%d: console %s\n", ((channel>>1)&1),
		       (i==1 ? "input" : (o==1 ? "output" : "WEIRD")));
	}
}
#endif

#ifndef CFG_P2001_SERIAL_FIQ 
static struct irqaction irq_usarta = {
	rs_interrupta, 0, 0, "uart", NULL, NULL
};
#endif

#if 0
static struct irqaction irq_usartb = {
	rs_interruptb, 0, 0, "usartb", NULL, NULL
};
#endif
extern int setup_arm_irq(int, struct irqaction *);

void interrupts_init(struct p2001_serial *info)
{
#ifndef CFG_P2001_SERIAL_FIQ
	irq_usarta.dev_id = info;
	setup_arm_irq(IRQ_UART, &irq_usarta);
#endif
}

/* rs_init inits the driver */
int rs_p2001_init(void)
{
	int flags,i;
	struct p2001_serial *info;
	/* Setup base handler, and timer table. */
	init_bh(SERIAL_BH, do_serial_bh);
	timer_table[RS_TIMER].fn = rs_timer;
	timer_table[RS_TIMER].expires = 0;

	/* Initialize the tty_driver structure */
	
	memset(&serial_driver, 0, sizeof(struct tty_driver));
	serial_driver.magic = TTY_DRIVER_MAGIC;
	serial_driver.name = "ttyS";
	serial_driver.major = TTY_MAJOR;
	serial_driver.minor_start = 64;
	serial_driver.num = 1;
	serial_driver.type = TTY_DRIVER_TYPE_SERIAL;
	serial_driver.subtype = SERIAL_TYPE_NORMAL;
	serial_driver.init_termios = tty_std_termios;

	serial_driver.init_termios.c_cflag =
		B57600 | CS8 | CREAD | HUPCL | CLOCAL;
	serial_driver.flags = TTY_DRIVER_REAL_RAW;
	serial_driver.refcount = &serial_refcount;
	serial_driver.table = &serial_table;
	serial_driver.termios = &serial_termios;
	serial_driver.termios_locked = &serial_termios_locked;

	serial_driver.open = rs_open;
	serial_driver.close = rs_close;
	serial_driver.write = rs_write;
	serial_driver.flush_chars = rs_flush_chars;
	serial_driver.write_room = rs_write_room;
	serial_driver.chars_in_buffer = rs_chars_in_buffer;
	serial_driver.flush_buffer = rs_flush_buffer;
	serial_driver.ioctl = rs_ioctl;
	serial_driver.throttle = rs_throttle;
	serial_driver.unthrottle = rs_unthrottle;
	serial_driver.set_termios = rs_set_termios;
	serial_driver.stop = rs_stop;
	serial_driver.start = rs_start;
	serial_driver.hangup = rs_hangup;
	serial_driver.set_ldisc = rs_set_ldisc;

	/*
	 * The callout device is just like normal device except for
	 * major number and the subtype code.
	 */
	callout_driver = serial_driver;
	callout_driver.name = "cua";
	callout_driver.major = TTYAUX_MAJOR;
	callout_driver.subtype = SERIAL_TYPE_CALLOUT;

	if (tty_register_driver(&serial_driver))
		panic("Couldn't register serial driver\n");
	if (tty_register_driver(&callout_driver))
		panic("Couldn't register callout driver\n");
	
	save_flags(flags); cli();
	i=0;
	info = &p2001_info;
	info->magic = SERIAL_MAGIC;
	info->uart = (volatile union uart_regs *) UART_BASE_ADDR;
	info->tty = 0;
#if 0
	info->irqmask = (i)?AIC_UB:AIC_UA;
#else
  #ifndef CFG_P2001_SERIAL_FIQ 
	info->irqmask = IRQ_UART;
  #endif
#endif
#ifndef CFG_P2001_SERIAL_FIQ 
	info->irq = IRQ_UART;
#endif
	info->port = i+1;
	set_ints_mode(0,info);
	info->custom_divisor = 16;
	info->close_delay = 50;
	info->closing_wait = 3000;
	info->cts_state = 1;
	info->x_char = 0;
	info->event = 0;
	info->count = 0;
	info->blocked_open = 0;
	info->tqueue.routine = do_softint;
	info->tqueue.data = info;
	info->tqueue_hangup.routine = do_serial_hangup;
	info->tqueue_hangup.data = info;
	info->callout_termios =callout_driver.init_termios;
	info->normal_termios = serial_driver.init_termios;
	info->open_wait = 0;
	info->close_wait = 0;
	info->line = 0;
	info->is_cons = (i)?0:1; /* Means shortcuts work */

	interrupts_init(info);
	restore_flags(flags);
// hack to do polling
	serialpoll.routine = serpoll;
	serialpoll.data = 0;
	
	return 0;
}

/*
 * register_serial and unregister_serial allows for serial ports to be
 * configured at run-time, to support PCMCIA modems.
 */
/* SPARC: Unused at this time, just here to make things link. */
int register_serial(struct serial_struct *req)
{
	return -1;
}

void unregister_serial(int line)
{
	return;
}

#if 0
/* Hooks for running a serial console.  con_init() calls this if the
 * console is being run over one of the ttya/ttyb serial ports.
 * 'chip' should be zero, as chip 1 drives the mouse/keyboard.
 * 'channel' is decoded as 0=TTYA 1=TTYB, note that the channels
 * are addressed backwards, channel B is first, then channel A.
 */
void
rs_cons_hook(int chip, int out, int channel)
{
	if(chip)
		panic("rs_cons_hook called with chip not zero");
	if(!p2001_chips[chip]) {
		p2001_chips[chip] = get_zs(chip);
		/* Two channels per chip */
		p2001_channels[(chip*2)] = &p2001_chips[chip]->channelA;
		p2001_channels[(chip*2)+1] = &p2001_chips[chip]->channelB;
	}
	p2001_info[channel].p2001_channel = p2001_channels[channel];
	p2001_info[channel].change_needed = 0;
	p2001_info[channel].clk_divisor = 16;
	p2001_info[channel].p2001_baud = get_zsbaud(&p2001_info[channel]);
	rs_cons_check(&p2001_info[channel], channel);
	if(out)
		p2001_cons_chanout = ((chip * 2) + channel);
	else
		p2001_cons_chanin = ((chip * 2) + channel);

}

/* This is called at boot time to prime the kgdb serial debugging
 * serial line.  The 'tty_num' argument is 0 for /dev/ttya and 1
 * for /dev/ttyb which is determined in setup_arch() from the
 * boot command line flags.
 */
void
rs_kgdb_hook(int tty_num)
{
	int chip = 0;

	if(!p2001_chips[chip]) {
		p2001_chips[chip] = get_zs(chip);
		/* Two channels per chip */
		p2001_channels[(chip*2)] = &p2001_chips[chip]->channelA;
		p2001_channels[(chip*2)+1] = &p2001_chips[chip]->channelB;
	}
	p2001_info[tty_num].p2001_channel = p2001_channels[tty_num];
	p2001_kgdbchan = p2001_info[tty_num].p2001_channel;
	p2001_info[tty_num].change_needed = 0;
	p2001_info[tty_num].clk_divisor = 16;
	p2001_info[tty_num].p2001_baud = get_zsbaud(&p2001_info[tty_num]);
	p2001_info[tty_num].kgdb_channel = 1;     /* This runs kgdb */
	p2001_info[tty_num ^ 1].kgdb_channel = 0; /* This does not */
	/* Turn on transmitter/receiver at 8-bits/char */
	kgdb_chaninit(&p2001_info[tty_num], 0, 9600);
	ZS_CLEARERR(p2001_kgdbchan);
	udelay(5);
	ZS_CLEARFIFO(p2001_kgdbchan);
}
#endif



#ifdef CFG_P2001_SERIAL_FIQ

/*


   RX flip buffer processing: see flush_to_ldisc() in tty_io.c

   TTY_FLIPBUF_SIZE   in tty.h  ( 640 bytes, was 512 byte ) 

*/

#define CFG_P2001_SERIAL_FIQ_FLIPBUF_HIGHWATER		128 /* high water threshold for Flip-Buffer */


#ifdef CFG_P2001_SERIAL_FIQ_RX_NOTIFY_INDIRECT

  #define M_P2001_SERIAL_FIQ_RX_NOTIFY()		p2001_serial_fiq_tc1_start()
 
#else

  #define M_P2001_SERIAL_FIQ_RX_NOTIFY()		queue_task_irq_off(&tty->flip.tqueue, &tq_timer)

#endif


#define PRD 
#define print(PRP) printk PRP


#ifdef CFG_P2001_SERIAL_DIAG

  #define M_DIAG_INCR(C,P)     v_diag[C][P]++
  #define M_DIAG_DECR(C,P)     v_diag[C][P]--
  #define M_DIAG_ADD(C,P,V)    v_diag[C][P]+=V
  #define M_DIAG_SUB(C,P,V)    v_diag[C][P]-=V
  #define M_DIAG_SET(C,P,V)    v_diag[C][P]=V
  #define M_DIAG_GET(C,P)      (v_diag[C][P])


  #define M_DIAG_ROWS		33
  #define M_DIAG_COLUMNS	3

  unsigned int v_diag[M_DIAG_COLUMNS][M_DIAG_ROWS];

#else

  #define M_DIAG_INCR(C,P)
  #define M_DIAG_DECR(C,P)
  #define M_DIAG_ADD(C,P,V)
  #define M_DIAG_SUB(C,P,V)
  #define M_DIAG_SET(C,P,V)
  #define M_DIAG_GET(C,P)

#endif


#ifdef CFG_P2001_SERIAL_FIQ_RX_NOTIFY_INDIRECT

static unsigned int tc1_hdlno=0;

void p2001_serial_fiq_tc1_init(void)
{
  tc1_hdlno=tc1_register( p2001_serial_fiq_tc1 );    
}

void p2001_serial_fiq_tc1_deinit(void)
{
  tc1_free( tc1_hdlno );    
}

void p2001_serial_fiq_tc1_start(void)
{
  tc1_mark_active( tc1_hdlno );
}

void p2001_serial_fiq_tc1(void)
{
  struct p2001_serial *info=&p2001_info;
  struct tty_struct *tty = info->tty;

  queue_task_irq_off(&tty->flip.tqueue, &tq_timer);
}
#endif


static unsigned int p2001_serial_fiq_rx_timeout=0;
static unsigned int p2001_serial_fiq_rts_asserted=0;

void p2001_serial_fiq(void)
{
  struct p2001_serial *info=&p2001_info;
  struct tty_struct *tty = info->tty;

  volatile union uart_regs *uart = info->uart;


  M_DIAG_INCR( 2, 0 );    /* count calls of this function */

/* ERROR */

{
   u_32 status;

   status = info->uart->r.irq_status & UART_IRQ_IRQ_MASK;

   if( status & UART_ERROR_INTS )  
   { 
        M_DIAG_INCR( 2, 25 );

	if( status & UART_IRQ_RX_FIFO_RAR_ERR )
	{
          M_DIAG_INCR( 2, 26 );
	}	 
	if( status & UART_IRQ_RX_BRK )
	{
          M_DIAG_INCR( 2, 27 );
	}	 
	if( status & UART_IRQ_ECHO_ERR )
	{
          M_DIAG_INCR( 2, 28 );
	}	 
	if( status & UART_IRQ_RX_LOST )
	{
          M_DIAG_INCR( 2, 29 );
	}	 

	uart->w.irq_status = uart->r.irq_status & ~status;	/* reset status */
   }
}


/* RX */

/* something to get from the HW-FiFo ? */

  {
    unsigned int count;
    unsigned int i;

    count = (uart->r.status & UART_STAT_RX_DELTA_MASK) >> UART_STAT_RX_DELTA_SHIFT;

    M_DIAG_INCR( 1, count );    /* store number of bytes in RX-fifo */

    if( count )
    {
        p2001_serial_fiq_rx_timeout=0;

#ifdef CFG_P2001_SERIAL_DIAG
	if( tty->flip.count >= M_DIAG_GET(2, 23) )
	{
                M_DIAG_SET( 2, 23, tty->flip.count );
	}
#endif

	if( tty->flip.count + count < TTY_FLIPBUF_SIZE )
	{
		for (i = count; i; i--)
		{
			//tty_insert_flip_char(struct tty_struct *tty, unsigned char ch, char flag)
   			*tty->flip.char_buf_ptr++ = (char) uart->r.rx1;
		}
	
		tty->flip.count += count;

		M_DIAG_ADD(2, 20, count);

	}

    }
    else
    {        
      /* timeout ? */		// TODO add configurable timeout ( n* FSCs)
      if( !p2001_serial_fiq_rx_timeout && tty->flip.count )
      {
        M_DIAG_INCR( 2, 22 );

        p2001_serial_fiq_rx_timeout=1;

	M_P2001_SERIAL_FIQ_RX_NOTIFY();
      }

    }


	if (tty->flip.count >= CFG_P2001_SERIAL_FIQ_FLIPBUF_HIGHWATER )
	{
                M_DIAG_INCR( 2, 21 );

		M_P2001_SERIAL_FIQ_RX_NOTIFY();
	}


#ifdef CFG_P2001_SERIAL_FIQ__SOFTWARE_RTS
	if( tty->flip.count > (TTY_FLIPBUF_SIZE-32) )
	{
		if( ! p2001_serial_fiq_rts_asserted )
		{
	  		/* Assert RTS to avoid Flip-Buffer overflow */
        	        M_RTS_SET();
	        	M_DIAG_INCR( 2, 5 );
	  		p2001_serial_fiq_rts_asserted=1;
	                M_DIAG_INCR( 2, 24 );
		}
	}
	else
	{
		if( p2001_serial_fiq_rts_asserted )
		{
			/* Reset RTS */
        	        M_RTS_RESET();
		        M_DIAG_INCR( 2, 6 );
  			p2001_serial_fiq_rts_asserted=0;
		}
	}
#endif

  }


/* TX */

/* something to send in the TX-buffer ? */
  {
    unsigned int count;

    if( info->xmit_cnt && !p2001_serial_fiq_rts_asserted )	/* something left to send ? */
    {
      M_DIAG_INCR( 2, 12 );	/* count */

      M_DIAG_SET( 2, 13, info->xmit_cnt );

      count = UART_TX_FIFO_SIZE - ((uart->r.status & UART_STAT_TX_DELTA_MASK)
				     >> UART_STAT_TX_DELTA_SHIFT);

      M_DIAG_INCR( 0, count );    /* store number of free bytes in TX-fifo */

      /* copy from buffer to TX-fifo */

	if (count > info->xmit_cnt)
		count = info->xmit_cnt;

	if (info->xmit_tail + count < SERIAL_XMIT_SIZE)
	{
                M_DIAG_ADD(2, 10, count);

//		xmit_string(info, info->xmit_buf + info->xmit_tail, count);
		{
			char *p = info->xmit_buf + info->xmit_tail;
			int len = count;
			while (len-- > 0)
			{
				uart->w.tx1 = (unsigned char) *p++;
			}
		}

		info->xmit_tail += count;
		info->xmit_cnt -= count;
	}
	else
	{
                M_DIAG_ADD(2, 11, SERIAL_XMIT_SIZE - info->xmit_tail);

//		xmit_string(info, info->xmit_buf + info->xmit_tail,
//			    SERIAL_XMIT_SIZE - info->xmit_tail);
		{
			char *p = info->xmit_buf + info->xmit_tail;
			int len = SERIAL_XMIT_SIZE - info->xmit_tail;
			while (len-- > 0)
			{
				uart->w.tx1 = (unsigned char) *p++;
			}
		}		

		info->xmit_cnt -= SERIAL_XMIT_SIZE - info->xmit_tail;
		info->xmit_tail = 0;
	}

    }

  }

}


#ifdef CFG_P2001_SERIAL_DIAG

void p2001_serial_diag_reset(void)
{

  unsigned int ch;
  unsigned int i;

  for(i=0; i<M_DIAG_ROWS; i++)
  {
    for(ch=0; ch<M_DIAG_COLUMNS; ch++)
    {
       v_diag[ch][i]=0;
    }    

  }
}

#endif


void p2001_serial_mon(void)
{
#ifdef CFG_P2001_SERIAL_DIAG

  unsigned int ch;
  unsigned int i;

  print(( PRD "p2001_serial_mon  3 \n"));

  print(( PRD "CFG_P2001_SERIAL_FIQ_FLIPBUF_HIGHWATER %d \n", CFG_P2001_SERIAL_FIQ_FLIPBUF_HIGHWATER ));
  print(( PRD "TTY_FLIPBUF_SIZE %d \n", TTY_FLIPBUF_SIZE ));
  print(( PRD "\n" ));
  print(( PRD "monitor runnning from address %p \n", p2001_serial_mon )); 
  print(( PRD "fiq-handler running from address %p \n", p2001_serial_fiq )); 
  print(( PRD "\n" ));

  for(i=0; i<=32; i++)
  {
    print(( PRD "v_diag[%d]   ", i ));

    for(ch=0; ch<=2; ch++)
    {
       print(( PRD " %6d ", v_diag[ch][i] ));
    }    
    print(( PRD "\n" ));

  }
#endif

}


#endif
