#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>
#include <linux/sched.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <linux/proc_fs.h>

#include <asm/hardware.h>
#include <asm/arch/triohook.h>
#include "triopio.h"

#define HOOK_MINOR	180

#define INTERVAL_20Hz		5		// 1 jiffies ~ 10ms (?), max detect 40ms
#define INTERVAL_MIN		1		// 1 jiffies ~ 10ms (?), max detect 40ms
#define INTERVAL_50MS		5		// 1 jiffies ~ 10ms (?), detect 100ms

static unsigned char last_hook;
static unsigned char hook;
static unsigned int detecting;
static unsigned int delayed;
static ulong lasttime;
static unsigned int hook_polling;

static unsigned char is_off_hook(void);
static int hook_interrupt(int polled);

struct tq_struct hook_poll;

int queuedcnt;

static struct pio_handler hook_handler;
static struct trioport *hookport;

struct wait_queue *hook_queue = NULL;

#define HOOK_MASK	PIO(3)

static unsigned char is_off_hook(){
	return(!(trioport_get_in(hookport) & HOOK_MASK));
}
int pio_hook_interrupt(struct trioport *p, u_32 changed){
	return((changed & HOOK_MASK)?hook_interrupt(0):0);
}
int hook_interrupt(int polled){
	int count=0;
	hook = is_off_hook();
	do{
		if (hook!=last_hook || !polled){
			last_hook = hook;
			delayed = 0;
			detecting = 1;
			if (!queuedcnt){
				queue_task_irq_off(&hook_poll, &tq_timer);
				queuedcnt++;
			}
			lasttime = jiffies;
		}else{
			if (detecting && hook_queue){
				if (jiffies - lasttime > INTERVAL_50MS){
					detecting = 0;
					delayed = 1;
					wake_up_interruptible(&hook_queue);
				}else{
					if (!queuedcnt){
						queue_task_irq_off(&hook_poll, &tq_timer);
						queuedcnt++;
					}
				}
			}else{
				if (hook_polling){
					if (!queuedcnt){
						queue_task_irq_off(&hook_poll, &tq_timer);
						queuedcnt++;
					}
				}
			}
		}
		hook = is_off_hook();
		count++;
	}while((count < 2) && (last_hook != hook));
	return 1;
}
static void hookpoll(void *data){
	if(queuedcnt)
		queuedcnt--;
	if(queuedcnt > 2){
		printk("!%d", queuedcnt);
	}
	hook_interrupt(1);
}
static int hook_read(struct inode *inode, struct file *file, char *buffer, int arg){
	if (hook){
		put_user(A_OFFHOOK, buffer);
	}else{
		put_user(A_ONHOOK, buffer);
	}
	last_hook = hook;
	delayed = 0;
	detecting = 0;
	if (hook_polling){
		queue_task_irq_off(&hook_poll, &tq_timer);
	}
	return 1;
}
static int hook_select(struct inode *inode, struct file *file, int sel_type, select_table *wait){
	if (sel_type == SEL_IN) {
		if ((hook == last_hook) && (!delayed)){
			select_wait(&hook_queue, wait);
			return 0;
		}
	}
	return 1;
}
static int hook_open(struct inode *inode, struct file *file)
{
	hook_polling=0;
	memset(&hook_handler, 0, sizeof(hook_handler));

	// first get the pointers and init IN/OUT
	if(pio_hook_driver(&hook_handler)<0){
		printk("Hook driver not initialized");
		return -1;
	}
	hookport = hook_handler.pa;
	trioport_change_out(hookport, HOOK_MASK, ~HOOK_MASK);
	
	if(!hook_polling){
		hook_handler.pio_interrupt = pio_hook_interrupt;
		hook_handler.pa_mask = HOOK_MASK;
	}
	if(pio_hook_driver(&hook_handler)<0){
		printk("Hook driver not initialized");
		return -1;
	}
	hook = last_hook = is_off_hook();
	delayed = detecting = 0;
	hook_poll.routine = hookpoll;
	hook_poll.data = 0;
	queuedcnt=0;
	if (hook_polling){
		queue_task_irq_off(&hook_poll, &tq_timer);
	}
	return 0;
}
static void hook_release(struct inode *inode, struct file *file){
	pio_remove_driver(&hook_handler);
}
static struct file_operations hook_fops = {
	NULL,
	hook_read,
	NULL,		/* No write */
	NULL,	/* No readdir */
	hook_select,
	NULL,
	NULL,		/* No mmap */
	hook_open,
	hook_release
};
static struct miscdevice hook_dev=
{
	HOOK_MINOR,
	"triohook",
	&hook_fops
};
int triohook_init(void)
{
	misc_register(&hook_dev);
	return 0;
}

