#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/config.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/arch-trio/irq.h>
#include <asm/arch-trio/trioport.h>
#include <asm/arch-trio/triokb.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <asm/bitops.h>
#include <asm/delay.h>
#include "triopio.h"
/*
 *	We sponge a minor off of the misc major. No need slurping
 *	up another valuable major dev number for this.
 */

#define TRIOKB_MINOR	178

static struct wait_queue *user_queue = NULL;
static struct tq_struct kb_poll;

#define MAX_KEYS 16
#define LOW_COUNT_LIMIT 1			// 40 ms minimum key validation time
#define DEF_COUNT_LIMIT	2			// 80 ms Default time out for a key		
#define AUTOREPEAT_DELAY_COUNT 6	// 240 ms delay before autorepeat 

#define POLL_PERIOD	10				// in ms */
#define BUF_SIZE	8

#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

struct pio_handler kb_handler;
static struct triokb_piomask kb_mask;
static unsigned int last_keys_msk;
static unsigned int last_keys_duration;
static int cols[63];
static int colnb;
static int rownb;
static int keynb;
static int key_number;
static unsigned int keybstat;
static int autorepeat_delay;
static unsigned char kb_buffer[BUF_SIZE];
static int kb_cnt;
static int kb_head;
static int kb_tail;


static struct skey kb_stat[MAX_KEYS]={
				 {-1, DEF_COUNT_LIMIT,0,0},			
				 {KEY_NAVY_DOWN, DEF_COUNT_LIMIT,0,0},			
				 {KEY_NAVY_UP,DEF_COUNT_LIMIT,0,0},
				 {-1, DEF_COUNT_LIMIT,0,0},
				 {KEY_NAVY_LEFT, DEF_COUNT_LIMIT,0,0},
				 {KEY_NAVY_RIGHT, DEF_COUNT_LIMIT,0,0},
				 {KEY_EXT, DEF_COUNT_LIMIT,0,0},
				 {KEY_ENTER, DEF_COUNT_LIMIT,0,0},
				 {KEY_NUM_SPECIAL, DEF_COUNT_LIMIT,0,0},
				 {KEY_FUNCTION, DEF_COUNT_LIMIT,0,0},
				 {KEY_NETCALL, DEF_COUNT_LIMIT,0,0},
				 {KEY_ALPHA, DEF_COUNT_LIMIT,0,0}
};


static struct rows {
	int row[31];
	int nb;
	int row_offset;
	struct trioport *p;
	u_32 mask;
} pa_rows,pb_rows,*p_rows;

static int cur_row;
u_32 col, cola, colb;



static int set_key_parms (struct skey * keyparms)
{
	int i;

	if (!keyparms->key_val) {
		for(i=0; i< keynb;i++) {
			kb_stat[i].timeout = keyparms->timeout/(POLL_PERIOD *rownb);
			if (kb_stat[i].timeout == 0) kb_stat[i].timeout = DEF_COUNT_LIMIT;
			kb_stat[i].autorepeat = keyparms->autorepeat/(POLL_PERIOD *rownb);
//			printk("key_parms = %d, %d, %d \n", kb_stat[i].key_val, kb_stat[i].timeout, kb_stat[i].autorepeat);
		}
	}
	else {
		for(i=0; i< keynb;i++) {
			if (kb_stat[i].key_val == keyparms->key_val) {
				kb_stat[i].timeout = keyparms->timeout/(POLL_PERIOD *rownb);
				if (kb_stat[i].timeout == 0) kb_stat[i].timeout = DEF_COUNT_LIMIT;
				kb_stat[i].autorepeat = keyparms->autorepeat/(POLL_PERIOD *rownb);
//				printk("key_parms = %d, %d, %d \n", kb_stat[i].key_val, kb_stat[i].timeout, kb_stat[i].autorepeat);
				return 0;
			}
		}
	}
	return 0;
}

static int get_duration(unsigned int key)
{
	int i;
	
	for(i=0; i< keynb;i++) {
		if (kb_stat[i].key_val == key) {
			return (kb_stat[i].duration *rownb * POLL_PERIOD);
		}
	}
	return 0;
}


static int calc_tables(void){
	int i,j;
	j=0;

	pa_rows.nb=pb_rows.nb=colnb=rownb=0;
	memset(pa_rows.row, 0, sizeof(pa_rows.row));
	memset(pb_rows.row, 0, sizeof(pb_rows.row));
	for(i=0;i<32;i++){
		if(kb_mask.pra_mask >> i & 1){
			pa_rows.row[j++] = 1<<i;
//			printk("a row%d = %x\n", j-1, pa_rows.row[j-1]);
			pa_rows.nb++;
		}
	}
	pa_rows.row_offset=0;
	pa_rows.p= kb_handler.pa;
	pa_rows.mask=kb_mask.pra_mask;
	j=0;
	for(i=0;i<32;i++){
		if(kb_mask.prb_mask >> i & 1){
			pb_rows.row[j++] = 1<<i;
//			printk("b row%d = %x\n", j-1, pb_rows.row[j-1]);
			pb_rows.nb++;
		}
	}
	pb_rows.row_offset=pa_rows.nb;
	pb_rows.p= kb_handler.pb;
	rownb=pb_rows.nb+pa_rows.nb;
	pb_rows.mask=kb_mask.prb_mask;
	j=1;
	for(i=0;i<32;i++){
		if(kb_mask.pca_mask >> i & 1){
			cols[i] = j++;
//			printk("col%d = %x\n", i, cols[i]);
			colnb++;
		}
		for(i=0;i<32;i++){
			if(kb_mask.pcb_mask >> i & 1){
				cols[i+32] = j++;
//				printk("col%d = %x\n", i+32, cols[i+32]);
				colnb++;
			}
		}
	}
	keynb = colnb*rownb;
	printk("arows %d brows %d columns %d\n", pa_rows.nb, pb_rows.nb, colnb);
	return(colnb);
}



static int kb_input_key(unsigned int * keys)
{
	int j;
	int keys_num=0;
	int key_pressed;

	trioport_change_out(p_rows->p,p_rows->mask,pa_rows.row[cur_row]);
	trioport_set_out(p_rows->p, p_rows->row[cur_row], 0);
	cola = ~(kb_handler.pa->base->pdsr) & kb_mask.pca_mask;
	colb = ~(kb_handler.pb->base->pdsr) & kb_mask.pcb_mask;
	trioport_set_out(p_rows->p, p_rows->row[cur_row], p_rows->row[cur_row]);
	if (cola || colb){
		for (j=0;j<32 && key_number < keynb ;j++) {
			if(cola && ((cola >> j) & 1)) {
				key_pressed = cols[j] + (cur_row+p_rows->row_offset)* colnb;
				keybstat |= 1 << (key_pressed);
				key_number++;
			}
			if(colb && ((colb >> j) & 1)) {
				key_pressed = cols[j+32] + (cur_row+p_rows->row_offset)* colnb;
				keybstat |= 1 << (key_pressed);
				key_number++;
			}
		}
	}

	if (++cur_row >= p_rows->nb || key_number > keynb ) {
		cur_row=0;
		p_rows = (p_rows == &pa_rows)? &pb_rows:&pa_rows;
		if(p_rows == &pa_rows) {
			/* all the rows are done*/
			keys_num=key_number;
			*keys = keybstat;
			key_number=0;
			keybstat=0;
			return keys_num;
		}
	}
	return -1;
}

int write_kb_buffer(int key)
{
	if(kb_cnt >= BUF_SIZE - 1) return 0;
	kb_buffer[kb_head++] = key;
	kb_head &= BUF_SIZE-1;
	kb_cnt++;
//	printk("k=%x  \n",key);
	return(key);
}



static void kbpoll(void *data)
{
	unsigned int keys_msk=0;
	int keysinfo_count=0;
	int wakeup=0;
	int i;

	keysinfo_count = kb_input_key(&keys_msk);
//	printk("keyinfo = %x, %d\n",keys_msk, keysinfo_count); 
	if (keysinfo_count >= 0) {
		if (keys_msk && (keys_msk == last_keys_msk)) {
			last_keys_duration++;
			for(i=0; i< keynb;i++) {
				if ((1<< kb_stat[i].key_val) & keys_msk) {
					kb_stat[i].duration++;
					/* validation time */
					if (kb_stat[i].duration == kb_stat[i].timeout){
						wakeup = write_kb_buffer(kb_stat[i].key_val);
					}
					/* autorepeat - only if we have one key -*/
					else {
						if ((1<< kb_stat[i].key_val) == keys_msk &&
							kb_stat[i].autorepeat && kb_stat[i].duration >= autorepeat_delay &&  
							kb_stat[i].duration % kb_stat[i].autorepeat == 0) {
								wakeup = write_kb_buffer(kb_stat[i].key_val);
						}
					}
				}
				else {
					kb_stat[i].duration=0;
				}
			}
		}	
		else {			// no more key, test if we are under the timeout val
			if(last_keys_msk) {
				last_keys_duration=0;
				for(i=0; i< keynb;i++) {
					if (kb_stat[i].duration > LOW_COUNT_LIMIT && kb_stat[i].duration < kb_stat[i].timeout) {
						wakeup = write_kb_buffer(kb_stat[i].key_val);
					}
					//reset the stuff
					kb_stat[i].duration=0;
				}
			}
		}
		last_keys_msk = keys_msk ;
		if (wakeup && user_queue){
			wake_up_interruptible(&user_queue);
		}
	}
	queue_task_irq_off(&kb_poll, &tq_timer);
}


static int triokb_read(struct inode *inode, struct file *file, char *buf, int count)
{
	int size=0;
	
	if(kb_cnt <= 0) return 0;

	size = MIN(count, kb_cnt);
	if(kb_tail + size < BUF_SIZE){
		memcpy_tofs(buf, (char *)(kb_buffer+kb_tail), size);
	}else{
		memcpy_tofs(buf, (char *)(kb_buffer+kb_tail), BUF_SIZE - kb_tail);
		memcpy_tofs(buf, kb_buffer, kb_tail + size - BUF_SIZE);
	}
	kb_tail = (kb_tail + size) & (BUF_SIZE-1);
	kb_cnt-=size;
	return (size);

}

static int triokb_select(struct inode *inode, struct file *file, int sel_type, select_table *wait){

	if (sel_type == SEL_IN) {
		if(kb_cnt <= 0) {
			select_wait(&user_queue, wait);
			return 0;
		}
	}
	return 1;
}
static int triokb_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg){
	int key_duration;
	struct skey keyparms;
	int i;
	
	switch(cmd){
		case TRIOKB_PIOMASK:
			kb_mask = *(struct triokb_piomask *)arg;
			if(!calc_tables()){
				return -EINVAL;
			}
			if(kb_mask.pra_mask){
				trioport_change_out(kb_handler.pa,kb_mask.pra_mask, 0);
			}
			if(kb_mask.prb_mask){
				trioport_change_out(kb_handler.pb, kb_mask.prb_mask, 0);
			}
			queue_task_irq_off(&kb_poll, &tq_timer);
			break;
		case TRIOKB_STATUS:
			key_duration = (last_keys_duration*rownb) * POLL_PERIOD ;
			put_user(last_keys_msk, &((unsigned int *)arg)[0]);
			put_user(key_duration,&((unsigned int *)arg)[1]);
			break;
		case TRIOKB_KEYDURATION:
			return (get_duration(arg));
		case TRIOKB_HIT:
			return kb_cnt;
		case TRIOKB_FLUSH:
			kb_cnt = kb_head = kb_tail = 0;
			break;
		case TRIOKB_SET_KEY:
			i= verify_area (VERIFY_READ, (void *)arg, sizeof (struct skey));
			if (i) return i;
			memcpy_fromfs ((void *)&keyparms, (void *)arg, sizeof (struct skey));
			set_key_parms (&keyparms);
			break;	
		case TRIOKB_SET_AUTOREPEAT_DELAY:
			autorepeat_delay = (arg/(POLL_PERIOD *rownb))?(arg/(POLL_PERIOD *rownb)):AUTOREPEAT_DELAY_COUNT;
			break;
		case TRIOKB_AUTOREPEAT:
		case TRIOKB_SETBREAK:
		default:
			return -EINVAL;
	}
	return 0;
}
static int triokb_open(struct inode *inode, struct file *file)
{
	return 0;
}
static void triokb_release(struct inode *inode, struct file *file)
{
}
static struct file_operations triokb_fops = {
	NULL,
	triokb_read,
	NULL,
	NULL,		   /* No readdir */
	triokb_select,
	triokb_ioctl,
	NULL,		/* No mmap */
	triokb_open,
	triokb_release
};
static struct miscdevice triokb_dev=
{
	TRIOKB_MINOR,
	"triokb",
	&triokb_fops
};

static int triokb_reinit(void){

	p_rows = &pa_rows;
	cur_row=0;
	keybstat=0;
	kb_cnt = kb_head = kb_tail = 0;
	
	kb_poll.routine = kbpoll;
	kb_poll.data = 0;
	kb_handler.pio_interrupt = NULL;
	if(	pio_hook_driver(&kb_handler) < 0){
		printk("Kb driver not initialized\n");
		return -1;
	}
	return 0;
}	
int triokb_init(void){
	if(triokb_reinit() < 0)
		printk("Kb driver not registered\n");
	else
		misc_register(&triokb_dev);
	return 0;
}

