/*
 *
 */

#define MUX_VERSION		"0.01"

/*
 *	Note that *all* calls to CMOS_READ and CMOS_WRITE are done with
 *	interrupts disabled. Due to the index-port/data-port (0x70/0x71)
 *	design of the RTC, we don't want two different things trying to
 *	get to it at once. (e.g. the periodic 11 min sync from time.c vs.
 *	this driver.)
 */

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>
#include <linux/sched.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>

#include <asm/hardware.h>
#include "triopio.h"
#include <asm/arch/trioled.h>

/*
 *	We sponge a minor off of the misc major. No need slurping
 *	up another valuable major dev number for this.
 */

#define LED_MINOR	177

// model dependent part
#define ALED_PSALOU_APLIO	(PIO(11))
#define ALL_PSALOU_LEDS		ALED_PSALOU_APLIO
#define ALED_PSALOU_NB	1


static struct led2led{
	u_32 locled;
	u_32 publed;
};

#define ALED_MAX_NB	1

static struct led2led psalou_led[ALED_PSALOU_NB]={ALED_PSALOU_APLIO, ALED_APLIO};

ushort led_mask[ALED_MAX_NB]={
	ALED_APLIO,
};
struct trioport *aled_addr;
u_32 all_leds;
u_32 aled_nb;
struct led2led *led_ptr;
struct pio_handler led_handler;


struct tq_struct ledpoll;
static void start_blink(struct aplioled *led);
static void stop_blink(struct aplioled *led);
static void led_blink(void *data);
static int led_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);

struct blink_info{
	struct aplioled led;
	u_32 trueled;
	ulong onstart;
	ulong offstart;
};
struct blink_info blink[ALED_MAX_NB];

static int queued_cnt;
static u_32 getlocled(u_32 al){
	int i;
	u_32 result = 0;
	for(i = 0;i < aled_nb; i++){
		if (led_ptr[i].publed & al)
			result |= led_ptr[i].locled;
	}
	if(!result)
		printk("\nLocal led not found\n");
	return result;
}
static u_32 getpubled(u_32 al){
	int i;
	u_32 result = 0;
	for(i = 0;i < aled_nb; i++){
		if (led_ptr[i].locled & al)
			result |= led_ptr[i].publed;
	}
	if(!result)
		printk("\nGlobal led not found\n");
	return result;
}
static void start_blink(struct aplioled *led){
	int i;
	for(i=0; i< ALED_MAX_NB; i++){
		if (led->al & led_mask[i]){
			blink[i].led = *led;
			blink[i].trueled = getlocled(led->al & led_mask[i]);
			if(blink[i].trueled){
				blink[i].offstart = jiffies;
				blink[i].onstart = 0;
				if (aled_addr->base->pdsr & blink[i].trueled){
//					printk("\n^%d\n",blink[i].trueled);
					trioport_set_out(aled_addr, blink[i].trueled, 0);
				}else{
//					printk("\nv%d\n",blink[i].trueled);
					trioport_set_out(aled_addr, blink[i].trueled, 1);
				}
			}
		}
	}
	if (led->al && !queued_cnt){
		queue_task(&ledpoll, &tq_timer);
		queued_cnt=1;
	}
}
static void stop_blink(struct aplioled *led){
	int i;
	for(i=0; i< ALED_MAX_NB; i++){
		if (led->al & led_mask[i]){
			blink[i].onstart = 0;
			blink[i].offstart = 0;
//			printk("\n$%d\n",blink[i].trueled);
			blink[i].led.al = 0;
			trioport_set_out(aled_addr, blink[i].trueled, 0);
			blink[i].trueled = 0;
		}
	}
}
static void led_blink(void *data){
	int i;
	if(queued_cnt)
		queued_cnt--;
	
	for(i=0; i< ALED_MAX_NB; i++){
		if (blink[i].onstart && (jiffies-blink[i].onstart > blink[i].led.on)){
//			printk("onstart:%d jiffies:%d interval %d\n",blink[i].onstart,jiffies,blink[i].led.on);
			blink[i].onstart = 0;
			blink[i].offstart = jiffies;
			trioport_set_out(aled_addr, blink[i].trueled, blink[i].trueled);
		}else{
			if (blink[i].offstart && (jiffies-blink[i].offstart > blink[i].led.off)){
//				printk("offstart:%d jiffies:%d interval %d\n",blink[i].offstart,jiffies,blink[i].led.off);
				blink[i].offstart = 0;
				blink[i].onstart = jiffies;
				trioport_set_out(aled_addr, blink[i].trueled, 0);
			}
		}
	}
	for(i=0; i< ALED_MAX_NB; i++){
		if (blink[i].trueled){
			if (!queued_cnt){
				queue_task(&ledpoll, &tq_timer);
				queued_cnt=1;
			}
			return;
		}
	}
}

static int led_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
						 unsigned long arg)
{
	switch (cmd) {
		case LED_READ:
		{
			struct aplioled *led = (struct aplioled *)arg;
			led->al = getpubled(aled_addr->base->pdsr & all_leds);
			return 0;
		}
		case LED_ON:
		{
			struct aplioled ll;
			ll.al = (u_32)arg;
			stop_blink(&ll);
			trioport_set_out(aled_addr, getlocled((u_32)arg),getlocled((u_32)arg) );
			return 0;
		}
		case LED_OFF:
		{
			struct aplioled ll;
			ll.al = (u_32)arg;
			stop_blink(&ll);
			trioport_set_out(aled_addr, getlocled((u_32)arg), 0);
			return 0;
		}
		case LED_BLINK:
		{
			struct aplioled *led = (struct aplioled *)arg;
			start_blink(led);
			return 0;
		}
		case LED_STOP_BLINK:
		{
			struct aplioled *led = (struct aplioled *)arg;
			stop_blink(led);
			return 0;
		}
		default:
			return -EINVAL;
	}
}
static int led_open(struct inode *inode, struct file *file)
{
	return 0;
}
#if 0
static void led_release(struct inode *inode, struct file *file)
{
	aled_addr |= all_leds;
	memset(blink, sizeof(blink),0);
}
#endif
static struct file_operations led_fops = {
	NULL,
	NULL,
	NULL,		/* No write */
	NULL,		/* No readdir */
	NULL,
	led_ioctl,
	NULL,		/* No mmap */
	led_open,
	NULL
};

static struct miscdevice led_dev=
{
	LED_MINOR,
	"leds",
	&led_fops
};
int trioled_init(void)
{
	led_handler.pio_interrupt = NULL;
	if(	pio_hook_driver(&led_handler) < 0){
		printk("Led driver not initialized\n");
		return -1;
	}
	ledpoll.routine = led_blink;
	ledpoll.data = (void *)blink;
	memset(blink, sizeof(blink),0);
	queued_cnt=0;
	
	// model dependent part, will be transferred to ioctl
	aled_addr = led_handler.pa;
	led_ptr = psalou_led;
	aled_nb = ALED_PSALOU_NB;
	all_leds = ALL_PSALOU_LEDS;
	trioport_change_out(aled_addr,all_leds, 0);
	
	misc_register(&led_dev);
	return 0;
}
