#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>
#include <linux/sched.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <linux/proc_fs.h>

#include <asm/hardware.h>
#include <asm/arch/triohook.h>
#include "triopio.h"

#define VOICE_MINOR	181

static struct pio_handler voice_handler;
static struct trioport *voiceport;

#define VOICE_MASK	PIO(5)

static int voice_read(struct inode *inode, struct file *file, char *buffer, int arg){
	put_user(((trioport_get_in(voiceport) & VOICE_MASK))?1:0, buffer);
	return 1;
}
static int voice_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg){
	u_32 voice;
	voice = (cmd)?VOICE_MASK:~VOICE_MASK;
	trioport_set_out(voiceport, VOICE_MASK, voice);
	return 1;
}
static int voice_open(struct inode *inode, struct file *file)
{
	memset(&voice_handler, 0, sizeof(voice_handler));

	// first get the pointers and init IN/OUT
	if(pio_hook_driver(&voice_handler)<0){
		printk("Voice driver not initialized");
		return -1;
	}
	voiceport = voice_handler.pa;
	trioport_change_out(voiceport, VOICE_MASK, VOICE_MASK);

	return 0;
}
static struct file_operations voice_fops = {
	NULL,
	voice_read,
	NULL,	/* No write */
	NULL,	/* No readdir */
	NULL,
	voice_ioctl,
	NULL,		/* No mmap */
	voice_open,
	NULL
};
static struct miscdevice voice_dev=
{
	VOICE_MINOR,
	"triovoice",
	&voice_fops
};
int triovoice_init(void)
{
	misc_register(&voice_dev);
	return 0;
}

