/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/config.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/netdevice.h>
#include <asm/segment.h>
#include <asm/arch/irqs.h>
#include <asm/arch/irq.h>
#include <linux/isdnif.h>

#include "bsp.h"
#ifdef PROJECT_FAMILY_ROUTERMODUL
#include "global.h"
#endif

#include "g_syscfg.h"
#include "g_comdef.h"
#include "g_os_int.h"
#include "g_io_int.h"

#include "g_hw_int.h"
#include "g_ph_int.h"

#include <linux/mpsdev.h>

#ifdef PROJECT_FAMILY_ROUTERMODUL
#include "g_pr_int.h"
#include "b_hdlc/h_bhdlc_conf.h"
#include "b_hdlc/h_bhdlc.h"
#include "b_hdlc/h_hdlcmc.h"
#include "eth_p2001/h_eth.h"
#endif

#define DEBUG_MPSDEV 0
#define EMV_TRAFFIC 0

#if EMV_TRAFFIC
#define EMV_PACKETS 12
#define EMV_DELAY (HZ / 20)
#define EMV_MSG_SIZE 0x400
#endif

#if DEBUG_MPSDEV
#define DEBOUT(X) OUTPUT("mpsdev", X)
#define ERROUT(X) OUTPUT("mpsdev err", X)
//#define HDLOUT(X) OUTPUT("mpsdev hdlc", X)#L@XX
#else
#define DEBOUT(X)
#define ERROUT(X) OUTPUT("mpsdev err", X)
#define HDLOUT(X)
#endif

#define OUTPUT(X,Y) do { \
			printk("%s (%d): ", X, (int) __LINE__); \
			printk Y; \
		    } while (0)



#if 0
static void *my_kmalloc(unsigned size, int mode, int line)
{
	void *ret = kmalloc(size, mode);

	printk("kmalloc(%d): %u\n", line, (unsigned) ret);
	return ret;
}

static void my_kfree(void *p)
{
	kfree(p);
	printk("kfree: %u\n", (unsigned) p);
}

#define kmalloc(X,Y) my_kmalloc(X,Y,__LINE__)
#define kfree my_kfree
#endif

#if EMV_TRAFFIC
static struct timer_list emv_timer;
static t_msg *emv_msg;
#endif


#define ETH_TRC 0

struct mps_msg {
	struct mps_msg *next;
	int free_it;
	void (*f)(void *);
	void *arg;
	t_msg *msg;
};

#define MSG_BUFS 128
#define MPS_MSG_BUFS 256

struct msg_buf {
	struct mps_msg *head;
	struct mps_msg *tail;
	struct mps_msg *buf;
};

struct client {
	struct client *next;
	struct client *prev;
	int registered;
	unsigned long id;
	unsigned minor;
	int (*cb)(unsigned, t_msg *, void (*)(void *), void *);
	void (*stat_cb)(int);
	int (*canput)(unsigned);
	int read_offset;
	struct msg_buf msgs;
	struct wait_queue *wq_read;
};

static int init_client(struct client *cl)
{
	cl->id = (unsigned long) -1;
	cl->minor = (unsigned) -1;
	cl->registered = 0;
	cl->wq_read = NULL;
	cl->cb = NULL;
	cl->stat_cb = NULL;
	cl->canput = NULL;
	cl->read_offset = 0;
	return 0;
}

static int init_client_bufs(struct client *cl, int num_bufs)
{
	struct mps_msg *buf;
	int i;

	if (!(buf = kmalloc(num_bufs * sizeof *buf, GFP_KERNEL))) {
		return -ENOMEM;
	}
	for (i = 0; i < num_bufs; i++) {
		buf[i].msg = NULL;
		buf[i].free_it = 0;
		buf[i].f = (void (*)(void *)) NULL;
		buf[i].arg = NULL;
		buf[i].next = i < num_bufs - 1 ? &buf[i + 1] : &buf[0];
	}
	cl->msgs.buf = cl->msgs.head = cl->msgs.tail = buf;
	return 0;
}

static void finit_client(struct client *cl)
{
	struct mps_msg *buf = cl->msgs.buf;

	if (!buf)
		return;
	do {
		if (buf->free_it) {
			buf->free_it = 0;
			kfree(buf->msg);
		}
		if (buf->f) {
			buf->f(buf->arg);
			buf->f = (void (*)(void *)) NULL;
			buf->arg = NULL;
		}
		buf = buf->next;
	} while (buf != cl->msgs.buf);
	kfree(cl->msgs.buf);
}

/*  Das richtige hier waere wohl ein AVL tree, aber bei den paar
 *  Clients, die wir haben, waere das wahrscheinlich Overkill
 */

static struct client client_list;
static struct client *mps_client;

static struct wait_queue *mps_wait;

static void init_client_list(void)
{
	client_list.next = client_list.prev = &client_list;
	mps_client = NULL;
	mps_wait = NULL;
}

static struct client *find_client(unsigned long id, unsigned minor)
{
	struct client *ret;

	if (id < Z_LAST_INTERNAL_ENTRY)
		return mps_client;
	for (ret = client_list.next; ret != &client_list; ret = ret->next) {
		if (ret->id == id && ret->minor == minor) {
			return ret;
		}
	}
	return NULL;
}

static void notify_clients(int status)
{
	struct client *cl;

	for (cl = client_list.next; cl != &client_list; cl = cl->next) {
		if (cl->stat_cb) {
			cl->stat_cb(status);
		}
	}
}

static int insert_client(struct client *c)
{
	if (find_client(c->id, c->minor))
		return -1;
	c->prev = &client_list;
	c->next = client_list.next;
	client_list.next->prev = c;
	client_list.next = c;
	return 0;
}

static void remove_client(struct client *c)
{
	c->next->prev = c->prev;
	c->prev->next = c->next;
}

#ifdef PROJECT_FAMILY_ROUTERMODUL

#define USE_ETH_TIMER 0

#define ETH_DEV_NAME "eth0"



#define HDLC_BUF_LOWAT 0

struct hdlc_cb_data {
	struct hdlc_cb_data *next;
	void (*f)(void *);
	void *arg;
};

struct hdlc_cb_machine {
	struct hdlc_cb_data cbs[P2001_ANZ_TX_DESCR_PER_CH];
	struct hdlc_cb_data *fill_pointer;
	struct hdlc_cb_data *read_pointer;
	int hdlc_buf_count;
	struct timer_list cb_timer;
};

static struct hdlc_cb_machine hdlc_cb_state[P2001_MAX_ANZ_BHDLCDMA + 1];

static void hdlc_machine_cb(unsigned long scc_id)
{
	struct hdlc_cb_machine *ps = &hdlc_cb_state[scc_id];

	HDLOUT(("HDLC MACHINE TIMER HERE for %lu and %d buffers!\n", scc_id,
		ps->hdlc_buf_count));
}

static void reinit_hdlc_machine(int scc_id)
{
	struct hdlc_cb_machine *pm = &hdlc_cb_state[scc_id];

	pm->fill_pointer = pm->cbs;
	pm->read_pointer = pm->cbs;
	pm->hdlc_buf_count = P2001_ANZ_TX_DESCR_PER_CH;
}

static void purge_hdlc_machine(int scc_id)
{
	struct hdlc_cb_machine *pm = &hdlc_cb_state[scc_id];

	pm->hdlc_buf_count = 0;
	while (pm->read_pointer->f) {
		pm->read_pointer->f(pm->read_pointer->arg);
		pm->read_pointer->f = (void (*)(void *)) NULL;
		pm->read_pointer = pm->read_pointer->next;
	}
}

static void handle_hdlc_packet_sent(int scc_id)
{
	struct hdlc_cb_machine *pm = &hdlc_cb_state[scc_id];

	if (pm->read_pointer->f) {
		pm->read_pointer->f(pm->read_pointer->arg);
		pm->read_pointer->f = (void (*)(void *)) NULL;
		pm->read_pointer = pm->read_pointer->next;
	}
	pm->hdlc_buf_count++;
/*	HDLOUT(("count incremented to %d\n", pm->hdlc_buf_count)); */
}

static int save_hdlc_cb(unsigned scc_id, void (*f)(void *), void *arg)
{
	struct hdlc_cb_machine *pm;

	if (scc_id == 0 || scc_id > P2001_MAX_ANZ_BHDLCDMA)
		return -2;
	pm = &hdlc_cb_state[scc_id];
	if (pm->hdlc_buf_count <= HDLC_BUF_LOWAT) {
		HDLOUT(("count < lowat\n"));
		return -1;
	}
	if (pm->fill_pointer->f) {
		HDLOUT(("FILL POINTER IS FULL!\n"));
		return -1;
	}
	pm->fill_pointer->f = f;
	pm->fill_pointer->arg = arg;
	pm->fill_pointer = pm->fill_pointer->next;
	pm->hdlc_buf_count--;
/*	HDLOUT(("count decremented to %d\n", pm->hdlc_buf_count)); */
	return 0;
}

static void init_hdlc_machines(void)
{
	int i;

	for (i = 0; i <= P2001_MAX_ANZ_BHDLCDMA; i++) {
		struct hdlc_cb_machine *pm = &hdlc_cb_state[i];
		int j;

		for (j = 0; j < P2001_ANZ_TX_DESCR_PER_CH; j++) {
			struct hdlc_cb_data *pd = &pm->cbs[j];

			pd->f = (void (*)(void *)) NULL;
			pd->arg = NULL;
			pd->next = (j == P2001_ANZ_TX_DESCR_PER_CH - 1 ?
				    pm->cbs : pd + 1);
		}

		pm->fill_pointer = pm->cbs;
		pm->read_pointer = pm->cbs;
		pm->hdlc_buf_count = 0;
		init_timer(&pm->cb_timer);
		pm->cb_timer.function = hdlc_machine_cb;
		pm->cb_timer.data = i;
	}
}


#if USE_ETH_TIMER
static struct timer_list eth_timer;
#else

static int eth_device_event_handler(struct notifier_block *,
				    unsigned long,
				    void *);

static struct notifier_block mpsdev_device_notifier = {
	eth_device_event_handler,
	NULL,
	0
};
#endif

static struct device *eth_dev;

static unsigned char header[] = {
	0xff,0xff,0xff,0xff,0xff,0xff,0x00,0x09,0x4f,0x12,0x10,0x01,0x88,0x88
};
#if 0
static int send_eth_data(const char *buf, int count)
{
	struct sk_buff *skb;
	unsigned char *snd_buf;
	int len = sizeof header + count;

	if (!(skb = dev_alloc_skb(len))) {
		printk("No memory in MPSDEV\n");
		return -1;
	}
	SET_SKB_FREE(skb);
	snd_buf = skb_put(skb, len);
	memcpy(snd_buf, header, sizeof header);
	memcpy_fromfs(snd_buf + sizeof header, buf, count);
/*	skb->dev = eth_dev; */
	skb->arp = 1;
#if ETH_TRC
 {
	 t_msg *m = (t_msg *) buf;

	 printk("XMT: %u->%u, info %u, len %u\n",
		m->source, m->dest, m->info, m->length);
 }
#endif
	dev_queue_xmit(skb, eth_dev, 1);
	return 0;
}

static int eth_hook(void *p, struct sk_buff *skb)
{
	struct client *dst;
	t_msg *orig_m = (t_msg *) skb->data;
	t_msg *m;
	unsigned long flags;
	struct msg_buf *msgs;
	int minor = 0;

	if (orig_m->info == M_PR_DATA_INDICATION)
		minor = orig_m->phys_id.port;
#if ETH_TRC
	printk("RCV: %u->%u, info %u, len %u\n",
	       orig_m->source, orig_m->dest,
	       orig_m->info, orig_m->length);
#endif
	save_flags_cli(flags);
	if (!(dst = find_client(orig_m->dest, minor))) {
		restore_flags(flags);
		ERROUT(("Unknown destination %u %u->%u\n",
			orig_m->info, orig_m->source, orig_m->dest));
		return 1;
	}
	if (dst->cb) {
		restore_flags(flags);
		
	} else {
		if (!(m = kmalloc(skb->len, GFP_ATOMIC))) {
			restore_flags(flags);
			printk("Out of memory in eth_hook (mpsdev)\n");
			dev_kfree_skb(skb);
			return 0;
		}
		msgs = &dst->msgs;

		if (msgs->head->next == msgs->tail) {
			printk("Full queue in mpsdev\n");
			kfree(m);
		} else {
			memcpy(m, orig_m, skb->len);
			msgs->head->msg = m;
			msgs->head->free_it = 1;
			msgs->head->f = NULL;
			msgs->head->arg = NULL;
		msgs->head = msgs->head->next;
		wake_up_interruptible(&dst->wq_read);
		}
		restore_flags(flags);
		dev_kfree_skb(skb,FREE_WRITE);
	}
	return 0;
}



#else

static int send_eth_data(const char *buf, int len)
{
	struct device *dp = eth_dev;
	struct sk_buff *skb;
	struct ethict_hdr icthdr;
	unsigned total_len = len + sizeof header + sizeof icthdr;
	unsigned char *snd_buf;

#if ETH_TRC
 {
	 t_msg *m = (t_msg *) buf;

	 printk("XMT: %u->%u, info %u, len %u\n",
		m->source, m->dest, m->info, m->length);
 }
#endif
	if (dp == NULL) {
#if !EMV_TRAFFIC
		printk("mpsdev: discarding packet\n");
#endif
		return -1;
	}
	if (total_len <= ICT_MAX_ETH_PKT) {
		skb = dev_alloc_skb(total_len);
		if (!skb) {
			printk("mpsdev: dev_alloc_skb failed\n");
			return -1;
		}
		SET_SKB_FREE(skb);
		skb->arp = 1;
		skb->dev = dp;
		snd_buf = skb_put(skb, total_len);
		memcpy(snd_buf, header, sizeof header);
		icthdr.type = ETHICT_TYPE_DATA;
		icthdr.seqno = 0x80;
		icthdr.length = len;
		memcpy(snd_buf + sizeof header, &icthdr, sizeof icthdr);
		memcpy(snd_buf + sizeof header + sizeof icthdr, buf, len);
		dev_queue_xmit( skb,dp,1 );
	} else if (len <= ETHICT_MAXSIZE) {
		unsigned step = ICT_MAX_ETH_PKT - sizeof icthdr - sizeof header;
		unsigned offset = 0;
		unsigned togo = len;
		unsigned long flags;

		icthdr.type = ETHICT_TYPE_DATA;
		icthdr.seqno = 0;
		save_flags_cli(flags);
		for (;;) {
			if (togo > step) {
				skb = dev_alloc_skb(ICT_MAX_ETH_PKT);
				if (!skb) {
					printk("mpsdev: dev_alloc_skb failed\n");
					restore_flags(flags);
					return -1;
				}
				SET_SKB_FREE(skb);
				skb->arp = 1;
				skb->dev = dp;
				snd_buf = skb_put(skb, ICT_MAX_ETH_PKT);
				memcpy(snd_buf, header, sizeof header);
				icthdr.length = step;
				memcpy(snd_buf + sizeof header, &icthdr,
				       sizeof icthdr);
				icthdr.seqno++;
				memcpy(snd_buf + sizeof header
				       + sizeof icthdr,
				       buf + offset, step);
				offset += step;
				togo -= step;
				dev_queue_xmit( skb,dp,1 );
			} else {
				unsigned rest = togo + sizeof header +
					sizeof icthdr;

				skb = dev_alloc_skb(rest);
				if (!skb) {
					printk("mpsdev: dev_alloc_skb failed\n");
					restore_flags(flags);
					return -1;
				}
				SET_SKB_FREE(skb);
				skb->arp = 1;
				skb->dev = dp;
				snd_buf = skb_put(skb, rest);
				memcpy(snd_buf, header, sizeof header);
				icthdr.length = togo;
				icthdr.seqno |= 0x80;
				memcpy(snd_buf + sizeof header, &icthdr,
				       sizeof icthdr);
				memcpy(snd_buf + sizeof header
				       + sizeof icthdr,
				       buf + offset, togo);
				dev_queue_xmit( skb,dp,1 );
				break;
			}
		}
		restore_flags(flags);
	} else {
		printk("mpsdev: packet too long (%u)\n", len);
	}
	return 0;
}

#define ETHICT_STATE_NULL 0
#define ETHICT_STATE_RCV 1
#define ETHICT_STATE_ACCUM 2

static int handle_packet(t_msg *orig_m, unsigned len)
{
	struct client *dst;
	t_msg *m;
	unsigned long flags;
	struct msg_buf *msgs;
	int minor = 0;

#if EMV_TRAFFIC
	if (orig_m->dest == Z_HDLC_C && orig_m->info == 0x4711) {
		return 0;
	}
#endif
	if (orig_m->info == M_PR_DATA_INDICATION)
		minor = orig_m->phys_id.port;
	else if (orig_m->dest == Z_HDLC_C
		 && orig_m->info == M_PR_COMMAND_INDICATION) {
		switch (orig_m->data) {
		case 1:
			reinit_hdlc_machine(orig_m->phys_id.port);
			break;
		case 2:
			purge_hdlc_machine(orig_m->phys_id.port);
			break;
		default:
			handle_hdlc_packet_sent(orig_m->phys_id.port);
			break;
		}
		return 0;
	}
#if ETH_TRC
	printk("RCV(%d): %u->%u, info %u, len %u\n", minor,
	       orig_m->source, orig_m->dest,
	       orig_m->info, orig_m->length);
#endif
	save_flags_cli(flags);
	if (!(dst = find_client(orig_m->dest, minor))) {
		restore_flags(flags);
		DEBOUT(("Unknown destination %u %u->%u\n",
			orig_m->info, orig_m->source, orig_m->dest));
		return 1;
	}
	if (!(m = kmalloc(len, GFP_ATOMIC))) {
		restore_flags(flags);
		printk("Out of memory in eth_hook (mpsdev)\n");
		return 0;
	}
	if (dst->cb) {
		memcpy(m, orig_m, len);
		restore_flags(flags);
		dst->cb(minor, m, kfree, m);
	} else {
		msgs = &dst->msgs;

		if (msgs->head->next == msgs->tail) {
			printk("Full queue in mpsdev\n");
			kfree(m);
		} else {
			memcpy(m, orig_m, len);
			msgs->head->msg = m;
			msgs->head->free_it = 1;
			msgs->head->f = NULL;
			msgs->head->arg = NULL;
			msgs->head = msgs->head->next;
			wake_up_interruptible(&dst->wq_read);
		}
		restore_flags(flags);
	}
	return 0;

}

static struct sk_buff_head backlog;
static int backlog_size = 0;

static int eth_hook(void *p, struct sk_buff *skb)
{
	static int dropping = 0;
	struct ethict_hdr *hdr;

	if (skb->len < sizeof *hdr) {
		return 1;
	}
	hdr = (struct ethict_hdr *) skb->data;
	if (hdr->type != ETHICT_TYPE_DATA) {
		return 1;
	}

	skb->sk = NULL;
	skb->free = 1;

	if (backlog_size < 50) {
  		dropping = 0;
	} else if (backlog_size > 300) {
		if (dropping == 0) {
			printk( KERN_WARNING"[WRN] mpsdev: dropping in eth_hook\n" );
		}
		dropping = 1;
	}

	if (dropping) {
		kfree_skb(skb, FREE_READ);
		mark_bh(MPSDEV_BH);
		return 0;
	}
	skb_queue_tail(&backlog,skb);
	backlog_size++;
	mark_bh(MPSDEV_BH);
	return 0;
}

static void eth_hook_bh(void)
{
	static unsigned char ethict_buf[ETHICT_MAXSIZE];
	static int ethict_state = ETHICT_STATE_NULL;
	static unsigned char next_seq;
	static unsigned short next_offset;
	unsigned long flags;
	struct sk_buff *skb;
	struct ethict_hdr *hdr;
	t_msg *m;

	WD_TRIGGER;

	while (!skb_queue_empty(&backlog)) {
		save_flags_cli(flags);
		skb = backlog.next;
		__skb_unlink(skb, &backlog);
  		if (backlog_size > 0) {
			backlog_size--;
		}
		restore_flags(flags);
#if 0
		printk("MPSDEV SKB %p, lock %d, free %d\n", skb,
		       (int) skb->lock, (int) skb->free);
#endif
		hdr = (struct ethict_hdr *) skb->data;
		m = (t_msg *) (skb->data + sizeof *hdr);
		switch (ethict_state) {
		case ETHICT_STATE_NULL: /* Wait for sync */
			switch (hdr->seqno) {
			case 0x00:
			case 0x80:
				ethict_state = ETHICT_STATE_RCV;
				goto state_recv;
			default:
				break;
			}
			break;
		case ETHICT_STATE_RCV:
	state_recv:
			switch (hdr->seqno) {
			case 0x80:
				handle_packet(m, skb->len - sizeof *hdr);
				break;
			case 0x00:
				if (hdr->length > ETHICT_MAXSIZE) {
					printk("mpsdev: Got long packet %u\n",
						hdr->length);
					ethict_state = ETHICT_STATE_NULL;
					break;
				}
				memcpy(ethict_buf, m, hdr->length);
				next_seq = 1;
				next_offset = hdr->length;
				ethict_state = ETHICT_STATE_ACCUM;
				break;
			default:
				printk("mpsdev: Unexpected seqno %u\n",
					hdr->seqno);
				ethict_state = ETHICT_STATE_NULL;
				break;
			}
			break;
		case ETHICT_STATE_ACCUM:
		{
			unsigned final = hdr->seqno & 0x80;
			unsigned real_no = hdr->seqno & 0x7F;

			if (hdr->length + next_offset > ETHICT_MAXSIZE) {
				printk("mpsdev: packet too large %u\n",
					hdr->length + next_offset);
				ethict_state = ETHICT_STATE_NULL;
				break;
			}
			if (real_no != next_seq) {
				printk("mpsdev: Got %u, expected %u\n",
					real_no, next_seq);
				ethict_state = ETHICT_STATE_NULL;
				break;
			}
			memcpy(ethict_buf + next_offset, m, hdr->length);
			if (final) {
				handle_packet((t_msg *) ethict_buf,
					      next_offset + hdr->length);
				ethict_state = ETHICT_STATE_RCV;
			} else {
				next_seq++;
				next_offset += hdr->length;
			}
			break;
		}
		default:
			printk("mpsdev: Unknown state %d\n", ethict_state);
			break;
		}
		dev_kfree_skb(skb, FREE_READ);
	}
	return;
}


#endif

#if EMV_TRAFFIC
static void emv_timer_cb(unsigned long arg)
{
	int i;

	for (i = 0; i < EMV_PACKETS; i++) {
		send_eth_data((char *) emv_msg, EMV_MSG_SIZE);
	}
	emv_timer.expires = jiffies + EMV_DELAY;
	add_timer(&emv_timer);
}
#endif

#if USE_ETH_TIMER
static void eth_timer_cb(unsigned long arg)
{
	if ((eth_dev = dev_get(ETH_DEV_NAME))) {
		struct client *cl;
		unsigned long flags;

		printk("MPSDEV got eth device\n");
		if (eth_dev->hook_register)
			eth_dev->hook_register(eth_dev->priv,
					       eth_hook, NULL);
		else panic("mpsdev: %s device has no hook!\n", ETH_DEV_NAME);
		if (!(cl = kmalloc(sizeof *cl, GFP_ATOMIC))) {
			panic("No memory for mps_client\n");
		}
		memset(cl, '\0', sizeof *cl);
		if (init_client(cl)) {
			panic("Error in init_client for MPSDEV\n");
		}
		cl->minor = 0;
		cl->registered = 1;
		mps_client = cl;
		wake_up_interruptible(&mps_wait);
		save_flags_cli(flags);
		notify_clients(1);
		restore_flags(flags);
		return;
	}
	eth_timer.expires = jiffies + HZ / 2;
	add_timer(&eth_timer);
}

#else

static int eth_device_event_handler(struct notifier_block *nbp,
				    unsigned long event,
				    void *arg)
{
	struct device *dp = arg;

	if (strcmp(dp->name, ETH_DEV_NAME))
		return NOTIFY_DONE;
	switch (event) {
	case NETDEV_UP:
	{
		struct client *cl;
		unsigned long flags;

		eth_dev = dp;
		printk("MPSDEV got eth device\n");
		if (eth_dev->hook_register)
			eth_dev->hook_register(eth_dev->priv,
					       eth_hook, NULL);
		else panic("mpsdev: %s device has no hook!\n", ETH_DEV_NAME);
		if (mps_client) {
#if EMV_TRAFFIC
			emv_timer.expires = jiffies + EMV_DELAY + (HZ * 60);
			add_timer(&emv_timer);
#endif
			break;
		}
		if (!(cl = kmalloc(sizeof *cl, GFP_ATOMIC))) {
			panic("No memory for mps_client\n");
		}
		memset(cl, '\0', sizeof *cl);
		if (init_client(cl)) {
			panic("Error in init_client for MPSDEV\n");
		}
		cl->minor = 0;
		cl->registered = 1;
		mps_client = cl;
		wake_up_interruptible(&mps_wait);
		save_flags_cli(flags);
		notify_clients(1);
		restore_flags(flags);
#if EMV_TRAFFIC
		emv_timer.expires = jiffies + EMV_DELAY + (HZ * 60);
		add_timer(&emv_timer);
#endif
		break;
	}
	case NETDEV_DOWN:
	{
		if (!eth_dev)
			break;
#if EMV_TRAFFIC
		del_timer(&emv_timer);
#endif
		if (eth_dev->hook_unregister) {
			eth_dev->hook_unregister(eth_dev->priv, eth_hook);
			mark_bh(MPSDEV_BH);
		}
		eth_dev = NULL;
		printk("MPSDEV lost eth device\n");
		break;
	}
	default:
		printk("mpsdev: Unknown event in device notifier 0x%08lX\n",
		       event);
		break;
	}
	return NOTIFY_DONE;
}

#endif /* USE_ETH_TIMER */

static int hdlc_callback(unsigned scc_id, t_msg *m,
			 void (*kf)(void *), void *arg)
{
#if 0
	int ret;

	MPSIHL(("HDLC packet for id %d\n", scc_id));
	ret = f_p2001_hdlc_tx_cb(scc_id, (unsigned char *) &m->data, m->length,
				 kf, arg);
	MPSIHL(("hdlc_tx returned %d\n", ret));
	if (ret != 1) {
		kf(arg);
	}
	return ret == 1 ? 0 : -1;
#endif
	if (save_hdlc_cb(scc_id, kf, arg)) {
		return -1;
	}
#if 0
	if (hdlc_cb_state[scc_id].hdlc_buf_count <= HDLC_BUF_LOWAT) {
#if 0
		if (kf) {
			kf(arg);
		}
#endif
		return -1;
	}
	hdlc_cb_state[scc_id].hdlc_buf_count--;
#endif
#if 0
	printk("%u buf_count decremented to %d\n",
	       scc_id, hdlc_buf_count[scc_id]);
#endif
	m->phys_id.port = scc_id;
	send_eth_data((const char *) m, sizeof *m + m->length);
#if 0
	if (kf) {
		kf(arg);
	}
#endif
	return 0;
}

static void hdlc_stat_callback(int status)
{
	;
}

int hdlc_canput(unsigned scc_id)
{
	return hdlc_cb_state[scc_id].hdlc_buf_count > HDLC_BUF_LOWAT ? 1 : 0;
}

#endif

static int mps_read(struct inode *inode, struct file *filep,
		    char *buf, int nbytes)
{
	struct wait_queue wait = { current, NULL };
	struct client *cl = filep->private_data;
	struct msg_buf *msgs;
	unsigned long flags;
	int count = 0, retval = 0;

	if (!cl || !cl->registered)
		return -EIO;
	msgs = &cl->msgs;
	if (nbytes < 0)
		return -EINVAL;
	if (nbytes == 0)
		return 0;
	add_wait_queue(&cl->wq_read, &wait);
	for (;;) {
		current->state = TASK_INTERRUPTIBLE;
		save_flags_cli(flags);
		if (msgs->head != msgs->tail) {
			t_msg *m = msgs->tail->msg;
			int msglen = m->length + sizeof *m;

			count = msglen - cl->read_offset;
			if (nbytes < count) {
				count = nbytes;
			}
			if ((retval = verify_area(VERIFY_WRITE,
						  buf, count))) {
				count = 0;
			} else {
				memcpy_tofs(buf, (unsigned char *) m + cl->read_offset,
					    count);
				if (cl->read_offset + count >= msglen) {
					if (msgs->tail->free_it) {
						msgs->tail->free_it = 0;
						kfree(msgs->tail->msg);
					}
					if (msgs->tail->f) {
						msgs->tail->f(msgs->tail->arg);
						msgs->tail->f = (void (*)(void *)) NULL;
						msgs->tail->arg = NULL;
					}
					msgs->tail = msgs->tail->next;
					cl->read_offset = 0;
					if (cl == mps_client)
						wake_up_interruptible(&mps_wait);
				} else {
					cl->read_offset += count;
				}
			}
			restore_flags(flags);
			break;
		} else {
			if (filep->f_flags & O_NONBLOCK) {
				retval = -EAGAIN;
				restore_flags(flags);
				break;
			}
			if (current->signal & ~current->blocked) {
				retval = -ERESTARTSYS;
				restore_flags(flags);
				break;
			}
		}
		restore_flags(flags);
		schedule();
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&cl->wq_read, &wait);
	if (inode && count)
		UPDATE_ATIME(inode);
	return count ? count : retval;
}

static int mps_write(struct inode *inode, struct file *filep,
		     const char *buf, int count)
{
	unsigned minor = MINOR(inode->i_rdev);
	t_msg *m;
	struct client *cl = filep->private_data;
	struct client *dst;
	struct wait_queue wait = { current, NULL };
	int ret = 0, msglen;
	unsigned long flags;

	if (!cl || !cl->registered)
		return -EIO;

	if (count < 0)
		return -EINVAL;
	if (count == 0)
		return 0;
	if ((ret = verify_area(VERIFY_READ, buf, count)))
		return -EINVAL;
	m = (t_msg *) buf;
	msglen = m->length + sizeof *m;
	if (sizeof *m > count || msglen > count)
		return -EINVAL;
	    /* Semaphore in der Inode hochsetzen.  Sonst warten andere
	     * darauf, dasz wir hieraus zurueckkehren bevor sie hier
	     * reinkommen... was u.U. bis zum Sankt Nimmerlein's Tag
	     * dauern kann, da wir umgekehrt auf sie warten...
	     */
	up(&inode->i_sem);
	save_flags_cli(flags);
	if (!(dst = find_client(m->dest, minor))) {
		restore_flags(flags);
		down(&inode->i_sem);
		DEBOUT(("%u %u->%u: cannot find dst\n",
			m->info, m->source, m->dest));
		return -EINVAL;
	}
	if (dst == mps_client) {
		add_wait_queue(&mps_wait, &wait);
#ifdef PROJECT_FAMILY_ROUTERMODUL
		restore_flags(flags);
		for (;;) {
			current->state = TASK_INTERRUPTIBLE;
			if (mps_client) {
				send_eth_data(buf, msglen);
				count = msglen;
				break;
			}
			schedule();
		}
#else
		for (;;) {
			current->state = TASK_INTERRUPTIBLE;
			    /* mps_client might change (theoretically) */
			if (mps_client) {
				struct msg_buf *msgs = &mps_client->msgs;

				if (msgs->head->next != msgs->tail) {
					if (!(m = kmalloc(msglen, GFP_KERNEL))) {
						restore_flags(flags);
						ret = -ENOMEM;
						break;
					}
					memcpy_fromfs(m, buf, msglen);
					msgs->head->msg = m;
					msgs->head->free_it = 1;
					msgs->head->f = (void (*)(void *)) NULL;
					msgs->head->arg = NULL;
					msgs->head = msgs->head->next;
					wake_up_interruptible(&mps_client->wq_read);
					restore_flags(flags);
					count = msglen;
					break;
				} else {
					DEBOUT(("%u %u->%u: dest queue full\n",
						m->info, m->source, m->dest));
							
					if (filep->f_flags & O_NONBLOCK) {
						ret = -EAGAIN;
						restore_flags(flags);
						break;
					}
					if (current->signal & ~current->blocked) {
						ret = -ERESTARTSYS;
						restore_flags(flags);
						break;
					}
				}
			}
			restore_flags(flags);
			schedule();
			save_flags_cli(flags);
		}
#endif
		current->state = TASK_RUNNING;
		remove_wait_queue(&mps_wait, &wait);
	} else {
		struct msg_buf *msgs = &dst->msgs;

		if (dst->cb || msgs->head->next != msgs->tail) {
			if (!(m = kmalloc(msglen, GFP_KERNEL))) {
				restore_flags(flags);
				down(&inode->i_sem);
				return -ENOMEM;
			}
			memcpy_fromfs(m, buf, msglen);
			if (dst->cb) { /* No blocking */
				if (dst->cb(minor, m, kfree, m)) {
					DEBOUT(("%u %u->%u: dst cb != 0\n",
						m->info, m->source, m->dest));
					ret = -EAGAIN;
				} else {
					count = msglen;
				}
			} else {
				msgs->head->msg = m;
				msgs->head->free_it = 1;
				msgs->head = msgs->head->next;
				wake_up_interruptible(&dst->wq_read);
				count = msglen;
			}
		} else {
			    /* Die MPS blockiert nicht wenn sie
			     * versucht, auf eine volle Client-Queue zu schreiben
			     */
			ret = -EAGAIN;
			DEBOUT(("%u %u->%u: dst queue full\n",
				m->info, m->source, m->dest));
		}
		restore_flags(flags);
	}
	down(&inode->i_sem);
	return ret ? ret : count;
}

static int mps_select(struct inode *inode, struct file *filep,
		      int sel_type, select_table *wait)
{
	struct client *cl = filep->private_data;
	unsigned minor = MINOR(inode->i_rdev);
	struct msg_buf *msgs;
	int ready;
	unsigned long flags;

	if (!cl || !cl->registered)
		return 1;
	switch (sel_type) {
	case SEL_IN :
		msgs = &cl->msgs;
		save_flags_cli(flags);
		ready = (msgs->head != msgs->tail);
		restore_flags(flags);
		if (ready)
			return 1;
		select_wait(&cl->wq_read, wait);
		break;
	case SEL_OUT :
		if (cl == mps_client) {
			    /* Wohin soll die msg gehen?
			     * Weisz man nicht.
			     * write returned mit EAGAIN wenn Ziel-Queue voll
			     */
			return 1;
		} else if (minor) {
			    /* minor hack: kann nur an HDLC-Treiber gehen */
			struct client *dst;

			save_flags_cli(flags);
			dst = find_client(Z_HDLC_C, minor);
			restore_flags(flags);
			if (dst && dst->canput && dst->canput(minor)) {
				return 1;
			}
		} else {
			if (mps_client == NULL) {
				    /* Kleines Feature: Wenn jemand warten will,
				     * bis die MPS gestartet ist, kann er es
				     * mit einem select auf write machen
				     */
				select_wait(&mps_wait, wait);
			} else {
				msgs = &mps_client->msgs;
				save_flags_cli(flags);
				ready = (msgs->tail != msgs->head->next);
				restore_flags(flags);
				if (ready)
					return 1;
				select_wait(&mps_wait, wait);
			}
		}
		break;
	}
	return 0;
}

static int mps_open(struct inode *ino, struct file *filep)
{
	struct client *cl;
	unsigned minor = MINOR(ino->i_rdev);
	int err;

	if (!(cl = kmalloc(sizeof *cl, GFP_KERNEL)))
		return -ENOMEM;
	memset(cl, '\0', sizeof *cl);
	if ((err = init_client(cl))) {
		kfree(cl);
		return err;
	}
	cl->minor = minor;
	filep->private_data = cl;
	return 0;
}

static void mps_close(struct inode *ino, struct file *filep)
{
	struct client *cl = filep->private_data;
	unsigned long flags;

	if (cl == NULL)
		return;
	
	save_flags_cli(flags);
	if (cl->registered) {
		if (cl == mps_client) {
			mps_client = NULL;
			wake_up_interruptible(&mps_wait);
			notify_clients(0);
		} else {
			remove_client(cl);
		}
	}
	filep->private_data = NULL;
	if (cl->registered)
		finit_client(cl);
	kfree(cl);
	restore_flags(flags);
}

static int mps_ioctl(struct inode *ino, struct file *filep,
		     unsigned int cmd, unsigned long arg)
{
	struct client *cl = filep->private_data;
	unsigned long flags;
	int err;

	switch (cmd) {
	case MPSIOCTL_SETID:
	{
		if (cl->registered) {
			if (cl->id != arg)
				return -EINVAL;
			else return 0;
		}
		save_flags_cli(flags);
		if (arg == 0) { /* MPS meldet sich an */
			if (mps_client) {
				restore_flags(flags);
				return -EINVAL;
			}
			if ((err = init_client_bufs(cl, MPS_MSG_BUFS))) {
				restore_flags(flags);
				return err;
			}
			cl->id = 0;
			cl->registered = 1;
			mps_client = cl;
			wake_up_interruptible(&mps_wait);
			notify_clients(1);
		} else if (arg >= Z_LAST_INTERNAL_ENTRY) {
			cl->id = arg;
			if (insert_client(cl)) {
				restore_flags(flags);
				return -EINVAL;
			}
			if ((err = init_client_bufs(cl, MSG_BUFS))) {
				remove_client(cl);
				restore_flags(flags);
				return err;
			}
			cl->registered = 1;
		} else {
			restore_flags(flags);
			return -EINVAL;
		}
		restore_flags(flags);
		break;
	}
	case FIONREAD:
	{
		int *plen = (int *) arg;
		struct msg_buf *msgs;

		if (!cl || !cl->registered)
			return -EIO;
		msgs = &cl->msgs;
		if (verify_area(VERIFY_WRITE, plen, sizeof *plen))
			return -EINVAL;
		save_flags_cli(flags);
		if (msgs->head != msgs->tail) {
			t_msg *m = msgs->tail->msg;
			int msglen = m->length + sizeof *m;

			*plen = msglen - cl->read_offset;
		} else {
			*plen = 0;
		}
		restore_flags(flags);
		break;
	}
	default :
		return -EINVAL;
	}
	return 0;
}

static struct file_operations mps_fops = {
        NULL,		/* lseek */
	mps_read,	/* read */
	mps_write,	/* write */
	NULL,		/* readdir */
	mps_select,	/* select */
	mps_ioctl,	/* ioctl */
	NULL,		/* mmap */
        mps_open,
        mps_close	/* release */
};

int init_mpsdev(void)
{
#ifdef PROJECT_FAMILY_ROUTERMODUL
	int i, ret;
#endif
	if (register_chrdev(MPSDEV_MAJOR, "mps", &mps_fops)) {
		printk("unable to get major %d for mps devices\n",
		       MPSDEV_MAJOR);
		return -EIO;
	}
	init_client_list();
#ifdef PROJECT_FAMILY_ROUTERMODUL
	init_hdlc_machines();
	for (i = 1; i <= P2001_MAX_ANZ_BHDLCDMA; i++)
		if ((ret = register_mps_callback(Z_HDLC_C, i, hdlc_callback,
						 hdlc_stat_callback,
						 hdlc_canput))) {
			printk("mpsdev: register_mps_callback failed "
			       "for hdlc %d with %d\n", i, ret);
		}
	eth_dev = NULL;
#if EMV_TRAFFIC
	init_timer(&emv_timer);
	if (!(emv_msg = kmalloc(EMV_MSG_SIZE, GFP_KERNEL))) {
		panic("Out of memory in MPS device");
	}
	emv_msg->source = Z_HDLC_C;
	emv_msg->dest = Z_HDLC_C;
	emv_msg->info = 0x4711;
	emv_msg->length = EMV_MSG_SIZE - sizeof *emv_msg;
	emv_timer.data = 0;
	emv_timer.function = emv_timer_cb;
#endif
#if USE_ETH_TIMER
	init_timer(&eth_timer);
	eth_timer.data = 0;
	eth_timer.function = eth_timer_cb;
	eth_timer_cb(0);
#else
	if (register_netdevice_notifier(&mpsdev_device_notifier)) {
		printk("MPSDEV: Could not register device notifier\n");
		return -EIO;
	}
	skb_queue_head_init(&backlog);
	init_bh(MPSDEV_BH, eth_hook_bh);
#endif
#endif
	return 0;
}

#ifdef PROJECT_FAMILY_ROUTERMODUL

static int do_send_msg(unsigned minor, t_msg *m,
		       int free_it, void (*f)(void *), void *arg)
{
	struct client *dst;
	unsigned long flags;
	int ret = 0;

	save_flags_cli(flags);
	if (!(dst = find_client(m->dest, minor))) {
		restore_flags(flags);
		return -EINVAL;
	}
	if (dst == mps_client) {
		restore_flags(flags);
		if (mps_client) {
			ret = send_eth_data((char *) m, sizeof *m + m->length);
			if (free_it) {
				kfree(m);
			} else if (f) {
				f(arg);
			}
		} else ret = -1;
	} else if (!dst->cb) {
		struct msg_buf *msgs = &dst->msgs;

		if (msgs->head->next == msgs->tail) {
			ret = -EAGAIN;
		} else {
			msgs->head->msg = m;
			msgs->head->free_it = free_it;
			msgs->head->f = f;
			msgs->head->arg = arg;
			msgs->head = msgs->head->next;
			wake_up_interruptible(&dst->wq_read);
		}
		restore_flags(flags);
	} else {
		restore_flags(flags);
		if (f)
			ret = dst->cb(minor, m, f, arg);
		else
			ret = dst->cb(minor, m, kfree, m);
	}
	return ret;
}

int send_mps_msg(unsigned minor, t_msg *m)
{
	return do_send_msg(minor, m, 1, (void (*)(void *)) NULL, NULL);
}

int send_mps_msg_notify(unsigned minor, t_msg *m, void (*f)(void *), void *arg)
{
	return do_send_msg(minor, m, 0, f, arg);
}

void unregister_mps_callback(unsigned id, unsigned minor)
{
	panic("MPSDEV kernel msgs not implemented\n");
}

int register_mps_callback(unsigned id, unsigned minor,
			  int (*cb)(unsigned , t_msg *,
				    void (*)(void *), void *),
			  void (*stat_cb)(int),
			  int (*canput)(unsigned))
{
	struct client *cl;
	int err;
	unsigned long flags;

	if (id < Z_LAST_INTERNAL_ENTRY)
		return -1;
	save_flags_cli(flags);
	if (find_client(id, minor)) {
		restore_flags(flags);
		return -2;
	}
	if (!(cl = kmalloc(sizeof *cl, GFP_KERNEL))) {
		restore_flags(flags);
		return -ENOMEM;
	}
	memset(cl, '\0', sizeof *cl);
	if ((err = init_client(cl))) {
		restore_flags(flags);
		kfree(cl);
		return err;
	}
	cl->cb = cb;
	cl->stat_cb = stat_cb;
	cl->canput = canput;
	cl->id = id;
	cl->minor = minor;
	cl->registered = 1;
	if (insert_client(cl)) {
		restore_flags(flags);
		kfree(cl);
		return -3;
	}
	if (mps_client) {
		cl->stat_cb(1);
	}
	restore_flags(flags);
	return 0;
}

#else

static int do_send_msg(unsigned minor, t_msg *m,
		       int free_it, void (*f)(void *), void *arg)
{
	struct client *dst;
	unsigned long flags;
	int ret = 0;

	save_flags_cli(flags);
	if (!(dst = find_client(m->dest, minor))) {
		restore_flags(flags);
		return -EINVAL;
	}
	if (dst == mps_client || !dst->cb) {
		struct msg_buf *msgs = &dst->msgs;

		if (msgs->head->next == msgs->tail) {
			ret = -EAGAIN;
		} else {
			msgs->head->msg = m;
			msgs->head->free_it = free_it;
			msgs->head->f = f;
			msgs->head->arg = arg;
			msgs->head = msgs->head->next;
			wake_up_interruptible(&dst->wq_read);
		}
	} else {
		if (f)
			ret = dst->cb(minor, m, f, arg);
		else
			ret = dst->cb(minor, m, kfree, m);
	}
	restore_flags(flags);
	return ret;
}

int send_mps_msg(unsigned minor, t_msg *m)
{
	return do_send_msg(minor, m, 1, (void (*)(void *)) NULL, NULL);
}

int send_mps_msg_notify(unsigned minor, t_msg *m, void (*f)(void *), void *arg)
{
	return do_send_msg(minor, m, 0, f, arg);
}

int register_mps_callback(unsigned id, unsigned minor,
			  int (*cb)(unsigned , t_msg *,
				    void (*)(void *), void *),
			  void (*stat_cb)(int),
			  int (*canput)(unsigned))
{
	struct client *cl;
	int err;
	unsigned long flags;

	if (id < Z_LAST_INTERNAL_ENTRY)
		return -1;
	save_flags_cli(flags);
	if (find_client(id, minor)) {
		restore_flags(flags);
		return -2;
	}
	if (!(cl = kmalloc(sizeof *cl, GFP_KERNEL))) {
		restore_flags(flags);
		return -ENOMEM;
	}
	memset(cl, '\0', sizeof *cl);
	if ((err = init_client(cl))) {
		restore_flags(flags);
		kfree(cl);
		return err;
	}
	cl->cb = cb;
	cl->stat_cb = stat_cb;
	cl->canput = canput;
	cl->id = id;
	cl->minor = minor;
	cl->registered = 1;
	if (insert_client(cl)) {
		restore_flags(flags);
		kfree(cl);
		return -3;
	}
	if (mps_client) {
		cl->stat_cb(1);
	}
	restore_flags(flags);
	return 0;
}

void unregister_mps_callback(unsigned id, unsigned minor)
{
	struct client *cl;
	unsigned long flags;

	save_flags_cli(flags);
	if (id == 0 || !(cl = find_client(id, minor))) {
		restore_flags(flags);
		printk("mpsdev: unregister for unknown client %u\n", id);
		return;
	}
	if (!cl->cb) {
		restore_flags(flags);
		printk("mpsdev: unregister for non-callback client %u\n", id);
		return;
	}
	remove_client(cl);
	restore_flags(flags);
	kfree(cl);
}
#endif
