/*
 *  linux/arch/arm/kernel/setup.c
 *
 *  Copyright (C) 1995-1998 Russell King
 *
 * This file obtains various parameters about the system that the kernel
 * is running on.
 */

/*
 * Modifications for P2001 Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/stddef.h>
#include <linux/unistd.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/user.h>
#include <linux/a.out.h>
#include <linux/tty.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/major.h>
#include <linux/utsname.h>
#include <linux/blk.h>

#include <asm/segment.h>
#include <asm/system.h>
#include <asm/hardware.h>
#include <asm/pgtable.h>
#include <asm/arch/mmu.h>
#include <asm/procinfo.h>
#include <asm/io.h>
#include <asm/setup.h>

struct drive_info_struct { char dummy[32]; } drive_info;
struct screen_info screen_info;
struct processor processor;

extern const struct processor arm2_processor_functions;
extern const struct processor arm250_processor_functions;
extern const struct processor arm3_processor_functions;
extern const struct processor arm6_processor_functions;
extern const struct processor arm7_processor_functions;
extern const struct processor sa110_processor_functions;

struct armversions armidlist[] = {
#if defined(CONFIG_CPU_ARM2) || defined(CONFIG_CPU_ARM3)
	{ 0x41560200, 0xfffffff0, F_MEMC	, "ARM/VLSI",	"arm2"		, &arm2_processor_functions   },
	{ 0x41560250, 0xfffffff0, F_MEMC	, "ARM/VLSI",	"arm250"	, &arm250_processor_functions },
	{ 0x41560300, 0xfffffff0, F_MEMC|F_CACHE, "ARM/VLSI",	"arm3"		, &arm3_processor_functions   },
#endif
#if defined(CONFIG_CPU_ARM6) || defined(CONFIG_CPU_SA110)
	{ 0x41560600, 0xfffffff0, F_MMU|F_32BIT	, "ARM/VLSI",	"arm6"		, &arm6_processor_functions   },
	{ 0x41560610, 0xfffffff0, F_MMU|F_32BIT	, "ARM/VLSI",	"arm610"	, &arm6_processor_functions   },
	{ 0x41007000, 0xffffff00, F_MMU|F_32BIT , "ARM/VLSI",   "arm7"		, &arm7_processor_functions   },
	/* ARM710 does not follow the spec */
	{ 0x41007100, 0xfff8ff00, F_MMU|F_32BIT , "ARM/VLSI",   "arm710"	, &arm7_processor_functions   },
	{ 0x4401a100, 0xfffffff0, F_MMU|F_32BIT	, "DEC",	"sa110"		, &sa110_processor_functions  },
#endif
#if defined(CONFIG_CPU_ARM7)
	{ 0x41009000, 0xffffff00, F_32BIT , "ELMEG GmbH & Co.KG", "P2001/ARM9TDMI", &arm7_processor_functions   },
	{ 0x41007000, 0xffffff00, F_32BIT , "ARM/APLIO",   "arm7"		, &arm7_processor_functions   },
#endif

	{ 0x00000000, 0x00000000, 0		, "***",	"*unknown*"	, NULL }
	
};


#ifdef CONFIG_ARCH_P2001
#include "bsp.h"
#include "display.h"

#define GPO_4_OUT(x)		\
	(*(volatile unsigned long *)(GPIO_BASE_ADDR + Adr_GPO4_OUT) = (x)); \
	(*(volatile unsigned long *)(GPIO_BASE_ADDR + Adr_GPO4_OUT) = (x))

#define GPO_4_IN(x)		\
	((x) = *(volatile unsigned long *)(GPIO_BASE_ADDR + Adr_GPO4_OUT))

static unsigned long	reg_GPO4;

#if defined(PROJECT_T240) || defined(PROJECT_T444)
static unsigned long    davosv2_led_latch    = 0x00000000; 
static unsigned long    davosv2_module_latch = 0x00000080; 
#endif

void	p2001_GPO4( unsigned long,unsigned long );
void	(*_func_disp_cb_1)( int ) = NULL;
void	(*_func_disp_cb_2)( int,int ) = NULL;
void	(*_func_disp_cb_3)( int,int,int ) = NULL;
void	(*_func_disp_cb_4)( int,int,int,int ) = NULL;
static void init_led_callbacks( void );
#if defined(PROJECT_FAMILY_COMPACTDSL) || defined(PROJECT_FAMILY_MODULARDSL) || defined(PROJECT_FAMILY_ROUTERMODUL)
#include "g_syscfg.h"
#undef LINEAR
#include "g_comdef.h"
#include "g_hw_int.h"
void	init_serno( void );
int	get_serno( char * );
int	get_hwinfo( char * );
void    davosv2_Latch( unsigned long, unsigned long );
#endif
#endif


#if defined(CONFIG_ARCH_TRIO) || defined(CONFIG_ARCH_P2001)

static struct param_struct *params = (struct param_struct *) PARAMS_BASE;

u_long trio_romdisk_addr = 3*1024*1024;

#else
struct param_struct trio_params = { { {
	4096,				// page_size
	4*1024*1024/4096,	// nr_pages
	0,					// ramdisk_size	
	0,					// flags
	0,					// rootdev
	
	80,					// video_num_cols
	24,					// video_num_rows
	0,					// video_x
	0,					// video_y
	0,					// bytes_per_char_h
	0,					// bytes_per_char_v

	0,  // initrd_start
	0,

	0
}
}
};

static struct param_struct *params = &trio_params;
#endif
	

unsigned long arm_id;
unsigned int vram_half_sam;
int armidindex;
int ioebpresent;
int memc_ctrl_reg;
int number_ide_drives;
int number_mfm_drives;

extern int bytes_per_char_h;
extern int bytes_per_char_v;
extern int root_mountflags;
extern int _etext, _edata, _end;
extern unsigned long real_end_mem;

/*
 * ram disk
 */
#ifdef CONFIG_BLK_DEV_RAM
extern int rd_doload;		/* 1 = load ramdisk, 0 = don't load */
extern int rd_prompt;		/* 1 = prompt for ramdisk, 0 = don't prompt */
extern int rd_image_start;	/* starting block # of image */

static void setup_ramdisk (struct param_struct *params)
{
	rd_image_start	= params->u1.s.rd_start;
	rd_prompt	= (params->u1.s.flags & FLAG_RDPROMPT) == 0;
	rd_doload	= (params->u1.s.flags & FLAG_RDLOAD) == 0;
}
#else
#define setup_ramdisk(p)
#endif

/*
 * initial ram disk
 */
#ifdef CONFIG_BLK_DEV_INITRD
static void setup_initrd (struct param_struct *params, unsigned long memory_end)
{
	initrd_start = params->u1.s.initrd_start;
	initrd_end   = params->u1.s.initrd_start + params->u1.s.initrd_size;

	if (initrd_end > memory_end) {
		printk ("initrd extends beyond end of memory "
			"(0x%08lx > 0x%08lx) - disabling initrd\n",
			initrd_end, memory_end);
		initrd_start = 0;
	}
}
#else
#define setup_initrd(p,m)
#endif

static inline void check_ioeb_present(void)
{
#if defined(CONFIG_ARCH_TRIO) || defined(CONFIG_ARCH_P2001)
#else
	if (((*IOEB_BASE) & 15) == 5)
		armidlist[armidindex].features |= F_IOEB;
#endif
}

static void get_processor_type (void)
{
	for (armidindex = 0; ; armidindex ++)
		if (!((armidlist[armidindex].id ^ arm_id) &
		      armidlist[armidindex].mask))
			break;

	if (armidlist[armidindex].id == 0) {
#if defined(CONFIG_ARCH_TRIO) || defined(CONFIG_ARCH_P2001)
#else
		int i;
		for (i = 0; i < 3200; i++)
			((unsigned long *)SCREEN2_BASE)[i] = 0x77113322;
#endif

		while (1);
	}

	processor = *armidlist[armidindex].proc;
}

#define COMMAND_LINE_SIZE 256

static char command_line[COMMAND_LINE_SIZE] = { 0, };
       char saved_command_line[COMMAND_LINE_SIZE];

void setup_arch(char **cmdline_p,
	unsigned long * memory_start_p, unsigned long * memory_end_p)
{
	static unsigned char smptrap;
	unsigned long memory_start, memory_end;
	char c = ' ', *to = command_line, *from;
	int len = 0;

	if (smptrap == 1)
		return;
	smptrap = 1;

#ifdef CONFIG_ARCH_P2001
	GPO_4_IN( reg_GPO4 );
	reg_GPO4 &= 0x00000fff;
	GPO_4_OUT( reg_GPO4 );
	init_led_callbacks();
#endif
	get_processor_type ();
	check_ioeb_present ();
	processor._proc_init ();

#ifdef CONFIG_CONSOLE	
	bytes_per_char_h  = params->u1.s.bytes_per_char_h;
	bytes_per_char_v  = params->u1.s.bytes_per_char_v;
#endif
	from		  = params->commandline;
	ROOT_DEV	  = to_kdev_t (params->u1.s.rootdev);
	ORIG_X		  = params->u1.s.video_x;
	ORIG_Y		  = params->u1.s.video_y;
	ORIG_VIDEO_COLS	  = params->u1.s.video_num_cols;
	ORIG_VIDEO_LINES  = params->u1.s.video_num_rows;


	setup_ramdisk (params);

	if (!(params->u1.s.flags & FLAG_READONLY))
		root_mountflags &= ~MS_RDONLY;

	memory_start = MAPTOPHYS((unsigned long)&_end);
	memory_end   = GET_MEMORY_END(params);
#if 1
#ifdef PROJECT_FAMILY_COMPACTDSL
#warning MEMORY_END TWEAKED FOR SRAM
	memory_end -= 0x10000;		/*##L@XXX*/
#endif
#ifdef PROJECT_FAMILY_MODULARDSL
#warning MEMORY_END TWEAKED FOR SRAM
	memory_end -= (0x18000 -0x104);/*CFG_MEM_SRAM_SIZE*/	/*##L@XXX*/
#endif
#endif

	*memory_start_p = memory_start;
	*memory_end_p = memory_end;

	init_task.mm->start_code = TASK_SIZE;
	init_task.mm->end_code	 = TASK_SIZE + (unsigned long) &_etext;
	init_task.mm->end_data	 = TASK_SIZE + (unsigned long) &_edata;
	init_task.mm->brk	 = TASK_SIZE + (unsigned long) &_end;

	/* Save unparsed command line copy for /proc/cmdline */
	memcpy(saved_command_line, from, COMMAND_LINE_SIZE);
	saved_command_line[COMMAND_LINE_SIZE-1] = '\0';

	for (;;) {
		if (c == ' ' &&
		    from[0] == 'm' &&
		    from[1] == 'e' &&
		    from[2] == 'm' &&
		    from[3] == '=') {
			memory_end = simple_strtoul(from+4, &from, 0);
			if (*from == 'K' || *from == 'k') {
				memory_end = memory_end << 10;
				from++;
			} else if (*from == 'M' || *from == 'm') {
				memory_end = memory_end << 20;
				from++;
			}
			memory_end = memory_end + PAGE_OFFSET;
		}
#if defined( CONFIG_ARCH_TRIO)
		else if (c == ' ' &&
				from[0] == 'r' &&
				from[1] == 'o' &&
				from[2] == 'm' &&
				from[3] == '=') {
				trio_romdisk_addr = simple_strtoul(from+4, &from, 0);
				if (*from == 'K' || *from == 'k') {
					trio_romdisk_addr = trio_romdisk_addr << 10;
					from++;
				} else if (*from == 'M' || *from == 'm') {
					trio_romdisk_addr = trio_romdisk_addr << 20;
					from++;
				}
			}
#endif						
		c = *from++;
		if (!c)
			break;
		if (COMMAND_LINE_SIZE <= ++len)
			break;
		*to++ = c;
	}

	*to = '\0';
	*cmdline_p = command_line;
	*memory_start_p = memory_start;
	*memory_end_p = memory_end;

	setup_initrd (params, memory_end);

	strcpy (system_utsname.machine, armidlist[armidindex].name);
#ifdef CONFIG_SERIAL_TRIO
	{
		extern void register_console(void (*proc)(const char *));
		extern void console_print_trio(const char * b);
		register_console(console_print_trio);
	}
#endif
#ifdef CONFIG_ARCH_P2001
#if defined(PROJECT_FAMILY_COMPACTDSL) || defined(PROJECT_FAMILY_MODULARDSL) || defined(PROJECT_FAMILY_ROUTERMODUL)
	init_serno();
#if defined(PROJECT_VARIANT_SERIAL_CONSOLE)
	{
		extern void register_console(void (*proc)(const char *));
		extern void console_print_p2001(const char * b);
		register_console(console_print_p2001);
	}
#endif
#endif
#endif
}

#define ISSET(bit) (armidlist[armidindex].features & bit)

int get_cpuinfo(char * buffer)
{
	int len;

	len = sprintf (buffer,  "CPU:\n"
				"Type\t\t: %s\n"
				"Revision\t: %d\n"
				"Manufacturer\t: %s\n"
				"32bit modes\t: %s\n"
				"BogoMips\t: %lu.%02lu\n",
				armidlist[armidindex].name,
				(int)arm_id & 15,
				armidlist[armidindex].manu,
				ISSET (F_32BIT) ? "yes" : "no",
				(loops_per_sec+2500) / 500000,
				((loops_per_sec+2500) / 5000) % 100);
	return len;
}

#ifdef CONFIG_ARCH_P2001
void
p2001_GPO4( unsigned long msk,unsigned long val )
{
	unsigned long	flags;

	msk &= 0x000000ff;
	val &= 0x000000ff;
	save_flags_cli( flags );
	reg_GPO4 = (msk & val) | (~msk & reg_GPO4);
	GPO_4_OUT( reg_GPO4 );
	restore_flags( flags );
}

static void
nop_led1( int x1 )
{
	printk( KERN_NOTICE"led: [WRN] nop_led1 called\n" );
}

static void
nop_led2( int x1,int x2 )
{
	printk( KERN_NOTICE"led: [WRN] nop_led2 called\n" );
}


static void
nop_led3( int x1,int x2,int x3 )
{
	printk( KERN_NOTICE"led: [WRN] nop_led3 called\n" );
}

static void
nop_led4( int x1,int x2,int x3,int x4 )
{
	printk( KERN_NOTICE"led: [WRN] nop_led4 called\n" );
}

static void
init_led_callbacks( void )
{
	_func_disp_cb_1 = nop_led1;
	_func_disp_cb_2 = nop_led2;
	_func_disp_cb_3 = nop_led3;
	_func_disp_cb_4 = nop_led4;
}

#if defined(PROJECT_FAMILY_COMPACTDSL) || defined(PROJECT_FAMILY_MODULARDSL) || defined(PROJECT_FAMILY_ROUTERMODUL)
static t_hw_pbx_data	pbx_data;

void
init_serno( void )
{
	flash_layout_t	*flash = (flash_layout_t *)FLASH_BASE_ADDR;

	memcpy( &pbx_data,&flash->hw_pbx_data,sizeof( pbx_data ));
}

int
get_serno( char *buf )
{
	t_hw_pbx_data	*p = &pbx_data;
	int		len;
	int		i;

	if (sizeof( p->cty_cod ) != sizeof( int )) {
		printk( KERN_CRIT"[ERR] get_serno: wrong size of enum\n" );
	}

	memcpy( buf,"serno\t",6 );
	len = 6;
	if (p->ser_len > 0 && p->ser_len <= G_MAX_SERIAL_NUM_LEN) {
		memcpy( buf+len,p->ser_num,p->ser_len );
		len += p->ser_len;
	} else {
		memcpy( buf+len,"---",3 );
		len += 3;
	}
	buf[len++] = '\n';

	len += sprintf( buf+len,"oem\t%d\n",p->oem_cod );

	for (i=0; i<6; i++) {
		if (p->mac_adr[i][0] == 0x00 &&
		    p->mac_adr[i][1] == 0x09 &&
		    p->mac_adr[i][2] == 0x4f)
		{
			len += sprintf( buf+len,"mac%d\t00:09:4f:%02x:%02x:%02x\n",i,p->mac_adr[i][3],p->mac_adr[i][4],p->mac_adr[i][5] );
		} else {
			len += sprintf( buf+len,"mac%d\t00:09:4f:00:00:%02d\n",i,i );
		}
	}
	return( len );
}

int
get_hwinfo( char *buf )
{
	return( sprintf( buf,"0x%08x\n",*(volatile unsigned int *)(HWID_ADDR) ));
}

#if defined(PROJECT_T240) || defined(PROJECT_T444)
void
davosv2_LedLatch( unsigned long msk,unsigned long val )
{
	unsigned long		flags;
	volatile unsigned short *LatchAddr; // 16 Bit access !!!

	msk &= 0x000000ff;
	val &= 0x000000ff;
	save_flags_cli( flags );
	davosv2_led_latch = (msk & val) | (~msk & davosv2_led_latch);
	LatchAddr  = (unsigned short *)0x08800000;
	*LatchAddr = (unsigned short)(davosv2_led_latch & 0x000000ff);
	restore_flags( flags );
}

void
davosv2_ModLatch( unsigned long msk,unsigned long val )
{
	unsigned long		flags;
	volatile unsigned short *LatchAddr; // 16 Bit access !!!

	msk &= 0x000000ff;
	val &= 0x000000ff;
	save_flags_cli( flags );
	davosv2_module_latch = (msk & val) | (~msk & davosv2_module_latch);
	LatchAddr  = (unsigned short *)0x08C00000;
	*LatchAddr = (unsigned short)(davosv2_module_latch & 0x000000ff);
	restore_flags( flags );
}
#endif
#endif
#endif
