/*
 *  linux/fs/binfmt_flat.c
 */

/*
 * Modifications for P2001 Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

#include <linux/module.h>

#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/a.out.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/user.h>
#include <linux/malloc.h>
#include <linux/binfmts.h>
#include <linux/personality.h>
#include <linux/flat.h>
#include <linux/config.h>

#include <asm/system.h>
#include <asm/semaphore.h>
#include <asm/segment.h>
#include <asm/pgtable.h>
#include <asm/byteorder.h>
#include <asm/unaligned.h>

extern int	decompress_exec( struct inode *inode,
                                 unsigned long offset,
                                 char *buffer,
                                 long len,
                                 int fd );

extern int	do_load_new_flat_binary( struct linux_binprm *bprm,
                                         struct pt_regs *regs,
                                         struct file *file,
                                         struct linux_binfmt *format );

extern int	do_load_old_flat_binary( struct linux_binprm *bprm,
                                         struct pt_regs *regs,
                                         struct file *file,
                                         struct linux_binfmt *format );

static int	load_flat_binary( struct linux_binprm *,struct pt_regs *regs );

struct semaphore sem_mutex;

static struct linux_binfmt flat_format = {
#ifndef MODULE
	NULL, NULL, load_flat_binary, NULL, NULL
#else
	NULL, &mod_use_count_, load_flat_binary, NULL, NULL
#endif
};

static int
load_flat_binary( struct linux_binprm *bprm,struct pt_regs *regs )
{
	static int	mutex = 0;
	struct flat_hdr	*hdr;
	struct file	*file;
	int		rc;

	MOD_INC_USE_COUNT;
	down( &sem_mutex );
	mutex++;
#if 0	/* jse debug */
	if (mutex > 1) {
#warning STOPPER ENABLED
		/* XXX */
		printk( KERN_CRIT"kernel: [ERR] load_flat_binary(%s,%s): mutex=%d sem=%d\n",current->comm,bprm->filename,mutex,sem_mutex.count );
		while(1);
	}
#endif
	current->personality = PER_LINUX;
	file = current->files->fd[open_inode( bprm->inode,O_RDONLY )];
	hdr = (struct flat_hdr *) bprm->buf;
	if (flush_old_exec( bprm )) {
		printk( KERN_WARNING"kernel: [WRN] binfmt_flat: unable to flush\n" );
		rc = -ENOMEM;
		goto out;
	}
	if (!strncmp( hdr->magic,"bFLT",4 )) {
		if (ntohl( hdr->rev ) == FLAT_VERSION) {
			rc = do_load_new_flat_binary( bprm,regs,file,&flat_format );
		} else if (ntohl( hdr->rev ) == OLD_FLAT_VERSION ||
		           ntohl( hdr->rev ) == 3)
		{
			rc = do_load_old_flat_binary( bprm,regs,file,&flat_format );
		} else {
			printk( KERN_NOTICE"kernel: [WRN] binfmt_flat: bad magic/rev (%ld, need %d, 3 or %d)\n",ntohl( hdr->rev ),(int)OLD_FLAT_VERSION,(int)FLAT_VERSION );
			rc = -ENOEXEC;
		}
	} else {
		rc = -ENOEXEC;
	}
out:
	mutex--;
	up( &sem_mutex );
	MOD_DEC_USE_COUNT;
	return( rc );
}

int
init_flat_binfmt( void )
{
	sem_mutex.count = 1;
	sem_mutex.waking = 0;
	sem_mutex.lock = 0;
	sem_mutex.wait = NULL;
	return( register_binfmt( &flat_format ));
}

#ifdef MODULE
int
init_module( void )
{
	return( init_flat_binfmt());
}

void
cleanup_module( void)
{
	unregister_binfmt( &flat_format );
}
#endif
