/*
 *  linux/fs/binfmt_flat.c
 */

/*
 * Modifications for P2001 Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

#include <linux/module.h>

#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/a.out.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/user.h>
#include <linux/malloc.h>
#include <linux/binfmts.h>
#include <linux/personality.h>
//#include <linux/flat.h>

#include <asm/system.h>
#include <asm/semaphore.h>
#include <asm/segment.h>
#include <asm/pgtable.h>
#include <asm/flat.h>


#include <asm/byteorder.h>


extern int	decompress_exec( struct inode *inode,
                                 unsigned long offset,
                                 char *buffer,
                                 long len,
                                 int fd );

#if 0
#define DEBUG
#define JSE_DEBUG
#define JSE_DEBUG2
#endif

static unsigned long putstring(unsigned long p, char * string)
{
	unsigned long l = strlen(string)+1;
#ifdef DEBUG
	printk("put_string '%s'\n", string);
#endif
	p -= l;
	memcpy((void*)p, string, l);
	return p;
}

static unsigned long putstringarray(unsigned long p, int count, char ** array)
{
#ifdef DEBUG
	/*p=putstring(p, "");*/
	/*printk("p1=%x, array=%x\n", p, array);
	printk("array[0]=%x\n", array[0]);*/
	printk("putstringarray(%d)\n", count);
#endif
	while(count) {
		p=putstring(p, array[--count]);
#ifdef DEBUG
		printk("p2=%x\n", (unsigned int)p);
#endif
	}
	return p;
}

static unsigned long stringarraylen(int count, char ** array)
{
	int l = 4;
	while(count) {
		l += strlen(array[--count]);
		l++;
		l+=4;
	}
	return l;
}

/*
 * create_flat_tables() parses the env- and arg-strings in new user
 * memory and creates the pointer tables from them, and puts their
 * addresses on the "stack", returning the new stack pointer value.
 */
static unsigned long create_flat_tables(unsigned long pp, struct linux_binprm * bprm)
{
	unsigned long *argv,*envp;
	unsigned long * sp;
	char * p = (char*)pp;
	int argc = bprm->argc;
	int envc = bprm->envc;

	sp = (unsigned long *) ((-(unsigned long)sizeof(char *)) & (unsigned long) p);
#ifdef __alpha__
/* whee.. test-programs are so much fun. */
	put_user(0, --sp);
	put_user(0, --sp);
	if (bprm->loader) {
		put_user(0, --sp);
		put_user(0x3eb, --sp);
		put_user(bprm->loader, --sp);
		put_user(0x3ea, --sp);
	}
	put_user(bprm->exec, --sp);
	put_user(0x3e9, --sp);
#endif
	sp -= envc+1;
	envp = sp;
	sp -= argc+1;
	argv = sp;
#if defined(__i386__) || defined(__mc68000__) || defined(__arm__)
	put_user(envp,--sp);
	put_user(argv,--sp);
#endif
	put_user(argc,--sp);
	current->mm->arg_start = (unsigned long) p;
	while (argc-->0) {
		put_user(p,argv++);
		while (get_user(p++)) /* nothing */ ;
	}
	put_user(NULL,argv);
	current->mm->arg_end = current->mm->env_start = (unsigned long) p;
	while (envc-->0) {
		put_user(p,envp++);
		while (get_user(p++)) /* nothing */ ;
	}
	put_user(NULL,envp);
	current->mm->env_end = (unsigned long) p;
	return (unsigned long)sp;
}

void do_reloc(struct flat_reloc * r)
{
#ifdef DEBUG
	char *segment[] = { "TEXT", "DATA", "BSS", "*UNKNOWN*" };
#endif
	unsigned long * ptr = (unsigned long*)
		(current->mm->start_data + r->offset);

#ifdef DEBUG
	printk("Relocation of variable at DATASEG+%x (address %p, currently %x) into segment %s\n",
		r->offset, ptr, (int)*ptr, segment[r->type]);
#endif
	
	switch (r->type) {
	case FLAT_RELOC_TYPE_TEXT:
		*ptr += current->mm->start_code;
		break;
	case FLAT_RELOC_TYPE_DATA:
		*ptr += current->mm->start_data;
		break;
	case FLAT_RELOC_TYPE_BSS:
		*ptr += current->mm->end_data;
		break;
	default:
		printk("ERROR: Unknown relocation\n");
	}

 
#ifdef DEBUG
	printk("Relocation became %x\n", (int)*ptr);
#endif
}		


/*
 * These are the functions used to load (z)flat style executables and shared
 * libraries.  There is no binary dependent code anywhere else.
 */

int
do_load_old_flat_binary( struct linux_binprm *bprm,
                         struct pt_regs *regs,
                         struct file *file,
                         struct linux_binfmt *format )
{
	struct flat_hdr * hdr;
	unsigned long result;
	unsigned long pos_code, pos_data, pos_bss, pos_stack;
	unsigned long p = bprm->p;
	unsigned long data_len, bss_len, stack_len, code_len;
	unsigned long fildatst;


#ifdef JSE_DEBUG
	printk("BINFMT_FLAT: Loading file: %s\n",bprm->filename );
	show_free_areas();
#endif

	hdr = (struct flat_hdr*)bprm->buf;
	
	code_len = ntohl(hdr->data_start) - ntohl(hdr->entry);
	data_len = ntohl(hdr->data_end) - code_len;
	bss_len = ntohl(hdr->bss_end) - ntohl(hdr->data_end);
	stack_len = ntohl(hdr->stack_size);
	
	/* Make room on stack for arguments & environment */
	stack_len += strlen(bprm->filename) + 1;
	stack_len += stringarraylen(bprm->envc, bprm->envp);
	stack_len += stringarraylen(bprm->argc, bprm->argv);
	
	/*
	 * allocate separate chunks of mem - one day I should
	 * make a smart allocator which groups small chunks
	 */
	if((pos_code = do_mmap(0, 0, code_len, PROT_READ|PROT_WRITE|PROT_EXEC, 0, 0)) >= (unsigned long)-4096)
		return pos_code;
	if((pos_data = do_mmap(0, 0, data_len + bss_len, PROT_READ|PROT_WRITE|PROT_EXEC, 0, 0)) >= (unsigned long)-4096)
	{
		do_munmap(pos_code, 0);
		return pos_data;
	}
	pos_bss = pos_data + data_len;
	if((pos_stack = do_mmap(0, 0, stack_len, PROT_READ|PROT_WRITE|PROT_EXEC, 0, 0)) >= (unsigned long)-4096)
	{
		do_munmap(pos_code, 0);
		do_munmap(pos_data, 0);
		return pos_stack;
	}


#ifdef JSE_DEBUG2
	printk( "execve(%s) code=%ld data=%ld bss=%ld stack=%ld sum=%ld\n",
	         bprm->filename,code_len,data_len,bss_len,stack_len,
	         code_len+data_len+bss_len+stack_len );
	printk( "execve(%s) code=%d data=%d bss=%d stack=%d sum=%d\n",
	         bprm->filename,code_len,data_len,bss_len,stack_len,
	         code_len+data_len+bss_len+stack_len );
	printk("BINFMT_FLAT: Allocated:\ncode\t0x%x\t%d\ndata\t0x%x\t%d\nbss\t0x%x\t%d\nstack\t0x%x\t%d\n", (int)pos_code, (int)code_len, (int)pos_data, (int)data_len, (int)pos_bss, (int)bss_len, (int)pos_stack, (int)stack_len);
	show_free_areas();
#endif

	// read code and data in memory
	if (hdr->flags & FLAT_GZIP_CODE) {
		result = decompress_exec( bprm->inode,ntohl( hdr->entry ),
		                          (char *) pos_code,
		                          ntohl( hdr->gz_codesize ),0 );
	} else {
		result = read_exec( bprm->inode,ntohl( hdr->entry ),
		                    (char *)pos_code,code_len,0 );
	}
	if(result >= (unsigned long)-4096) {
		do_munmap(pos_code, 0);
		do_munmap(pos_data, 0);
		do_munmap(pos_stack, 0);
		return result;
	}

	if (hdr->flags & FLAT_GZIP_CODE) {
		fildatst = ntohl( hdr->entry ) + ntohl( hdr->gz_codesize );
		result = decompress_exec( bprm->inode,fildatst,
		                          (char *)pos_data,data_len,0 );
	} else {
		result = read_exec( bprm->inode,ntohl( hdr->data_start ),
		                    (char *)pos_data,data_len,0 );
	}
	if(result >= (unsigned long)-4096)
	{
		do_munmap(pos_code, 0);
		do_munmap(pos_data, 0);
		do_munmap(pos_stack, 0);
		return result;
	}
	memset((void*)pos_bss, 0, bss_len);
	memset((void*)pos_stack, 0xee, stack_len);
	*(unsigned long *)pos_stack = STACK_MAGIC;	/* jse added XXX */
	current->mm->start_stack_page = pos_stack;

	current->mm->start_code = pos_code;
	current->mm->end_code = pos_code + code_len;
	current->mm->start_data = pos_data;
	current->mm->end_data = pos_data + data_len;
	current->mm->brk = pos_bss + bss_len;
	if (bprm->inode->i_sb->s_flags & MS_SYNCHRONOUS) {
#ifdef DEBUG
		printk("Retaining inode\n");
#endif
		current->mm->executable = bprm->inode;
		bprm->inode->i_count++;
	}
	else
		current->mm->executable = 0;
	
#ifdef JSE_DEBUG
	printk( "execve(%s): make fix TEXT=%x DATA=%x BSS=%x STACK(%d)=%x\n",
		bprm->filename,
		(int)current->mm->start_code,
		(int)current->mm->start_data,
		(int)current->mm->end_data,
		stack_len,pos_stack );
#endif

	{
		int r;
		int rcnt = ntohl(hdr->reloc_count);

		for(r=0;r<rcnt;r++) {
			union {
				struct flat_reloc reloc;
				unsigned long lword;
			} rr;
			unsigned long result = read_exec(bprm->inode, 
				ntohl(hdr->reloc_start) + (sizeof(struct flat_reloc)*r), 
				(char *)&rr,
			  	sizeof(struct flat_reloc), 
			  	0);
			if (result >= (unsigned long)-4096) {
				printk("Failure reloading relocation\n");
			} else {
				rr.lword = ntohl(rr.lword);
				do_reloc(&rr.reloc);
			}
		}
	}

	current->mm->rss = 0;
	current->suid = current->euid = current->fsuid = bprm->e_uid;
	current->sgid = current->egid = current->fsgid = bprm->e_gid;
 	current->flags &= ~PF_FORKNOEXEC;
 
	if (current->exec_domain && current->exec_domain->use_count)
		(*current->exec_domain->use_count)--;
	if (current->binfmt && current->binfmt->use_count)
		(*current->binfmt->use_count)--;
	current->exec_domain = lookup_exec_domain(current->personality);
	current->binfmt = format;
	if (current->exec_domain && current->exec_domain->use_count)
		(*current->exec_domain->use_count)++;
	if (current->binfmt && current->binfmt->use_count)
		(*current->binfmt->use_count)++;

	/*set_brk(current->mm->start_brk, current->mm->brk);*/

	p = pos_stack + stack_len - 4;
	
#ifdef DEBUG
	printk("p=%x\n", (int)p);
#endif
	
	p = putstringarray(p, 1, &bprm->filename);

#ifdef DEBUG
	printk("p(filename)=%x\n", (int)p);
#endif

	p = putstringarray(p, bprm->envc, bprm->envp);

#ifdef DEBUG
	printk("p(envp)=%x\n", (int)p);
#endif

	p = putstringarray(p, bprm->argc, bprm->argv);

#ifdef DEBUG
	printk("p(argv)=%x\n", (int)p);
#endif

	p = create_flat_tables(p, bprm);

#ifdef DEBUG
	printk("p(create_flat_tables)=%x\n", (int)p);
	
	printk("arg_start = %x\n", (int)current->mm->arg_start);
	printk("arg_end = %x\n", (int)current->mm->arg_end);
	printk("env_start = %x\n", (int)current->mm->env_start);
	printk("env_end = %x\n", (int)current->mm->env_end);
#endif

	current->mm->start_stack = p;

#ifdef JSE_DEBUG
	printk("start_stack: %x\n", current->mm->start_stack);
	show_free_areas();
	printk( "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<\n\n" ); /* XXX jse */
#endif

	start_thread(regs, current->mm->start_code, current->mm->start_data /*- hdr->data_start*/, p);
	
	if (current->flags & PF_PTRACED)
		send_sig(SIGTRAP, current, 0);
	return 0;
}
