/*
 * linux/include/asm-arm/proc-armv/system.h
 *
 * Copyright (C) 1996 Russell King
 */

/*
 * Modifications for P2001 Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH> 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

#ifndef __ASM_PROC_SYSTEM_H
#define __ASM_PROC_SYSTEM_H

extern const char xchg_str[];
extern void arm_invalidptr(const char *, int);

extern __inline__ unsigned long __xchg(unsigned long x, volatile void *ptr, int size)
{
	switch (size) {
		case 1:	__asm__ __volatile__ ("swpb %0, %1, [%2]" : "=r" (x) : "r" (x), "r" (ptr) : "memory");
			break;
		case 2:	abort ();
		case 4:	__asm__ __volatile__ ("swp %0, %1, [%2]" : "=r" (x) : "r" (x), "r" (ptr) : "memory");
			break;
		default: arm_invalidptr(xchg_str, size);
	}
	return x;
}

/*
 * This processor does not need anything special before reset,
 * but RPC may do...
 */
extern __inline__ void proc_hard_reset(void)
{
}

/*
 * We can wait for an interrupt...
 */
#if 0
#define proc_idle()			\
	do {				\
	__asm__ __volatile__(		\
"	mcr	p15, 0, %0, c15, c8, 2"	\
	  : : "r" (0));			\
	} while (0)
#else
#define proc_idle()
#endif
/*
 * A couple of speedups for the ARM
 */

/*
 * Save the current interrupt enable state & disable IRQs
 */
#define save_flags_cli(x)		\
	do {				\
	  unsigned long temp;		\
	  __asm__ __volatile__(		\
	"mrs	%1, cpsr\n"		\
"	and	%0, %1, #192\n"		\
"	orr	%1, %1, #128\n"		\
"	msr	cpsr, %1"		\
	  : "=&r" (x), "=r" (temp)	\
	  :				\
	  : "memory");			\
	} while (0)
	
/*
 * Enable IRQs
 */
#define sti()				\
	do {				\
	  unsigned long temp;		\
	  __asm__ __volatile__(		\
	"mrs	%0, cpsr\n"		\
"	bic	%0, %0, #128\n"		\
"	msr	cpsr, %0"		\
	  : "=r" (temp)			\
	  :				\
	  : "memory");			\
	} while(0)

/*
 * Disable IRQs
 */
#define cli()				\
	do {				\
	  unsigned long temp;		\
	  __asm__ __volatile__(		\
	"mrs	%0, cpsr\n"		\
"	orr	%0, %0, #128\n"		\
"	msr	cpsr, %0"		\
	  : "=r" (temp)			\
	  :				\
	  : "memory");			\
	} while(0)

/*
 * Enable FIQs
 */
#define stf()				\
	do {				\
	  unsigned long temp;		\
	  __asm__ __volatile__(		\
	"mrs	%0, cpsr\n"		\
"	bic	%0, %0, #64\n"		\
"	msr	cpsr, %0"		\
	  : "=r" (temp)			\
	  :				\
	  : "memory");			\
	} while(0)

/*
 * Disable FIQs
 */
#define clf()				\
	do {				\
	  unsigned long temp;		\
	  __asm__ __volatile__(		\
	"mrs	%0, cpsr\n"		\
"	orr	%0, %0, #64\n"		\
"	msr	cpsr, %0"		\
	  : "=r" (temp)			\
	  :				\
	  : "memory");			\
	} while(0)


/*
 * save current IRQ & FIQ state
 */
#define save_flags(x)			\
	do {				\
	  __asm__ __volatile__(		\
	"mrs	%0, cpsr\n"		\
"	and	%0, %0, #192"		\
	  : "=r" (x));			\
	} while (0)

/*
 * restore saved IRQ & FIQ state
 */
#define restore_flags(x)		\
	do {				\
	  unsigned long temp;		\
	  __asm__ __volatile__(		\
	"mrs	%0, cpsr\n"		\
"	bic	%0, %0, #192\n"		\
"	orr	%0, %0, %1\n"		\
"	msr	cpsr, %0"		\
	  : "=&r" (temp)			\
	  : "r" (x)			\
	  : "memory");			\
	} while (0)

#endif
