#ifndef TS_H
#define TS_H

#include <linux/types.h>

enum A_TST_CMD{
	A_SET_DURATION=1,
	A_SET_OFF_DURATION,
	A_SET_SYNC_ASYNC_MODE,
	A_SET_CODEC,
	A_SET_DTMF_LEVEL,
	A_START_DTMF_DETECT,
	A_STOP_DTMF_DETECT,
	A_GET_TONE_STATE,
};

enum A_TSS_CMD{
	A_START_RECORD=1,
	A_STOP_RECORD,
	A_START_PLAY,
	A_STOP_PLAY,
	A_SET_R_COMP,
	A_SET_P_COMP,
	A_SET_BIGENDIAN,
	A_SET_FRAMENB,
	A_CHANGE_PLAY_VOL,
	A_CHANGE_REC_VOL,
	A_PLAY_ENDED,
	A_GET_PLAY_DEPTH,			// frames to be played
	A_GET_FREE_FOR_PLAY_DEPTH,	// free frames for play
	A_GET_MAX_PLAY_DEPTH,		// max frames in buffer
	A_DETECT_SILENCE,
	A_GET_BIGENDIAN,
	A_PLAY_STARTED,
	A_REC_STARTED,
	A_IDLE,
	A_LOCK_OUTPUT,
	A_LOCK_INPUT,
	A_IS_INPUT_LOCKED,
	A_IS_OUTPUT_LOCKED,
	A_UNLOCK_OUTPUT,
	A_UNLOCK_INPUT,
	A_SET_PLAY_VOL,
	A_SET_REC_VOL,
	A_GET_PLAY_VOL,
	A_GET_REC_VOL,
	A_START_TRAIN_AEC,
	A_STOP_TRAIN_AEC,
	A_GET_AEC,
};

enum A_COMP_TYPE{
	A_TS_41,
	A_TS_53,
	A_TS_63,
	A_MAX_CMP,
};

// DTMF/tone playing and detecting
struct TSPollInfo{
	unsigned char status;
	unsigned char dtmfCode;
	unsigned char toneCode;
};
enum TONE_STATUS{
	TS_GOT_DTMF=1,
	TS_GOT_1300=2,
	TS_GOT_BUSY=4,
	TS_TONE_PLAYED=8,
	TS_SEE_DTMF=16,	// the driver sees DTMF but it can't confirm the key yet
	TS_SEE_TONE=32,	// the driver sees the tone but it can't confirm its timing
	TS_GOT_2100=64,
	TS_GOT_OTHER_TONE=128
};

#define TS_GOT_MASK  (TS_GOT_DTMF | TS_GOT_BUSY | TS_GOT_1300 | TS_GOT_2100)
//#define TS_GOT_MASK  (64+4+2+1)

// Sound playing/recording states
enum SOUND_STATUS{
 TS_NEXT_PLAY_BUFF	=	1,
 TS_NEXT_REC_BUFF	=	2,
 TS_NEXT_TONE_PLAY	=	4,	
 TS_TONE_PLAYING	=	8,
 TS_PLAY_UNDERRUN	=	16, 
 TS_PLAY_EMPTY		=	32,
 TS_REC_OVERRUN		=	64
};

// ON_MAX = 0 means that the tone detection should be signaled during the frequence presence
// and depends on the ON_MIN value -> if 0 , signal immediately

#define TS_1300_ON_MIN	800  // in ms
#define TS_1300_ON_MAX	1200  // in ms
#define TS_1300_OFF_MIN	800  // in ms

#define TS_2100_ON_MIN	0  // in ms
#define TS_2100_ON_MAX	2000  // in ms
#define TS_2100_OFF_MIN	0  // in ms

#define TS_1100_ON_MIN	500X  // in ms
#define TS_1100_ON_MAX	500  // in ms
#define TS_1100_OFF_MIN	3000  // in ms
#define TS_1100_OFF_MAX	3000  // in ms

// DTMF codes
#define FT_DTMF_0			0
#define FT_DTMF_9			9
#define FT_DTMF_A			10
#define FT_DTMF_B			11
#define FT_DTMF_C			12
#define FT_DTMF_D			13
#define FT_DTMF_HASH		14
#define FT_DTMF_STAR		15
#define FT_BUSY_TONE		16
#define FT_1300				17
#define FT_2100				18
#define FT_440				19

#define FT_NONE				0xFF



#endif


