
typedef struct erase_info_t {
    u_long      Offset;
    u_long      Size;
} erase_info_t;


#define MTD_ERASE_PENDING      	0x01
#define MTD_ERASING		0x02
#define MTD_ERASE_SUSPEND	0x04
#define MTD_ERASE_DONE          0x08
#define MTD_ERASE_FAILED        0x10

struct erase_info {
	struct mtd_info *mtd;
	u_long addr;
	u_long len;
	u_long time;
	u_long retries;
	u_int dev;
	u_int cell;
	void (*callback) (struct erase_info *self);
	u_long priv;
	u_char state;
	struct erase_info *next;
};



#define MAX_MTD_DEVICES 16



#define MTD_ABSENT		0
#define MTD_RAM			1
#define MTD_ROM			2
#define MTD_FLASH		3
#define MTD_REVFLASH		4
#define MTD_PEROM		5
#define MTD_OTHER		14
#define MTD_UNKNOWN		15


#define MTD_CLEAR_BITS		1
#define MTD_SET_BITS		2
#define MTD_ERASEABLE		4
#define MTD_WRITEB_WRITEABLE	8
#define MTD_VOLATILE		16

#define MTD_WRITEABLE		3

#ifdef __KERNEL__
struct mtd_info_user {
#else
struct mtd_info {
#endif
	char name[32];
	u_long size;	
	u_char flags;
	u_char mfr;
	u_char id;
	u_char interleave;
	u_char devtype;
	u_long erasesize;
	u_long devsize;
	int usage;
};


#ifdef __KERNEL__
struct mtd_info {
	char name[32];
	u_long size;	
	u_char flags;
	u_char mfr;
	u_char id;
	u_char interleave;
	u_char devtype;
	u_long erasesize;
	u_long devsize;
	int usage;
	int (*erase) (struct mtd_info *mtd, struct erase_info *instr);
	int (*point) (struct mtd_info *mtd, u_long from, u_long len, u_long *retlen, u_char **mtdbuf);
	void (*unpoint) (struct mtd_info *mtd, u_long addr);
	int (*read) (struct mtd_info *mtd, u_long from, u_long len, u_long *retlen, u_char *buf, int userspace);
	int (*write) (struct mtd_info *mtd, u_long to, u_long len, u_long *retlen, const u_char *buf, int userspace);
	void (*openclose) (struct mtd_info *mtd, int which);
	void (*sync) (struct mtd_info *mtd);
	void *priv;
};


#define MEMGETINFO              _IOR('M', 1, struct mtd_info_user)
#else
#define MEMGETINFO              _IOR('M', 1, struct mtd_info)
#endif

#define MEMERASE                _IOW('M', 2, erase_info_t)
#define MEMADD					 _IO('M', 3)
#define MEMREMOVE                _IO('M', 4)


#define MTD_MAJOR 90


#ifdef __KERNEL__



struct mtd_notifier {
	void (*func)(struct mtd_info *mtd, int num);
	struct mtd_notifier *next;
};

extern int add_mtd_device(struct mtd_info *mtd);
extern struct mtd_info *get_mtd_device(u_char num);
extern int del_mtd_device (struct mtd_info *mtd);
extern void register_mtd_notifier (struct mtd_notifier *new);
extern int unregister_mtd_notifier (struct mtd_notifier *old);



#ifndef MTDC
#define MTD_ERASE(mtd, args...) (*(mtd->erase))(mtd, args)
#define MTD_POINT(mtd, a,b,c,d) (*(mtd->point))(mtd, a,b,c, (u_char **)(d))
#define MTD_UNPOINT(mtd, args...) (*(mtd->unpoint))(mtd, args)
#define MTD_READ(mtd, args...) (*(mtd->read))(mtd, args)
#define MTD_WRITE(mtd, args...) (*(mtd->write))(mtd, args)
#define MTD_OPENCLOSE(mtd, args...) (*(mtd->openclose))(mtd, args)
#define MTD_SYNC(mtd) do { if (mtd->sync) (*(mtd->sync))(mtd);  } while (0) 
#endif

#endif


