/* 
 * Herein lies all the functions/variables that are "exported" for linkage
 * with dynamically loaded kernel modules.
 *			Jon.
 *
 * - Stacked module support and unified symbol table added (June 1994)
 * - External symbol table support added (December 1994)
 * - Versions on symbols added (December 1994)
 * by Bjorn Ekwall <bj0rn@blox.se>
 */

/*
 * Modifications for P2001 Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/smp.h>
#include <linux/fs.h>
#include <linux/blkdev.h>
#include <linux/sched.h>
#include <linux/kernel_stat.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/ptrace.h>
#include <linux/sys.h>
#include <linux/utsname.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/timer.h>
#include <net/sock.h>
#include <linux/net.h>
#include <net/udp.h>
#include <linux/binfmts.h>
#include <linux/personality.h>
#include <linux/termios.h>
#include <linux/tqueue.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/locks.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/sem.h>
#include <linux/minix_fs.h>
#include <linux/ext2_fs.h>
#include <linux/random.h>
#include <linux/mount.h>
#include <linux/pagemap.h>
#include <linux/sysctl.h>
#include <linux/hdreg.h>
#include <linux/skbuff.h>
#include <linux/genhd.h>
#include <linux/swap.h>
#include <linux/ctype.h>
#include <linux/file.h>
//DW
#include <net/route.h>
#include <linux/firewall.h>
#include <linux/ip_fw.h>
//DW

#ifndef CONFIG_ARM
extern unsigned char aux_device_present, kbd_read_mask;
#endif

#ifdef CONFIG_PCI
#include <linux/bios32.h>
#include <linux/pci.h>
#endif
#if defined(CONFIG_PROC_FS)
#include <linux/proc_fs.h>
#endif
#ifdef CONFIG_KERNELD
#include <linux/kerneld.h>
#endif
#include <asm/irq.h>
#ifdef __SMP__
#include <linux/smp.h>
#endif

#ifdef CONFIG_ARCH_P2001
#if (defined(PROJECT_FAMILY_COMPACTDSL) || defined(PROJECT_FAMILY_MODULARDSL) || defined(PROJECT_FAMILY_ROUTERMODUL))
#include "display.h"

int send_mps_msg(void*);
int send_mps_msg_notify(void*, void (*f)(void *), void *arg);
int register_mps_callback(unsigned id, void (*f)(void *));
void unregister_mps_callback(unsigned id);

/* fiq */
extern int  register_fscfiq( void (*fct) (void));
extern int  free_fscfiq(unsigned fsc_hdlno);
extern int  register_fiq(unsigned num, void (*fct) (void));
extern int  free_fiq(unsigned num);
extern void save_fiq_clf(int * flags);
extern void restore_fiq(int flags);
extern void disable_fiq(unsigned fiq_no);
extern void enable_fiq(unsigned fiq_no);

/* tc1 */
extern void tc1_stop(void);
extern int  tc1_register( void (*fct) (void));
extern int  tc1_free(unsigned tc1_hdlno);
extern int  tc1_mark_active(unsigned tc1_hdlno);


/* sn */
//extern int f_sn_adr(t_interface_id *, int) ;
extern int f_sn_adr(void);
extern int f_sn_port(int,int) ;
extern int f_sn_get_highway(int, int*, int*) ;
extern int f_dsv_send_sms(unsigned short v_sms_index, unsigned char *W_sms_string);
extern int f_dsv_sms_get_data(unsigned short v_sms_index);
extern int f_sn_offset(int W_pr);

//extern t_mfve  mfve_tab;
extern int mfve_tab;
extern int ab_offset; 

/* dtmf */
extern int     f_mfv_active( void );
extern void    f_mfv_send( int index, int dest_adr, char * data );
extern void    f_mfv_stop( int index );
extern unsigned int   f_dtmf_active( void );


/* kf_p2001 */
extern int f_kf_switch( int switch_type, int src_hw, int src_ts, int dest_hw, int dest_ts );

/* fsk */
extern void f_fsk_stop( int index );
extern int  f_fsk_active( void );
extern void f_fsk_send;
extern int f_fsk_offhook( int switch_type, int src_hw, int src_ts );
extern void f_fsk_dtas( int index, int dest_adr );
//extern unsigned char MFV_CHAR[MAX_MFVE];
extern unsigned char MFV_CHAR;
extern int           mfve_dbit;

extern int f_p2001_hdlc_tx_cb(unsigned scc_id, unsigned char *buf, unsigned len, void (*cb)(void *), void *arg);
extern int hdlc_canput(unsigned scc_id);
extern int  f_p2001_bhdlc_register_handler;
extern int  f_p2001_bhdlc_reset_handler( unsigned int W_scc_id );

extern void p2001_GPO4( unsigned long,unsigned long );
extern void davosv2_LedLatch( unsigned long,unsigned long );
extern void davosv2_ModLatch( unsigned long,unsigned long );

#if (defined(PROJECT_FAMILY_COMPACTDSL) || defined(PROJECT_FAMILY_MODULARDSL) )
extern void f_dtmf_available( int mfe_type );
#endif

/* PROJECT_FAMILY_MODULARDSL */
extern void f_monitor_init(void);
extern void f_mon_rx_abort_cnt_set( unsigned char W_cnt );
extern void f_monitor_close(void);
extern int f_monitor_transmit( void );
extern int f_monitor_rxsize( void );
extern int f_monitor_txsize( void );
extern int f_monitor_get( void );
extern int f_mon_sw( void );
extern int f_ci_sw( void );
extern int f_ci_iom_read( void );
extern int f_ci_iom_transmit( void );
extern int f_hw_module_exists( void );
extern void f_hw_module_clr(unsigned int W_mod_nr );
extern void F_p2001_l1_init_data( void );
extern void F_p2001_l1_enable_l1( void );
extern int  F_p2001_l1_act_req( void );
extern int  F_p2001_l1_deact_req( void );
extern int  F_p2001_l1_ready_for_data( void );
extern int  F_p2001_l1_data_req( void );
extern void f_register_init_ab_port( void );
extern int  f_hw_read_country_code( void );   

/* VoIP QoS */
extern int	qos_add_udp_port( unsigned short port );
extern int	qos_rem_udp_port( unsigned short port );
#endif
#endif

extern const unsigned long irq_masks[32];
extern void set_fiq_vec(void (*)(void));


extern char *get_options(char *str, int *ints);
extern void set_device_ro(kdev_t dev,int flag);
extern struct file_operations * get_blkfops(unsigned int);
extern void blkdev_release(struct inode * inode);

extern void *sys_call_table;

extern struct timezone sys_tz;
extern int request_dma(unsigned int dmanr, char * deviceID);
extern void free_dma(unsigned int dmanr);

extern void hard_reset_now(void);

extern void select_free_wait(select_table * p);
extern int select_check(int flag, select_table * wait, struct file * file);


extern struct socket *sockfd_lookup(int fd, int *err);
extern void sockfd_put(struct socket *sock);


struct symbol_table symbol_table = {
#include <linux/symtab_begin.h>
#ifdef MODVERSIONS
	{ (void *)1 /* Version version :-) */,
		SYMBOL_NAME_STR (Using_Versions) },
#endif

	/* stackable module support */
	X(register_symtab_from),
	X(get_module_symbol),
#ifdef CONFIG_KERNELD
	X(kerneld_send),
#endif
	X(get_options),

	/* system info variables */
	/* These check that they aren't defines (0/1) */
#ifndef EISA_bus__is_a_macro
	X(EISA_bus),
#endif
#ifndef MCA_bus__is_a_macro
	X(MCA_bus),
#endif
#ifndef wp_works_ok__is_a_macro
	X(wp_works_ok),
#endif

#ifdef CONFIG_PCI
	/* PCI BIOS support */
	X(pcibios_present),
	X(pcibios_find_class),
	X(pcibios_find_device),
	X(pcibios_read_config_byte),
	X(pcibios_read_config_word),
	X(pcibios_read_config_dword),
    	X(pcibios_strerror),
	X(pcibios_write_config_byte),
	X(pcibios_write_config_word),
	X(pcibios_write_config_dword),
#endif

	/* process memory management */
	X(verify_area),
	X(do_mmap),
	X(do_munmap),
	X(exit_mm),

	/* internal kernel memory management */
	X(__get_free_pages),
	X(free_pages),
	X(kmalloc),
	X(kfree),
	X(vmalloc),
	X(vremap),
	X(vfree),
 	X(mem_map),
#ifndef NO_MM	
 	X(remap_page_range),
#endif
	X(high_memory),
	X(max_mapnr),
	X(update_vm_cache),

	/* filesystem internal functions */
	X(getname),
	X(putname),
	X(__iget),
	X(iput),
	X(namei),
	X(lnamei),
	X(open_namei),
	X(sys_close),
	X(close_fp),
	X(check_disk_change),
	X(invalidate_buffers),
	X(invalidate_inodes),
	X(invalidate_inode_pages),
	X(fsync_dev),
	X(permission),
	X(inode_setattr),
	X(inode_change_ok),
	X(set_blocksize),
	X(getblk),
	X(bread),
	X(breada),

	X(select_check),
	X(select_free_wait),

	X(__brelse),
	X(__bforget),
	X(ll_rw_block),
	X(brw_page),
	X(__wait_on_buffer),
	X(mark_buffer_uptodate),
	X(unlock_buffer),
	X(dcache_lookup),
	X(dcache_add),
	X(add_blkdev_randomness),
	X(generic_file_read),
	X(generic_file_mmap),
	X(generic_readpage),
	X(__fput),
	X(make_bad_inode),

	/* device registration */
	X(register_chrdev),
	X(unregister_chrdev),
	X(register_blkdev),
	X(unregister_blkdev),
	X(tty_register_driver),
	X(tty_unregister_driver),
	X(tty_std_termios),

	/* block device driver support */
	X(block_read),
	X(block_write),
	X(block_fsync),
	X(wait_for_request),
	X(blksize_size),
	X(hardsect_size),
	X(blk_size),
	X(blk_dev),
#ifndef NO_MM	
	X(max_sectors),
	X(max_segments),
#endif
	X(is_read_only),
	X(set_device_ro),
	X(bmap),
	X(sync_dev),
	X(get_blkfops),
	X(blkdev_open),
	X(blkdev_release),
	X(gendisk_head),
	X(resetup_one_dev),
	X(unplug_device),
	X(make_request),
	X(tq_disk),

#ifdef CONFIG_SERIAL	
	/* Module creation of serial units */
	X(register_serial),
	X(unregister_serial),
#endif
	/* tty routines */
	X(tty_hangup),
	X(tty_wait_until_sent),
	X(tty_check_change),
	X(tty_hung_up_p),
	X(do_SAK),
#ifdef CONFIG_CONSOLE
	X(console_print),
#endif
	/* filesystem registration */
	X(register_filesystem),
	X(unregister_filesystem),

	/* executable format registration */
	X(register_binfmt),
	X(unregister_binfmt),
	X(search_binary_handler),
	X(prepare_binprm),
#ifndef NO_MM
	X(remove_arg_zero),
#endif
	
	/* execution environment registration */
	X(lookup_exec_domain),
	X(register_exec_domain),
	X(unregister_exec_domain),

	/* sysctl table registration */
	X(register_sysctl_table),
	X(unregister_sysctl_table),
	X(sysctl_string),
	X(sysctl_intvec),
	X(proc_dostring),
	X(proc_dointvec),
	X(proc_dointvec_minmax),

	/* interrupt handling */
	X(request_irq),
	X(free_irq),
	X(enable_irq),
	X(disable_irq),
	X(probe_irq_on),
	X(probe_irq_off),
	X(bh_active),
	X(bh_mask),
	X(bh_mask_count),
	X(bh_base),
	X(add_timer),
	X(del_timer),
	X(tq_timer),
	X(tq_immediate),
	X(tq_scheduler),
	X(timer_active),
	X(timer_table),
 	X(intr_count),

	/* autoirq from  drivers/net/auto_irq.c */
#ifdef CONFIG_NET
	X(autoirq_setup),
	X(autoirq_report),
#endif

	/* dma handling */
	X(request_dma),
	X(free_dma),
#ifdef HAVE_DISABLE_HLT
	X(disable_hlt),
	X(enable_hlt),
#endif

	/* IO port handling */
	X(check_region),
	X(request_region),
	X(release_region),

	/* process management */
	X(wake_up),
	X(wake_up_interruptible),
	X(sleep_on),
	X(interruptible_sleep_on),
	X(schedule),
	X(current_set),
	X(jiffies),
	X(xtime),
	X(do_gettimeofday),
	X(loops_per_sec),
	X(need_resched),
	X(kstat),
	X(kill_proc),
	X(kill_pg),
	X(kill_sl),
	X(force_sig),

	/* Network interfaces */
	X(ip_rt_dev),

	/* misc */
	X(panic),
	X(printk),
	X(sprintf),
	X(vsprintf),
	X(kdevname),
	X(simple_strtoul),
	X(system_utsname),
	X(sys_call_table),
	X(hard_reset_now),
	X(_ctype),
	X(_ctmp),
	X(get_random_bytes),

	/* Signal interfaces */
	X(send_sig),

	/* Program loader interfaces */
#ifndef NO_MM
	X(setup_arg_pages),
	X(copy_strings),
#endif
	X(do_execve),
	X(flush_old_exec),
	X(open_inode),
	X(read_exec),

	/* Miscellaneous access points */
	X(si_meminfo),

	/* Added to make file system as module */
	X(set_writetime),
	X(sys_tz),
	X(__wait_on_super),
	X(file_fsync),
	X(clear_inode),
	X(refile_buffer),
	X(nr_async_pages),
	X(___strtok),
	X(init_fifo),
	X(super_blocks),
	X(fifo_inode_operations),
	X(chrdev_inode_operations),
	X(blkdev_inode_operations),
	X(read_ahead),
	X(get_hash_table),
	X(get_empty_inode),
	X(insert_inode_hash),
	X(event),
	X(__down),
	X(__up),
	X(securelevel),
/* all busmice */
	X(add_mouse_randomness),
	X(fasync_helper),
#if  !defined(CONFIG_ARM) && !defined(__mc68000__)
/* psaux mouse */
	X(aux_device_present),
	X(kbd_read_mask),
#endif

#ifdef CONFIG_BLK_DEV_IDE_PCMCIA
	X(ide_register),
	X(ide_unregister),
#endif

#ifdef CONFIG_BLK_DEV_MD
	X(disk_name),	/* for md.c */
#endif
 	
	/* binfmt_aout */
	X(get_write_access),
	X(put_write_access),

#ifdef CONFIG_PROC_FS
	X(proc_dir_inode_operations),
#endif

#ifdef CONFIG_ARCH_P2001
	X(_func_disp_cb_1),
	X(_func_disp_cb_2),
	X(_func_disp_cb_3),
	X(_func_disp_cb_4),
	X(p2001_GPO4),

#if (defined(PROJECT_FAMILY_COMPACTDSL) || defined(PROJECT_FAMILY_MODULARDSL) || defined (PROJECT_FAMILY_ROUTERMODUL))
	X(irq_masks),
	X(set_fiq_vec),
	X(send_mps_msg),
	X(send_mps_msg_notify),
	X(register_mps_callback),
	X(unregister_mps_callback),
#ifdef UDP_SWITCHING
	X(udp_register_hook),
	X(udp_send),
	X(udp_de_register_hook),
	X(__release_sock),
	X(sockfd_lookup),
	X(sockfd_put),
#endif
        
#if defined(CONFIG_ELMEG_VOIP_RTP_QOS)
	X(qos_add_udp_port),
	X(qos_rem_udp_port),
#endif

#ifdef PROJECT_VARIANT_MODKERN	
	/* fiq */
	X(register_fscfiq),
	X(free_fscfiq),
	X(save_fiq_clf),
	X(restore_fiq),
	X(disable_fiq),
	X(enable_fiq),

	/* tc1 */
	X(tc1_stop),
	X(tc1_register),
	X(tc1_free),
	X(tc1_mark_active),

	/* switching network lxsn*/
	X(f_sn_adr),
	X(f_sn_get_highway),
	X(f_sn_port),
	X(f_dsv_send_sms),
	X(f_dsv_sms_get_data),
	X(f_sn_offset),

	X(f_dtmf_active),
	X(f_mfv_active),
	X(f_mfv_send),
	X(mfve_tab),
	X(f_mfv_stop),
	X(mfve_dbit),
	X(MFV_CHAR),

	X(f_kf_switch),
	X(f_fsk_stop),
	X(f_fsk_send),
	X(f_fsk_active),
	X(f_fsk_offhook),
	X(f_fsk_dtas),
	X(ab_offset),

	/* b_hdlc */
	X(f_p2001_hdlc_tx_cb),
	X(hdlc_canput),
	X(f_p2001_bhdlc_register_handler),
	X(f_p2001_bhdlc_reset_handler),


//#if defined(PROJECT_FAMILY_MODULARDSL)
	X(f_monitor_init),
	X(f_mon_rx_abort_cnt_set),
	X(f_monitor_close),
	X(f_ci_sw),
	X(f_ci_iom_read),
	X(f_ci_iom_transmit),
	X(f_mon_sw),
	X(f_monitor_transmit),
	X(f_monitor_rxsize),
	X(f_monitor_txsize),
	X(f_monitor_get),
	X(f_hw_read_country_code),
#if (defined(PROJECT_FAMILY_COMPACTDSL) || defined(PROJECT_FAMILY_MODULARDSL))
    X(f_dtmf_available),
#endif
#if defined(PROJECT_FAMILY_MODULARDSL)
	X(f_hw_module_exists),
	X(f_hw_module_clr),
	X(F_p2001_l1_init_data),
	X(F_p2001_l1_enable_l1),
	X(F_p2001_l1_act_req),
	X(F_p2001_l1_deact_req),
	X(F_p2001_l1_ready_for_data),
	X(F_p2001_l1_data_req),
#endif
#endif

#if defined(PROJECT_T240) || defined(PROJECT_T444)
	X(davosv2_LedLatch),
	X(davosv2_ModLatch),
#endif

#endif
#endif

	/* Modular sound */
	X(sys_open),
	X(sys_read),
	/********************************************************
	 * Do not add anything below this line,
	 * as the stacked modules depend on this!
	 */
#include <linux/symtab_end.h>
};

/*
int symbol_table_size = sizeof (symbol_table) / sizeof (symbol_table[0]);
*/
