/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 * Written by Juergen Selent (j.selent@elmeg.de)
 *
 * ip_masq_h323.c - H.323 nat app
 * ------------------------------
 *
 * $RCSfile: ip_masq_h323.c,v $
 * $Source: /opt/funkwerk/cvsroot/source/linux/uclinux2.0.38/net/ipv4/ip_masq_h323.c,v $
 * $Author: jselent $
 * $Revision: 1.6 $
 * $Date: 2008-05-19 06:44:54 $
 * $State: Exp $
 * $Name:  $
 *
 * $Log: ip_masq_h323.c,v $
 * Revision 1.6  2008-05-19 06:44:54  jselent
 * CHG: copyright info
 *
 * Revision 1.5  2006-02-10 10:12:17  jselent
 * CHG: debug output disabled
 *
 * Revision 1.4  2006/02/09 17:42:35  jselent
 * FIX: bug in tpkt-framer fixed which could detect tpkt-frames with negative length
 *
 * Revision 1.3  2006/02/09 13:09:40  jselent
 * FIX: masquerading of h.245 outgoing packets
 *
 * Revision 1.2  2004/06/10 16:19:04  selent
 * no NAT entries for H.245/OpenLogicalChannel & DataApplicationCapability T.120
 *
 * Revision 1.1  2004/06/09 14:48:46  selent
 * imported
 *
 * Initial Version	Thu Jun  3 16:08:36 CEST 2004 - jse
 * Written
 */

/*
 * NOTE! Requires a little patch in ip_masq_app.c to work.
 * mapp->n_attach must be decremented by ip_masq_bind_chg()
 * before masq_done_1() is called in function ip_masq_unbind_app().
 */

/*
 * Known problems:
 * ---------------
 * - no ASN.1
 * - did not desegment the tcp stream
 * - searches the ip-addresses with memcpm()
 * - tunneling of H.245 PDUs
 * - multiple PDUs between two TPKTs
 */

/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 *      
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */



/* ************************************************************************** */
/* *** INCLUDES ************************************************************* */

#include <linux/module.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/skbuff.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <net/protocol.h>
#include <net/udp.h>
#include <net/tcp.h>
#include <net/ip_masq.h>



/* ************************************************************************** */
/* *** STUFF **************************************************************** */

/* utils.c */
extern char	*in_ntoa( unsigned long );


/*
 * Debug macros
 */
#if 0
#define DEB(fmt, args...) \
	printk( KERN_DEBUG/**/"ip_masq_h323: [DEB] "/**/fmt, ##args );
#else
#define DEB(fmt, args...)
#endif
#define INF(fmt, args...) \
	printk( KERN_NOTICE/**/"ip_masq_h323: [INF] "/**/fmt, ##args );
#define WRN(fmt, args...) \
	printk( KERN_WARNING/**/"ip_masq_h323: [WRN] "/**/fmt, ##args );
#define ERR(fmt, args...) \
	printk( KERN_ERR/**/"ip_masq_h323: [ERR] "/**/fmt, ##args );



/* ************************************************************************** */
/* *** TPKT ***************************************************************** */

/*
 * TPKT framing stuff
 */
typedef struct {
	int			(*f)(struct ip_masq *,struct device *,unsigned char *,int);
	int 			state;
	int 			len;
} tpkt_framer_t;

typedef struct {
	tpkt_framer_t		tx;
	tpkt_framer_t		rx;
	int			used;
} tpkt_app_t;

#define ST_TPKT_SEARCH		0
#define ST_TPKT_F1		1
#define ST_TPKT_F2		2
#define ST_TPKT_F3		3
#define ST_TPKT_SKIP		4
#define ST_TPKT_FOUND		5

#define TPKT_MAX_APPS		128

static tpkt_app_t		tpktApps[TPKT_MAX_APPS];

static void
tpktInit( void )
{
	int	i;

	for (i=0; i<TPKT_MAX_APPS; i++) {
		tpktApps[i].used = 0;
	}
}

static tpkt_app_t *
tpktAppAlloc( int (*txcb)(struct ip_masq *,struct device *,unsigned char *,int),
              int (*rxcb)(struct ip_masq *,struct device *,unsigned char *,int))
{
	int	i;

	for (i=0; i<TPKT_MAX_APPS; i++) {
		if (tpktApps[i].used == 0) {
			tpktApps[i].used = 1;
			tpktApps[i].tx.state = ST_TPKT_SEARCH;
			tpktApps[i].tx.f = txcb;
			tpktApps[i].rx.state = ST_TPKT_SEARCH;
			tpktApps[i].rx.f = rxcb;
			return( &tpktApps[i] );
		}
	}
	return( NULL );
}

static void
tpktAppRelease( tpkt_app_t *ap )
{
	if (ap != NULL) {
		ap->used = 0;
	}
}

static int
tpktFrameSKB( tpkt_framer_t *fp,
              struct ip_masq *ms,
              struct device *dev,
              struct sk_buff **skb_p )
{
	struct sk_buff	*skb;
	struct iphdr	*iph;
	struct tcphdr	*th;
	unsigned char	*data;
	unsigned char	*data_limit;
	int		len;

	skb = *skb_p;
	iph = skb->h.iph;
	th = (struct tcphdr *)&(((char *)iph)[iph->ihl * 4]);
	data = (char *)th + (th->doff * 4);
	data_limit = skb->h.raw + skb->len;

	while (data < data_limit) {
		switch (fp->state) {
		case ST_TPKT_SEARCH:
			if (*data++ == 0x03) {
				fp->state = ST_TPKT_F1;
			}
			break;
		case ST_TPKT_F1:
			if (*data++ == 0x00) {
				fp->state = ST_TPKT_F2;
			} else {
				fp->state = ST_TPKT_SEARCH;
			}
			break;
		case ST_TPKT_F2:
			fp->len = *data++ * 256;
			fp->state = ST_TPKT_F3;
			break;
		case ST_TPKT_F3:
			fp->len = (fp->len + *data++) - 4;
			if (fp->len > 0) {
				DEB( "tpktFrameSKB: found frame len=%d\n",fp->len );
				fp->state = ST_TPKT_FOUND;
			} else {
				fp->state = ST_TPKT_SEARCH;
			}
			break;
		case ST_TPKT_FOUND:
			len = data_limit - data;
			if (fp->len <= len) {
				fp->f( ms,dev,data,fp->len );
				fp->state = ST_TPKT_SEARCH;
				data += fp->len;
			} else {
				WRN( "tpktFrameSKB: short frame!\n" );
				fp->f( ms,dev,data,len );
				data += len;
				fp->len -= len;
				fp->state = ST_TPKT_SKIP;
			}
			break;
		case ST_TPKT_SKIP:
			data++;
			if (--fp->len <= 0) {
				fp->state = ST_TPKT_SEARCH;
			}
			break;
		}
	}
	return( 0 );
}



/* ************************************************************************** */
/* *** H.245 **************************************************************** */

/*
 * T.120 portno
 */
#define PORT_T120	1503
#define PORT_T120_H	0x05
#define PORT_T120_L	0xdf



static int
masq_h245_out( struct ip_masq_app *mapp,
                     struct ip_masq *ms,
                     struct sk_buff **skb_p,
                     struct device *dev )
{
	tpkt_app_t	*ap;

	DEB( "masq_h245_out: ENTER\n" );
	if ((ap = ms->app_data) != NULL) {
		tpktFrameSKB( &ap->tx,ms,dev,skb_p );
	}
	return( 0 );
}

static int
masq_h245_out_2( struct ip_masq *ms,
                 struct device *dev,
                 unsigned char *data,
                 int len )
{
	unsigned char	*data_limit;
	struct ip_masq	*nms;

	DEB( "masq_h245_out_2: ENTER len=%d\n",len );
	DEB( ">>> caller: %s:%d\n",in_ntoa( ms->saddr ),ntohs( ms->sport ));
	DEB( ">>> router: %s:%d\n",in_ntoa( ms->maddr ),ntohs( ms->mport ));
	DEB( ">>> callee: %s:%d\n",in_ntoa( ms->daddr ),ntohs( ms->dport ));

	data_limit = data + len;

	/* H.245 Request  + OpenLogicalChannel */
	if (len > 5 && (data[0] & 0x6f) == 0x03) {
		INF( "masq_h245_out: H.245/OpenLogicalChannel\n" );
		data += 5; /* advance into ForwardLogicalChannelParameters */

	/* H.245 Request  + OpenLogicalChannelAck */
	} else if (len > 7 && (data[0] & 0x6f) == 0x22 && (data[1] & 0x80) == 0x80) {
		INF( "masq_h245_out: H.245/OpenLogicalChannelAck\n" );
		data += 7; /* advance into ReverseLogicalChannelParameters */
	} else {
		return( 0 );
	}

	while ((data + 5) < data_limit) {
		if (!memcmp( data,&ms->saddr,4 )) {
			if (data[4] != PORT_T120_H && data[5] != PORT_T120_L) {
				INF( "masq_h245_out: found RTCP/RTP %d.%d.%d.%d:%d\n",data[0],data[1],data[2],data[3],data[4]*256+data[5] );
				if ((nms = ip_masq_out_get_2( IPPROTO_UDP,ms->saddr,htons( data[4]*256+data[5] ),ms->daddr,0 )) != NULL) {
					ip_masq_set_expire( nms,0 );
				} else {
					if ((nms = ip_masq_new( dev,IPPROTO_UDP,ms->saddr,htons( data[4]*256+data[5] ),ms->daddr,0,IP_MASQ_F_NO_DPORT )) == NULL) {
						ERR( "masq_h245_out: ip_masq_new failed!\n" );
						return( 0 );
					}
				}
				memcpy( data,&nms->maddr,4 );
				memcpy( data+4,&nms->mport,2 );
			}
			data += 6;
			continue;
		}
		data++;
	}
	return( 0 );
}

static int
masq_h245_in( struct ip_masq_app *mapp,
                     struct ip_masq *ms,
                     struct sk_buff **skb_p,
                     struct device *dev )
{
#if 0
	tpkt_app_t	*ap;

	if ((ap = ms->app_data) != NULL) {
		tpktFrameSKB( &ap->rx,ms,dev,skb_p );
	}
#endif
	DEB( "masq_h245_in: ENTER\n" );
	return( 0 );
}

static int
masq_h245_in_2( struct ip_masq *ms,
                struct device *dev,
                unsigned char *data,
                int len )
{
#if 1
	DEB( "masq_h245_in_2: ENTER len=%d\n",len );
	DEB( ">>> caller: %s:%d\n",in_ntoa( ms->saddr ),ntohs( ms->sport ));
	DEB( ">>> router: %s:%d\n",in_ntoa( ms->maddr ),ntohs( ms->mport ));
	DEB( ">>> callee: %s:%d\n",in_ntoa( ms->daddr ),ntohs( ms->dport ));
#endif
	return( 0 );
}

static int
masq_h245_init_1( struct ip_masq_app *mapp,struct ip_masq *ms )
{
	DEB( "masq_h245_init_1: ENTER\n" );
	DEB( ">>> caller: %s:%d\n",in_ntoa( ms->saddr ),ntohs( ms->sport ));
	DEB( ">>> router: %s:%d\n",in_ntoa( ms->maddr ),ntohs( ms->mport ));
	DEB( ">>> callee: %s:%d\n",in_ntoa( ms->daddr ),ntohs( ms->dport ));
	if (ms->app_data != NULL) {
		WRN( "masq_h245_init_1: app_data != NULL\n" );
	}
	ms->app_data = tpktAppAlloc( masq_h245_out_2,masq_h245_in_2 );
	return( 0 );
}

static int
masq_h245_done_1( struct ip_masq_app *mapp,struct ip_masq *ms )
{
	DEB( "masq_h245_done_1: ENTER\n" );
	DEB( ">>> caller: %s:%d\n",in_ntoa( ms->saddr ),ntohs( ms->sport ));
	DEB( ">>> router: %s:%d\n",in_ntoa( ms->maddr ),ntohs( ms->mport ));
	DEB( ">>> callee: %s:%d\n",in_ntoa( ms->daddr ),ntohs( ms->dport ));
	tpktAppRelease( ms->app_data );
	ms->app_data = NULL;
	if (unregister_ip_masq_app( mapp ) == 0) {
		DEB( "masq_h245_done_1: kfreeing ip_masq_app\n" );
		kfree( mapp );
	}
	return( 0 );
}

static int
masqMonitorH245Port( unsigned short port )
{
	struct ip_masq_app	*mapp;
	int			rc;

	DEB( "masqMonitorH245Port: ENTER port=%d\n",port );
	rc = 0;

	if (ip_masq_app_get( IPPROTO_TCP,port ) == NULL) {
		if ((mapp = kmalloc( sizeof( *mapp ),GFP_ATOMIC )) == NULL) {
			ERR( "masqMonitorH245Port: out of memory\n" );
			return( -ENOMEM );
		}
		mapp->next = NULL;
		mapp->name = "H.245";
		mapp->type = 0;
		mapp->masq_init_1 = masq_h245_init_1;
		mapp->masq_done_1 = masq_h245_done_1;
		mapp->pkt_out = masq_h245_out;
		mapp->pkt_in = masq_h245_in;
		if ((rc = register_ip_masq_app( mapp,IPPROTO_TCP,port )) != 0) {
			ERR( "masqMonitorH245Port: register_ip_masq_app for port=%d failed\n",port );
			kfree( mapp );
		}
	}
	return( rc );
}



/* ************************************************************************** */
/* *** H.225 **************************************************************** */

/*
 * H.225 portno
 */
#define PORT_H225	1720
#define PORT_H225_H	0x06
#define PORT_H225_L	0xb8

/*
 * Q.931 messages
 */
#define Q931_ALERT		0x01
#define Q931_CALLPROC		0x02
#define Q931_PROGRESS		0x03
#define Q931_SETUP		0x05
#define Q931_CONNECT		0x07
#define Q931_CONNECT		0x07



static char *
toStr_q931msg( unsigned char c )
{
	switch (c) {
	case 0x01: return( "ALERT" );
	case 0x02: return( "CALL-PROC" );
	case 0x03: return( "PROGRESS" );
	case 0x05: return( "SETUP" );
	case 0x07: return( "CONNECT" );
	case 0x0d: return( "SETUP-ACK" );
	case 0x0F: return( "CONNECT-ACK" );
	case 0x20: return( "USER-INFO" );
	case 0x21: return( "SUSPEND-REJ" );
	case 0x22: return( "RESUME-REJ" );
	case 0x25: return( "SUSPEND" );
	case 0x26: return( "RESUME" );
	case 0x2d: return( "SUSPEND-ACK" );
	case 0x2e: return( "RESUME-ACK" );
	case 0x45: return( "DISCONNECT" );
	case 0x46: return( "RESTART" );
	case 0x4d: return( "RELEASE" );
	case 0x4e: return( "RESTART-ACK" );
	case 0x5a: return( "RELEASE-COMPLETE" );
	case 0x60: return( "SEGMENT" );
	case 0x6e: return( "NOTIFY" );
	case 0x75: return( "STATUS-ENQ" );
	case 0x79: return( "CONGESTION-CTRL" );
	case 0x7b: return( "INFO" );
	case 0x7d: return( "STATUS" );
	default:   return( "UNKOWN" );
	}
}

static unsigned char *
findIE( unsigned char ie,unsigned char *p,unsigned char *lim )
{
	while (p < lim) {
		DEB( "findIE: 0x%02x\n",*p );
		if (*p == ie) {
			return( p );
		} else if ((*p & 0x80) == 0) {
			p = p + p[1] + 2;
		} else {
			p++;
		}
	}
	return( NULL );
}

static int
masq_h225_out( struct ip_masq_app *mapp,
               struct ip_masq *ms,
               struct sk_buff **skb_p,
               struct device *dev )
{
	tpkt_app_t	*ap;

	DEB( "masq_h225_out: ENTER\n" );
	if ((ap = ms->app_data) != NULL) {
		tpktFrameSKB( &ap->tx,ms,dev,skb_p );
	}
	return( 0 );
}

static int
masq_h225_out_2( struct ip_masq *ms,
                 struct device *dev,
                 unsigned char *data,
                 int len )
{
	unsigned char	*data_limit;
	unsigned char	*p;
	int		q931MsgType;
	unsigned char	callerAddr[6];

	DEB( "masq_h225_out_2: ENTER len=%d\n",len );
	DEB( ">>> caller: %s:%d\n",in_ntoa( ms->saddr ),ntohs( ms->sport ));
	DEB( ">>> router: %s:%d\n",in_ntoa( ms->maddr ),ntohs( ms->mport ));
	DEB( ">>> callee: %s:%d\n",in_ntoa( ms->daddr ),ntohs( ms->dport ));

	data_limit = data + len;

	/* q.931 protocol discr, callref len, callref, msgtype */
	if (data[0] != 0x08 || len <  2) {
		return( -EINVAL );
	}
	if ((p = data + data[1] + 2) >= data_limit) {
		return( -EINVAL );
	}

	q931MsgType = *p++;
	INF( "masq_h225_out: Q.931/%s\n",toStr_q931msg( q931MsgType ));

	switch (q931MsgType) {
	case Q931_ALERT:
	case Q931_CALLPROC:
	case Q931_PROGRESS:
	case Q931_SETUP:
	case Q931_CONNECT:
		break;
	default:
		return( 0 );
	}

	memcpy( &callerAddr[0],&ms->saddr,4 );
	memcpy( &callerAddr[4],&ms->sport,2 );

	/* find user-user IE */
	if ((p = findIE( 0x7e,p,data_limit )) != NULL) {
		p += 4;		/* begin of h323_uu_pdu */
		while ((p + 5) < data_limit) {
			if (q931MsgType == Q931_SETUP) {
				/* ipaddr, portno */
				if (!memcmp( p,callerAddr,6 )) {
					DEB( "masq_h225_out: patching srcCallSignalAddress\n" );
					memcpy( p,&ms->maddr,4 );
					memcpy( p+4,&ms->mport,2 );
					p += 6;
					continue;
				}
			}
			if (ntohs( ms->mport ) < PORT_MASQ_BEGIN &&
			    !memcmp( p,callerAddr,4 ) &&
			    p[4] != PORT_H225_H && p[5] != PORT_H225_L)
			{
				struct ip_masq	*nms;

				INF( "masq_h225_out: found h245Address %d.%d.%d.%d:%d\n",p[0],p[1],p[2],p[3],p[4]*256+p[5] );
				if ((nms = ip_masq_out_get_2( IPPROTO_TCP,ms->saddr,htons( p[4]*256+p[5] ),ms->daddr,0 )) != NULL) {
					ip_masq_set_expire( nms,0 );
				} else {
					if ((nms = ip_masq_new( dev,IPPROTO_TCP,ms->saddr,htons( p[4]*256+p[5] ),ms->daddr,0,IP_MASQ_F_NO_DPORT )) == NULL) {
						ERR( "masq_h225_out: ip_masq_new failed!\n" );
						return( 0 );
					}
				}
				masqMonitorH245Port( p[4]*256+p[5] );
				ip_masq_bind_app( nms );
				DEB( "masq_h225_out: patching h245Address\n" );
				memcpy( p,&nms->maddr,4 );
				memcpy( p+4,&nms->mport,2 );
				p += 6;
				continue;
			}
			p++;
		}
	}
	return( 0 );
}

static int
masq_h225_in( struct ip_masq_app *mapp,
              struct ip_masq *ms,
              struct sk_buff **skb_p,
              struct device *dev )
{
	tpkt_app_t	*ap;

	DEB( "masq_h225_in: ENTER\n" );
	if ((ap = ms->app_data) != NULL) {
		tpktFrameSKB( &ap->rx,ms,dev,skb_p );
	}
	return( 0 );
}

static int
masq_h225_in_2( struct ip_masq *ms,
                struct device *dev,
                unsigned char *data,
                int len )
{
	unsigned char	*data_limit;
	unsigned char	*p;
	int		q931MsgType;
	unsigned char	routerAddr[6];

	DEB( "masq_h225_in_2: ENTER len=%d\n",len );
	DEB( ">>> caller: %s:%d\n",in_ntoa( ms->saddr ),ntohs( ms->sport ));
	DEB( ">>> router: %s:%d\n",in_ntoa( ms->maddr ),ntohs( ms->mport ));
	DEB( ">>> callee: %s:%d\n",in_ntoa( ms->daddr ),ntohs( ms->dport ));

	data_limit = data + len;

	/* q.931 protocol discr, callref len, callref, msgtype */
	if (data[0] != 0x08 || len <  2) {
		return( -EINVAL );
	}
	if ((p = data + data[1] + 2) >= data_limit) {
		return( -EINVAL );
	}

	q931MsgType = *p++;
	INF( "masq_h225_in: Q.931/%s\n",toStr_q931msg( q931MsgType ));

	switch (q931MsgType) {
	case Q931_ALERT:
	case Q931_CALLPROC:
	case Q931_PROGRESS:
	case Q931_SETUP:
	case Q931_CONNECT:
		break;
	default:
		return( 0 );
	}

	memcpy( &routerAddr[0],&ms->maddr,4 );
	memcpy( &routerAddr[4],&ms->mport,2 );

	/* find user-user IE */
	if ((p = findIE( 0x7e,p,data_limit )) != NULL) {
		p += 4;		/* begin of h323_uu_pdu */
		while ((p + 5) < data_limit) {
			if (q931MsgType == Q931_SETUP) {
				/* ipaddr, portno */
				if (!memcmp( p,routerAddr,6 )) {
					DEB( "masq_h225_in: patching destCallSignalAddress\n" );
					memcpy( p,&ms->saddr,4 );
					memcpy( p+4,&ms->sport,2 );
					p += 6;
					continue;
				}
			}
			if (ntohs( ms->mport ) >= PORT_MASQ_BEGIN &&
			    !memcmp( p,&ms->daddr,4 ) &&
			    p[4] != PORT_H225_H && p[5] != PORT_H225_L)
			{
				INF( "masq_h225_in: found h245Address %d.%d.%d.%d:%d\n",p[0],p[1],p[2],p[3],p[4]*256+p[5] );
				masqMonitorH245Port( p[4]*256+p[5] );
				p += 6;
				continue;
			}
			p++;
		}
	}
	return( 0 );
}

static int
masq_h225_init_1( struct ip_masq_app *mapp,struct ip_masq *ms )
{
	DEB( "masq_h225_init_1: ENTER\n" );
	DEB( ">>> caller: %s:%d\n",in_ntoa( ms->saddr ),ntohs( ms->sport ));
	DEB( ">>> router: %s:%d\n",in_ntoa( ms->maddr ),ntohs( ms->mport ));
	DEB( ">>> callee: %s:%d\n",in_ntoa( ms->daddr ),ntohs( ms->dport ));
	if (ms->app_data != NULL) {
		WRN( "masq_h225_init_1: app_data != NULL\n" );
	}
	ms->app_data = tpktAppAlloc( masq_h225_out_2,masq_h225_in_2 );
	return( 0 );
}

static int
masq_h225_done_1( struct ip_masq_app *mapp,struct ip_masq *ms )
{
	DEB( "masq_h225_done_1: ENTER\n" );
	DEB( ">>> caller: %s:%d\n",in_ntoa( ms->saddr ),ntohs( ms->sport ));
	DEB( ">>> router: %s:%d\n",in_ntoa( ms->maddr ),ntohs( ms->mport ));
	DEB( ">>> callee: %s:%d\n",in_ntoa( ms->daddr ),ntohs( ms->dport ));
	tpktAppRelease( ms->app_data );
	ms->app_data = NULL;
	return( 0 );
}

struct ip_masq_app	ip_masq_h225_app = {
	NULL,			/* next */
	"H.225",
	0,                      /* type */
	0,                      /* n_attach */
	masq_h225_init_1,	/* ip_masq_init_1 */
	masq_h225_done_1,	/* ip_masq_done_1 */
	masq_h225_out,		/* pkt_out */
	masq_h225_in    	/* pkt_in */
};

int
ip_masq_h323_init( void )
{
	int	rc;

	DEB( "ip_masq_h323_init: ENTER\n" );
	tpktInit();
	if ((rc = register_ip_masq_app( &ip_masq_h225_app,
	                                IPPROTO_TCP,
	                                PORT_H225 )))
	{
		ERR( "initalizing H.225 support on port=%d",PORT_H225 );

	}
	DEB( "ip_masq_h323_init: LEAVE(%d)\n",rc );
	return( rc );
}

int
ip_masq_h323_finit( void )
{
	DEB( "ip_masq_h323_finit: ENTER\n" );
	DEB( "ip_masq_h323_finit: LEAVE\n" );
	return( 0 );
}

#ifdef MODULE
#error TODO
#endif
