/* -*- linux-c -*-
 * sysctl_net_ipv4.c: sysctl interface to net IPV4 subsystem.
 *
 * Begun April 1, 1996, Mike Shaver.
 * Added /proc/sys/net/ipv4 directory entry (empty =) ). [MS]
 */

/*
 * Modifications for P2001 Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 *      
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

#include <linux/mm.h>
#include <linux/sysctl.h>
#include <net/ip.h>

/* From arp.c */
extern int sysctl_arp_res_time;
extern int sysctl_arp_dead_res_time;
extern int sysctl_arp_max_tries;
extern int sysctl_arp_timeout;
extern int sysctl_arp_check_interval;
extern int sysctl_arp_confirm_interval;
extern int sysctl_arp_confirm_timeout;

extern int sysctl_ip_forward;
extern int sysctl_ip_dynaddr;
#ifndef CONFIG_NO_PATH_MTU_DISCOVERY
int sysctl_ip_pmtu = 1;
#endif

static int proc_doipforward(ctl_table *ctl, int write, struct file *filp,
			    void *buffer, size_t *lenp) 
{
    int val = sysctl_ip_forward;    
    int retv;

    retv = proc_dointvec(ctl, write, filp, buffer, lenp);
    if (write) {
	if (sysctl_ip_forward && !val) {
	    printk(KERN_INFO "sysctl: ip forwarding enabled\n");
	    ip_statistics.IpForwarding = 1; 
	}
	if (!sysctl_ip_forward && val) {
	    printk(KERN_INFO "sysctl: ip forwarding off\n");
	    ip_statistics.IpForwarding = 2;
	}
    }
    return retv; 
}

#ifndef CONFIG_NO_PATH_MTU_DISCOVERY
static int
proc_dopmtu( ctl_table *ctl,int write,struct file *filp,
             void *buffer, size_t *lenp )
{
	int	val = sysctl_ip_pmtu;    
	int	rc;

	rc = proc_dointvec( ctl,write,filp,buffer,lenp );
	if (write) {
		if (sysctl_ip_pmtu && !val) {
			printk( KERN_INFO"sysctl: pmtu discovery enabled\n" );
		}
		if (!sysctl_ip_pmtu && val) {
			printk( KERN_INFO"sysctl: pmtu discovery disabled\n" );
		}
	}
	return( rc );
}
#endif

ctl_table ipv4_table[] = {
        {NET_IPV4_ARP_RES_TIME, "arp_res_time",
         &sysctl_arp_res_time, sizeof(int), 0644, NULL, &proc_dointvec},
        {NET_IPV4_ARP_DEAD_RES_TIME, "arp_dead_res_time",
         &sysctl_arp_dead_res_time, sizeof(int), 0644, NULL, &proc_dointvec},
        {NET_IPV4_ARP_MAX_TRIES, "arp_max_tries",
         &sysctl_arp_max_tries, sizeof(int), 0644, NULL, &proc_dointvec},
        {NET_IPV4_ARP_TIMEOUT, "arp_timeout",
         &sysctl_arp_timeout, sizeof(int), 0644, NULL, &proc_dointvec},
        {NET_IPV4_ARP_CHECK_INTERVAL, "arp_check_interval",
         &sysctl_arp_check_interval, sizeof(int), 0644, NULL, &proc_dointvec},
        {NET_IPV4_ARP_CONFIRM_INTERVAL, "arp_confirm_interval",
         &sysctl_arp_confirm_interval, sizeof(int), 0644, NULL,
         &proc_dointvec},
        {NET_IPV4_ARP_CONFIRM_TIMEOUT, "arp_confirm_timeout",
         &sysctl_arp_confirm_timeout, sizeof(int), 0644, NULL,
         &proc_dointvec},
	{NET_IPV4_FORWARD, "ip_forward", &sysctl_ip_forward, sizeof(int),
	 0644, NULL, &proc_doipforward },
        {NET_IPV4_DYNADDR, "ip_dynaddr",
         &sysctl_ip_dynaddr, sizeof(int), 0644, NULL, &proc_dointvec},
#ifndef CONFIG_NO_PATH_MTU_DISCOVERY
        {NET_IPV4_PMTU_DISCOVERY, "pmtu_discovery",
         &sysctl_ip_pmtu, sizeof(int), 0644, NULL, &proc_dopmtu},
#endif
	{0}
};
